/****************************************************************************
 *
 * FILENAME:        $RCSfile: factory.c,v $
 *
 * LAST REVISION:   $Revision: 1.11 $
 * LAST MODIFIED:   $Date: 2015/02/25 01:58:21 $
 *
 * DESCRIPTION:     factory data support
 *
 * vi: set ts=4:
 *
 * Copyright (c) 2013 by Grandstream Networks, Inc.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 *
 ***************************************************************************/

//=======================
//  Includes
//=======================
#include <linux/types.h>
#include <linux/vmalloc.h>
#include <linux/mtd/mtd.h>
#include <linux/module.h>
#include <gs/factory.h>
#include <provision.h>
#include <gs/gs_sec_module.h>

//=======================
//  Defines
//=======================
//must be updated for every new model inthe table
#define HW_MODEL_TABLE_SIZE     (sizeof(hw_ids) / (3 * sizeof(uint16_t)) )
#define ATE_MAGIC (47954745)

//=======================
//  Globals
//=======================

//=======================
//  Typedefs
//=======================

//=======================
//  Locals
//=======================

    /*
     * Private Functions
     */

    /*
     * Private Data
     */
static hw_info_t hw_info;
static sec_info_t sec_info;
static char factory_para_buf[GS_FACTORY_CFG_SIZE];
static GS_FACTORYPARA * factory_para = NULL;
static int feature_table_loaded = 0;
static int factory_para_dirty = 0;
static int ate_mode_cache = -1;

//=======================
//  Functions
//=======================

//===========================================================================
int
ate_mode(
    void
)
// Description: Cache the initial value of the production flag.  Right now
// this is only used for MTD locks.  If the device boots up as non-production
// then allow access to sensitive MTD parts until the device is rebooted.
//===========================================================================
{
    if ( ate_mode_cache < 0 ) {
        printk("ate_mode triggers fact_para_get\n");
        fact_para_get();
    }

    return ( ate_mode_cache == ATE_MAGIC );
}
EXPORT_SYMBOL(ate_mode);

//===========================================================================
void
enter_ate_mode(
    void
)
//===========================================================================
{
    ate_mode_cache = ATE_MAGIC;
}
EXPORT_SYMBOL(enter_ate_mode);

//===========================================================================
static void
load_factory_para(
    void
)
//===========================================================================
{
    int len;
    struct mtd_info *mtd;

    mtd = get_mtd_device_nm( "factory" );
    if ( mtd <= 0 )
    {
        printk("load_factory_para: unable to open mtd device\n");
        return;
    }

    mtd_read( mtd, GS_FACTORY_CFG_OFFSET, GS_FACTORY_CFG_SIZE, &len, factory_para_buf );
    if ( len != GS_FACTORY_CFG_SIZE )
    {
        printk("load_factory_para: mtd_read read size != GS_FACTORY_CFG_SIZE\n");
    }

    put_mtd_device( mtd );
}

//===========================================================================
uint16_t
fact_para_checksum(
    GS_FACTORYPARA * factPara
)
//===========================================================================
{
    GS_FACTORYPARA tmpData;
    uint8_t *p = (uint8_t*)&tmpData;
    uint16_t sum = 0;
    uint16_t checksum = 0;
    unsigned int length = 0;
    int i;

    length = ( factPara->length[1] << 16 ) | factPara->length[0];
    if ( length > GS_FACTORY_CFG_SIZE )
    {
        printk("fact_para_checksum: length exceeds GS_FACTORY_CFG_SIZE\n");
        return 0xDEAD;
    }
    p = ( unsigned char * )vmalloc( length );
	if (p == NULL) {
        printk("fact_para_checksum: vmalloc failed\n");
        return 0xDEAD;
    }
    memcpy( p, factPara, length );
    ((GS_FACTORYPARA*)p)->checksum = 0;

    for ( i = 0; i < length; i+= 2 )
    {
         sum += ((p[i] << 8) | p[i+1]);
    }
    checksum = 0xffff - sum + 0x01;

    vfree( p );

    return ((checksum >> 8) | (checksum << 8));
}


//===========================================================================
GS_FACTORYPARA *
fact_para_get(
    void
)
//===========================================================================
{
    if ( factory_para == NULL || factory_para_dirty )
    {
        factory_para = (GS_FACTORYPARA *)factory_para_buf;
        load_factory_para();
        factory_para_dirty = 0;
    }
    if ( ate_mode_cache < 0 ) {
        if ( factory_para->hw_production != 1 ) {
            enter_ate_mode();
        }
    }
    return factory_para;
}
EXPORT_SYMBOL(fact_para_get);

//===========================================================================
void
free_memory(
    void
)
//===========================================================================
{
    if ( hw_info.hw_logo_data )
        vfree( hw_info.hw_logo_data );

    if ( sec_info.certificate.data )
        vfree( sec_info.certificate.data );

    if ( sec_info.private_key.data )
        vfree( sec_info.private_key.data );

    if ( sec_info.sn.data )
        vfree( sec_info.sn.data );

    if ( sec_info.hash.data )
        vfree( sec_info.hash.data );

#ifdef CONFIG_CFG80211
    if ( sec_info.country_code.data )
        vfree( sec_info.country_code.data );
#endif

    if ( sec_info.password.data )
        vfree( sec_info.password.data );

    if ( sec_info.xml_key.data )
        vfree( sec_info.xml_key.data );

    if ( sec_info.checksum.data )
        vfree( sec_info.checksum.data );
}

//===========================================================================
void
set_factory_para_dirty(
    void
)
//===========================================================================
{
    factory_para_dirty = 1;
    feature_table_loaded = 0;
}

//===========================================================================
sec_info_t*
sec_info_get(
    void
)
//===========================================================================
{
    hw_get_hw_feature();
    return &sec_info;
}

//===========================================================================
hw_info_t *
hw_get_hw_feature(
    void
)
// Description: Get the hardware logic features from HW feature table in flash factory
//              parameter sector
//===========================================================================
{
    factory_param_sector_t *fact_para;
    hw_feature_table_header_t *hwTableHdr;
    hw_feature_attrib_t *featurePtr;
    uint16_t tableLength, parsedLength, index, featureLen;
    static int has_security_table = 0;

    if ( feature_table_loaded )
    {
        return &hw_info;
    }

    if ( factory_para_dirty )
        free_memory();

    fact_para = (factory_param_sector_t*)fact_para_get();
    memset( &hw_info, 0, sizeof ( hw_info ) );
    memset( &sec_info, 0, sizeof ( sec_info ) );

    hw_info.poe_class = 0xFF; /* Set so we can test for existence in flash */
    hw_info.num_linekeys = 0xFF;	// Initial NULL value for models that do not have it in flash.

	// make sure there is at least one HW feature attribute in the flash
   if( !(fact_para->factory_para.length[0] >= sizeof(factory_param_sector_t)) )
   {
       printk("Warning: No factory table\n");
       return &hw_info;
   }

    // make sure the flash sector contains a complete HW feature table
    tableLength = fact_para->hw_feature_table_header.hw_feature_table_len;
    tableLength += sizeof(GS_FACTORYPARA);

    if ( fact_para->factory_para.length[0] >= tableLength )
    {
        has_security_table = 1;
    }

    // locate the HW feature table in the sector, in case there are other tables in it
    hwTableHdr = &fact_para->hw_feature_table_header;
    while ( hwTableHdr->hw_feature_table_version != HW_FEATURE_TABLE_VERSION )
    {
        hwTableHdr = (hw_feature_table_header_t*)( (uint16_t*)hwTableHdr + hwTableHdr->hw_feature_table_len);
        if ((uint16_t*)hwTableHdr-(uint16_t*)fact_para >= fact_para->factory_para.length[0])
        {  // did not find the HW feature table in the flash sector
            hwTableHdr = NULL;
            break;
        }
    }
    if ( hwTableHdr == NULL )
    {
        printk("Warning: Can't find feature table\n");
        return &hw_info;
    }

    tableLength = sizeof(uint16_t) * hwTableHdr->hw_feature_table_len;
    featurePtr = &fact_para->hw_feature;
    parsedLength = 6;

    for (index=0; index<hwTableHdr->hw_feature_attrib_count; index++)
    {
        switch (featurePtr->hw_feature_type)
        {
            case HW_FEATURE_VENDOR:
                memcpy( &hw_info.vendor, &featurePtr->hw_feature_value, featurePtr->hw_feature_len * sizeof(uint16_t) );
                break;
            case HW_FEATURE_MODELSTR:
                memcpy( &hw_info.modelstr, &featurePtr->hw_feature_value, featurePtr->hw_feature_len * sizeof(uint16_t) );
                break;
            case HW_FEATURE_MODELALIAS:
                memcpy( &hw_info.modelalias, &featurePtr->hw_feature_value, featurePtr->hw_feature_len * sizeof(uint16_t) );
                break;
            case HW_FEATURE_OEM_ID:
                hw_info.oem_id = featurePtr->hw_feature_value;
                break;
            case HW_FEATURE_VENDOR_FULLNAME:
                memcpy( &hw_info.vendor_fullname, &featurePtr->hw_feature_value, featurePtr->hw_feature_len * sizeof(uint16_t) );
                break;
            case HW_FEATURE_FWKEY:
                memcpy( &hw_info.fwkey, &featurePtr->hw_feature_value, featurePtr->hw_feature_len * sizeof(uint16_t) );
                break;
            case HW_FEATURE_ETH:
                hw_info.eth = featurePtr->hw_feature_value;
                break;
			case HW_FEATURE_ETH_INTERFACE:
				hw_info.eth_interface = featurePtr->hw_feature_value;
				break;
            case HW_FEATURE_HANDSET:
                hw_info.handset = featurePtr->hw_feature_value;
                break;
			case HW_FEATURE_HEADSET_MIC_DETECT:
				hw_info.headset_mic_detect = featurePtr->hw_feature_value;
            case HW_FEATURE_NUM_ACCTS:
				hw_info.num_accts = featurePtr->hw_feature_value;
                break;
            case HW_FEATURE_EXT:
                hw_info.ext = featurePtr->hw_feature_value;
                break;
            case HW_FEATURE_NUM_MPK:
				hw_info.num_mpk = featurePtr->hw_feature_value;
                break;
            case HW_FEATURE_CPU:
                memcpy( &hw_info.cpu, &featurePtr->hw_feature_value, featurePtr->hw_feature_len * sizeof ( uint16_t ) );
                break;
            case HW_FEATURE_POE_CLASS:
                hw_info.poe_class = featurePtr->hw_feature_value;
                break;
            case HW_FEATURE_FLASH:
                hw_info.flash_device = featurePtr->hw_feature_value;
                break;
            case HW_FEATURE_FLASH_SIZE:
                memcpy( &hw_info.flash_size, &featurePtr->hw_feature_value, featurePtr->hw_feature_len * sizeof ( uint16_t ) );
                break;
            case HW_FEATURE_BLUETOOTH:
                hw_info.bluetooth = featurePtr->hw_feature_value;
                break;
            case HW_FEATURE_BLUETOOTH_CHIP:
                hw_info.bluetooth_chip = featurePtr->hw_feature_value;
                break;
            case HW_FEATURE_OEM_FILE_PREFIX:
                memcpy( &hw_info.oem_file_prefix, &featurePtr->hw_feature_value, featurePtr->hw_feature_len * sizeof ( uint16_t ) );
                break;
			case HW_FEATURE_MODEL_BASE:
				memcpy( &hw_info.gs_model, &featurePtr->hw_feature_value, featurePtr->hw_feature_len * sizeof ( uint16_t ) );
				break;
            case HW_FEATURE_OEM_BASE:
                memcpy( &hw_info.oem_base, &featurePtr->hw_feature_value, featurePtr->hw_feature_len * sizeof ( uint16_t ) );
                break;
            case HW_FEATURE_GS_MODEL_NAME:
                memcpy( &hw_info.gs_model_name, &featurePtr->hw_feature_value, featurePtr->hw_feature_len * sizeof ( uint16_t ) );
                break;
            case HW_FEATURE_NUM_LINEKEYS:
				hw_info.num_linekeys = featurePtr->hw_feature_value;
                break;
			case HW_FEATURE_NUM_SOFTKEYS:
				hw_info.num_softkeys = featurePtr->hw_feature_value;
				break;
            case HW_FEATURE_LOGODEF:
                hw_info.hw_logo_definition = featurePtr->hw_feature_value;
                break;
            case HW_FEATURE_LOGODATA:
                hw_info.hw_logo_data = vmalloc( featurePtr->hw_feature_len * sizeof ( uint16_t ) );
                memcpy( hw_info.hw_logo_data, &featurePtr->hw_feature_value, featurePtr->hw_feature_len * sizeof ( uint16_t ) );
                hw_info.hw_logo_len = featurePtr->hw_feature_len * sizeof ( uint16_t );
                break;
            case HW_FEATURE_MINI_LOGODEF:
                hw_info.hw_mini_logo_definition = featurePtr->hw_feature_value;
                break;
            case HW_FEATURE_MINI_LOGODATA:
                hw_info.hw_mini_logo_data = vmalloc( featurePtr->hw_feature_len * sizeof ( uint16_t ) );
                memcpy( hw_info.hw_mini_logo_data, &featurePtr->hw_feature_value, featurePtr->hw_feature_len * sizeof ( uint16_t ) );
                hw_info.hw_mini_logo_len = featurePtr->hw_feature_len * sizeof ( uint16_t );
                break;
            case HW_FEATURE_NO_USB:
                hw_info.usb_unsupported = featurePtr->hw_feature_value;
                break;
            case HW_FEATURE_LCD_PANEL:
                hw_info.hw_lcd_panel = featurePtr->hw_feature_value;
                break;
            case HW_FEATURE_LCD_XRES:
                memcpy( &hw_info.hw_lcd_xres, &featurePtr->hw_feature_value, featurePtr->hw_feature_len * sizeof ( uint16_t ) );
                break;
            case HW_FEATURE_LCD_YRES:
                memcpy( &hw_info.hw_lcd_yres, &featurePtr->hw_feature_value, featurePtr->hw_feature_len * sizeof ( uint16_t ) );
                break;
            case HW_FEATURE_REQUIRE_PEM:
                hw_info.require_pem = featurePtr->hw_feature_value;
                break;
            case HW_FEATURE_REQUIRE_SN:
                hw_info.require_sn = featurePtr->hw_feature_value;
                break;
            case HW_FEATURE_REQUIRE_ADMIN_PWD:
                hw_info.require_admin_pwd = featurePtr->hw_feature_value;
                break;
            case HW_FEATURE_REQUIRE_XML_KEY:
                hw_info.require_xml_key = featurePtr->hw_feature_value;
                break;
            case HW_FEATURE_REQUIRE_HASH:
                hw_info.require_hash = featurePtr->hw_feature_value;
                break;
			case HW_FEATURE_LCD_BACKLIGHT_DRIVER:
				hw_info.lcd_backlight_driver = featurePtr->hw_feature_value;
				break;
            case HW_FEATURE_LCD_SECOND_PANEL:
				hw_info.second_panel = featurePtr->hw_feature_value;
                break;
            case HW_FEATURE_PHONEBOOK_KEY:
				hw_info.phonebook_key = featurePtr->hw_feature_value;
                break;
#ifdef CONFIG_CFG80211
            case HW_FEATURE_WIFI_CHIP:
                hw_info.wifi_chip = featurePtr->hw_feature_value;
                break;
			case HW_FEATURE_WIFI_COUNTRY_CODE:
				memcpy( &hw_info.wifi_country_code, &featurePtr->hw_feature_value, featurePtr->hw_feature_len * sizeof ( uint16_t ) );
				break;
#endif
            case HW_FEATURE_CONSOLE_UART:
                hw_info.console_uart = featurePtr->hw_feature_value;
                break;
            default:
                break;
        }

        // move to the next attribute
        featureLen = sizeof(hw_feature_attrib_t) + sizeof(uint16_t)*( featurePtr->hw_feature_len - 1 );
        parsedLength += featureLen;
        featurePtr = (hw_feature_attrib_t*)((char*)featurePtr + featureLen);
        if (parsedLength >= tableLength)
        {
            break;
        }
    }

    feature_table_loaded = 1;
    
    // locate the security table
    hwTableHdr = &fact_para->hw_feature_table_header;
    while ( hwTableHdr->hw_feature_table_version != HW_SECURITY_TABLE_VERSION )
    {
        hwTableHdr = (hw_feature_table_header_t*) ( (uint16_t*)hwTableHdr + hwTableHdr->hw_feature_table_len );
        if ((uint16_t*)hwTableHdr-(uint16_t*)fact_para >= fact_para->factory_para.length[0] || hwTableHdr->hw_feature_table_version == 0)
        {  // did not find the HW feature table in the flash sector
            hwTableHdr = NULL;
            break;
        }
    }
    if ( hwTableHdr == NULL )
    {
        //dump_stack();
        printk("Warning: Can't find security table\n");
        return &hw_info;
    }

    tableLength = sizeof(uint16_t) * hwTableHdr->hw_feature_table_len;
    featurePtr = (hw_feature_attrib_t*)( (uint8_t*)hwTableHdr + sizeof ( hw_feature_table_header_t ) );
    parsedLength = 6;

    if ( fact_para->factory_para.hw_production == 1 ) 
    {
        /* Perform decryption on the security table */
        printk("Decrypting security table\n");
        if ( gs_sec_fact_mod_cipher( GS_SEC_METHOD_DECRYPT, (char*)featurePtr, 
                                     tableLength - sizeof ( hw_feature_table_header_t ) ) == 0 )
        {
            fact_para->factory_para.checksum = fact_para_checksum( &fact_para->factory_para );
        }
        else
        {
            printk("WARNING!!! Security table decryption failed\n");
        	return &hw_info;
        }
    }

    printk( "Loading Security Table: tableLength=%d attribCount=%d\n",
            tableLength, hwTableHdr->hw_feature_attrib_count );
    
    for (index=0; index < hwTableHdr->hw_feature_attrib_count; index++)
    {
        // printk( "Security-table feature: type=%d len=%d\n", featurePtr->hw_feature_type, 
        //         featurePtr->hw_feature_len * sizeof ( uint16_t ) );
        if ( featurePtr->hw_feature_len == 0 )
        {
            printk( "Skipping security-table feature %d with lenght 0\n", featurePtr->hw_feature_type );
            goto skip;
        }

        switch (featurePtr->hw_feature_type)
        {
            case SEC_FEATURE_SN:
                sec_info.sn.data = vmalloc( featurePtr->hw_feature_len * sizeof ( uint16_t ) );
                memcpy( sec_info.sn.data, &featurePtr->hw_feature_value, featurePtr->hw_feature_len * sizeof ( uint16_t ) );
                sec_info.sn.len = featurePtr->hw_feature_len * sizeof ( uint16_t  );
                break;
            case SEC_FEATURE_PRIVATE_KEY:
                sec_info.private_key.data = vmalloc( featurePtr->hw_feature_len * sizeof ( uint16_t ) );
                memcpy( sec_info.private_key.data, &featurePtr->hw_feature_value, featurePtr->hw_feature_len * sizeof ( uint16_t ) );
                sec_info.private_key.len = featurePtr->hw_feature_len * sizeof ( uint16_t );
                //printk("Feature Size: %d\n", sec_info.private_key.len);
                break;
            case SEC_FEATURE_CERTIFICATE:
                sec_info.certificate.data = vmalloc( featurePtr->hw_feature_len * sizeof ( uint16_t ) );
                memcpy( sec_info.certificate.data, &featurePtr->hw_feature_value, featurePtr->hw_feature_len * sizeof ( uint16_t ) );
                sec_info.certificate.len = featurePtr->hw_feature_len * sizeof ( uint16_t );
                //printk("Feature Size: %d\n", sec_info.certificate.len );
                break;
            case SEC_FEATURE_CHECKSUM:
                sec_info.checksum.data = vmalloc( featurePtr->hw_feature_len * sizeof ( uint16_t ) );
                memcpy( sec_info.checksum.data, &featurePtr->hw_feature_value, featurePtr->hw_feature_len * sizeof ( uint16_t ) );
                sec_info.checksum.len = featurePtr->hw_feature_len * sizeof ( uint16_t );
                //printk("Feature Size: %d\n", sec_info.checksum.len );
                break;
            case SEC_FEATURE_HASH:
                sec_info.hash.data = vmalloc( featurePtr->hw_feature_len * sizeof ( uint16_t ) );
                memcpy( sec_info.hash.data, &featurePtr->hw_feature_value, featurePtr->hw_feature_len * sizeof ( uint16_t ) );
                sec_info.hash.len = featurePtr->hw_feature_len * sizeof ( uint16_t );
                break;
#ifdef CONFIG_CFG80211
            case SEC_FEATURE_COUNTRY_CODE:
                sec_info.country_code.data = vmalloc( featurePtr->hw_feature_len * sizeof ( uint16_t ) );
                memcpy( sec_info.country_code.data, &featurePtr->hw_feature_value, featurePtr->hw_feature_len * sizeof ( uint16_t ) );
                sec_info.country_code.len = featurePtr->hw_feature_len * sizeof ( uint16_t );
                break;
#endif
            case SEC_FEATURE_PASSWORD:
                sec_info.password.data = vmalloc( featurePtr->hw_feature_len * sizeof ( uint16_t ) );
                memcpy( sec_info.password.data, &featurePtr->hw_feature_value, featurePtr->hw_feature_len * sizeof ( uint16_t ) );
                sec_info.password.len = featurePtr->hw_feature_len * sizeof ( uint16_t );
                break;
            case SEC_FEATURE_XML_KEY:
                sec_info.xml_key.data = vmalloc( featurePtr->hw_feature_len * sizeof ( uint16_t ) );
                memcpy( sec_info.xml_key.data, &featurePtr->hw_feature_value, featurePtr->hw_feature_len * sizeof ( uint16_t ) );
                sec_info.xml_key.len = featurePtr->hw_feature_len * sizeof ( uint16_t );
                break;
        }
skip:
        // move to the next attribute
        featureLen = sizeof(hw_feature_attrib_t) + sizeof(uint16_t)*( featurePtr->hw_feature_len - 1 );
        parsedLength += featureLen;
        featurePtr = (hw_feature_attrib_t*)((char*)featurePtr + featureLen);
        if (parsedLength >= tableLength)
        {
            break;
        }
    }

    return &hw_info;
}

//===========================================================================
void
hw_update_hw_feature(
	hw_feature_table_header_t* hwTableHdr,
	uint16_t feature_type,
    char* value,
	unsigned int length
)
// Description: Update the hardware logic features from HW feature table 
// in flash factory parameter sector
//===========================================================================
{
    hw_feature_table_header_t *new_hwTableHdr = NULL;
    hw_feature_attrib_t *featurePtr;
    hw_feature_attrib_t *new_featurePtr;
    uint16_t tableLength, parsedLength, index, featureLen;
	uint8_t feature_found_flag = 0;

    tableLength = sizeof(uint16_t) * hwTableHdr->hw_feature_table_len;
    featurePtr = (hw_feature_attrib_t*)( hwTableHdr + sizeof( hw_feature_table_header_t ) );
    parsedLength = 6;
    new_hwTableHdr = ( hw_feature_table_header_t* )vmalloc( 
	    sizeof(uint16_t) * hwTableHdr->hw_feature_table_len + ( length + 1 )/ sizeof(uint16_t) );
	if ( new_hwTableHdr == NULL )
	{
		printk( "hw_update_hw_feature: vmalloc failed\n" );
		return;
	}

	new_hwTableHdr->hw_feature_table_len = hwTableHdr->hw_feature_table_len;
	new_hwTableHdr->hw_feature_table_version = hwTableHdr->hw_feature_table_version;
	new_hwTableHdr->hw_feature_attrib_count = hwTableHdr->hw_feature_attrib_count;
	new_featurePtr = (hw_feature_attrib_t*)( new_hwTableHdr + sizeof( hw_feature_table_header_t ) );

    for (index=0; index<hwTableHdr->hw_feature_attrib_count; index++)
    {
        printk("featurePtr->hw_feature_type = %d\n", featurePtr->hw_feature_type);
		new_featurePtr->hw_feature_type = featurePtr->hw_feature_type;
		if ( feature_type != featurePtr->hw_feature_type )
		{
			new_featurePtr->hw_feature_len = featurePtr->hw_feature_len;
			memcpy( &new_featurePtr->hw_feature_value, 
			        &featurePtr->hw_feature_value, 
					featurePtr->hw_feature_len * sizeof(uint16_t) );
		}
		else
		{
			if ( feature_found_flag )
				printk( "Duplicate feature %x\n", feature_type );
		    feature_found_flag++;
		}

        // move to the next attribute
        featureLen = sizeof(hw_feature_attrib_t) + sizeof(uint16_t)*( featurePtr->hw_feature_len - 1 );
        parsedLength += featureLen;
        featurePtr = (hw_feature_attrib_t*)((char*)featurePtr + featureLen);
    }
	parsedLength += 4;
	new_featurePtr->hw_feature_type = feature_type;
	new_featurePtr->hw_feature_len = ( length + 1 )/ sizeof( uint16_t );
	memcpy( ( char* )&(new_featurePtr->hw_feature_value), value, length );
	parsedLength += length;
	new_hwTableHdr->hw_feature_table_len = ( parsedLength + 1 )/ sizeof( uint16_t );
    if ( 0 == feature_found_flag )
        new_hwTableHdr->hw_feature_attrib_count++;

    memcpy( hwTableHdr, new_hwTableHdr, parsedLength );
	vfree( new_hwTableHdr );

    printk("Done\n");

    return;
}

//===========================================================================
void
hw_dump_hw_feature(
    factory_param_sector_t *fact_para
)
// Description: Get the hardware logic features from HW feature table in flash factory
//              parameter sector
//===========================================================================
{
    hw_feature_table_header_t *hwTableHdr;
    hw_feature_attrib_t *featurePtr;
    uint16_t tableLength, parsedLength, index, featureLen;
    static int has_security_table = 0;

    printk( "%s %s %d %x %x\n", __FILE__, __FUNCTION__, __LINE__, fact_para->factory_para.length[0], sizeof(factory_param_sector_t) );

	// make sure there is at least one HW feature attribute in the flash
    if( !(fact_para->factory_para.length[0] >= sizeof(factory_param_sector_t)) )
    {
        printk("Warning: No factory table\n");
        return;
    }
    printk( "length0 %x length1 %x\n", fact_para->factory_para.length[0], fact_para->factory_para.length[1] );

    // make sure the flash sector contains a complete HW feature table
    tableLength = fact_para->hw_feature_table_header.hw_feature_table_len;
    printk( "hw feature table length %x\n", fact_para->hw_feature_table_header.hw_feature_table_len );
    tableLength += sizeof(GS_FACTORYPARA);
    printk( "hw feature tablelength %x\n", tableLength );

    if ( fact_para->factory_para.length[0] >= tableLength )
    {
        has_security_table = 1;
    }
    printk( "has security table %x\n", has_security_table );

    // locate the HW feature table in the sector, in case there are other tables in it
    hwTableHdr = &fact_para->hw_feature_table_header;
    while (HW_FEATURE_TABLE_VERSION != hwTableHdr->hw_feature_table_version)
    {
    printk( "hw feature table length %x\n", hwTableHdr->hw_feature_table_len );
        hwTableHdr = (hw_feature_table_header_t*)( (uint16_t*)hwTableHdr + hwTableHdr->hw_feature_table_len);
        if ((uint16_t*)hwTableHdr-(uint16_t*)fact_para >= fact_para->factory_para.length[0])
        {  // did not find the HW feature table in the flash sector
            hwTableHdr = NULL;
            break;
        }
    }
    if ( !(NULL!= hwTableHdr) )
    {
        printk("Warning: Can't find feature table\n");
        return;
    }

    tableLength = sizeof(uint16_t) * hwTableHdr->hw_feature_table_len;
    printk( "hw feature tablelength %x\n", tableLength );
    featurePtr = &fact_para->hw_feature;
    parsedLength = 6;
    printk( "parsedLength %x\n", parsedLength );

    printk("attrib count %x \n", hwTableHdr->hw_feature_attrib_count);
    for (index=0; index<hwTableHdr->hw_feature_attrib_count; index++)
    {
    printk("featurePtr->hw_feature_type = %d length %x \n", featurePtr->hw_feature_type, featurePtr->hw_feature_len);

        // move to the next attribute
        featureLen = sizeof(hw_feature_attrib_t) + sizeof(uint16_t)*( featurePtr->hw_feature_len - 1 );
        parsedLength += featureLen;
    printk( "featureLen %x parsedLength %x\n", featureLen, parsedLength );
        featurePtr = (hw_feature_attrib_t*)((char*)featurePtr + featureLen);
        if (parsedLength >= tableLength)
        {
            break;
        }
    }

    printk("Done\n");

    return;
}

#if 0
EXPORT_SYMBOL(hw_get_hw_feature);
EXPORT_SYMBOL(fact_para_get);
EXPORT_SYMBOL(fact_para_checksum);
#endif

/* EOF */
