/****************************************************************************
 *
 * FILENAME:            $RCSfile: factory_proc.c,v $
 *
 * LAST REVISION:       $Revision: 1.22 $
 * LAST MODIFIED:   $Date: 2015/02/25 01:58:21 $
 *
 * DESCRIPTION:         sipp55 application entry point
 *
 * vi: set ts=4:
 *
 * Copyright (c) 2009 by Grandstream Networks, Inc.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 *
 ***************************************************************************/

//=======================
//      Includes
//=======================
#include <linux/types.h>
#include <linux/proc_fs.h>
#include <linux/uaccess.h>
#include <linux/module.h>
#include <linux/mtd/mtd.h>
#include <linux/vmalloc.h>
#include <linux/delay.h>
#include <linux/fb.h>
#include <linux/seq_file.h>
#include <linux/ctype.h>
#if defined( CONFIG_FB_DSPG )
#include <video/dspgfb.h>
#endif
#if defined( CONFIG_FB_ST7735 )
#include <video/st7789.h>
#endif
#include <gs/factory.h>
#include <provision_funcs.h>
#include <gs/gs_sec_module.h>


//=======================
//     Globals
//=======================
#define GS_MANIFEST_PROC

#ifdef GS_MANIFEST_PROC
#include <linux/toml.h>
#endif

//=======================
//      Defines
//=======================
// #define SECURE_BOOT_OTP_ROM_DEBUG

#define DISABLE_SECURITY_PROCFS_WRITE
#if defined( CONFIG_FB_UC1628 )
#define DEFAULT_CONTRAST_LEVEL 3
#endif


#define DECLARE_READ_PROC( var, type )\
static int \
var##_read_proc(\
    struct seq_file *seq, \
    void *filp \
)\
{\
    int *data = (int*)(get_proc_data(seq->private)); \
    return type##_read_proc( seq, filp, data, &var );\
}

#define DECLARE_WRITE_PROC( var, type, accessor )\
static int \
var##_write_proc(\
    struct file *file,\
    const char __user *buffer,\
    size_t count,\
    loff_t *offset\
)\
{\
    void *data = (void*)(get_proc_data(file)); \
    return type##_write_proc( file, buffer, count, offset, data, &var, accessor );\
}

#define DECLARE_OPEN_PROC( var ) \
static int \
var##_open_proc( \
    struct inode *inod, \
    struct file *file \
) \
{ \
    return single_open( file, var##_read_proc, file ); \
} 

#define DECLARE_FOPS( var ) \
    DECLARE_OPEN_PROC( var ); \
    static struct file_operations var##_fops = { \
        .owner = THIS_MODULE, \
        .open = var##_open_proc, \
        .read = seq_read, \
        .write = var##_write_proc, \
        .release = single_release \
    } \


//=======================
//      Typedefs
//=======================
struct gs_procfs_entry
{
    char * name;
    struct file_operations *fops;
    void ( *init )( void );
    int value;
};

//=======================
//      Locals
//=======================

        /*
         * Private Functions
         */
static int erase_helper( struct mtd_info* mtd, loff_t offset, size_t len );
static void erase_callback( struct erase_info* info );
void lcd_brightness_set( int value );
void mpk_brightness_set( int value );
extern void gs_lcd_set_contrast( int percent );
extern unsigned short gs_contrast_level_to_pm( int level );
static void lcd_brightness_init(void);
static void mpk_brightness_init(void);
static int hw_feature_proc_read( struct seq_file *seq, void *filp );
#ifdef GS_MANIFEST_PROC
static int manifest_proc_read( struct seq_file *seq, void *filp );
#endif
static int devinfo_proc_read( struct seq_file *seq, void *filp );
#ifdef GS_MANIFEST_PROC
static int __init gs_manifest_init( void );
#endif

        /*
         * Private Data
         */
static struct {
    int enabled;
    unsigned int value;
} oem_override;

static unsigned int lcd_on_off;
static unsigned int mpk_on_off;
static unsigned int lcd_brightness;
static unsigned int mpk_brightness;
#if defined( CONFIG_FB_UC1628 )
static unsigned int lcd_contrast;
static unsigned int lcd_contrast_pm;
#endif

#ifdef GS_MANIFEST_PROC
static struct {
    char *full;
    struct toml_object *root;
} manifest;
#endif

//=======================
//      Functions
//=======================

// #define DEBUG
#ifdef DEBUG
static void print_buf(char* heading, unsigned char* buf, int size)
{
	int i;

    if (heading)
        printk("\n%s\n", heading);
    else
        printk("\n");
    
	for (i = 0; i < size; i += 16)
	{
		if ( (i+16) < size )
			printk("%08X: %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X\n",
                i, buf[i], buf[i+1], buf[i+2], buf[i+3], buf[i+4], buf[i+5], buf[i+6], buf[i+7],
                buf[i+8], buf[i+9], buf[i+10], buf[i+11], buf[i+12], buf[i+13], buf[i+14], buf[i+15]);

        //FIXME: Add logic to print remaining bytes that are less than 16 bytes.
    }
	printk("\n");
}
#endif

//=============================================================================
static int
erase_helper(
    struct mtd_info* mtd,
    loff_t offset,
    size_t len
)
//=============================================================================
{
    struct erase_info info;
    DECLARE_COMPLETION_ONSTACK(complete);
    int ret;

    memset( &info, 0, sizeof ( info ) );
    info.mtd = mtd;
    info.addr = offset;
    info.len = len;
    info.callback = erase_callback;
    info.priv = (long)&complete;

    ret = mtd_erase( mtd, &info );
    if ( ret )
        return -EIO;

    wait_for_completion( &complete );
    if ( info.state != MTD_ERASE_DONE )
        return -EIO;

    return 0;
}

//=============================================================================
static void
erase_callback(
    struct erase_info* info
)
//=============================================================================
{
    complete( (struct completion *)info->priv );
}

//=============================================================================
static int
integer_read_proc(
    struct seq_file *seq,
    void *filp,
    int *data,
    int *integer
)
//=============================================================================
{
    seq_printf( seq, "%d\n", *integer );

    return 0;
}

//=============================================================================
static int
integer_write_proc(
    struct file *file,
    const char __user *buffer,
    size_t count,
    loff_t *offset,
    void *data,
    int *integer,
    void (*set)( int val )
)
//=============================================================================
{
    char state_string[ 12 ] = {'\0'};
    char * endp;
    unsigned long new_value = 0;

    if ( ( count > sizeof ( state_string ) - 1 ) )
    {
        return -EINVAL;
    }

    if ( copy_from_user( state_string, buffer, count ) )
    {
        return -EFAULT;
    }

    state_string[ count ] = '\0';

    new_value = simple_strtoul( state_string, &endp, 10 /* Base */ );
    *integer = new_value;
    set( *integer );

    return count;
}

//=============================================================================
static void
lcd_on_off_init(
    void
)
//=============================================================================
{
    lcd_on_off = 1; // lcd is enabled by default
}

//=============================================================================
void
lcd_on_off_set(
    int value
)
//=============================================================================
{
}

DECLARE_WRITE_PROC( lcd_on_off, integer, lcd_on_off_set );
DECLARE_READ_PROC( lcd_on_off, integer );
DECLARE_FOPS( lcd_on_off );

//=============================================================================
static void
lcd_brightness_init(
    void
)
//=============================================================================
{
    lcd_brightness = 80;
}

//=============================================================================
void
lcd_brightness_set(
    int value
)
// LCD brightness value range: [0:100]
//=============================================================================
{
}

DECLARE_WRITE_PROC( lcd_brightness, integer, lcd_brightness_set );
DECLARE_READ_PROC( lcd_brightness, integer );
DECLARE_FOPS( lcd_brightness );

//=============================================================================
static void
mpk_on_off_init(
    void
)
//=============================================================================
{
    mpk_on_off = 1; // lcd is enabled by default
}

//=============================================================================
void
mpk_on_off_set(
    int value
)
//=============================================================================
{
}

DECLARE_WRITE_PROC( mpk_on_off, integer, mpk_on_off_set );
DECLARE_READ_PROC( mpk_on_off, integer );
DECLARE_FOPS( mpk_on_off );

//=============================================================================
static void
mpk_brightness_init(
    void
)
//=============================================================================
{
    mpk_brightness = 80;
}

//=============================================================================
void
mpk_brightness_set(
    int value
)
// LCD brightness value range: [0:100]
//=============================================================================
{
}

DECLARE_WRITE_PROC( mpk_brightness, integer, mpk_brightness_set );
DECLARE_READ_PROC( mpk_brightness, integer );
DECLARE_FOPS( mpk_brightness );

#if defined( CONFIG_FB_UC1628 )
//=============================================================================
static void
lcd_contrast_init(
    void
)
//=============================================================================
{
    lcd_contrast = DEFAULT_CONTRAST_LEVEL;
}

//=============================================================================
void
lcd_contrast_set(
    int level
)
//=============================================================================
{
    if ( level < 1 || level > 5 )
        level = 1;

    lcd_contrast = level;
    lcd_contrast_pm = gs_contrast_level_to_pm( level );

    gs_lcd_set_contrast( lcd_contrast_pm );
}

DECLARE_WRITE_PROC( lcd_contrast, integer, lcd_contrast_set );
DECLARE_READ_PROC( lcd_contrast, integer );
DECLARE_FOPS( lcd_contrast );

//=============================================================================
static void
lcd_contrast_pm_init(
    void
)
//=============================================================================
{
    lcd_contrast_pm = gs_contrast_level_to_pm( DEFAULT_CONTRAST_LEVEL );
}

//=============================================================================
void
lcd_contrast_set_pm(
    int value
)
//=============================================================================
{
    lcd_contrast_pm = value;
    gs_lcd_set_contrast( (u8)lcd_contrast_pm );
}

DECLARE_WRITE_PROC( lcd_contrast_pm, integer, lcd_contrast_set_pm );
DECLARE_READ_PROC( lcd_contrast_pm, integer );
DECLARE_FOPS( lcd_contrast_pm );
#endif

//=============================================================================
static int
gs_oem_override(
    struct file *file,
    const char __user *buffer,
    size_t count,
    loff_t *pos
)
//=============================================================================
{
    int i;

    for ( i = 0; i < count - 1; i++ )
    {
        if ( !( buffer[i] >= '0' && buffer[i] <= '9' ) )
        {
            printk("%s: value must be a positive integer\n", __func__);
            return -EINVAL;
        }
    }

    oem_override.value = (unsigned int)simple_strtoul( buffer, NULL, 10 );
    oem_override.enabled = 1;

    return count;
}


/*
 * ProcFS Entries
 */
struct gs_procfs_entry gs_procfs_entry[] =
{
#define GS_PROCFS_ENTRY( name ) { #name, &name##_fops, name##_init }
    GS_PROCFS_ENTRY( lcd_on_off ),      /* 1 = on; 0 = off */
    GS_PROCFS_ENTRY( mpk_on_off ),      /* 1 = on; 0 = off */
    GS_PROCFS_ENTRY( lcd_brightness ),
    GS_PROCFS_ENTRY( mpk_brightness ),
#ifdef CONFIG_FB_UC1628
    GS_PROCFS_ENTRY( lcd_contrast ),
    GS_PROCFS_ENTRY( lcd_contrast_pm ),
#endif
#undef GS_PROCFS_ENTRY
};
#define NUM_PROCFS_ENTRIES ( sizeof ( gs_procfs_entry ) / sizeof ( struct gs_procfs_entry ) )

struct gs_procfs_entry gs_devinfo_entry[] =
{
#define GS_DEVINFO_ENTRY( name ) { #name, NULL, NULL, 0 }
#define iPN 0
    GS_DEVINFO_ENTRY( PN ),
#define idev_alias 1
    GS_DEVINFO_ENTRY( dev_alias ),
#define idev_data 2
    GS_DEVINFO_ENTRY( dev_data ),
#define idev_id 3
    GS_DEVINFO_ENTRY( dev_id ),
#define idev_mac 4
    GS_DEVINFO_ENTRY( dev_mac ),
#define idev_rev 5
    GS_DEVINFO_ENTRY( dev_rev ),
#define irequired_bits 6
    GS_DEVINFO_ENTRY( required_bits ),
#define imodel_key 7
    GS_DEVINFO_ENTRY( model_key ),
#define idev_prefix 8
        GS_DEVINFO_ENTRY( dev_prefix ),
#undef GS_DEVINFO_ENTRY
};
#define NUM_DEVINFO_ENTRIES ( sizeof ( gs_devinfo_entry ) / sizeof ( struct gs_procfs_entry ) )

struct gs_procfs_entry hw_feature_entry[] =
{
#define HW_FEATURE_ENTRY( name, value ) { #name, NULL, NULL, value }
    HW_FEATURE_ENTRY( vendor, HW_FEATURE_VENDOR ),
    HW_FEATURE_ENTRY( model_str, HW_FEATURE_MODELSTR ),
    HW_FEATURE_ENTRY( model_alias, HW_FEATURE_MODELALIAS ),
    HW_FEATURE_ENTRY( default_fwkey, HW_FEATURE_FWKEY ),
    HW_FEATURE_ENTRY( oem_id, HW_FEATURE_OEM_ID ),
    HW_FEATURE_ENTRY( oem_base, HW_FEATURE_OEM_BASE  ),
    HW_FEATURE_ENTRY( gs_model_name, HW_FEATURE_GS_MODEL_NAME ),
    HW_FEATURE_ENTRY( vendor_fullname, HW_FEATURE_VENDOR_FULLNAME ),
    HW_FEATURE_ENTRY( eth, HW_FEATURE_ETH ),
    HW_FEATURE_ENTRY( eth_interface, HW_FEATURE_ETH_INTERFACE ),
    HW_FEATURE_ENTRY( handset, HW_FEATURE_HANDSET ),
    HW_FEATURE_ENTRY( headset_mic_detect, HW_FEATURE_HEADSET_MIC_DETECT ),
    HW_FEATURE_ENTRY( num_accts, HW_FEATURE_NUM_ACCTS ),
    HW_FEATURE_ENTRY( ext, HW_FEATURE_EXT ),
    HW_FEATURE_ENTRY( num_mpk, HW_FEATURE_NUM_MPK ),
    HW_FEATURE_ENTRY( cpu, HW_FEATURE_CPU ),
    HW_FEATURE_ENTRY( flash, HW_FEATURE_FLASH ),
    HW_FEATURE_ENTRY( flash_size, HW_FEATURE_FLASH_SIZE ),
    HW_FEATURE_ENTRY( oem_file_prefix, HW_FEATURE_OEM_FILE_PREFIX ),
    HW_FEATURE_ENTRY( poe_class, HW_FEATURE_POE_CLASS ),
    HW_FEATURE_ENTRY( gs_model, HW_FEATURE_MODEL_BASE ),
    HW_FEATURE_ENTRY( num_linekeys, HW_FEATURE_NUM_LINEKEYS ),
    HW_FEATURE_ENTRY( bluetooth, HW_FEATURE_BLUETOOTH ),
    HW_FEATURE_ENTRY( bluetooth_chip, HW_FEATURE_BLUETOOTH_CHIP ),
    HW_FEATURE_ENTRY( num_softkeys, HW_FEATURE_NUM_SOFTKEYS ),
    HW_FEATURE_ENTRY( no_usb, HW_FEATURE_NO_USB ),
    HW_FEATURE_ENTRY( hw_lcd_panel, HW_FEATURE_LCD_PANEL ),
    HW_FEATURE_ENTRY( require_pem, HW_FEATURE_REQUIRE_PEM ),
    HW_FEATURE_ENTRY( require_sn, HW_FEATURE_REQUIRE_SN ),
    HW_FEATURE_ENTRY( require_admin_pwd, HW_FEATURE_REQUIRE_ADMIN_PWD ),
    HW_FEATURE_ENTRY( require_xml_key, HW_FEATURE_REQUIRE_XML_KEY ),
    HW_FEATURE_ENTRY( require_hash, HW_FEATURE_REQUIRE_HASH ),
    HW_FEATURE_ENTRY( lcd_backlight_driver, HW_FEATURE_LCD_BACKLIGHT_DRIVER ),
    HW_FEATURE_ENTRY( second_panel, HW_FEATURE_LCD_SECOND_PANEL ),
    HW_FEATURE_ENTRY( hw_lcd_xres, HW_FEATURE_LCD_XRES ),
    HW_FEATURE_ENTRY( hw_lcd_yres, HW_FEATURE_LCD_YRES ),
    HW_FEATURE_ENTRY( phonebook_key, HW_FEATURE_PHONEBOOK_KEY ),
#ifdef CONFIG_WIRELESS
    HW_FEATURE_ENTRY( wifi_chip, HW_FEATURE_WIFI_CHIP ),
    HW_FEATURE_ENTRY( wifi_country_code, HW_FEATURE_WIFI_COUNTRY_CODE ),
#endif
    HW_FEATURE_ENTRY( console_uart, HW_FEATURE_CONSOLE_UART ),
};
#define NUM_HW_FEATURE_ENTRIES ( sizeof ( hw_feature_entry ) / sizeof ( struct gs_procfs_entry ) )

struct gs_procfs_entry sec_feature_entry[] =
{
#define SEC_FEATURE_ENTRY( name, value ) { #name, NULL, NULL, value }
    SEC_FEATURE_ENTRY( private_key, SEC_FEATURE_PRIVATE_KEY ),
    SEC_FEATURE_ENTRY( certificate, SEC_FEATURE_CERTIFICATE ),
    SEC_FEATURE_ENTRY( checksum, SEC_FEATURE_CHECKSUM ),
    SEC_FEATURE_ENTRY( sn, SEC_FEATURE_SN ),
    SEC_FEATURE_ENTRY( password, SEC_FEATURE_PASSWORD ),
    SEC_FEATURE_ENTRY( xml_key, SEC_FEATURE_XML_KEY ),
    SEC_FEATURE_ENTRY( hash, SEC_FEATURE_HASH ),
#ifdef CONFIG_WIRELESS
    SEC_FEATURE_ENTRY( country_code, SEC_FEATURE_COUNTRY_CODE ),
#endif
};
#define NUM_SEC_FEATURE_ENTRIES ( sizeof ( sec_feature_entry ) / sizeof ( struct gs_procfs_entry ) )

//=============================================================================
hw_info_t *
hw_feature_get(
    void
)
//=============================================================================
{
    static hw_info_t * hw_info = NULL;

    if ( hw_info == NULL )
    {
        GS_FACTORYPARA * factory_para;

        factory_para = fact_para_get();
        if ( factory_para->checksum != fact_para_checksum( factory_para ) )
        {
            printk("INVALID FACTORY CHECKSUM\n");
        }

        hw_info = hw_get_hw_feature();
    }

    return hw_info;
}

#ifdef GS_MANIFEST_PROC
//=============================================================================
static int manifest_proc_read(
    struct seq_file *seq,
    void *filp
)
//=============================================================================
{
    struct toml_object *obj = (struct toml_object*)(get_proc_data(seq->private));

    if ( !obj ) {
        printk("error - %s - proc data is null\n", __func__);
        return -EINVAL;
    }

    if ( obj->type != TOML_TYPE_KV ) {
        printk("error - %s - proc data type is not kv\n", __func__);
        return -EINVAL;
    }
    seq_printf( seq, "%s", obj->u.toml_kv.value );

    return 0;
}
#endif

//=============================================================================
static int hw_feature_proc_read(
    struct seq_file *seq,
    void *filp
)
//=============================================================================
{
    int len = 0;
    char str[128];
    hw_info_t * hw_info = hw_feature_get();
    int data = (int)(get_proc_data(seq->private));

    switch ( data )
    {
        case HW_FEATURE_VENDOR:
            strcpy( str, hw_info->vendor );
            break;
        case HW_FEATURE_VENDOR_FULLNAME:
            strcpy( str, hw_info->vendor_fullname );
            break;
        case HW_FEATURE_MODELSTR:
            strcpy( str, hw_info->modelstr );
            break;
        case HW_FEATURE_MODELALIAS:
            strcpy( str, hw_info->modelalias );
            break;
        case HW_FEATURE_OEM_ID:
            if ( oem_override.enabled )
            {
                sprintf( str, "%d", oem_override.value );
            }
            else
            {
                /* SCF - %x is correct to produce the correct
                 * value in base 10. The OEM ID in the factory
                 * tools is incorrectly specified as a base 16 field
                 * which people have been filling in assuming it was
                 * base 10. */
                sprintf( str, "%x", hw_info->oem_id );
            }
            break;
        case HW_FEATURE_FWKEY:
            if ( hw_info->fwkey[0] == 0 )
            {
                strcpy( str, "nokey" );
            }
            else
            {
                strcpy( str, hw_info->fwkey );
            }
            break;
        case HW_FEATURE_ETH:
            switch ( hw_info->eth ) {
                case ETH_TYPE_KSZ8873:
                    sprintf( str, "KSZ8873" );
                    break;
                case ETH_TYPE_KSZ8863:
                    sprintf( str, "KSZ8863" );
                    break;
                case ETH_TYPE_ICPLUS:
                    sprintf( str, "IC+" );
                    break;
                case ETH_TYPE_KSZ8051:
                    sprintf( str, "KSZ8051" );
                    break;
                case ETH_TYPE_KSZ8031:
                    sprintf( str, "KSZ8031" );
                    break;
                case ETH_TYPE_RTL8363SB:
                    sprintf( str, "RTL8363SB" );
                    break;
                case ETH_TYPE_QCA8334:
                    sprintf( str, "QCA8334" );
                    break;
                case ETH_TYPE_RTL8304:
                    sprintf( str, "RTL8304" );
                    break;
                default:
                    sprintf( str, "Unknown" );
                    break;
            }
            break;
        case HW_FEATURE_ETH_INTERFACE:
            switch ( hw_info->eth_interface ) {
                case ETH_IF_RGMII:
                    sprintf( str, "RGMII" );
                    break;
                case ETH_IF_RMII:
                    sprintf( str, "RMII" );
                    break;
                default:
                    sprintf( str, "Unknown" );
                    break;
            }
            break;
        case HW_FEATURE_HANDSET:
            switch ( hw_info->handset ) {
            case HANDSET_WIDEBAND:
                sprintf( str, "wideband" );
                break;
            case HANDSET_NARROWBAND:
                sprintf( str, "narrowband" );
                break;
            }
            break;
        case HW_FEATURE_HEADSET_MIC_DETECT:
            sprintf( str, "%d", hw_info->headset_mic_detect );
            break;
        case HW_FEATURE_NUM_ACCTS:
            sprintf( str, "%d", hw_info->num_accts );
            break;
        case HW_FEATURE_EXT:
            sprintf( str, "%d", hw_info->ext );
            break;
        case HW_FEATURE_NUM_MPK:
            sprintf( str, "%d", hw_info->num_mpk );
            break;
        case HW_FEATURE_CPU:
            strcpy( str, hw_info->cpu );
            break;
        case HW_FEATURE_FLASH:
            switch ( hw_info->flash_device )
            {
                case FLASH_K9F1G08U0D:
                    sprintf( str, "K9F1G08U0D" );
                    break;
                case FLASH_K9F1G08U0E:
                    sprintf( str, "K9F1G08U0E" );
                    break;
                case FLASH_F59L1G81LA:
                    sprintf( str, "F59L1G81LA" );
                    break;
                case FLASH_GD5F1GQ4U:
                    sprintf( str, "GD5F1GQ4U" );
                    break;
                default:
                    sprintf( str, "Unknown" );
                    break;
            }
            break;
        case HW_FEATURE_FLASH_SIZE:
            strcpy( str, hw_info->flash_size );
            break;
        case HW_FEATURE_BLUETOOTH:
            sprintf( str, "%d", hw_info->bluetooth);
            break;
        case HW_FEATURE_BLUETOOTH_CHIP:
            switch ( hw_info->bluetooth_chip )
            {
                case BT_RDA5876A:
                    sprintf( str, "RDA5876A" );
                    break;
                case BT_AP6256:
                    sprintf( str, "AP6256" );
                    break;
                default:
                    sprintf( str, "Unknown" );
                    break;
            }
            break;
        case HW_FEATURE_OEM_FILE_PREFIX:
            strcpy( str, hw_info->oem_file_prefix );
            break;
        case HW_FEATURE_POE_CLASS:
            sprintf( str, "%d", hw_info->poe_class );
            break;
        case HW_FEATURE_MODEL_BASE:
            if ( strlen( hw_info->gs_model ) )
                strcpy( str, hw_info->gs_model );
            else
                strcpy( str, hw_info->modelstr );
            break;
        case HW_FEATURE_OEM_BASE:
            strcpy( str, hw_info->oem_base );
            break;
        case HW_FEATURE_GS_MODEL_NAME:
            strcpy( str, hw_info->gs_model_name );
            break;
        case HW_FEATURE_NUM_LINEKEYS:
            sprintf( str, "%d", hw_info->num_linekeys );
            break;
        case HW_FEATURE_NUM_SOFTKEYS:
            sprintf( str, "%d", hw_info->num_softkeys );
            break;
        case HW_FEATURE_NO_USB:
            sprintf( str, "%d", hw_info->usb_unsupported );
            break;
        case HW_FEATURE_LCD_PANEL:
            sprintf( str, "%d", hw_info->hw_lcd_panel );
            break;
        case HW_FEATURE_REQUIRE_PEM:
            sprintf( str, "%d", hw_info->require_pem );
            break;
        case HW_FEATURE_REQUIRE_SN:
            sprintf( str, "%d", hw_info->require_sn );
            break;
        case HW_FEATURE_REQUIRE_ADMIN_PWD:
            sprintf( str, "%d", hw_info->require_admin_pwd );
            break;
        case HW_FEATURE_REQUIRE_XML_KEY:
            sprintf( str, "%d", hw_info->require_xml_key );
            break;
        case HW_FEATURE_REQUIRE_HASH:
            sprintf( str, "%d", hw_info->require_hash );
            break;
        case HW_FEATURE_PHONEBOOK_KEY:
            sprintf( str, "%d", hw_info->phonebook_key );
            break;
        case HW_FEATURE_LCD_BACKLIGHT_DRIVER:
            switch ( hw_info->lcd_backlight_driver )
            {
                case LCD_BL_MP1488:
                    sprintf( str, "MP1488" );
                    break;
                case LCD_BL_AP3031:
                    sprintf( str, "AP3031" );
                    break;
                case LCD_BL_MOSFET:
                    sprintf( str, "MOSFET" );
                    break;
                case LCD_BL_ST7789S:
                    sprintf( str, "ST7789S" );
                    break;
                case LCD_BL_TT43WQ:
                    sprintf( str, "TT43WQ" );
                    break;
                default:
                    sprintf( str, "Unknown" );
                    break;
            }
            break;
        case HW_FEATURE_LCD_SECOND_PANEL:
            sprintf( str, "%d", hw_info->second_panel );
            break;
        case HW_FEATURE_LCD_XRES:
            if ( strlen( hw_info->hw_lcd_xres ) )
                strcpy( str, hw_info->hw_lcd_xres );
            break;
        case HW_FEATURE_LCD_YRES:
            if ( strlen( hw_info->hw_lcd_yres ) )
                strcpy( str, hw_info->hw_lcd_yres );
            break;
#ifdef CONFIG_WIRELESS
        case HW_FEATURE_WIFI_CHIP:
            switch ( hw_info->wifi_chip ) {
                case WIFI_NONE:
                    sprintf( str, "None" );
                    break;
                case WIFI_AP6330:
                    sprintf( str, "AP6330" );
                    break;
                case WIFI_AP6255:
                    sprintf( str, "AP6255" );
                    break;
                case WIFI_AP6212A:
                    sprintf( str, "AP6212A" );
                    break;
                case WIFI_AP6256:
                    sprintf( str, "AP6256" );
                    break;
                default:
                    sprintf( str, "Unknown" );
                    break;
            }
            break;
        case  HW_FEATURE_WIFI_COUNTRY_CODE:
            strcpy( str, hw_info->wifi_country_code );
            break;
#endif
        case HW_FEATURE_CONSOLE_UART:
            sprintf( str, "%d", hw_info->console_uart );
            break;
    }

    len = strlen( str );
    seq_printf( seq, "%s", str );

    return 0;
}

//=============================================================================
static int sec_feature_proc_read(
    struct seq_file *seq,
    void *filp
)
//=============================================================================
{
    int len = 0;
    sec_info_t * sec_info = sec_info_get();
    int data = (int)(get_proc_data(seq->private));

    switch ( data )
    {
        case SEC_FEATURE_SN:
            if ( sec_info->sn.len == 0 )
            {
                // not set
                break;
            }
            len = strlen( sec_info->sn.data );
            if( len > sec_info->sn.len )
                printk( "sn length is greater than the stored length\n" );
            seq_printf( seq, "%s", sec_info->sn.data );
            break;
        case SEC_FEATURE_PRIVATE_KEY:
            if ( sec_info->private_key.len == 0 )
            {
                // not set
                break;
            }
            len = strlen( sec_info->private_key.data );
            if( len > sec_info->private_key.len )
                printk( "private key length is greater than the stored length\n" );
            seq_printf( seq, "%s", sec_info->private_key.data );
            break;
        case SEC_FEATURE_CERTIFICATE:
            if ( sec_info->certificate.len == 0 )
            {
                // not set
                break;
            }
            len = strlen( sec_info->certificate.data );
            if( len > sec_info->certificate.len )
                printk( "certificate length is greater than the stored length\n" );
            seq_printf( seq, "%s", sec_info->certificate.data );
            break;
        case SEC_FEATURE_CHECKSUM:
            if ( sec_info->checksum.len == 0 )
            {
                // not set
                break;
            }
            len = strlen( sec_info->checksum.data );
            if( len > sec_info->checksum.len )
                printk( "checksum length is greater than the stored length\n" );
            seq_printf( seq, "%s", sec_info->checksum.data );
            break;
        case SEC_FEATURE_HASH:
            if ( sec_info->hash.len == 0 )
            {
                // not set
                break;
            }
            len = strlen( sec_info->hash.data );
            if( len > sec_info->hash.len )
                printk( "hash length is greater than the stored length\n" );
            seq_printf( seq, "%s", sec_info->hash.data );
            break;
#ifdef CONFIG_WIRELESS
        case SEC_FEATURE_COUNTRY_CODE:
            if ( sec_info->country_code.len == 0 )
            {
                // not set
                break;
            }
            len = strlen( sec_info->country_code.data );
            if( len > sec_info->country_code.len )
                printk( "country_code length is greater than the stored length\n" );
            seq_printf( seq, "%s", sec_info->country_code.data );
            break;
#endif
        case SEC_FEATURE_PASSWORD:
            if ( sec_info->password.len == 0)
            {
                // not set
                break;
            }
            len = strlen( sec_info->password.data );
            if( len > sec_info->password.len )
                printk( "password length is greater than the stored length\n" );
            seq_printf( seq, "%s", sec_info->password.data );
            break;
        case SEC_FEATURE_XML_KEY:
            if ( sec_info->xml_key.len == 0)
            {
                // not set
                break;
            }
            len = strlen( sec_info->xml_key.data );
            if( len > sec_info->xml_key.len )
                printk( "xml key length is greater than the stored length\n" );
            seq_printf( seq, "%s", sec_info->xml_key.data );
            break;
        default:
            return -EINVAL;
    }

    return 0;
}

//=============================================================================
static int devinfo_proc_read(
    struct seq_file *seq,
    void *filp
)
//=============================================================================
{
    int len = 0;
    int checksum_valid = 1;
    char * nickstr = HW_UNKNOWN;
    char * default_mac = "00:0B:82:FF:FF:FF";
    char * pn;
    char mac_str[64];
    unsigned short int required_bits[4];
    GS_FACTORYPARA * factory_para;
    hw_info_t * hw_info;
    int data = (int)(get_proc_data(seq->private));

    factory_para = fact_para_get();

    if ( factory_para->checksum != fact_para_checksum( factory_para ) )
    {
        printk("INVALID CHECKSUM\n");
        checksum_valid = 0;
    }

    hw_info = hw_feature_get();

    switch ( data )
    {
        case imodel_key:
            sprintf( mac_str, "%04x", factory_para->hw_model );
            seq_printf( seq, "%s", mac_str );
            break;
        case idev_prefix:
            printk("WARNING -- DEPRECATED PROC ENTRY PREFIX ACCESSED\n");
            strcpy( mac_str, hw_info->modelstr );
            seq_printf( seq, "%s", mac_str );
            break;
        case irequired_bits:
            if ( checksum_valid )
            {
                required_bits[0] = factory_para->RequiredBits[0];
                required_bits[1] = factory_para->RequiredBits[1];
                required_bits[2] = factory_para->RequiredBits[2];
                required_bits[3] = factory_para->RequiredBits[3];
                /* This gets byte swapped in flash by hexpnxbin, we want all instances
                 * of required bits to look the same for convenience */
                required_bits[0] = ( ( required_bits[0] & 0x00FF ) << 8 ) | ( ( required_bits[0] & 0xFF00 ) >> 8 );
                required_bits[1] = ( ( required_bits[1] & 0x00FF ) << 8 ) | ( ( required_bits[1] & 0xFF00 ) >> 8 );
                required_bits[2] = ( ( required_bits[2] & 0x00FF ) << 8 ) | ( ( required_bits[2] & 0xFF00 ) >> 8 );
                required_bits[3] = ( ( required_bits[3] & 0x00FF ) << 8 ) | ( ( required_bits[3] & 0xFF00 ) >> 8 );
            }
            else
            {
                memset( required_bits, 0, sizeof ( required_bits ) );
            }
            sprintf( mac_str, "%04x%04x%04x%04x",
                required_bits[0],
                required_bits[1],
                required_bits[2],
                required_bits[3] );
            seq_printf( seq, "%s", mac_str );
            break;
        case iPN:
            if ( checksum_valid )
            {
                pn = (char *)factory_para->PartNumber;
                seq_printf( seq, "%s", pn );
            }
            else
            {
                seq_printf( seq, "%s", HW_UNKNOWN );
            }
            break;

        case idev_alias:
            nickstr = hw_info->modelalias;
            seq_printf( seq, "%s", nickstr );
            break;

        case idev_id:
            nickstr = hw_info->modelstr;
            seq_printf( seq, "%s", nickstr );
            break;

        case idev_data:
            len = ( ( factory_para->length[1] << 16 ) | factory_para->length[0] );
            if ( len > sizeof ( GS_FACTORYPARA ) )
                len = sizeof ( GS_FACTORYPARA ); /* How did we pass checksum if this is true? */
            if ( checksum_valid )
                seq_write( seq, factory_para, len );
            else
                seq_write( seq, 0, sizeof ( GS_FACTORYPARA ) );
            break;

        case idev_mac:
            if ( checksum_valid )
            {
                sprintf( mac_str, "%02x:%02x:%02x:%02x:%02x:%02x",
                    factory_para->mac[0] & 0xFF, (factory_para->mac[0] >> 8) & 0xFF,
                    factory_para->mac[1] & 0xFF, (factory_para->mac[1] >> 8) & 0xFF,
                    factory_para->mac[2] & 0xFF, (factory_para->mac[2] >> 8) & 0xFF );
                seq_printf( seq, "%s", mac_str );
            }
            else
            {
                seq_printf( seq, "%s", default_mac );
            }
            break;
        case idev_rev:
            if ( checksum_valid )
            {
                pn = (char *)factory_para->PartNumber;
                sprintf(mac_str, "V%c.%c%c", pn[8], pn[9], pn[10]);
                seq_printf( seq, "%s", mac_str );
            }
            else
            {
                seq_printf( seq, "%s", HW_UNKNOWN );
            }
            break;
        default:
            return -EINVAL;
    }

    return 0;
}

//=============================================================================
static int
devinfo_proc_write (
    struct file *file,
    const char __user *buffer,
    size_t count,
    loff_t *offset
)
//=============================================================================
{
    GS_FACTORYPARA * factory_para;
    unsigned short mac[3];
    struct mtd_info* mtd;
    int len;
    uint8_t* factory_buf;

    if ( count != sizeof ( mac ) )
    {
        printk("devinfo_proc_write: invalid size\n");
        return -EINVAL;
    }

    if ( copy_from_user( mac, buffer, 6 ) )
    {
        printk("devinfo_proc_write: copy error\n");
        return -EFAULT;
    }

    factory_para = fact_para_get();
    if ( factory_para->checksum != fact_para_checksum( factory_para ) )
    {
        printk("devinfo_proc_write: checksum fail\n");
        return -EINVAL;
    }

    if ( !( factory_para->mac[0] == 0x0b00 &&
            factory_para->mac[1] == 0x0082 &&
            factory_para->mac[2] == 0x0000 ) &&
         !( factory_para->mac[0] == 0x74c0 &&
            factory_para->mac[1] == 0x00ad &&
            factory_para->mac[2] == 0x0000 ) )
    {
        printk("devinfo_proc_write: mac already set\n");
        return -EFAULT;
    }

    mtd = get_mtd_device_nm( "factory" );
    if ( mtd <= 0 )
    {
        printk("devinfo_proc_write: can't get mtd device\n");
        return -EFAULT;
    }

    factory_buf = vmalloc( mtd->size );
    if ( factory_buf == NULL )
    {
        printk("devinfo_proc_write: can't allocate buffer\n");
        put_mtd_device( mtd );
        return -EFAULT;
    }

    printk("devinfo_proc_write: setting MAC %04x:%04x:%04x\n", mac[0], mac[1], mac[2]);

    factory_para->mac[0] = mac[0];
    factory_para->mac[1] = mac[1];
    factory_para->mac[2] = mac[2];

    factory_para->checksum = fact_para_checksum( factory_para );

    mtd_read( mtd, 0, mtd->size, &len, factory_buf );
    if ( len != mtd->size )
    {
        printk("devinfo_proc_write: error reading factory part\n");
        put_mtd_device( mtd );
        vfree( factory_buf );
        return -EFAULT;
    }

    memcpy( factory_buf + GS_FACTORY_CFG_OFFSET, factory_para, GS_FACTORY_CFG_SIZE_FLASH );
    if ( erase_helper( mtd, 0, mtd->size ) < 0 )
    {
        printk("devinfo_proc_write: error erasing factory part\n");
        put_mtd_device( mtd );
        vfree( factory_buf );
        return -EFAULT;
    }

    mtd_write( mtd, 0, mtd->size, &len, factory_buf );
    if ( len != mtd->size )
    {
        printk("devinfo_proc_write: error writing factory part\n");
        put_mtd_device( mtd );
        vfree( factory_buf );
        return -EFAULT;
    }

    set_factory_para_dirty();
    vfree( factory_buf );
    put_mtd_device( mtd );

    return count;
}

//=============================================================================
static int production_proc_read(
    struct seq_file *seq,
    void *filp
)
//=============================================================================
{
    int checksum_valid = 1;
    GS_FACTORYPARA * factory_para;
 
    factory_para = fact_para_get();

    if ( factory_para->checksum != fact_para_checksum( factory_para ) )
    {
        printk("INVALID CHECKSUM\n");
        checksum_valid = 0;
    }

    if ( checksum_valid )
    {
        seq_printf( seq, "%d", factory_para->hw_production );
    }
    else
    {
        seq_printf( seq, "%s", HW_UNKNOWN );
    }

    return 0;
}

#ifdef SECURE_BOOT_OTP_ROM_DEBUG
//=============================================================================
static int otpdump_proc_read(
    struct seq_file *seq,
    void *filp
)
//=============================================================================
{
    dspg_crypto_hw_otp_dump();
    return 0;
}
#endif

//=============================================================================
static int
production_lock(
    struct file       *file,
    const char __user *buffer,
    size_t            count,
    loff_t            *pos
)
//=============================================================================
{
#define PRODUCTION_LOCK     "SetProductionUnit!"
    char lock[32] = { '\0' };
    GS_FACTORYPARA * factory_para = NULL;
    struct mtd_info* mtd;
    int len;
    uint8_t* factory_buf;
    factory_param_sector_t *fact_para;
    char * ptr_sec_table = NULL;
    int sec_table_size = 0;
    hw_feature_table_header_t *hwTableHdr;
    unsigned int gs_factory_cfg_size_flash = GS_FACTORY_CFG_SIZE_FLASH;
    //int ret;

    len = count > 32 ? 32 : count;
    if ( copy_from_user( lock, buffer, len  ) )
    {
        printk("Error in copy_from_user\n");
        return -EINVAL;
    }
    // Last char is LF, so replace it with a NULL char.
    lock[len-1] = '\0';
    // printk("%s: count=%d lock=%s\n", __func__, count, lock);
    if ( strcmp(lock, PRODUCTION_LOCK) != 0 )
    {
        printk("Error: Invalid operation!\n");
        return -EINVAL;
    }

    /*Invalidate the cached fact cfg table to make sure we have correct fact cfg values */
    set_factory_para_dirty();

    factory_para = fact_para_get();
    if ( factory_para->checksum != fact_para_checksum( factory_para ) )
    {
        printk("Error: fact_para_checksum fails.\n");
        return -EINVAL;
    }

    if ( factory_para->hw_production == 1 )
    {
        printk("Error: Production is already locked on this device.\n");
        return -EINVAL;
    }

    // Secure device by burning OTP ROM.
    /*ret = dspg_crypto_hw_secure_device();
    if ( ret != 0 )
    {
        printk("Error: OTP ROM burning fails %d\n", ret);
        return ret;
    }*/

    /* Locate security table */
    fact_para = (factory_param_sector_t *) factory_para;
    hwTableHdr = &fact_para->hw_feature_table_header;
    while ( hwTableHdr->hw_feature_table_version != HW_SECURITY_TABLE_VERSION )
    {
        hwTableHdr = (hw_feature_table_header_t*) ( (uint16_t*) hwTableHdr + hwTableHdr->hw_feature_table_len );
        if ( (uint16_t*) hwTableHdr - (uint16_t*) fact_para >= fact_para->factory_para.length[0] ||                                    hwTableHdr->hw_feature_table_version == 0 )
        {
            hwTableHdr = NULL;
            break;
        }
    }
    if ( hwTableHdr == NULL )
    {
        printk("Error: Can't find security table.\n");
        return -EINVAL;
    }

    sec_table_size = sizeof(uint16_t) * hwTableHdr->hw_feature_table_len - sizeof(hw_feature_table_header_t);
    ptr_sec_table = (char *) hwTableHdr + sizeof(hw_feature_table_header_t);

#ifdef DEBUG
    print_buf("**** Security table dump (plain)", ptr_sec_table, 128);
#endif

    printk("Encrypting security table...\n");
    if ( gs_sec_fact_mod_cipher( GS_SEC_METHOD_ENCRYPT, ptr_sec_table, sec_table_size ) != 0 )
    {
        printk("Error: security table encryption failed.\n");
        return -EFAULT;
    }
    printk("Encrypting security table...done!\n");

#ifdef DEBUG
    print_buf("**** Security table dump (encrypted)", ptr_sec_table, 128);

    printk("Decrypting security table...\n");
    if ( gs_sec_fact_mod_cipher( GS_SEC_METHOD_DECRYPT, ptr_sec_table, sec_table_size ) != 0 )
    {
        printk("Error: security table decryption failed.\n");
        return -EFAULT;
    }
    print_buf("**** Security table dump (decrypted)", ptr_sec_table, 128);
#endif

    factory_para->hw_production = 1;
    factory_para->checksum = fact_para_checksum( factory_para );
    set_factory_para_dirty();

    /* Reprogram factory cfg partition in Flash */
    mtd = get_mtd_device_nm( "factory" );
    if ( mtd <= 0 )
    {
        printk("Can't get mtd device.\n");
        return -EFAULT;
    }
#if defined( CONFIG_MTD_NAND )
    gs_factory_cfg_size_flash = mtd->size;
#else
    gs_factory_cfg_size_flash = GS_FACTORY_CFG_SIZE_FLASH;
#endif
    factory_buf = vmalloc( gs_factory_cfg_size_flash );
    if ( factory_buf == NULL )
    {
        printk("Error: Can't allocate buffer.\n");
        put_mtd_device( mtd );
        return -EFAULT;
    }
    mtd_read( mtd, 0, gs_factory_cfg_size_flash, &len, factory_buf );
    if ( len != gs_factory_cfg_size_flash )
    {
        printk("Error in reading factory partition.\n");
        put_mtd_device( mtd );
        vfree( factory_buf );
        return -EFAULT;
    }
    if ( erase_helper( mtd, 0, gs_factory_cfg_size_flash ) < 0 )
    {
        printk("Error in erasing factory partition.\n");
        put_mtd_device( mtd );
        vfree( factory_buf );
        return -EFAULT;
    }
    memcpy( factory_buf + GS_FACTORY_CFG_OFFSET, factory_para, GS_FACTORY_CFG_SIZE );
    mtd_write( mtd, 0, gs_factory_cfg_size_flash, &len, factory_buf );
    if ( len != gs_factory_cfg_size_flash )
    {
        printk("Error in writing factory partition.\n");
        put_mtd_device( mtd );
        vfree( factory_buf );
        return -EFAULT;
    }

    set_factory_para_dirty();
    put_mtd_device( mtd );
    vfree( factory_buf );

    factory_para = fact_para_get();
    printk("hw_production=%d\n", factory_para->hw_production);

    // Force NVRAM commit and hence the NVRAM encryption.
    //printk("Force NVRAM encryption\n");
    //nvram_trigger_encrypt();

    printk("Production lock is successful!\n");
    return count;
}

//=============================================================================
static int
cfg_reload_write(
    struct file *file,
    const char __user *buffer,
    size_t count,
    loff_t *pos
)
//=============================================================================
{
    // use this to trigger factory reload after writing security table by factory tool
    set_factory_para_dirty();
    return count;
}

// For security table item certificate and private key,
// their length may exceed one page, use seq_file

//=============================================================================
static void *
private_key_seq_start(
        struct seq_file *s,
        loff_t *pos
)
//=============================================================================
{
    sec_info_t * sec_info = sec_info_get();

    if ( sec_info == NULL ||
         sec_info->private_key.len == 0 ||
         *pos > 0 )
        return NULL;
    else
        return (void *)( sec_info->private_key.data );

}

//=============================================================================
static void *
private_key_seq_next(
        struct seq_file *s,
        void *v,
        loff_t *pos
)
//=============================================================================
{
    (*pos)++;
    return private_key_seq_start(s, pos);
}

//=============================================================================
static void
private_key_seq_stop(
        struct seq_file *s,
        void *v
)
//=============================================================================
{
    return;
}

//=============================================================================
static int private_key_seq_show(
        struct seq_file *s,
        void *v
)
//=============================================================================
{
    seq_printf(s, "%s", (char *) v);

    return 0;
}

static struct seq_operations private_key_seq_ops={
    .start = private_key_seq_start,
    .next = private_key_seq_next,
    .stop = private_key_seq_stop,
    .show = private_key_seq_show
};

//=============================================================================
static int
private_key_proc_open(
        struct inode *inode,
        struct file *file
)
//=============================================================================
{
    return seq_open(file, &private_key_seq_ops);
}

static struct file_operations sec_private_key_fops = {
    .owner = THIS_MODULE,
    .open = private_key_proc_open,
    .read = seq_read,
    .llseek = seq_lseek,
    .release = seq_release
};

//=============================================================================
static void *
certificate_seq_start(
        struct seq_file *s,
        loff_t *pos
)
//=============================================================================
{
    sec_info_t * sec_info = sec_info_get();

    if ( sec_info == NULL ||
         sec_info->certificate.len == 0 ||
         *pos > 0 )
        return NULL;
    else
        return (void *)( sec_info->certificate.data );

}

//=============================================================================
static void *
certificate_seq_next(
        struct seq_file *s,
        void *v,
        loff_t *pos
)
//=============================================================================
{
    (*pos)++;
    return certificate_seq_start(s, pos);
}

//=============================================================================
static void
certificate_seq_stop(
        struct seq_file *s,
        void *v
)
//=============================================================================
{
    return;
}

//=============================================================================
static int certificate_seq_show(
        struct seq_file *s,
        void *v
)
//=============================================================================
{
    seq_printf(s, "%s", (char *) v);
    return 0;
}

static struct seq_operations certificate_seq_ops = {
    .start = certificate_seq_start,
    .next = certificate_seq_next,
    .stop = certificate_seq_stop,
    .show = certificate_seq_show
};

//=============================================================================
static int
certificate_proc_open(
        struct inode *inode,
        struct file *file
)
//=============================================================================
{
    return seq_open(file, &certificate_seq_ops);
}

//=================
// File Operations
//=================
//=============================================================================
static int
devinfo_proc_open(
    struct inode *inode,
    struct file *filp
)
//=============================================================================
{
    return single_open( filp, devinfo_proc_read, filp );
}

#ifdef GS_MANIFEST_PROC
//=============================================================================
static int
manifest_proc_open(
    struct inode *inode,
    struct file *filp
)
//=============================================================================
{
    return single_open( filp, manifest_proc_read, filp );
}
#endif

//=============================================================================
static int
hw_feature_proc_open(
    struct inode *inode,
    struct file *filp
)
//=============================================================================
{
    return single_open( filp, hw_feature_proc_read, filp );
}

//=============================================================================
static int
sec_feature_proc_open(
    struct inode *inode,
    struct file *filp
)
//=============================================================================
{
    return single_open( filp, sec_feature_proc_read, filp );
}

//=============================================================================
static int
production_proc_open(
    struct inode *inode,
    struct file *filp
)
//=============================================================================
{
    return single_open( filp, production_proc_read, filp );
}

#ifdef SECURE_BOOT_OTP_ROM_DEBUG
//=============================================================================
static int
otpdump_proc_open(
    struct inode *inode,
    struct file *filp
)
//=============================================================================
{
    return single_open( filp, otpdump_proc_read, filp );
}
#endif

static struct file_operations sec_certificate_fops = {
    .owner = THIS_MODULE,
    .open = certificate_proc_open,
    .read = seq_read,
    .llseek = seq_lseek,
    .release = seq_release
};

static struct file_operations gs_devinfo_fops = {
    .owner = THIS_MODULE,
    .open = devinfo_proc_open,
    .read = seq_read,
    .write = devinfo_proc_write,
    .release = single_release
};

static struct file_operations gs_reload_fops = {
    .owner = THIS_MODULE,
    .write = cfg_reload_write
};

static struct file_operations gs_production_fops = {
    .owner = THIS_MODULE,
    .open = production_proc_open,
    .read = seq_read,
    .write = production_lock,
    .release = single_release
};

#ifdef SECURE_BOOT_OTP_ROM_DEBUG
static struct file_operations gs_otpdump_fops = {
    .owner = THIS_MODULE,
    .open = otpdump_proc_open,
    .read = seq_read,
    .release = single_release
};
#endif

static struct file_operations gs_hw_feature_fops = {
    .owner = THIS_MODULE,
    .open = hw_feature_proc_open,
    .read = seq_read,
    .release = single_release
};

#ifdef GS_MANIFEST_PROC
static struct file_operations gs_manifest_fops = {
    .owner = THIS_MODULE,
    .open = manifest_proc_open,
    .read = seq_read,
    .release = single_release,
};
#endif

static struct file_operations sec_feature_fops = {
    .owner = THIS_MODULE,
    .open = sec_feature_proc_open,
    .read = seq_read,
    .release = single_release
};

static struct file_operations gs_oem_fops = {
    .owner = THIS_MODULE,
    .write = gs_oem_override,
};
    
#ifdef GS_MANIFEST_PROC
//=============================================================================
static void
manifest_build_proc(
    struct toml_object *obj,
    struct proc_dir_entry *dir
)
//=============================================================================
{
    struct toml_object *o = obj->children;

    while ( o ) {
        if ( o->type == TOML_TYPE_TABLE ) {
            printk("toml parse table %s (array=%d)\n", o->u.toml_table.tag, o->array);
        }

        if ( o->type == TOML_TYPE_TABLE && !o->array ) {
            struct proc_dir_entry *d;

            if ( !o->u.toml_table.tag ) {
                printk("error - missing table tag\n");
                return;
            }
            d = proc_mkdir( o->u.toml_table.tag, dir );
            if ( !d ) {
                printk("error - unable to create table entry %s\n", o->u.toml_table.tag );
            }

            manifest_build_proc( o, d );
        }
        else if ( o->type == TOML_TYPE_KV ) {
            if ( !o->u.toml_kv.key ) {
                printk("error - toml kv missing key\n");
                return;
            }

            if ( !proc_create_data( o->u.toml_kv.key,
                                    S_IFREG | S_IRUGO, 
                                    dir, 
                                    &gs_manifest_fops,
                                    (int*)o ) ) {
                printk("error - unable to create proc entry for toml kv %s\n", o->u.toml_kv.key);
            }
        }
        else if ( o->type == TOML_TYPE_TABLE && o->array ) {
            if ( strcmp( o->u.toml_table.tag, "image" ) == 0 ) {
                struct toml_object *name = NULL;
                struct toml_object *type = NULL;
                struct toml_object *j = o->children;

                while ( j ) {
                    if ( j->type == TOML_TYPE_KV && j->u.toml_kv.key && j->u.toml_kv.value ) {
                        if ( strcmp( j->u.toml_kv.key, "name" ) == 0 ) {
                            name = j;
                        }
                        else if ( strcmp( j->u.toml_kv.key, "type" ) == 0 ) {
                            type = j;
                        }
                    }
                    j = j->next;
                }

                if ( name && type ) {
                    struct proc_dir_entry *d = proc_mkdir( name->u.toml_kv.value, dir );
                    if ( !d ) {
                        printk("error - unable to create array entry %s\n", name->u.toml_kv.value);
                    }
                    else {
                        if ( !proc_create_data( type->u.toml_kv.key,
                                                S_IFREG | S_IRUGO,
                                                d,
                                                &gs_manifest_fops,
                                                (int*)type ) ) {
                            printk("error - unable to create proc entry for toml kv %s", type->u.toml_kv.key);
                        }
                    }
                }
            }
            else if ( strcmp( o->u.toml_table.tag, "compatible" ) == 0 ) {
                struct toml_object *gs_model = NULL;
                struct toml_object *j = o->children;

                printk("toml parse compatible array\n");

                while ( j ) {
                    if ( j->type == TOML_TYPE_KV && j->u.toml_kv.key && j->u.toml_kv.value ) {
                        if ( strcmp( j->u.toml_kv.key, "gsmodel" ) == 0 ) {
                            gs_model = j;
                        }
                    }
                    j = j->next;
                }

                if ( gs_model ) {
                    const hw_info_t* hw = hw_get_hw_feature();
                    char *gs_model_fact = kstrdup( hw->gs_model, GFP_KERNEL );
                    char *p;

                    for ( p = gs_model_fact; *p; p++ ) {
                        *p = tolower( *p );
                    }

                    printk("check compatible table (%s-%s)\n", gs_model->u.toml_kv.value,
                                                                       gs_model_fact );

                    
                    if ( strcmp( gs_model->u.toml_kv.value, gs_model_fact ) == 0 ) {
                        struct proc_dir_entry *d;

                        d = proc_mkdir( "compatible", dir );
                        if ( !d ) {
                            printk("error - unable to create 'compatible' proc dir entry\n");
                        }
                        else {
                            manifest_build_proc( o, d );
                        }
                    }

                    kfree( gs_model_fact );
                }
            }
            else if ( strcmp( o->u.toml_table.tag, "volume" ) == 0 ) {
                struct toml_object *name = NULL;
                struct toml_object *image = NULL;
                struct toml_object *version = NULL;
                struct toml_object *j = o->children;

                while ( j ) {
                    if ( j->type == TOML_TYPE_KV && j->u.toml_kv.key && j->u.toml_kv.value ) {
                        if ( strcmp( j->u.toml_kv.key, "name" ) == 0 ) {
                            name = j;   
                        }
                        else if ( strcmp( j->u.toml_kv.key, "image" ) == 0 ) {
                            image = j;
                        }
                        else if ( strcmp( j->u.toml_kv.key, "version" ) == 0 ) {
                            version = j;
                        }
                    }
                    j = j->next;
                }

                if ( name && version ) {
                    struct proc_dir_entry *d;

                    d = proc_mkdir( name->u.toml_kv.value, dir );
                    if ( !d ) {
                        printk("error - unable to create %s proc dir entry\n", name->u.toml_kv.value);
                    }
                    else {
                        if ( !proc_create_data( "name",
                                                S_IFREG | S_IRUGO,
                                                d,
                                                &gs_manifest_fops,
                                                (int*)name ) ) {
                            printk("error - unable to create proc entry for toml kv name\n");
                        }
                        if ( !image ) {
                            if ( !proc_create_data( "image",
                                                    S_IFREG | S_IRUGO,
                                                    d,
                                                    &gs_manifest_fops,
                                                    (int*)name ) ) {
                                printk("error - unable to create proc entry for toml kv image\n");
                            }
                        }
                        else {
                            if ( !proc_create_data( "image",
                                                    S_IFREG | S_IRUGO,
                                                    d,
                                                    &gs_manifest_fops,
                                                    (int*)image ) ) {
                                printk("error - unable to create proc entry for toml kv image\n");
                            }
                        }
                        if ( !proc_create_data( "version",
                                                S_IFREG | S_IRUGO,
                                                d,
                                                &gs_manifest_fops,
                                                (int*)version ) ) {
                            printk("error - unable to create proc entry for toml kv version\n");
                        }
                    }
                }
            }
        }
        o = o->next;
    }
}

//=============================================================================
static int __init
gs_manifest_init(
    void
)
//=============================================================================
{
    struct proc_dir_entry *mdir = proc_mkdir( "manifest", NULL );
    struct mtd_info *mtd;
    ImageHeader header;
    size_t read_len;
    struct toml_object *root;
    struct toml_object *full_obj;

    memset( &manifest, 0, sizeof ( manifest ) );

    mtd = get_mtd_device_nm( "ubi0:manifest" );
    if ( IS_ERR(mtd) ) {
        printk("error - unable to locate manifest volume\n");
        return -EINVAL;
    }

    /* gluebi part - no bad block handling */
    if ( mtd_read( mtd, 0, sizeof ( header ), &read_len, (unsigned char*)&header ) ) {
        printk("error - unable to read manifest volume header\n");
        put_mtd_device( mtd );
        return -EINVAL;
    }

    printk("READ HEADER size=%d\n", header.SizeUsed);

    manifest.full = kmalloc( header.SizeUsed + 1, GFP_KERNEL );
    if ( !manifest.full ) {
        return -ENOMEM;
    }
    full_obj = (struct toml_object*)kmalloc( sizeof ( struct toml_object ), GFP_KERNEL );
    if ( !full_obj ) {
        kfree( manifest.full );
        manifest.full = NULL;
        return -ENOMEM;
    }
    if ( mtd_read( mtd, HEADER_BLOCK_SIZE, header.SizeUsed, &read_len, manifest.full ) ) {
        printk("error - unable to read manifest volume\n");
        put_mtd_device( mtd );
        kfree( manifest.full );
        kfree( full_obj );
        manifest.full = NULL;
        return -EINVAL;
    }
    manifest.full[header.SizeUsed] = 0;
    put_mtd_device( mtd );

    full_obj->type = TOML_TYPE_KV;
    full_obj->u.toml_kv.key = "full";
    full_obj->u.toml_kv.value = kstrdup( manifest.full, GFP_KERNEL );

    if ( !proc_create_data( "full", S_IFREG | S_IRUGO, mdir, &gs_manifest_fops, (int*)full_obj  ) ) {
        printk("error - unable to create proc entry manifest/full\n");
        put_mtd_device( mtd );
        kfree( manifest.full );
        kfree( full_obj );
        manifest.full = NULL;
        goto out;
    }

    root = toml_init();
    if ( toml_parse( root, manifest.full ) < 0 ) {
        printk("error - could not parse manifest\n");
        toml_free( root );
        kfree( manifest.full );
        manifest.root = NULL;
        return -1;
    }
    kfree( manifest.full );

    if ( !root->children ) {
        printk("error - manifest contained no nodes\n");
        toml_free( root );
        manifest.root = NULL;
        return  -1;
    }


    manifest_build_proc( root, mdir );

out:
    return 0;
}
#endif

//=============================================================================
static int 
__init gs_procfs_init( 
    void
)
//=============================================================================
{
    int i;
    struct proc_dir_entry *gs_procfs_root;
    struct proc_dir_entry *gs_devinfo_root;
    struct proc_dir_entry *gs_hw_feature_root;
    struct proc_dir_entry *gs_sec_feature_root;

    memset( &oem_override, 0, sizeof ( oem_override ) );

    gs_procfs_root = proc_mkdir( "gxp", NULL );

    if ( !proc_create( "oem_override", S_IFREG | S_IRUGO, gs_procfs_root, &gs_oem_fops ) )
        goto err_out;
    if ( !proc_create( "cfg_reload", S_IFREG | S_IRUGO, gs_procfs_root, &gs_reload_fops ) )
        goto err_out;
    if ( !proc_create( "production", S_IFREG | S_IRUGO, gs_procfs_root, &gs_production_fops ) )
        goto err_out;
#ifdef SECURE_BOOT_OTP_ROM_DEBUG
    if ( !proc_create( "otpdump", S_IFREG | S_IRUGO, gs_procfs_root, &gs_otpdump_fops ) )
        goto err_out;
#endif

    for ( i = 0; i < NUM_PROCFS_ENTRIES; i++ )
    {
        if ( !proc_create( gs_procfs_entry[i].name, 
                           S_IFREG | S_IRUGO, 
                           gs_procfs_root, 
                           gs_procfs_entry[i].fops ) )
            goto err_out;       
        gs_procfs_entry[i].init();
    }

    gs_devinfo_root = proc_mkdir( "dev_info", gs_procfs_root );
    for ( i = 0; i < NUM_DEVINFO_ENTRIES; i++ )
    {
        if ( !proc_create_data( gs_devinfo_entry[i].name, 
                                S_IFREG | S_IRUGO, 
                                gs_devinfo_root, 
                                &gs_devinfo_fops,
                                (int*)i ) )
            goto err_out;
    }

    gs_hw_feature_root = proc_mkdir( "hw_features", gs_devinfo_root );
    for ( i = 0; i < NUM_HW_FEATURE_ENTRIES; i++ )
    {
        if ( !proc_create_data( hw_feature_entry[i].name, 
                                S_IFREG | S_IRUGO, 
                                gs_hw_feature_root, 
                                &gs_hw_feature_fops,
                                (int*)hw_feature_entry[i].value ) )
            goto err_out;
    }

    gs_sec_feature_root = proc_mkdir( "security", gs_devinfo_root );
    for ( i = 0; i < NUM_SEC_FEATURE_ENTRIES; i++ )
    {
        switch ( sec_feature_entry[i].value ) {
            case SEC_FEATURE_PRIVATE_KEY:
                if ( !proc_create( sec_feature_entry[i].name,
                                   S_IFREG | S_IRUGO,
                                   gs_sec_feature_root,
                                   &sec_private_key_fops ) )
                    goto err_out;
                break;
            case SEC_FEATURE_CERTIFICATE:
                if ( !proc_create( sec_feature_entry[i].name,
                                   S_IFREG | S_IRUGO,
                                   gs_sec_feature_root,
                                   &sec_certificate_fops ) )
                    goto err_out;
                break;
            default:
                if ( !proc_create_data( sec_feature_entry[i].name,
                                        S_IFREG | S_IRUGO,
                                        gs_sec_feature_root,
                                        &sec_feature_fops,
                                        (int*)sec_feature_entry[i].value ) )
                    goto err_out;
                break;
        }
    }

    return 0;
err_out:
    printk("%s: creation error, system will not work\n", __func__);
    panic( "cannot create gs procfs entries\n" );
    return -1;
}

static void __exit gs_procfs_exit( void )
{

}

EXPORT_SYMBOL(hw_feature_get);
EXPORT_SYMBOL(lcd_on_off_set);
EXPORT_SYMBOL(mpk_on_off_set);
EXPORT_SYMBOL(lcd_brightness_set);
EXPORT_SYMBOL(mpk_brightness_set);

subsys_initcall(gs_procfs_init);
#ifdef GS_MANIFEST_PROC
late_initcall(gs_manifest_init);
#endif
module_exit(gs_procfs_exit);

