/**
 * @file sensor_gc2023.c
 * @brief camera driver file
 * Copyright (C) 2011 Anyka (Guangzhou) Microelectronics Technology Co., Ltd
 * @author ye_guohong 
 * @date
 * @version 1.0
 * @ref
 */ 
#include "ak_sensor_common.h"

#define SENSOR_PWDN_LEVEL		1 
#define SENSOR_RESET_LEVEL		0
#define SENSOR_I2C_ADDR         0x6e 
#define SENSOR_ID               0x2023
#define SENSOR_MCLK             24		//dvp 27
#define SENSOR_REGADDR_BYTE     1
#define SENSOR_DATA_BYTE        1
#define MAX_FPS					25
#define SENSOR_OUTPUT_WIDTH		1920
#define SENSOR_OUTPUT_HEIGHT	1080
#define SENSOR_VALID_OFFSET_X	0
#define SENSOR_VALID_OFFSET_Y	0
#define SENSOR_BUS_TYPE		BUS_TYPE_RAW
#define SENSOR_IO_INTERFACE     MIPI_INTERFACE
#define SENSOR_IO_LEVEL         SENSOR_IO_LEVEL_1V8

#define SKIP_NUM 		2
#define EXP_EFFECT_FRAMES 		1
#define VSYNC_ACTIVE_MS			31
#define MIPI_MHZ		768		// 192 for 2lane
#define MIPI_LANE		1

static int gc2023_cmos_updata_d_gain(const unsigned int d_gain);
static int gc2023_cmos_updata_a_gain(const unsigned int a_gain);
static int gc2023_cmos_updata_exp_time(unsigned int exp_time);
u16 gc2023_setgain(u16 iGain);
static int gc2023_cmos_update_a_gain_timer(void);

static int g_fps = MAX_FPS;
static int to_fps = MAX_FPS;
static int to_fps_value = 0;

static int gc2023_sensor_read_register(int reg)
{
	struct ak_sensor_i2c_data i2cdata;

	i2cdata.u8DevAddr           = SENSOR_I2C_ADDR;
	i2cdata.u32RegAddr          = reg;
	i2cdata.u32RegAddrByteNum   = SENSOR_REGADDR_BYTE;
	i2cdata.u32Data             = 0;
	i2cdata.u32DataByteNum      = SENSOR_DATA_BYTE;

	return ak_sensor_read_register(&i2cdata);
}

static int gc2023_sensor_write_register(int reg, int data)
{
	struct ak_sensor_i2c_data i2cdata;

	i2cdata.u8DevAddr           = SENSOR_I2C_ADDR;
	i2cdata.u32RegAddr          = reg;
	i2cdata.u32RegAddrByteNum   = SENSOR_REGADDR_BYTE;
	i2cdata.u32Data             = data;
	i2cdata.u32DataByteNum      = SENSOR_DATA_BYTE;

	return ak_sensor_write_register(&i2cdata);
}

static int gc2023_init(const AK_ISP_SENSOR_INIT_PARA *para)
{
	int i;
	AK_ISP_SENSOR_REG_INFO *preg_info;
	//	int value;

	preg_info = para->reg_info;
	for (i = 0; i < para->num; i++) {
		//int value=gc2023_sensor_read_register(preg_info->reg_addr);
		//printk(KERN_ERR "before w reg:0x%.x, val:0x%.x, to write:0x%.x\n", preg_info->reg_addr, value, preg_info->value);
		gc2023_sensor_write_register(preg_info->reg_addr, preg_info->value);
		//value=gc2023_sensor_read_register(preg_info->reg_addr);
		//printk(KERN_ERR "after w reg:0x%.x, val:0x%.x\n", preg_info->reg_addr, value);
		preg_info++;
	}
	//gc2023_sensor_write_register(0x24,0xff);
	//gc2023_sensor_write_register(0x8c,0x00);
	//gc2023_sensor_write_register(0xfc,0xc4);
	//gc2023_sensor_write_register(0x23,0x32);

	g_fps = MAX_FPS;

	return 0;
}

static int gc2023_read_id(void)
{
	return SENSOR_ID;
}

static int gc2023_probe_id(void)
{
	u8 value;
	u32 id;

	value = gc2023_sensor_read_register(0xf0);
	id = value << 8;
	value = gc2023_sensor_read_register(0xf1);
	id |= value;
	
	printk(KERN_ERR "%s id:0x%x\n", __func__, id);

	if (id == SENSOR_ID)
		return SENSOR_ID;

	return 0;
}

static int gc2023_get_resolution(int *width, int *height)
{
	*width = SENSOR_OUTPUT_WIDTH;
	*height = SENSOR_OUTPUT_HEIGHT;

	return 0;
}

static int gc2023_get_mclk(void)
{
	return SENSOR_MCLK;
}

static int gc2023_get_fps(void)
{
	return g_fps;
}

static int gc2023_get_valid_coordinate(int *x, int *y)
{
	*x = SENSOR_VALID_OFFSET_X;
	*y = SENSOR_VALID_OFFSET_Y;

	return 0;
}

static enum sensor_bus_type gc2023_get_bus_type(void)
{
	return SENSOR_BUS_TYPE;
}

static int gc2023_get_parameter(int param, void *value)
{
	int ret = 0;
	enum sensor_get_param name = (enum sensor_get_param)param;

	switch (name) {
		case GET_MIPI_MHZ:
			*((int *)value) = MIPI_MHZ;
			break;

		case GET_VSYNC_ACTIVE_MS:
			*((int *)value) = VSYNC_ACTIVE_MS;
			break;

		case GET_CUR_FPS:
			*((int *)value) = g_fps;
			break;

		case GET_MIPI_LANE:
			*((int *)value) = MIPI_LANE;
			break;

		case GET_INTERFACE:
			*((int *)value) = SENSOR_IO_INTERFACE;
			break;

		case GET_SENSOR_IO_LEVEL:
			*((int *)value) = SENSOR_IO_LEVEL;
			break;

		default:
			ret = -1;
			break;
	}

	return ret;
}

static int gc2023_set_power_on(const int pwdn_pin, const int reset_pin)
{
	ak_sensor_set_pin_as_gpio(pwdn_pin);
	ak_sensor_set_pin_dir(pwdn_pin, 1);
	ak_sensor_set_pin_level(pwdn_pin, !SENSOR_PWDN_LEVEL);    
	ak_sensor_mdelay(10);
	ak_sensor_set_pin_as_gpio(reset_pin);
	ak_sensor_set_pin_dir(reset_pin, 1);
	ak_sensor_set_pin_level(reset_pin, SENSOR_RESET_LEVEL);
	ak_sensor_mdelay(10);
	ak_sensor_set_pin_level(reset_pin, !SENSOR_RESET_LEVEL);
	ak_sensor_mdelay(20);

	ak_sensor_print("%s\n",__func__);
	return 0;
}

static int gc2023_set_power_off(const int pwdn_pin, const int reset_pin)
{
	ak_sensor_set_pin_level(pwdn_pin, SENSOR_PWDN_LEVEL);
	ak_sensor_set_pin_dir(reset_pin, 0);

	return 0;
}

static int gc2023_set_fps(const int fps)
{
//#if 0
	int ret = 0;
	int tmp = 0;

    /*MIPI  tmp = (96M / 3200 / fps) - 1088 -16 */
	switch (fps) {
		case 25:
			tmp = 0x0060;
			break;

		case 15:
			tmp = 0x0380;
			break;

		case 12:
			tmp = 0x0510;
			break;

		case 10:
			tmp = 0x0768;
			break;

		case 8:
			tmp = 0x0a56;
			break;

		case 5:
			tmp = 0x1320;
			break;

		default:
			printk(KERN_ERR "%s set fps fail\n", __func__);
			ret = -EINVAL;
			break;
	}

	if (!ret) {
#if 0
		gc2023_sensor_write_register(0x07, tmp >> 8);
		gc2023_sensor_write_register(0x08, tmp & 0xff);
#endif
		//g_fps = fps;
		to_fps = fps;
		to_fps_value = tmp;
	}

	return ret;
//#endif
}

static int gc2023_set_standby_in(const int pwdn_pin, const int reset_pin)
{
	return 0;
}

static int gc2023_set_standby_out(const int pwdn_pin, const int reset_pin)
{
	return 0;
}

static AK_ISP_SENSOR_CB gc2023_callback = 
{
	.sensor_init_func 				= gc2023_init,
	.sensor_read_reg_func			= gc2023_sensor_read_register,
	.sensor_write_reg_func			= gc2023_sensor_write_register,
	.sensor_read_id_func			= gc2023_read_id,
	.sensor_update_a_gain_func		= gc2023_cmos_updata_a_gain,
	.sensor_update_d_gain_func		= gc2023_cmos_updata_d_gain,
	.sensor_updata_exp_time_func	= gc2023_cmos_updata_exp_time,
	.sensor_timer_func				= gc2023_cmos_update_a_gain_timer,

	.sensor_probe_id_func			= gc2023_probe_id,	//use IIC bus
	.sensor_get_resolution_func		= gc2023_get_resolution,
	.sensor_get_mclk_func			= gc2023_get_mclk,
	.sensor_get_fps_func			= gc2023_get_fps,
	.sensor_get_valid_coordinate_func	= gc2023_get_valid_coordinate,
	.sensor_get_bus_type_func		= gc2023_get_bus_type,
	.sensor_get_parameter_func		= gc2023_get_parameter,

	.sensor_set_power_on_func		= gc2023_set_power_on,
	.sensor_set_power_off_func		= gc2023_set_power_off,
	.sensor_set_fps_func			= gc2023_set_fps,
	.sensor_set_standby_in_func		= gc2023_set_standby_in,
	.sensor_set_standby_out_func	= gc2023_set_standby_out
};

AK_SENSOR_MODULE(gc2023_callback, gc2023)

///////////////////////////////////////////////////////////////////////////////////////////

#define ANALOG_GAIN_1 64 // 1.00x
#define ANALOG_GAIN_2 91 // 1.42x
#define ANALOG_GAIN_3 127 // 1.99x
#define ANALOG_GAIN_4 183 // 2.86x
#define ANALOG_GAIN_5 259 // 4.05x
#define ANALOG_GAIN_6 371 // 5.77x
#define ANALOG_GAIN_7 531 // 8.06x
#define ANALOG_GAIN_8 750 // 11.53x
#define ANALOG_GAIN_9 1059 // 16.42x
#define ANALOG_GAIN_10 1490 //23.3X
#define ANALOG_GAIN_11 2129 //32.57X

static void gc2023_set_fps_async(void)
{
	if (to_fps != g_fps) {
		gc2023_sensor_write_register(0x07, to_fps_value >> 8);
		gc2023_sensor_write_register(0x08, to_fps_value  & 0xff);

		g_fps = to_fps;
	}
}

u16 gc2023_setgain(u16 iGain)
{
	u16 iReg,temp;
//	if(GC2033_Lock)
//	return;
//#ifdef GC2033_DRIVER_TRACE
//	SENSORDB("GC2033_SetGain iGain = %d \n",iGain);
//#endif
	gc2023_sensor_write_register(0xfe, 0x00);
	gc2023_sensor_write_register(0xb1, 0x01);
	gc2023_sensor_write_register(0xb2, 0x00);
	iReg = iGain;
	if(iReg < 0x40)
		iReg = 0x40;
	else if((ANALOG_GAIN_1<= iReg)&&(iReg < ANALOG_GAIN_2))
	{
		//analog gain
		gc2023_sensor_write_register(0xb6, 0x00);
		gc2023_sensor_write_register(0x21, 0x28);
		temp = iReg;
		gc2023_sensor_write_register(0xb1, temp>>6);
		gc2023_sensor_write_register(0xb2, (temp<<2)&0xfc);
		//SENSORDB("GC2033 analogic gain 1x , GC2033 add pregain = %d\n",temp);
	}
	else if((ANALOG_GAIN_2<= iReg)&&(iReg < ANALOG_GAIN_3))
	{
		gc2023_sensor_write_register(0xb6, 0x01);
		gc2023_sensor_write_register(0x21, 0x28);
		temp = 64*iReg/ANALOG_GAIN_2;
		gc2023_sensor_write_register(0xb1, temp>>6);
		gc2023_sensor_write_register(0xb2, (temp<<2)&0xfc);
		//SENSORDB("GC2033 analogic gain 1.4x , GC2033 add pregain = %d\n",temp);
	}
	else if((ANALOG_GAIN_3<= iReg)&&(iReg < ANALOG_GAIN_4))
	{
		gc2023_sensor_write_register(0xb6, 0x02);
		gc2023_sensor_write_register(0x21, 0x28);
		temp = 64*iReg/ANALOG_GAIN_3;
		gc2023_sensor_write_register(0xb1, temp>>6);
		gc2023_sensor_write_register(0xb2, (temp<<2)&0xfc);
		//SENSORDB("GC2033 analogic gain 2.5x , GC2033 add pregain = %d\n",temp);
	}
	else if((ANALOG_GAIN_4<= iReg)&&(iReg < ANALOG_GAIN_5))
	{
		gc2023_sensor_write_register(0xb6, 0x03);
		gc2023_sensor_write_register(0x21, 0x28);
		temp = 64*iReg/ANALOG_GAIN_4;
		gc2023_sensor_write_register(0xb1, temp>>6);
		gc2023_sensor_write_register(0xb2, (temp<<2)&0xfc);
		//SENSORDB("GC2033 analogic gain 3.54x , GC2033 add pregain = %d\n",temp);
	}
	else if((ANALOG_GAIN_5<= iReg)&&(iReg < ANALOG_GAIN_6))
	{
		gc2023_sensor_write_register(0xb6, 0x04);
		gc2023_sensor_write_register(0x21, 0x28);
		temp = 64*iReg/ANALOG_GAIN_5;
		gc2023_sensor_write_register(0xb1, temp>>6);
		gc2023_sensor_write_register(0xb2, (temp<<2)&0xfc);
		//SENSORDB("GC2033 analogic gain 4.9x , GC2033 add pregain = %d\n",temp);
	}
	else if((ANALOG_GAIN_6<= iReg)&&(iReg < ANALOG_GAIN_7))
	{
		gc2023_sensor_write_register(0xb6, 0x05);//
		gc2023_sensor_write_register(0x21, 0x28);
		temp = 64*iReg/ANALOG_GAIN_6;
		gc2023_sensor_write_register(0xb1, temp>>6);
		gc2023_sensor_write_register(0xb2, (temp<<2)&0xfc);
		//SENSORDB("GC2033 analogic gain6.9x, GC2033 add pregain = %d\n",temp);
	}
	else if((ANALOG_GAIN_7<= iReg)&&(iReg < ANALOG_GAIN_8))
	{
		gc2023_sensor_write_register(0xb6, 0x06);//
		gc2023_sensor_write_register(0x21, 0x28);
		temp = 64*iReg/ANALOG_GAIN_7;
		gc2023_sensor_write_register(0xb1, temp>>6);
		gc2023_sensor_write_register(0xb2, (temp<<2)&0xfc);
		//SENSORDB("GC2033 analogic gain 9.7x, GC2033 add pregain = %d\n",temp);
	}
	else if((ANALOG_GAIN_8<= iReg)&&(iReg < ANALOG_GAIN_9))
	{
		gc2023_sensor_write_register(0xb6, 0x07);
		gc2023_sensor_write_register(0x21, 0x28);
		temp = 64*iReg/ANALOG_GAIN_8;
		gc2023_sensor_write_register(0xb1, temp>>6);
		gc2023_sensor_write_register(0xb2, (temp<<2)&0xfc);
		//SENSORDB("GC2033 analogic gain 13.6x,GC2033 add pregain = %d\n",temp);
	}
	else if(ANALOG_GAIN_9<= iReg)
	{
		gc2023_sensor_write_register(0xb6, 0x08);
		gc2023_sensor_write_register(0x21, 0x2c);
		temp = 64*iReg/ANALOG_GAIN_9;
		gc2023_sensor_write_register(0xb1, temp>>6);
		gc2023_sensor_write_register(0xb2, (temp<<2)&0xfc);
		//SENSORDB("GC2033 analogic gain 39x ,GC2033 add pregain = %d\n",temp);
	}

	gc2023_sensor_write_register(0xfe, 0x00);

	return 0;
}

static int gc2023_cmos_updata_d_gain(const unsigned int d_gain)
{
	//?????????????Ļص?????
	return 0;
}

static int gc2023_cmos_updata_a_gain(const unsigned int a_gain)
{
	//??????ģ???????Ļص?????
	unsigned int tmp_a_gain;
	unsigned int tmp_d_gain;

	tmp_a_gain = a_gain>>2;
	tmp_d_gain = 0;

	gc2023_set_fps_async();

	gc2023_setgain(tmp_a_gain);
	return 0;
}


static int gc2023_cmos_updata_exp_time(unsigned int exp_time)
{
    //?????ع?ʱ???Ļص?????
//    unsigned char exposure_time_msb;
//	unsigned char exposure_time_lsb;
//	printk("exp_time=%d\n",exp_time);
//    exposure_time_msb =(exp_time>>8)&0xff;
 //   exposure_time_lsb = exp_time&0xff;
//	printk("msb = %d\n",exposure_time_msb);
//	printk("lsb = %d\n",exposure_time_lsb);
	//gc2023_sensor_write_register(0x02,exposure_time_msb);
	//gc2023_sensor_write_register(0x01,exposure_time_lsb);
//	printk("0x16=%d\n",gc2023_sensor_read_register(0x16));

	gc2023_set_fps_async();
	
	if(exp_time < 1) exp_time = 1;
	if(exp_time > 8191) exp_time = 8191;//2^13
	//Update Shutter
	gc2023_sensor_write_register(0x04, (exp_time) & 0xff);
	gc2023_sensor_write_register(0x03, (exp_time >> 8) & 0x1f);
    
    return EXP_EFFECT_FRAMES;
}

static int gc2023_cmos_update_a_gain_timer(void)
{
#define UPDATE_FPS_TIMER_PERIOD  (2) //s
	static int start = 0;
	static struct timeval stv = {.tv_sec = 0,};
	struct timeval tv;

	if ((to_fps != g_fps) && !start) {
		start = 1;
		do_gettimeofday(&stv);
	}

	if (start == 1) {
		do_gettimeofday(&tv);
		if (tv.tv_sec >= stv.tv_sec + UPDATE_FPS_TIMER_PERIOD) {
			gc2023_set_fps_async();
			stv.tv_sec = tv.tv_sec;
			start = 0;
		}
	}

	return 0;
}
