/**
 * @file camera_hm2140.c
 * @brief camera driver file
 * Copyright (C) 2011 Anyka (Guangzhou) Microelectronics Technology Co., Ltd
 * @author dong_feilong
 * @date 2015-11-05
 * @version 1.0
 * @ref
 */ 
#include "ak_sensor_common.h"
#include <linux/delay.h>


#define SENSOR_PWDN_LEVEL		0
#define SENSOR_RESET_LEVEL		0
#define SENSOR_I2C_ADDR         0x48 
#define SENSOR_ID               0x2140
#define SENSOR_MCLK             24
#define SENSOR_REGADDR_BYTE     2
#define SENSOR_DATA_BYTE        1
#define MAX_FPS					25
#define SENSOR_OUTPUT_WIDTH		1920
#define SENSOR_OUTPUT_HEIGHT	1080
#define SENSOR_VALID_OFFSET_X	0
#define SENSOR_VALID_OFFSET_Y	0
#define SENSOR_BUS_TYPE		BUS_TYPE_RAW
#define SENSOR_IO_INTERFACE     DVP_INTERFACE
#define SENSOR_IO_LEVEL         SENSOR_IO_LEVEL_1V8

#define OV_MAX(a, b)            (((a) < (b) ) ?  (b) : (a))
#define OV_MIN(a, b)            (((a) > (b) ) ?  (b) : (a))
#define OV_CLIP3(low, high, x)  (OV_MAX(OV_MIN((x), high), low))

#define SKIP_NUM 		2
#define EXP_EFFECT_FRAMES 		1
#define VSYNC_ACTIVE_MS			36
#define MIPI_MHZ		768
#define MIPI_LANE		1

static int hm2140_cmos_updata_d_gain(const unsigned int d_gain);
static int hm2140_cmos_updata_a_gain(const unsigned int a_gain);
static int hm2140_cmos_updata_exp_time(unsigned int exp_time);
static int hm2140_cmos_update_a_gain_timer(void);

static int g_fps = MAX_FPS;
static int to_fps = MAX_FPS;
static int to_fps_value = 0;

static int hm2140_sensor_read_register(int reg)
{
	struct ak_sensor_i2c_data i2cdata;

	i2cdata.u8DevAddr           = SENSOR_I2C_ADDR;
	i2cdata.u32RegAddr          = reg;
	i2cdata.u32RegAddrByteNum   = SENSOR_REGADDR_BYTE;
	i2cdata.u32Data             = 0;
	i2cdata.u32DataByteNum      = SENSOR_DATA_BYTE;

	return ak_sensor_read_register(&i2cdata);
}

static int hm2140_sensor_write_register(int reg, int data)
{
	struct ak_sensor_i2c_data i2cdata;

	i2cdata.u8DevAddr           = SENSOR_I2C_ADDR;
	i2cdata.u32RegAddr          = reg;
	i2cdata.u32RegAddrByteNum   = SENSOR_REGADDR_BYTE;
	i2cdata.u32Data             = data;
	i2cdata.u32DataByteNum      = SENSOR_DATA_BYTE;

	return ak_sensor_write_register(&i2cdata);
}

static int hm2140_init(const AK_ISP_SENSOR_INIT_PARA *para)
{
	int i;
//	int value;
	AK_ISP_SENSOR_REG_INFO *preg_info;

	preg_info = para->reg_info;

	hm2140_sensor_write_register(0x0103, 0x00);
	ak_sensor_mdelay(5);
	hm2140_sensor_write_register(0x5227, 0x6F);
	hm2140_sensor_write_register(0x030B, 0x01);
	ak_sensor_mdelay(100);

	for (i = 0; i < para->num; i++) {
//value=hm2140_sensor_read_register(preg_info->reg_addr);
//printk(KERN_ERR "before w reg:0x%.x, val:0x%.x, to write:0x%.x\n", preg_info->reg_addr, value, preg_info->value);
		hm2140_sensor_write_register(preg_info->reg_addr, preg_info->value);
//value=hm2140_sensor_read_register(preg_info->reg_addr);
//printk(KERN_ERR "after w reg:0x%.x, val:0x%.x\n", preg_info->reg_addr, value);
		preg_info++;
	}

    g_fps = MAX_FPS;
	return 0;
}

static int hm2140_read_id(void)
{
    return SENSOR_ID;
}

static int hm2140_probe_id(void)
{
    u8 value;
    u32 id;

    value = hm2140_sensor_read_register(0x0000);
    id = value << 8;

    value = hm2140_sensor_read_register(0x0001);
    id |= value;

	printk(KERN_ERR "%s id:0x%x,id:0x%x\n", __func__, id, SENSOR_ID);
	
	if (id != SENSOR_ID)
		goto fail;
           
    return SENSOR_ID;

fail:
	return 0;
}

static int hm2140_get_resolution(int *width, int *height)
{
	*width = SENSOR_OUTPUT_WIDTH;
	*height = SENSOR_OUTPUT_HEIGHT;

	return 0;
}

static int hm2140_get_mclk(void)
{
	return SENSOR_MCLK;
}

static int hm2140_get_fps(void)
{
	return g_fps;
}

static int hm2140_get_valid_coordinate(int *x, int *y)
{
	*x = SENSOR_VALID_OFFSET_X;
	*y = SENSOR_VALID_OFFSET_Y;

	return 0;
}

static enum sensor_bus_type hm2140_get_bus_type(void)
{
	return SENSOR_BUS_TYPE;
}

static int hm2140_get_parameter(int param, void *value)
{
	int ret = 0;
	enum sensor_get_param name = (enum sensor_get_param)param;

	switch (name) {
		case GET_MIPI_MHZ:
			*((int *)value) = MIPI_MHZ;
			break;
			
		case GET_VSYNC_ACTIVE_MS:
			*((int *)value) = VSYNC_ACTIVE_MS;
			break;

		case GET_CUR_FPS:
			*((int *)value) = g_fps;
			break;

		case GET_MIPI_LANE:
			*((int *)value) = MIPI_LANE;
			break;

		case GET_INTERFACE:
			*((int *)value) = SENSOR_IO_INTERFACE;
			break;

		case GET_SENSOR_IO_LEVEL:
			*((int *)value) = SENSOR_IO_LEVEL;
			break;

		default:
			ret = -1;
			break;
	}

	return ret;
}
static int hm2140_set_power_on(const int pwdn_pin, const int reset_pin)
{  
    ak_sensor_set_pin_as_gpio(pwdn_pin);
    ak_sensor_set_pin_dir(pwdn_pin, 1);
    ak_sensor_set_pin_level(pwdn_pin, !SENSOR_PWDN_LEVEL);    
    ak_sensor_mdelay(10);
    ak_sensor_set_pin_as_gpio(reset_pin);
    ak_sensor_set_pin_dir(reset_pin, 1);
    ak_sensor_set_pin_level(reset_pin, SENSOR_RESET_LEVEL);
    ak_sensor_mdelay(30);   //modify reset time, used to solve no camera interrupt problem
    ak_sensor_set_pin_level(reset_pin, !SENSOR_RESET_LEVEL);
    ak_sensor_mdelay(20);
	
	return 0;
}

static int hm2140_set_power_off(const int pwdn_pin, const int reset_pin)
{
	//sccb software standby mode
    
    ak_sensor_set_pin_level(pwdn_pin, SENSOR_PWDN_LEVEL);
    ak_sensor_set_pin_dir(reset_pin, 0);

    return 0;
}

static int hm2140_set_fps(int fps)
{
	int ret = 0;
	int tmp=0;

    /* tmp = 76.8M / (reg0x342~3=2300) / fps */
	switch (fps) {
		case 30:
			tmp = 0x0458;
			break;
		case 25:
			tmp = 0x0518;
			break;
		case 20:
			tmp = 0x0685;
			break;
		case 15:
			tmp = 0x08b2;
			break;
		case 12:
			tmp = 0x0a6f;
			break;
		case 10:
			tmp = 0x0d0b;
			break;
		case 8:
			tmp = 0x104d;
			break;
		default:
			printk(KERN_ERR"%s set fps=%d fail\n", __func__, fps);
			ret = -1;
			break;
	}

    if (!ret) {
		to_fps = fps;
		to_fps_value = tmp;
    }
	
	return ret;
}

static int hm2140_set_standby_in(const int pwdn_pin, const int reset_pin)
{
	return 0;
}

static int hm2140_set_standby_out(const int pwdn_pin, const int reset_pin)
{
	return 0;
}

static AK_ISP_SENSOR_CB hm2140_callback = 
{
	.sensor_init_func 				= hm2140_init,
	.sensor_read_reg_func			= hm2140_sensor_read_register,
	.sensor_write_reg_func			= hm2140_sensor_write_register,
	.sensor_read_id_func			= hm2140_read_id,
	.sensor_update_a_gain_func		= hm2140_cmos_updata_a_gain,
	.sensor_update_d_gain_func		= hm2140_cmos_updata_d_gain,
	.sensor_updata_exp_time_func	= hm2140_cmos_updata_exp_time,
	.sensor_timer_func				= hm2140_cmos_update_a_gain_timer,

	.sensor_probe_id_func			= hm2140_probe_id,	//use IIC bus
	.sensor_get_resolution_func		= hm2140_get_resolution,
	.sensor_get_mclk_func			= hm2140_get_mclk,
	.sensor_get_fps_func			= hm2140_get_fps,
	.sensor_get_valid_coordinate_func	= hm2140_get_valid_coordinate,
	.sensor_get_bus_type_func		= hm2140_get_bus_type,
	.sensor_get_parameter_func		= hm2140_get_parameter,

	.sensor_set_power_on_func		= hm2140_set_power_on,
	.sensor_set_power_off_func		= hm2140_set_power_off,
	.sensor_set_fps_func			= hm2140_set_fps,
	.sensor_set_standby_in_func		= hm2140_set_standby_in,
	.sensor_set_standby_out_func	= hm2140_set_standby_out
};

AK_SENSOR_MODULE(hm2140_callback, hm2140)
///////////////////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////////////////

static int hm2140_cmos_updata_d_gain(const unsigned int d_gain)
{
	return 0;
}

static void hm2140_set_fps_async(void)
{
	if (to_fps != g_fps) {
		hm2140_sensor_write_register(0x0340, to_fps_value>>8);
	    hm2140_sensor_write_register(0x0341, to_fps_value&0xff);
		//ak_sensor_mdelay(10);
		hm2140_sensor_write_register(0x0104,0x1); //update 

		g_fps = to_fps;
	}
}


static unsigned int _cmos_gains_convert(unsigned int a_gain, unsigned int d_gain,
		unsigned int *a_gain_out ,unsigned int *d_gain_out)
{
    //Ä£??????????×ª???????????ã·??????????×ª????sensor?????????
    #define _2XGAIN	2048
	unsigned int  fine_gain = 0;
	int tmp = 0;
	int hm2140_curr_again_level = 0;
	//int ret = 0;

	//ak_sensor_print("a_gain =%d\n",a_gain);

	if (a_gain < _2XGAIN) {// 1 -1.9375
		tmp = a_gain - 1024;
		hm2140_curr_again_level = 0;
	} else if (a_gain < 4096) {// 2 - 3.875
		tmp = a_gain/2 - 1024;
		hm2140_curr_again_level = 1;
	} else if (a_gain < 8192) {//4-7.75
		tmp = a_gain/4 - 1024;
		hm2140_curr_again_level = 2;
	} else if (a_gain < 16384) {//8 - 15.5
		tmp = a_gain/8 - 1024;
		hm2140_curr_again_level = 3;
	} else if (a_gain < 32768) {//16 - 31
		tmp = a_gain/16 - 1024;
		hm2140_curr_again_level = 4;
	} else {//32 - 62
		if (a_gain > 63488)
			a_gain = 63488;
		tmp = a_gain/32 - 1024;
		hm2140_curr_again_level = 5;
	}

	if (tmp >= 0){
		fine_gain = tmp/64;
	} else {
		fine_gain = 0;
	}

	fine_gain &= 0x0f;
	fine_gain |= hm2140_curr_again_level << 4;

	//printk(KERN_ERR"again:%d, fine_gain:%x******************************\n",a_gain>>2, fine_gain);
	hm2140_sensor_write_register(0x0205, fine_gain);
	//ak_sensor_mdelay(10);
	hm2140_sensor_write_register(0x0104,0x1); //update 
	return 0;
}

static int hm2140_cmos_updata_a_gain(const unsigned int a_gain)
{
	unsigned int tmp_a_gain;
	unsigned int tmp_d_gain;
	unsigned int tmp_a_gain_out;
	unsigned int tmp_d_gain_out;

	tmp_a_gain = a_gain<<2;
	tmp_d_gain = 0;

	return _cmos_gains_convert(tmp_a_gain, tmp_d_gain,  &tmp_a_gain_out ,&tmp_d_gain_out);
}

static int hm2140_cmos_updata_exp_time(unsigned int exp_time)
{
	unsigned int exposure_time_msb;
	unsigned int exposure_time_lsb;

	//printk(KERN_ERR"exp_time=0x%x****************************\n",exp_time);
	hm2140_set_fps_async();
	
	if((to_fps_value-2) < exp_time)
	{
		exp_time = to_fps_value-2;
	}
	
	//exp_time = to_fps_value -1 -exp_time;
	exposure_time_msb =(exp_time>>8)&0xf;
	exposure_time_lsb =(exp_time)&0xff;
	hm2140_sensor_write_register(0x0202,exposure_time_msb);
	hm2140_sensor_write_register(0x0203,exposure_time_lsb);
	//ak_sensor_mdelay(10);
	hm2140_sensor_write_register(0x0104,0x01);
    return EXP_EFFECT_FRAMES;
}

static int hm2140_cmos_update_a_gain_timer(void)
{
	return 0;
}

