/**
 * @file camera_ov2643.c
 * @brief camera driver file
 * Copyright (C) 2011 Anyka (Guangzhou) Microelectronics Technology Co., Ltd
 * @author luweichun
 * @date 2015-11-05
 * @version 1.0
 * @ref
 */ 
#include "ak_sensor_common.h"

#define SENSOR_PWDN_LEVEL		1 
#define SENSOR_RESET_LEVEL		0
#define SENSOR_I2C_ADDR         0x60 
#define SENSOR_ID               0x2643
#define SENSOR_MCLK             24
#define SENSOR_REGADDR_BYTE     1
#define SENSOR_DATA_BYTE        1
#define MAX_FPS					25
#define SENSOR_OUTPUT_WIDTH		1280
#define SENSOR_OUTPUT_HEIGHT	720
#define SENSOR_VALID_OFFSET_X	0
#define SENSOR_VALID_OFFSET_Y	0
#define SENSROR_BUS_TYPE		BUS_TYPE_YUV

#define OV_MAX(a, b)            (((a) < (b) ) ?  (b) : (a))
#define OV_MIN(a, b)            (((a) > (b) ) ?  (b) : (a))
#define OV_CLIP3(low, high, x)  (OV_MAX(OV_MIN((x), high), low))

static int ov2643_cmos_updata_d_gain(const unsigned int d_gain);
static int ov2643_cmos_updata_a_gain(const unsigned int a_gain);
static int ov2643_cmos_updata_exp_time(unsigned int exp_time);
static int ov2643_cmos_update_a_gain_timer(void);

static int g_fps = MAX_FPS;

static int ov2643_sensor_read_register(int reg)
{
	struct ak_sensor_i2c_data i2cdata;

	i2cdata.u8DevAddr           = SENSOR_I2C_ADDR;
	i2cdata.u32RegAddr          = reg;
	i2cdata.u32RegAddrByteNum   = SENSOR_REGADDR_BYTE;
	i2cdata.u32Data             = 0;
	i2cdata.u32DataByteNum      = SENSOR_DATA_BYTE;

	return ak_sensor_read_register(&i2cdata);
}

static int ov2643_sensor_write_register(int reg, int data)
{
	struct ak_sensor_i2c_data i2cdata;

	i2cdata.u8DevAddr           = SENSOR_I2C_ADDR;
	i2cdata.u32RegAddr          = reg;
	i2cdata.u32RegAddrByteNum   = SENSOR_REGADDR_BYTE;
	i2cdata.u32Data             = data;
	i2cdata.u32DataByteNum      = SENSOR_DATA_BYTE;

	return ak_sensor_write_register(&i2cdata);
}

static int ov2643_init(const AK_ISP_SENSOR_INIT_PARA *para)
{
	int i;
	AK_ISP_SENSOR_REG_INFO *preg_info;

	preg_info = para->reg_info;
	for (i = 0; i < para->num; i++) {
//value=sc1144_sensor_read_register(preg_info->reg_addr);
//ak_sensor_print("before w reg:0x%.x, val:0x%.x, to write:0x%.x\n", preg_info->reg_addr, value, preg_info->value);
		ov2643_sensor_write_register(preg_info->reg_addr, preg_info->value);
//value=ov2643_sensor_read_register(preg_info->reg_addr);
//ak_sensor_print("after w reg:0x%.x, val:0x%.x\n", preg_info->reg_addr, value);
		preg_info++;
	}

	return 0;
}

static int ov2643_read_id(void)
{
    return SENSOR_ID;
}

static int ov2643_probe_id(void)
{
    u8 value;
    u32 id;

    value = ov2643_sensor_read_register(0x0a);
    id = value << 8;

    value = ov2643_sensor_read_register(0x0b);
    id |= value;

	ak_sensor_print("%s id:0x%x\n", __func__, id);
	if (id != SENSOR_ID)
		goto fail;

    return SENSOR_ID;

fail:
	return 0;
}

static int ov2643_get_resolution(int *width, int *height)
{
	*width = SENSOR_OUTPUT_WIDTH;
	*height = SENSOR_OUTPUT_HEIGHT;

	return 0;
}

static int ov2643_get_mclk(void)
{
	return SENSOR_MCLK;
}

static int ov2643_get_fps(void)
{
	return g_fps;
}

static int ov2643_get_valid_coordinate(int *x, int *y)
{
	*x = SENSOR_VALID_OFFSET_X;
	*y = SENSOR_VALID_OFFSET_Y;

	return 0;
}

static enum sensor_bus_type ov2643_get_bus_type(void)
{
	return SENSROR_BUS_TYPE;
}

static int ov2643_set_power_on(const int pwdn_pin, const int reset_pin)
{
    ak_sensor_set_pin_as_gpio(pwdn_pin);
    ak_sensor_set_pin_dir(pwdn_pin, 1);
    ak_sensor_set_pin_level(pwdn_pin, !SENSOR_PWDN_LEVEL);    
    ak_sensor_mdelay(10);
    ak_sensor_set_pin_as_gpio(reset_pin);
    ak_sensor_set_pin_dir(reset_pin, 1);
    ak_sensor_set_pin_level(reset_pin, SENSOR_RESET_LEVEL);
    ak_sensor_mdelay(30);   //modify reset time, just in case 1145 had the same problem with 1135
    ak_sensor_set_pin_level(reset_pin, !SENSOR_RESET_LEVEL);
    ak_sensor_mdelay(20);

	g_fps = MAX_FPS;
	ak_sensor_print("%s\n",__func__);
	return 0;
}

static int ov2643_set_power_off(const int pwdn_pin, const int reset_pin)
{
	//sccb software standby mode
    
    ak_sensor_set_pin_level(pwdn_pin, SENSOR_PWDN_LEVEL);
    ak_sensor_set_pin_dir(reset_pin, 0);

    return 0;
}

static int ov2643_set_fps(int fps)
{
	int ret = 0;

	ak_sensor_print("%s set fps:%d\n", __func__, fps);

	switch (fps) {
	case 25:
		break;
	default:
		ak_sensor_print("%s set fps fail\n", __func__);
		ret = -EINVAL;
		break;
	}

	if (!ret)
		g_fps = fps;

	return ret;
}

static int ov2643_set_standby_in(const int pwdn_pin, const int reset_pin)
{
	return 0;
}

static int ov2643_set_standby_out(const int pwdn_pin, const int reset_pin)
{
	return 0;
}

static AK_ISP_SENSOR_CB ov2643_callback = 
{
	.sensor_init_func 				= ov2643_init,
	.sensor_read_reg_func			= ov2643_sensor_read_register,
	.sensor_write_reg_func			= ov2643_sensor_write_register,
	.sensor_read_id_func			= ov2643_read_id,
	.sensor_update_a_gain_func		= ov2643_cmos_updata_a_gain,
	.sensor_update_d_gain_func		= ov2643_cmos_updata_d_gain,
	.sensor_updata_exp_time_func	= ov2643_cmos_updata_exp_time,
	.sensor_timer_func				= ov2643_cmos_update_a_gain_timer,

	.sensor_probe_id_func			= ov2643_probe_id,	//use IIC bus
	.sensor_get_resolution_func		= ov2643_get_resolution,
	.sensor_get_mclk_func			= ov2643_get_mclk,
	.sensor_get_fps_func			= ov2643_get_fps,
	.sensor_get_valid_coordinate_func	= ov2643_get_valid_coordinate,
	.sensor_get_bus_type_func		= ov2643_get_bus_type,

	.sensor_set_power_on_func		= ov2643_set_power_on,
	.sensor_set_power_off_func		= ov2643_set_power_off,
	.sensor_set_fps_func			= ov2643_set_fps,
	.sensor_set_standby_in_func		= ov2643_set_standby_in,
	.sensor_set_standby_out_func	= ov2643_set_standby_out
};

AK_SENSOR_MODULE(ov2643_callback, ov2643)

///////////////////////////////////////////////////////////////////////////////////////////
static int ov2643_cmos_updata_d_gain(const unsigned int d_gain)
{
	return 0;
}

static int ov2643_cmos_updata_a_gain(const unsigned int a_gain)
{
	return 0;
}

static int ov2643_cmos_updata_exp_time(unsigned int exp_time)
{
    return 0;
}

static int ov2643_cmos_update_a_gain_timer(void)
{
	return 0;
}
