/**
 * @file camera_ov9732.c
 * @brief camera driver file
 * Copyright (C) 2011 Anyka (Guangzhou) Microelectronics Technology Co., Ltd
 * @author luweichun
 * @date 2015-11-05
 * @version 1.0
 * @ref
 */ 
#include "ak_sensor_common.h"

#define SENSOR_PWDN_LEVEL		1 
#define SENSOR_RESET_LEVEL		0
#define SENSOR_I2C_ADDR         0x6c 
#define SENSOR_ID               0x9732
#define SENSOR_MCLK             24
#define SENSOR_REGADDR_BYTE     2
#define SENSOR_DATA_BYTE        1
#define MAX_FPS					30
#define DEF_FPS					25
#define SENSOR_OUTPUT_WIDTH		1280
#define SENSOR_OUTPUT_HEIGHT	720
#define SENSOR_VALID_OFFSET_X	0
#define SENSOR_VALID_OFFSET_Y	0
#define SENSROR_BUS_TYPE		BUS_TYPE_RAW

#define OV_MAX(a, b)            (((a) < (b) ) ?  (b) : (a))
#define OV_MIN(a, b)            (((a) > (b) ) ?  (b) : (a))
#define OV_CLIP3(low, high, x)  (OV_MAX(OV_MIN((x), high), low))

static int ov9732_cmos_updata_d_gain(const unsigned int d_gain);
static int ov9732_cmos_updata_a_gain(const unsigned int a_gain);
static int ov9732_cmos_updata_exp_time(unsigned int exp_time);
static int ov9732_cmos_update_a_gain_timer(void);

static int g_fps = MAX_FPS;
//static int g_a_gain = 0;

static int ov9732_sensor_read_register(int reg)
{
	struct ak_sensor_i2c_data i2cdata;

	i2cdata.u8DevAddr           = SENSOR_I2C_ADDR;
	i2cdata.u32RegAddr          = reg;
	i2cdata.u32RegAddrByteNum   = SENSOR_REGADDR_BYTE;
	i2cdata.u32Data             = 0;
	i2cdata.u32DataByteNum      = SENSOR_DATA_BYTE;

	return ak_sensor_read_register(&i2cdata);
}

static int ov9732_sensor_write_register(int reg, int data)
{
	struct ak_sensor_i2c_data i2cdata;

	i2cdata.u8DevAddr           = SENSOR_I2C_ADDR;
	i2cdata.u32RegAddr          = reg;
	i2cdata.u32RegAddrByteNum   = SENSOR_REGADDR_BYTE;
	i2cdata.u32Data             = data;
	i2cdata.u32DataByteNum      = SENSOR_DATA_BYTE;

	return ak_sensor_write_register(&i2cdata);
}

static int ov9732_init(const AK_ISP_SENSOR_INIT_PARA *para)
{
	int i;
	AK_ISP_SENSOR_REG_INFO *preg_info;

	preg_info = para->reg_info;
	for (i = 0; i < para->num; i++) {
//value=ov9732_sensor_read_register(preg_info->reg_addr);
//ak_sensor_print("before w reg:0x%.x, val:0x%.x, to write:0x%.x\n", preg_info->reg_addr, value, preg_info->value);
		ov9732_sensor_write_register(preg_info->reg_addr, preg_info->value);
//value=ov9732_sensor_read_register(preg_info->reg_addr);
//ak_sensor_print("after w reg:0x%.x, val:0x%.x\n", preg_info->reg_addr, value);
		preg_info++;
	}

	return 0;
}

static int ov9732_read_id(void)
{
    return SENSOR_ID;
}

static int ov9732_probe_id(void)
{
    u8 value;
    u32 id;

    value = ov9732_sensor_read_register(0x300a);
    id = value << 8;

    value = ov9732_sensor_read_register(0x300b);
    id |= value;

	ak_sensor_print("%s id:0x%x\n", __func__, id);
	if (id != SENSOR_ID)
		goto fail;
            
    return SENSOR_ID;

fail:
	return 0;
}

static int ov9732_get_resolution(int *width, int *height)
{
	*width = SENSOR_OUTPUT_WIDTH;
	*height = SENSOR_OUTPUT_HEIGHT;

	return 0;
}

static int ov9732_get_mclk(void)
{
	return SENSOR_MCLK;
}

static int ov9732_get_fps(void)
{
	return g_fps;
}

static int ov9732_get_valid_coordinate(int *x, int *y)
{
	*x = SENSOR_VALID_OFFSET_X;
	*y = SENSOR_VALID_OFFSET_Y;

	return 0;
}

static enum sensor_bus_type ov9732_get_bus_type(void)
{
	return SENSROR_BUS_TYPE;
}

static int ov9732_set_power_on(const int pwdn_pin, const int reset_pin)
{  
    ak_sensor_set_pin_as_gpio(pwdn_pin);
    ak_sensor_set_pin_dir(pwdn_pin, 1);
    ak_sensor_set_pin_level(pwdn_pin, !SENSOR_PWDN_LEVEL);    
    ak_sensor_mdelay(10);
    ak_sensor_set_pin_as_gpio(reset_pin);
    ak_sensor_set_pin_dir(reset_pin, 1);
    ak_sensor_set_pin_level(reset_pin, SENSOR_RESET_LEVEL);
    ak_sensor_mdelay(10);
    ak_sensor_set_pin_level(reset_pin, !SENSOR_RESET_LEVEL);
    ak_sensor_mdelay(20);

	g_fps = DEF_FPS;
	ak_sensor_print("%s\n",__func__);
	return 0;
}

static int ov9732_set_power_off(const int pwdn_pin, const int reset_pin)
{
	//sccb software standby mode
    
    ak_sensor_set_pin_level(pwdn_pin, SENSOR_PWDN_LEVEL);
    ak_sensor_set_pin_dir(reset_pin, 0);

    return 0;
}

static int ov9732_set_fps(const int fps)
{
	int ret = 0;

	switch (fps) {
	case 30:
		ov9732_sensor_write_register(0x380e, 0x3);
		ov9732_sensor_write_register(0x380f, 0x2b);
		break;
	case 25:
		ov9732_sensor_write_register(0x380e, 0x3);
		ov9732_sensor_write_register(0x380f, 0xce);
		break;
	case 13:	//actual: 12.5fps
	case 12:
		ov9732_sensor_write_register(0x380e, 0x7);
		ov9732_sensor_write_register(0x380f, 0x9c);
		break;
	case 10:
		ov9732_sensor_write_register(0x380e, 0x9);
		ov9732_sensor_write_register(0x380f, 0x83);
		break;
	case 8:
		ov9732_sensor_write_register(0x380e, 0xb);
		ov9732_sensor_write_register(0x380f, 0xe4);
		break;

	default:
		ak_sensor_print("%s set fps fail\n", __func__);
		ret = -EINVAL;
		break;
	}

	if (!ret)
		g_fps = fps;

	return ret;
}

static int ov9732_set_standby_in(const int pwdn_pin, const int reset_pin)
{
	return 0;
}

static int ov9732_set_standby_out(const int pwdn_pin, const int reset_pin)
{
	return 0;
}

static AK_ISP_SENSOR_CB ov9732_callback = 
{
	.sensor_init_func 				= ov9732_init,
	.sensor_read_reg_func			= ov9732_sensor_read_register,
	.sensor_write_reg_func			= ov9732_sensor_write_register,
	.sensor_read_id_func			= ov9732_read_id,
	.sensor_update_a_gain_func		= ov9732_cmos_updata_a_gain,
	.sensor_update_d_gain_func		= ov9732_cmos_updata_d_gain,
	.sensor_updata_exp_time_func	= ov9732_cmos_updata_exp_time,
	.sensor_timer_func				= ov9732_cmos_update_a_gain_timer,

	.sensor_probe_id_func			= ov9732_probe_id,	//use IIC bus
	.sensor_get_resolution_func		= ov9732_get_resolution,
	.sensor_get_mclk_func			= ov9732_get_mclk,
	.sensor_get_fps_func			= ov9732_get_fps,
	.sensor_get_valid_coordinate_func	= ov9732_get_valid_coordinate,
	.sensor_get_bus_type_func		= ov9732_get_bus_type,

	.sensor_set_power_on_func		= ov9732_set_power_on,
	.sensor_set_power_off_func		= ov9732_set_power_off,
	.sensor_set_fps_func			= ov9732_set_fps,
	.sensor_set_standby_in_func		= ov9732_set_standby_in,
	.sensor_set_standby_out_func	= ov9732_set_standby_out
};

AK_SENSOR_MODULE(ov9732_callback, ov9732)

///////////////////////////////////////////////////////////////////////////////////////////

static int ov9732_cmos_updata_d_gain(const unsigned int d_gain)
{
#if 0
	int msb, lsb;

	msb = (d_gain >> 8) & 0x3;
	lsb = d_gain & 0xff;

	ov9732_sensor_write_register(0x5102, msb);
	ov9732_sensor_write_register(0x5103, lsb);
#endif
	return 0;
}

static int ov9732_cmos_updata_a_gain(const unsigned int a_gain)
{
	int tmp_a_gain;
	int msb, lsb;

	tmp_a_gain = a_gain >> 4;
	msb = (tmp_a_gain >> 8) & 0x3;
	lsb = tmp_a_gain & 0xff;

	ov9732_sensor_write_register(0x350a, msb);
	ov9732_sensor_write_register(0x350b, lsb);
	//ak_sensor_print("%s a_gain:%d, msb:%d, lsb:%d\n", __func__, a_gain, msb, lsb);

	return 0;
}


static int ov9732_cmos_updata_exp_time(unsigned int exp_time)
{
	int tmp;
	int tmp_exp_time;
	int msb, isb, lsb;

	tmp_exp_time = exp_time << 4;
	msb = (tmp_exp_time >> 16) & 0xf;
	isb = (tmp_exp_time >> 8) & 0xff;
	lsb = tmp_exp_time & 0xff;

	tmp = ov9732_sensor_read_register(0x3500);
	tmp &= ~0xf;
	msb |= tmp;
	ov9732_sensor_write_register(0x3500, msb);
	ov9732_sensor_write_register(0x3501, isb);
	ov9732_sensor_write_register(0x3502, lsb);
	//ak_sensor_print("%s exp_time:%d, msb:%d, isb:%d, lsb:%d\n", __func__, exp_time, msb, isb, lsb);

    return 0;
}

static int ov9732_cmos_update_a_gain_timer(void)
{
	#define UPDATE_A_GAIN_TIMER_PERIOD	(5)	//s
	static struct timeval stv = {.tv_sec = 0,};
	struct timeval tv;

	do_gettimeofday(&tv);

	if (tv.tv_sec >= stv.tv_sec + UPDATE_A_GAIN_TIMER_PERIOD) {
		stv.tv_sec = tv.tv_sec;
		//ov9732_cmos_updata_a_gain(g_a_gain);
	}

	return 0;
}
