/**
 * @file camera_sc1035.c
 * @brief camera driver file
 * Copyright (C) 2011 Anyka (Guangzhou) Microelectronics Technology Co., Ltd
 * @author luweichun
 * @date 2015-11-05
 * @version 1.0
 * @ref
 */ 
#include "ak_sensor_common.h"

#define SENSOR_PWDN_LEVEL		1 
#define SENSOR_RESET_LEVEL		0
#define SENSOR_I2C_ADDR         0x60 
#define SENSOR_ID               0x1035
#define SENSOR_MCLK             27
#define SENSOR_REGADDR_BYTE     2
#define SENSOR_DATA_BYTE        1
#define MAX_FPS					25
#define SENSOR_OUTPUT_WIDTH		1280
#define SENSOR_OUTPUT_HEIGHT	960
#define SENSOR_VALID_OFFSET_X	0
#define SENSOR_VALID_OFFSET_Y	0
#define SENSROR_BUS_TYPE		BUS_TYPE_RAW

#define OV_MAX(a, b)            (((a) < (b) ) ?  (b) : (a))
#define OV_MIN(a, b)            (((a) > (b) ) ?  (b) : (a))
#define OV_CLIP3(low, high, x)  (OV_MAX(OV_MIN((x), high), low))

static inline int sc1035_denoise(void);
static int sc1035_cmos_updata_d_gain(const unsigned int d_gain);
static int sc1035_cmos_updata_a_gain(const unsigned int a_gain);
static int sc1035_cmos_updata_exp_time(unsigned int exp_time);
static int sc1035_cmos_update_a_gain_timer(void);

static int g_fps = MAX_FPS;
static int read_id_ok = 0;
static int g_a_gain = 0;

static int sc1035_sensor_read_register(int reg)
{
	struct ak_sensor_i2c_data i2cdata;

	i2cdata.u8DevAddr           = SENSOR_I2C_ADDR;
	i2cdata.u32RegAddr          = reg;
	i2cdata.u32RegAddrByteNum   = SENSOR_REGADDR_BYTE;
	i2cdata.u32Data             = 0;
	i2cdata.u32DataByteNum      = SENSOR_DATA_BYTE;

	return ak_sensor_read_register(&i2cdata);
}

static int sc1035_sensor_write_register(int reg, int data)
{
	struct ak_sensor_i2c_data i2cdata;

	i2cdata.u8DevAddr           = SENSOR_I2C_ADDR;
	i2cdata.u32RegAddr          = reg;
	i2cdata.u32RegAddrByteNum   = SENSOR_REGADDR_BYTE;
	i2cdata.u32Data             = data;
	i2cdata.u32DataByteNum      = SENSOR_DATA_BYTE;

	return ak_sensor_write_register(&i2cdata);
}

static int sc1035_init(const AK_ISP_SENSOR_INIT_PARA *para)
{
	int i;
	AK_ISP_SENSOR_REG_INFO *preg_info;

	preg_info = para->reg_info;
	for (i = 0; i < para->num; i++) {
//value=sc1035_sensor_read_register(preg_info->reg_addr);
//ak_sensor_print("before w reg:0x%.x, val:0x%.x, to write:0x%.x\n", preg_info->reg_addr, value, preg_info->value);
		sc1035_sensor_write_register(preg_info->reg_addr, preg_info->value);
//value=sc1035_sensor_read_register(preg_info->reg_addr);
//ak_sensor_print("after w reg:0x%.x, val:0x%.x\n", preg_info->reg_addr, value);
		preg_info++;
	}

	return 0;
}

static int sc1035_read_id(void)
{
    return SENSOR_ID;
}

static int sc1035_probe_id(void)
{
	/* sc1035 have no id register(s) */
    u8 value;

	if (read_id_ok)
		goto success;

#if 1

    value = sc1035_sensor_read_register(0x3780);
	if (value != 0x10)
		goto fail;

    value = sc1035_sensor_read_register(0x3781);
	if (value != 0x10)
		goto fail;

    value = sc1035_sensor_read_register(0x5000);
	if (value != 0x99)
		goto fail;

    value = sc1035_sensor_read_register(0x5800);
	if (value != 0x90)
		goto fail;
#endif

	read_id_ok = 1;

success:
	ak_sensor_print("%s ok\n",__func__);
    return SENSOR_ID;

fail:
	ak_sensor_print("%s fail\n",__func__);
	return 0;
}

static int sc1035_get_resolution(int *width, int *height)
{
	*width = SENSOR_OUTPUT_WIDTH;
	*height = SENSOR_OUTPUT_HEIGHT;

	return 0;
}

static int sc1035_get_mclk(void)
{
	return SENSOR_MCLK;
}

static int sc1035_get_fps(void)
{
	return g_fps;
}

static int sc1035_get_valid_coordinate(int *x, int *y)
{
	*x = SENSOR_VALID_OFFSET_X;
	*y = SENSOR_VALID_OFFSET_Y;

	return 0;
}

static enum sensor_bus_type sc1035_get_bus_type(void)
{
	return SENSROR_BUS_TYPE;
}

static int sc1035_set_power_on(const int pwdn_pin, const int reset_pin)
{  
    ak_sensor_set_pin_as_gpio(pwdn_pin);
    ak_sensor_set_pin_dir(pwdn_pin, 1);
    ak_sensor_set_pin_level(pwdn_pin, !SENSOR_PWDN_LEVEL);    
    ak_sensor_mdelay(10);
    ak_sensor_set_pin_as_gpio(reset_pin);
    ak_sensor_set_pin_dir(reset_pin, 1);
    ak_sensor_set_pin_level(reset_pin, SENSOR_RESET_LEVEL);
    ak_sensor_mdelay(10);
    ak_sensor_set_pin_level(reset_pin, !SENSOR_RESET_LEVEL);
    ak_sensor_mdelay(20);

	g_fps = MAX_FPS;
	ak_sensor_print("%s\n",__func__);
	return 0;
}

static int sc1035_set_power_off(const int pwdn_pin, const int reset_pin)
{
	//sccb software standby mode
    
    ak_sensor_set_pin_level(pwdn_pin, SENSOR_PWDN_LEVEL);
    ak_sensor_set_pin_dir(reset_pin, 0);

    return 0;
}

static int sc1035_set_fps(const int fps)
{
	int ret = 0;

	switch (fps) {
	case 30:
		sc1035_sensor_write_register(0x320e, 0x03);
		sc1035_sensor_write_register(0x320f, 0xe8);
		break;

	case 25:
		sc1035_sensor_write_register(0x320e, 0x04);
		sc1035_sensor_write_register(0x320f, 0xb0);
		//sc1035_sensor_write_register(0x3338, 0x03);
		//sc1035_sensor_write_register(0x3339, 0xe8);
		//sc1035_sensor_write_register(0x3336, 0x00);
		//sc1035_sensor_write_register(0x3337, 0x96);
		break;
	case 13:	//actual: 12.5fps
	case 12:
		sc1035_sensor_write_register(0x320e, 0x09);
		sc1035_sensor_write_register(0x320f, 0x60);
		//sc1035_sensor_write_register(0x3338, 0x07);
		//sc1035_sensor_write_register(0x3339, 0xd0);
		//sc1035_sensor_write_register(0x3336, 0x04);
		//sc1035_sensor_write_register(0x3337, 0x1a);
		break;
	case 10:
		sc1035_sensor_write_register(0x320e, 0x0b);
		sc1035_sensor_write_register(0x320f, 0xb8);
		break;
	case 8:
		sc1035_sensor_write_register(0x320e, 0x0e);
		sc1035_sensor_write_register(0x320f, 0xb8);
		break;

	default:
		ak_sensor_print("%s set fps fail\n", __func__);
		ret = -EINVAL;
		break;
	}

	if (!ret) {
		g_fps = fps;
		sc1035_denoise();
		//ak_sensor_print("%s r320e:%d, r320f:%d\n", __func__, sc1035_sensor_read_register(0x320e), sc1035_sensor_read_register(0x320f));
	}

	return ret;
}

static int sc1035_set_standby_in(const int pwdn_pin, const int reset_pin)
{
	return 0;
}

static int sc1035_set_standby_out(const int pwdn_pin, const int reset_pin)
{
	return 0;
}

static AK_ISP_SENSOR_CB sc1035_callback = 
{
	.sensor_init_func 				= sc1035_init,
	.sensor_read_reg_func			= sc1035_sensor_read_register,
	.sensor_write_reg_func			= sc1035_sensor_write_register,
	.sensor_read_id_func			= sc1035_read_id,
	.sensor_update_a_gain_func		= sc1035_cmos_updata_a_gain,
	.sensor_update_d_gain_func		= sc1035_cmos_updata_d_gain,
	.sensor_updata_exp_time_func	= sc1035_cmos_updata_exp_time,
	.sensor_timer_func				= sc1035_cmos_update_a_gain_timer,

	.sensor_probe_id_func			= sc1035_probe_id,	//use IIC bus
	.sensor_get_resolution_func		= sc1035_get_resolution,
	.sensor_get_mclk_func			= sc1035_get_mclk,
	.sensor_get_fps_func			= sc1035_get_fps,
	.sensor_get_valid_coordinate_func	= sc1035_get_valid_coordinate,
	.sensor_get_bus_type_func		= sc1035_get_bus_type,

	.sensor_set_power_on_func		= sc1035_set_power_on,
	.sensor_set_power_off_func		= sc1035_set_power_off,
	.sensor_set_fps_func			= sc1035_set_fps,
	.sensor_set_standby_in_func		= sc1035_set_standby_in,
	.sensor_set_standby_out_func	= sc1035_set_standby_out
};

AK_SENSOR_MODULE(sc1035_callback, sc1035)

///////////////////////////////////////////////////////////////////////////////////////////

//for SC1035
#define SENSOR_BLC_TOP_VALUE (0x58)
#define SENSOR_BLC_BOT_VALUE (0x45)
#define SENSOR_AGAIN_ADAPT_STEP (1)
#define SENSOR_MAX_AGAIN (0xff)
static const unsigned short sensor_gain_map[62] = {
	0x82, 0x83, 0x84,0x85, 0x86, 0x87, 0x88, 0x89,
	0x8a, 0x8b, 0x8c,0x8d, 0x8e, 0x8f, 0x90, 0x91,
	0x92, 0x93, 0x94, 0x95,0x96, 0x97, 0x98, 0x99,
	0x9a, 0x9b, 0x9c, 0x9d,0x9e, 0x9f, 0xb0, 0xb1,
	0xb2, 0xb3, 0xb4, 0xb5,0xb6, 0xb7, 0xb8, 0xb9,
	0xba, 0xbb, 0xbc, 0xbd,0xbe, 0xbf, 0xf0, 0xf1,
	0xf2, 0xf3, 0xf4, 0xf5,0xf6, 0xf7, 0xf8, 0xf9,
	0xfa, 0xfb, 0xfc, 0xfd,0xfe, 0xff
};

static unsigned short gainmap_gain2index(unsigned short Gain)
{
	unsigned short i = 0;
	for(i = 0; i < sizeof(sensor_gain_map)/sizeof(sensor_gain_map[0]); i++){
		if(sensor_gain_map[i] == Gain){
			break;
		}
	}
	return i;
}

//return limited Again
//param:Again that ISP calculated
unsigned int Again_limit(unsigned int Again)
{
	unsigned int ret_Again;
	unsigned short BLC_top = SENSOR_BLC_TOP_VALUE, BLC_bot = SENSOR_BLC_BOT_VALUE, BLC_reg/*, Again_step = SENSOR_AGAIN_ADAPT_STEP*/, gain_index;
	static unsigned short MaxAgain = SENSOR_MAX_AGAIN;

	//I2C read reg of sensor
	BLC_reg = sc1035_sensor_read_register(0x3911);

	gain_index = gainmap_gain2index(MaxAgain);

	if(BLC_reg > BLC_top){//>0x58
		if(gain_index>0){
			//limit max Again by step
			gain_index -= SENSOR_AGAIN_ADAPT_STEP;
		}
		MaxAgain = sensor_gain_map[gain_index];
	}else if(BLC_reg < BLC_bot){//<0x45
		//release Again limit by step
		gain_index += SENSOR_AGAIN_ADAPT_STEP;
		if(gain_index > sizeof(sensor_gain_map)/sizeof(sensor_gain_map[0])-1){
			gain_index = sizeof(sensor_gain_map)/sizeof(sensor_gain_map[0]) - 1;
		}
		MaxAgain = sensor_gain_map[gain_index];
	}else{//0x45 < BLC_reg < 0x58
		//do nothing
	}
	ret_Again = Again > MaxAgain ? MaxAgain : Again;
	//ak_sensor_print("limit gain:ret_Again=%d;Again=%d;MaxAgain=%d\n", ret_Again, Again, MaxAgain);
	return ret_Again;
}

static int last_dg_value = 0;
static unsigned int _cmos_gains_convert(unsigned int a_gain, unsigned int d_gain,
		unsigned int *a_gain_out ,unsigned int *d_gain_out)
{
	//ģ??????????ת?????????????????????ת????sensor?????????

    unsigned int dig_gain= 1, ana_gain= 1, dcg_gain= 256, fine_gain= 1024;
    unsigned int ag_value= 0x00, dg_value= 0x14;
    int tmp;
    
    if(a_gain<=1927)
    {
    	fine_gain = a_gain;
    }
    else if(a_gain>1927 && a_gain<=3411)
    {
    	dcg_gain = 453;
    	fine_gain = (a_gain*256)/dcg_gain;
    }
    else if(a_gain>3411 && a_gain<=6823)
    {
    	dcg_gain = 453;
    	ana_gain = 2;
    	fine_gain = (a_gain*256)/(dcg_gain*ana_gain);
    }
    else if(a_gain>6823 && a_gain<=14070)
    {
        dcg_gain = 453;
    	ana_gain = 4;
    	fine_gain = (a_gain*256)/(dcg_gain*ana_gain);
    }
    else if(a_gain>14070 && a_gain<=28144)
    {
    	dcg_gain = 453;
    	ana_gain = 8;
    	fine_gain = (a_gain*256)/(dcg_gain*ana_gain);
    }
    else if(a_gain>28144 && a_gain<=56289)
    {
    	dcg_gain = 453;
    	ana_gain = 8;
    	dig_gain = 2;
    	fine_gain = (a_gain*256)/(dcg_gain*ana_gain*dig_gain);
    }
    else if(a_gain>56289)
    {
    	dcg_gain = 453;
    	ana_gain = 8;
    	dig_gain = 4;
    	fine_gain = (a_gain*256)/(dcg_gain*ana_gain*dig_gain);
    }
    
	tmp = fine_gain-1024+30;
	if(tmp<0)
	{
		tmp = 0;
	}
	if(tmp>903)
	{
		tmp = 903;
	}
	tmp=tmp/60;
	ag_value = ag_value | tmp;
	if(dcg_gain == 453)
	{
		ag_value = ag_value|(1<<7);
		ag_value = ag_value|((ana_gain-1)<<4);
	}
	sc1035_sensor_write_register(0x3e09, Again_limit(ag_value));
	
	dg_value = dg_value|(dig_gain-1);
	sc1035_sensor_write_register(0x3e0f, dg_value);
	
	if(a_gain/1024>=80)
	{
		sc1035_sensor_write_register(0x5000, 0x66);
	}
	else
	{
		sc1035_sensor_write_register(0x5000, 0x00);
	}
	//ak_sensor_print(" 0x3e09 = %d\n", sc1035_sensor_read_register(0x3e09));
	//ak_sensor_print(" 0x3e0f = %d\n", sc1035_sensor_read_register(0x360f));
	if(dg_value!=last_dg_value)
	{
		last_dg_value = dg_value;
		return 1;
	}
	
	return 0;
}

static int sc1035_cmos_updata_d_gain(const unsigned int d_gain)
{
	//?????????????Ļص?????
	return 0;
}

#define SKIP_FRAME_NUM	2
static int sc1035_cmos_updata_a_gain(const unsigned int a_gain)
{
	//??????ģ???????Ļص?????
	unsigned int tmp_a_gain;
	unsigned int tmp_d_gain;
	unsigned int tmp_a_gain_out;
	unsigned int tmp_d_gain_out;
	int ret;

	g_a_gain = a_gain;

	tmp_a_gain = a_gain<<2;
	tmp_d_gain = 0;

	ret = _cmos_gains_convert(tmp_a_gain, tmp_d_gain,  &tmp_a_gain_out ,&tmp_d_gain_out);
	
	//ak_sensor_print(" 0x3e09 = %d\n", sc1035_sensor_read_register(0x3e09));
	//ak_sensor_print(" 0x3610 = %d\n", sc1035_sensor_read_register(0x3610));
	//ak_sensor_print(" 0x3605 = %d\n", sc1035_sensor_read_register(0x3605));
	//ak_sensor_print(" tmp_a_gain = %d ",tmp_a_gain);
	//ak_sensor_print(" tmp_a_gain_out = %d\n", tmp_a_gain_out);
	//updata_auto_exposure_num++;
	//isp->aec_param.current_exposure_time = tmp_exposure_time;
	if(ret)
		return SKIP_FRAME_NUM;
	else
		return 0;
}

static inline int sc1035_denoise(void)
{
	int reg_value_hight = sc1035_sensor_read_register(0x320e);
	int reg_value_low = sc1035_sensor_read_register(0x320f);
	int ex_time_max = ((reg_value_hight<<8)|reg_value_low)-4;
	int msb = sc1035_sensor_read_register(0x3e01);
	int lsb = sc1035_sensor_read_register(0x3e02);
	int exp_time = (msb << 4) | (lsb & 0xf);

	if(exp_time<ex_time_max*179/256)
	{
		sc1035_sensor_write_register(0x331e,0xe0);
	}
	else
	{
		int tmp;
		tmp = 224 - (exp_time - ex_time_max * 179/256)*198/(ex_time_max*77/256);
		if (tmp < 0x1a)
			tmp = 0x1a;
		sc1035_sensor_write_register(0x331e, tmp);
	}

	return 0;
}

static int sc1035_cmos_updata_exp_time(unsigned int exp_time)
{
    //?????ع?ʱ???Ļص?????	
    unsigned char exposure_time_msb;
	unsigned char exposure_time_lsb;
	unsigned char tem;
	unsigned int  ex_time_max, reg_value_hight, reg_value_low;
 	//ak_sensor_print("exp_time=%d ",exp_time);
    exposure_time_msb =(exp_time>>4)&0xff;
    exposure_time_lsb =((exp_time)&0xf)<<4;
    tem=sc1035_sensor_read_register(0x3e02)&0x0f;
    exposure_time_lsb =exposure_time_lsb|tem;
	//ak_sensor_print("msb = %d  ",exposure_time_msb);
	//ak_sensor_print("lsb = %d\n",exposure_time_lsb);
	sc1035_sensor_write_register(0x3e01,exposure_time_msb);
	sc1035_sensor_write_register(0x3e02,exposure_time_lsb);

	reg_value_hight = sc1035_sensor_read_register(0x320e);
	reg_value_low = sc1035_sensor_read_register(0x320f);
	ex_time_max = ((reg_value_hight<<8)|reg_value_low)-4;
	
	if(exp_time<ex_time_max*179/256)
	{
		sc1035_sensor_write_register(0x331e,0xe0);
	}
	else
	{
		int tmp;
		tmp = 224 - (exp_time - ex_time_max * 179/256)*198/(ex_time_max*77/256);
		if (tmp < 0x1a)
			tmp = 0x1a;
		sc1035_sensor_write_register(0x331e, tmp);
	}
	
	/*if(exp_time<=ex_time_max-300)
	{
		sc1035_sensor_write_register(0x331e,0xe0);
	}
	else
	{
		sc1035_sensor_write_register(0x331e,224-(exp_time-(ex_time_max-300))*198/300);
	}*/
    return 0;
}

static int sc1035_cmos_update_a_gain_timer(void)
{
	#define UPDATE_A_GAIN_TIMER_PERIOD	(5)	//s
	static struct timeval stv = {.tv_sec = 0,};
	struct timeval tv;

	do_gettimeofday(&tv);

	if (tv.tv_sec >= stv.tv_sec + UPDATE_A_GAIN_TIMER_PERIOD) {
		stv.tv_sec = tv.tv_sec;
		sc1035_cmos_updata_a_gain(g_a_gain);
	}

	return 0;
}
