/**
 * @file camera_sc1135.c
 * @brief camera driver file
 * Copyright (C) 2011 Anyka (Guangzhou) Microelectronics Technology Co., Ltd
 * @author luweichun
 * @date 2015-11-05
 * @version 1.0
 * @ref
 */ 
#include "ak_sensor_common.h"

#define SENSOR_PWDN_LEVEL		1 
#define SENSOR_RESET_LEVEL		0
#define SENSOR_I2C_ADDR         0x60 
#define SENSOR_ID               0x1135
#define SENSOR_MCLK             27
#define SENSOR_REGADDR_BYTE     2
#define SENSOR_DATA_BYTE        1
#define MAX_FPS					30
#define DETAULT_FPS				25
#define SENSOR_OUTPUT_WIDTH		1280
#define SENSOR_OUTPUT_HEIGHT	960
#define SENSOR_VALID_OFFSET_X	0
#define SENSOR_VALID_OFFSET_Y	0
#define SENSROR_BUS_TYPE		BUS_TYPE_RAW

#define OV_MAX(a, b)            (((a) < (b) ) ?  (b) : (a))
#define OV_MIN(a, b)            (((a) > (b) ) ?  (b) : (a))
#define OV_CLIP3(low, high, x)  (OV_MAX(OV_MIN((x), high), low))

static int sc1135_cmos_updata_d_gain(const unsigned int d_gain);
static int sc1135_cmos_updata_a_gain(const unsigned int a_gain);
static int sc1135_cmos_updata_exp_time(unsigned int exp_time);
static int sc1135_cmos_update_a_gain_timer(void);

static int g_fps = DETAULT_FPS;
static int read_id_ok = 0;
static int g_a_gain = 0;

static const char sc1135_version[] = "sensor_sc1135 V2.0.00";

static int sc1135_sensor_read_register(int reg)
{
	struct ak_sensor_i2c_data i2cdata;

	i2cdata.u8DevAddr           = SENSOR_I2C_ADDR;
	i2cdata.u32RegAddr          = reg;
	i2cdata.u32RegAddrByteNum   = SENSOR_REGADDR_BYTE;
	i2cdata.u32Data             = 0;
	i2cdata.u32DataByteNum      = SENSOR_DATA_BYTE;

	return ak_sensor_read_register(&i2cdata);
}

static int sc1135_sensor_write_register(int reg, int data)
{
	struct ak_sensor_i2c_data i2cdata;

	i2cdata.u8DevAddr           = SENSOR_I2C_ADDR;
	i2cdata.u32RegAddr          = reg;
	i2cdata.u32RegAddrByteNum   = SENSOR_REGADDR_BYTE;
	i2cdata.u32Data             = data;
	i2cdata.u32DataByteNum      = SENSOR_DATA_BYTE;

	return ak_sensor_write_register(&i2cdata);
}

static int sc1135_init(const AK_ISP_SENSOR_INIT_PARA *para)
{
	int i;
	AK_ISP_SENSOR_REG_INFO *preg_info;

	preg_info = para->reg_info;
	for (i = 0; i < para->num; i++) {
//value=sc1135_sensor_read_register(preg_info->reg_addr);
//ak_sensor_print("before w reg:0x%.x, val:0x%.x, to write:0x%.x\n", preg_info->reg_addr, value, preg_info->value);
		sc1135_sensor_write_register(preg_info->reg_addr, preg_info->value);
//value=sc1135_sensor_read_register(preg_info->reg_addr);
//ak_sensor_print("after w reg:0x%.x, val:0x%.x\n", preg_info->reg_addr, value);
		preg_info++;
	}

	return 0;
}

static int sc1135_read_id(void)
{
    return SENSOR_ID;
}

static int sc1135_probe_id(void)
{
	/* sc1135 virtual-id registers are same as sc1037*/
    u8 value;

	if (read_id_ok)
		goto success;

#if 1

    value = sc1135_sensor_read_register(0x3107);
	if (value != 0x00)
		goto fail;

    value = sc1135_sensor_read_register(0x3108);
	if (value != 0x00)
		goto fail;

    value = sc1135_sensor_read_register(0x3109);
	if (value != 0x00)
		goto fail;

    value = sc1135_sensor_read_register(0x3c05);
	if (value != 0x00)
		goto fail;

    value = sc1135_sensor_read_register(0x580b);
	if (value != 0x00)
		goto fail;
#endif

	read_id_ok = 1;

success:
	ak_sensor_print("%s ok\n",__func__);
    return SENSOR_ID;

fail:
	ak_sensor_print("%s fail\n",__func__);
	return 0;
}

static int sc1135_get_resolution(int *width, int *height)
{
	*width = SENSOR_OUTPUT_WIDTH;
	*height = SENSOR_OUTPUT_HEIGHT;

	return 0;
}

static int sc1135_get_mclk(void)
{
	return SENSOR_MCLK;
}

static int sc1135_get_fps(void)
{
	return g_fps;
}

static int sc1135_get_valid_coordinate(int *x, int *y)
{
	*x = SENSOR_VALID_OFFSET_X;
	*y = SENSOR_VALID_OFFSET_Y;

	return 0;
}

static enum sensor_bus_type sc1135_get_bus_type(void)
{
	return SENSROR_BUS_TYPE;
}

static int sc1135_set_power_on(const int pwdn_pin, const int reset_pin)
{  
    ak_sensor_set_pin_as_gpio(pwdn_pin);
    ak_sensor_set_pin_dir(pwdn_pin, 1);
    ak_sensor_set_pin_level(pwdn_pin, !SENSOR_PWDN_LEVEL);    
    ak_sensor_mdelay(10);
    ak_sensor_set_pin_as_gpio(reset_pin);
    ak_sensor_set_pin_dir(reset_pin, 1);
    ak_sensor_set_pin_level(reset_pin, SENSOR_RESET_LEVEL);
    ak_sensor_mdelay(30);   //modify reset time, used to solve no camera interrupt problem
    ak_sensor_set_pin_level(reset_pin, !SENSOR_RESET_LEVEL);
    ak_sensor_mdelay(20);

	g_fps = DETAULT_FPS;
	ak_sensor_print("%s, version: %s\n", __func__, sc1135_version);
	
	return 0;
}

static int sc1135_set_power_off(const int pwdn_pin, const int reset_pin)
{
	//sccb software standby mode
    
    ak_sensor_set_pin_level(pwdn_pin, SENSOR_PWDN_LEVEL);
    ak_sensor_set_pin_dir(reset_pin, 0);

    return 0;
}

static int sc1135_set_fps(const int fps)
{
	int ret = 0;

	switch (fps) {
	case 30:
		sc1135_sensor_write_register(0x320e, 0x03);
		sc1135_sensor_write_register(0x320f, 0xf2);

		sc1135_sensor_write_register(0x3336, (0x3f2 - 0x2e8) >> 8);
		sc1135_sensor_write_register(0x3337, (0x3f2 - 0x2e8) & 0xff);
		sc1135_sensor_write_register(0x3338, 0x03);
		sc1135_sensor_write_register(0x3339, 0xf2);
		break;
	case 25:
		sc1135_sensor_write_register(0x320e, 0x04);
		sc1135_sensor_write_register(0x320f, 0xbc);

		sc1135_sensor_write_register(0x3336, (0x4bc - 0x2e8) >> 8);
		sc1135_sensor_write_register(0x3337, (0x4bc - 0x2e8) & 0xff);
		sc1135_sensor_write_register(0x3338, 0x04);
		sc1135_sensor_write_register(0x3339, 0xbc);
		break;
	case 15:
		sc1135_sensor_write_register(0x320e, 0x07);
		sc1135_sensor_write_register(0x320f, 0xe4);

		sc1135_sensor_write_register(0x3336, (0x7e4 - 0x2e8) >> 8);
		sc1135_sensor_write_register(0x3337, (0x7e4 - 0x2e8) & 0xff);
		sc1135_sensor_write_register(0x3338, 0x07);
		sc1135_sensor_write_register(0x3339, 0xe4);
		break;
	case 13:	//actual: 12.5fps
	case 12:
		sc1135_sensor_write_register(0x320e, 0x09);
		sc1135_sensor_write_register(0x320f, 0x78);

		sc1135_sensor_write_register(0x3336, (0x978 - 0x2e8) >> 8);
		sc1135_sensor_write_register(0x3337, (0x978 - 0x2e8) & 0xff);
		sc1135_sensor_write_register(0x3338, 0x09);
		sc1135_sensor_write_register(0x3339, 0x78);
		break;
	case 10:
		sc1135_sensor_write_register(0x320e, 0x0b);
		sc1135_sensor_write_register(0x320f, 0xd6);

		sc1135_sensor_write_register(0x3336, (0xbd6 - 0x2e8) >> 8);
		sc1135_sensor_write_register(0x3337, (0xbd6 - 0x2e8) & 0xff);
		sc1135_sensor_write_register(0x3338, 0x0b);
		sc1135_sensor_write_register(0x3339, 0xd6);
		break;
	case 8:
		sc1135_sensor_write_register(0x320e, 0x0e);
		sc1135_sensor_write_register(0x320f, 0xcb);

		sc1135_sensor_write_register(0x3336, (0xecb - 0x2e8) >> 8);
		sc1135_sensor_write_register(0x3337, (0xecb - 0x2e8) & 0xff);
		sc1135_sensor_write_register(0x3338, 0x0e);
		sc1135_sensor_write_register(0x3339, 0xcb);
		break;
	case 5:
		sc1135_sensor_write_register(0x320e, 0x17);
		sc1135_sensor_write_register(0x320f, 0xac);

		sc1135_sensor_write_register(0x3336, (0x17ac - 0x2e8) >> 8);
		sc1135_sensor_write_register(0x3337, (0x17ac - 0x2e8) & 0xff);
		sc1135_sensor_write_register(0x3338, 0x17);
		sc1135_sensor_write_register(0x3339, 0xac);
		break;
	default:
		ak_sensor_print("%s set fps fail\n", __func__);
		ret = -EINVAL;
		break;
	}

	if (!ret)
		g_fps = fps;

	return ret;
}

static int sc1135_set_standby_in(const int pwdn_pin, const int reset_pin)
{
	return 0;
}

static int sc1135_set_standby_out(const int pwdn_pin, const int reset_pin)
{
	return 0;
}


static AK_ISP_SENSOR_CB sc1135_callback = 
{
	.sensor_init_func 				= sc1135_init,
	.sensor_read_reg_func			= sc1135_sensor_read_register,
	.sensor_write_reg_func			= sc1135_sensor_write_register,
	.sensor_read_id_func			= sc1135_read_id,
	.sensor_update_a_gain_func		= sc1135_cmos_updata_a_gain,
	.sensor_update_d_gain_func		= sc1135_cmos_updata_d_gain,
	.sensor_updata_exp_time_func	= sc1135_cmos_updata_exp_time,
	.sensor_timer_func				= sc1135_cmos_update_a_gain_timer,

	.sensor_probe_id_func			= sc1135_probe_id,	//use IIC bus
	.sensor_get_resolution_func		= sc1135_get_resolution,
	.sensor_get_mclk_func			= sc1135_get_mclk,
	.sensor_get_fps_func			= sc1135_get_fps,
	.sensor_get_valid_coordinate_func	= sc1135_get_valid_coordinate,
	.sensor_get_bus_type_func		= sc1135_get_bus_type,

	.sensor_set_power_on_func		= sc1135_set_power_on,
	.sensor_set_power_off_func		= sc1135_set_power_off,
	.sensor_set_fps_func			= sc1135_set_fps,
	.sensor_set_standby_in_func		= sc1135_set_standby_in,
	.sensor_set_standby_out_func	= sc1135_set_standby_out
};

AK_SENSOR_MODULE(sc1135_callback, sc1135)

///////////////////////////////////////////////////////////////////////////////////////////

//for SC1135
#define SENSOR_BLC_TOP_VALUE (0x58)
#define SENSOR_BLC_BOT_VALUE (0x45)
#define SENSOR_AGAIN_ADAPT_STEP (1)
#define SENSOR_MAX_AGAIN (0x7f)
static const unsigned short sensor_gain_map[] = {
	0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08, 0x09, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e, 0x0f, 
	0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17, 0x18, 0x19, 0x1a, 0x1b, 0x1c, 0x1d, 0x1e, 0x1f, 
	0x20, 0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27, 0x28, 0x29, 0x2a, 0x2b, 0x2c, 0x2d, 0x2e, 0x2f, 
	0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37, 0x38, 0x39, 0x3a, 0x3b, 0x3c, 0x3d, 0x3e, 0x3f, 
	0x40, 0x41, 0x42, 0x43, 0x44, 0x45, 0x46, 0x47, 0x48, 0x49, 0x4a, 0x4b, 0x4c, 0x4d, 0x4e, 0x4f, 
	0x50, 0x51, 0x52, 0x53, 0x54, 0x55, 0x56, 0x57, 0x58, 0x59, 0x5a, 0x5b, 0x5c, 0x5d, 0x5e, 0x5f, 
	0x60, 0x61, 0x62, 0x63, 0x64, 0x65, 0x66, 0x67, 0x68, 0x69, 0x6a, 0x6b, 0x6c, 0x6d, 0x6e, 0x6f, 
	0x70, 0x71, 0x72, 0x73, 0x74, 0x75, 0x76, 0x77, 0x78, 0x79, 0x7a, 0x7b, 0x7c, 0x7d, 0x7e, 0x7f
};

static unsigned short gainmap_gain2index(unsigned short Gain)
{
	unsigned short i = 0;
	for(i = 0; i < sizeof(sensor_gain_map)/sizeof(sensor_gain_map[0]); i++){
		if(sensor_gain_map[i] == Gain){
			break;
		}
	}
	return i;
}

//return limited Again
//param:Again that ISP calculated
unsigned int Again_limit(unsigned int Again)
{
	unsigned int ret_Again;
	unsigned short BLC_top = SENSOR_BLC_TOP_VALUE, BLC_bot = SENSOR_BLC_BOT_VALUE, BLC_reg/*, Again_step = SENSOR_AGAIN_ADAPT_STEP*/, gain_index;
	static unsigned short MaxAgain = SENSOR_MAX_AGAIN;

	//I2C read reg of sensor
	BLC_reg = sc1135_sensor_read_register(0x3911);

	gain_index = gainmap_gain2index(MaxAgain);

	if(BLC_reg > BLC_top){//>0x58
		if(gain_index>0){
			//limit max Again by step
			gain_index -= SENSOR_AGAIN_ADAPT_STEP;
		}
		MaxAgain = sensor_gain_map[gain_index];
	}else if(BLC_reg < BLC_bot){//<0x45
		//release Again limit by step
		gain_index += SENSOR_AGAIN_ADAPT_STEP;
		if(gain_index > sizeof(sensor_gain_map)/sizeof(sensor_gain_map[0])-1){
			gain_index = sizeof(sensor_gain_map)/sizeof(sensor_gain_map[0]) - 1;
		}
		MaxAgain = sensor_gain_map[gain_index];
	}else{//0x45 < BLC_reg < 0x58
		//do nothing
	}
	ret_Again = Again > MaxAgain ? MaxAgain : Again;
	//ak_sensor_print("limit gain:ret_Again=%d;Again=%d;MaxAgain=%d\n", ret_Again, Again, MaxAgain);
	return ret_Again;
}

#define SKIP_FRAME_NUM	2
static int sc1135_curr_2x_again = 0;
static unsigned int _cmos_gains_convert(unsigned int a_gain, unsigned int d_gain,
		unsigned int *a_gain_out ,unsigned int *d_gain_out)
{
	//ģ??????????ת?????????????????????ת????sensor?????????

	unsigned int  fine_gain= 0;
	unsigned int ag_value= 0x00;
	unsigned int corse_gain=0,sensor_d_gain = 0,sensor_2x_again=0;
	int tmp = 0;
	//ak_sensor_print("a_gain =%d\n",a_gain);
	if(a_gain<=1980)                           //1 - 1.8824
	{
		corse_gain = 0;
		tmp = a_gain-1024;
    }
    else if(a_gain>1980 && a_gain<=3960)       // 2 - 3.7648
    {
    	corse_gain = 1;
		tmp = a_gain/2-1024;
    }
    else if(a_gain>3960 && a_gain<=7920)       //4-7.5296
    {
    	corse_gain = 3;
		tmp = a_gain/4-1024;
    }
    else if(a_gain>7920 && a_gain<=15840)       //8 - 15.0592
    {
        corse_gain = 7;
		tmp = a_gain/8-1024;
    }
    else if(a_gain>15840 && a_gain<=31680)    // 16 - 30.11
    {
    	corse_gain = 7;
		sensor_2x_again =1;
		a_gain=a_gain>>1;
		tmp = a_gain/8-1024;
    }
    else if(a_gain>31680)   				// 30 - 60.23
    {
    	corse_gain = 7;
		sensor_2x_again =1;
		a_gain = 15840;
		tmp = 956;
    }
    
    if(tmp>=0){
		fine_gain = tmp/60;
	}
	else{			
		fine_gain = 0;
	}
	
	*a_gain_out = a_gain;
	*d_gain_out = sensor_d_gain;
    ag_value = corse_gain<<4|(fine_gain&0xf);
	
	//ak_sensor_print("corse=%d fine=%d\n",corse_gain,fine_gain);
	sc1135_sensor_write_register(0x3e09, Again_limit(ag_value));
	//sc1135_sensor_write_register(0x3e08, sensor_d_gain);
	if(sensor_2x_again==1)
		sc1135_sensor_write_register(0x3635, 0x0c);		//2x again
	else
		sc1135_sensor_write_register(0x3635, 0x0);		
		
	if(corse_gain<1)      //2 
	{
		sc1135_sensor_write_register(0x3630, 0xb8);
		sc1135_sensor_write_register(0x3631, 0x82);
	}
	else if(sensor_2x_again<1)      //2-16
	{
		sc1135_sensor_write_register(0x3630, 0x70);
		sc1135_sensor_write_register(0x3631, 0x8e);
	}
	else
	{
		sc1135_sensor_write_register(0x3630, 0x70);
		sc1135_sensor_write_register(0x3631, 0x8c);
	}
	//ak_sensor_print(" 0x3e09 = %d\n", sc1135_sensor_read_register(0x3e09));
	//ak_sensor_print(" 0x3e0f = %d\n", sc1135_sensor_read_register(0x360f));
	
	if(sc1135_curr_2x_again!=sensor_2x_again)
	{
		sc1135_curr_2x_again=sensor_2x_again;
		return SKIP_FRAME_NUM;
	}
	else 
		return 0;
}

static int sc1135_cmos_updata_d_gain(const unsigned int d_gain)
{
	//?????????????Ļص?????
	return 0;
}

static int sc1135_cmos_updata_a_gain(const unsigned int a_gain)
{
	//??????ģ???????Ļص?????
	unsigned int tmp_a_gain;
	unsigned int tmp_d_gain;
	unsigned int tmp_a_gain_out;
	unsigned int tmp_d_gain_out;

	g_a_gain = a_gain;

	tmp_a_gain = a_gain<<2;
	tmp_d_gain = 0;

	return _cmos_gains_convert(tmp_a_gain, tmp_d_gain,  &tmp_a_gain_out ,&tmp_d_gain_out);
	
	
	//ak_sensor_print(" 0x3e09 = %d\n", sc1135_sensor_read_register(0x3e09));
	//ak_sensor_print(" 0x3610 = %d\n", sc1135_sensor_read_register(0x3610));
	//ak_sensor_print(" 0x3605 = %d\n", sc1135_sensor_read_register(0x3605));
	//ak_sensor_print(" tmp_a_gain = %d ",tmp_a_gain);
	//ak_sensor_print(" tmp_a_gain_out = %d\n", tmp_a_gain_out);
	//updata_auto_exposure_num++;
	//isp->aec_param.current_exposure_time = tmp_exposure_time;
	//return 0;
}


static int sc1135_cmos_updata_exp_time(unsigned int exp_time)
{
    //?????ع?ʱ???Ļص?????	
    unsigned char exposure_time_msb;
	unsigned char exposure_time_lsb;
	unsigned char tem;
	unsigned int  ex_time_max, reg_value_hight, reg_value_low;
 	//ak_sensor_print("exp_time=%d ",exp_time);
    exposure_time_msb =(exp_time>>4)&0xff;
    exposure_time_lsb =((exp_time)&0xf)<<4;
    tem=sc1135_sensor_read_register(0x3e02)&0x0f;
    exposure_time_lsb =exposure_time_lsb|tem;
	//ak_sensor_print("msb = %d  ",exposure_time_msb);
	//ak_sensor_print("lsb = %d\n",exposure_time_lsb);
	sc1135_sensor_write_register(0x3e01,exposure_time_msb);
	sc1135_sensor_write_register(0x3e02,exposure_time_lsb);

	reg_value_hight = sc1135_sensor_read_register(0x320e);
	reg_value_low = sc1135_sensor_read_register(0x320f);
	ex_time_max = ((reg_value_hight<<8)|reg_value_low)-4;
	/*
	if(exp_time<=ex_time_max-300)
	{
		sc1135_sensor_write_register(0x331e,0xe0);
	}
	else
	{
		sc1135_sensor_write_register(0x331e,224-(exp_time-(ex_time_max-300))*198/300);
	}
	*/
    return 0;
}

static int sc1135_cmos_update_a_gain_timer(void)
{
	#define UPDATE_A_GAIN_TIMER_PERIOD	(5)	//s
	static struct timeval stv = {.tv_sec = 0,};
	struct timeval tv;

	do_gettimeofday(&tv);

	if (tv.tv_sec >= stv.tv_sec + UPDATE_A_GAIN_TIMER_PERIOD) {
		stv.tv_sec = tv.tv_sec;
		sc1135_cmos_updata_a_gain(g_a_gain);
	}

	return 0;
}
