/**
 * @file camera_sc1145.c
 * @brief camera driver file
 * Copyright (C) 2011 Anyka (Guangzhou) Microelectronics Technology Co., Ltd
 * @author luweichun
 * @date 2015-11-05
 * @version 1.0
 * @ref
 */ 
#include "ak_sensor_common.h"

#define SENSOR_PWDN_LEVEL		1 
#define SENSOR_RESET_LEVEL		0
#define SENSOR_I2C_ADDR         0x60 
#define SENSOR_ID               0x1145
#define SENSOR_MCLK             24
#define SENSOR_REGADDR_BYTE     2
#define SENSOR_DATA_BYTE        1
#define MAX_FPS					25
#define SENSOR_OUTPUT_WIDTH		1280
#define SENSOR_OUTPUT_HEIGHT	720
#define SENSOR_VALID_OFFSET_X	0
#define SENSOR_VALID_OFFSET_Y	0
#define SENSROR_BUS_TYPE		BUS_TYPE_RAW

#define OV_MAX(a, b)            (((a) < (b) ) ?  (b) : (a))
#define OV_MIN(a, b)            (((a) > (b) ) ?  (b) : (a))
#define OV_CLIP3(low, high, x)  (OV_MAX(OV_MIN((x), high), low))

static int sc1145_cmos_updata_d_gain(const unsigned int d_gain);
static int sc1145_cmos_updata_a_gain(const unsigned int a_gain);
static int sc1145_cmos_updata_exp_time(unsigned int exp_time);
static int sc1145_cmos_update_a_gain_timer(void);


static int g_fps = MAX_FPS;
static int g_a_gain = 0;

static int sc1145_sensor_read_register(int reg)
{
	struct ak_sensor_i2c_data i2cdata;

	i2cdata.u8DevAddr           = SENSOR_I2C_ADDR;
	i2cdata.u32RegAddr          = reg;
	i2cdata.u32RegAddrByteNum   = SENSOR_REGADDR_BYTE;
	i2cdata.u32Data             = 0;
	i2cdata.u32DataByteNum      = SENSOR_DATA_BYTE;

	return ak_sensor_read_register(&i2cdata);
}

static int sc1145_sensor_write_register(int reg, int data)
{
	struct ak_sensor_i2c_data i2cdata;

	i2cdata.u8DevAddr           = SENSOR_I2C_ADDR;
	i2cdata.u32RegAddr          = reg;
	i2cdata.u32RegAddrByteNum   = SENSOR_REGADDR_BYTE;
	i2cdata.u32Data             = data;
	i2cdata.u32DataByteNum      = SENSOR_DATA_BYTE;

	return ak_sensor_write_register(&i2cdata);
}

static int sc1145_init(const AK_ISP_SENSOR_INIT_PARA *para)
{
	int i;
	AK_ISP_SENSOR_REG_INFO *preg_info;

	preg_info = para->reg_info;
	for (i = 0; i < para->num; i++) {
//value=sc1145_sensor_read_register(preg_info->reg_addr);
//ak_sensor_print("before w reg:0x%.x, val:0x%.x, to write:0x%.x\n", preg_info->reg_addr, value, preg_info->value);
		sc1145_sensor_write_register(preg_info->reg_addr, preg_info->value);
//value=sc1145_sensor_read_register(preg_info->reg_addr);
//ak_sensor_print("after w reg:0x%.x, val:0x%.x\n", preg_info->reg_addr, value);
		preg_info++;
	}

	return 0;
}

static int sc1145_read_id(void)
{
    return SENSOR_ID;
}

static int sc1145_probe_id(void)
{
    u8 value;
    u32 id;

    value = sc1145_sensor_read_register(0x3107);
    id = value << 8;

    value = sc1145_sensor_read_register(0x3108);
    id |= value;

	//ak_sensor_print("%s id:0x%x\n", __func__, id);
	if (id != SENSOR_ID)
		goto fail;

    value = sc1145_sensor_read_register(0x3109);
	//ak_sensor_print("%s reg[0x3109]:0x%x\n", __func__, value);
	if (value != 0x01)
		goto fail;
            
    return SENSOR_ID;

fail:
	return 0;
}

static int sc1145_get_resolution(int *width, int *height)
{
	*width = SENSOR_OUTPUT_WIDTH;
	*height = SENSOR_OUTPUT_HEIGHT;

	return 0;
}

static int sc1145_get_mclk(void)
{
	return SENSOR_MCLK;
}

static int sc1145_get_fps(void)
{
	return g_fps;
}

static int sc1145_get_valid_coordinate(int *x, int *y)
{
	*x = SENSOR_VALID_OFFSET_X;
	*y = SENSOR_VALID_OFFSET_Y;

	return 0;
}

static enum sensor_bus_type sc1145_get_bus_type(void)
{
	return SENSROR_BUS_TYPE;
}

static int sc1145_set_power_on(const int pwdn_pin, const int reset_pin)
{
    ak_sensor_set_pin_as_gpio(pwdn_pin);
    ak_sensor_set_pin_dir(pwdn_pin, 1);
    ak_sensor_set_pin_level(pwdn_pin, !SENSOR_PWDN_LEVEL);    
    ak_sensor_mdelay(10);
    ak_sensor_set_pin_as_gpio(reset_pin);
    ak_sensor_set_pin_dir(reset_pin, 1);
    ak_sensor_set_pin_level(reset_pin, SENSOR_RESET_LEVEL);
    ak_sensor_mdelay(30);   //modify reset time, just in case 1145 had the same problem with 1135
    ak_sensor_set_pin_level(reset_pin, !SENSOR_RESET_LEVEL);
    ak_sensor_mdelay(20);

	g_fps = MAX_FPS;
	ak_sensor_print("%s\n",__func__);
	return 0;
}

static int sc1145_set_power_off(const int pwdn_pin, const int reset_pin)
{
	//sccb software standby mode
    
    ak_sensor_set_pin_level(pwdn_pin, SENSOR_PWDN_LEVEL);
    ak_sensor_set_pin_dir(reset_pin, 0);

    return 0;
}

static int sc1145_set_fps(int fps)
{
	int ret = 0;

	switch (fps) {
	case 30:
		sc1145_sensor_write_register(0x320e, 0x02);
		sc1145_sensor_write_register(0x320f, 0xee);
		sc1145_sensor_write_register(0x3338, 0x02);
		sc1145_sensor_write_register(0x3339, 0xee);
		sc1145_sensor_write_register(0x3336, 0x00);
		sc1145_sensor_write_register(0x3337, 0x00);
		break;
	case 25:
		sc1145_sensor_write_register(0x320e, 0x03);
		sc1145_sensor_write_register(0x320f, 0x84);
		sc1145_sensor_write_register(0x3338, 0x03);
		sc1145_sensor_write_register(0x3339, 0x84);
		sc1145_sensor_write_register(0x3336, 0x00);
		sc1145_sensor_write_register(0x3337, 0x96);
		break;
	case 15:
		sc1145_sensor_write_register(0x320e, 0x05);
		sc1145_sensor_write_register(0x320f, 0xdc);
		sc1145_sensor_write_register(0x3338, 0x05);
		sc1145_sensor_write_register(0x3339, 0xdc);
		sc1145_sensor_write_register(0x3336, 0x02);
		sc1145_sensor_write_register(0x3337, 0xee);
		break;
	case 13:	//actual: 12.5fps
	case 12:
		sc1145_sensor_write_register(0x320e, 0x07);
		sc1145_sensor_write_register(0x320f, 0x08);
		sc1145_sensor_write_register(0x3338, 0x07);
		sc1145_sensor_write_register(0x3339, 0x08);
		sc1145_sensor_write_register(0x3336, 0x04);
		sc1145_sensor_write_register(0x3337, 0x1a);
		break;
	case 10:
		sc1145_sensor_write_register(0x320e, 0x08);
		sc1145_sensor_write_register(0x320f, 0xca);
		sc1145_sensor_write_register(0x3338, 0x08);
		sc1145_sensor_write_register(0x3339, 0xca);
		sc1145_sensor_write_register(0x3336, 0x05);
		sc1145_sensor_write_register(0x3337, 0xdc);
		break;
	case 8:
		sc1145_sensor_write_register(0x320e, 0x0a);
		sc1145_sensor_write_register(0x320f, 0x8c);
		sc1145_sensor_write_register(0x3338, 0x0a);
		sc1145_sensor_write_register(0x3339, 0x8c);
		sc1145_sensor_write_register(0x3336, 0x07);
		sc1145_sensor_write_register(0x3337, 0x9e);
		break;

	default:
		ak_sensor_print("%s set fps fail\n", __func__);
		ret = -EINVAL;
		break;
	}

	if (!ret)
		g_fps = fps;

	return ret;
}

static int sc1145_set_standby_in(const int pwdn_pin, const int reset_pin)
{
	return 0;
}

static int sc1145_set_standby_out(const int pwdn_pin, const int reset_pin)
{
	return 0;
}

static AK_ISP_SENSOR_CB sc1145_callback = 
{
	.sensor_init_func 				= sc1145_init,
	.sensor_read_reg_func			= sc1145_sensor_read_register,
	.sensor_write_reg_func			= sc1145_sensor_write_register,
	.sensor_read_id_func			= sc1145_read_id,
	.sensor_update_a_gain_func		= sc1145_cmos_updata_a_gain,
	.sensor_update_d_gain_func		= sc1145_cmos_updata_d_gain,
	.sensor_updata_exp_time_func	= sc1145_cmos_updata_exp_time,
	.sensor_timer_func				= sc1145_cmos_update_a_gain_timer,

	.sensor_probe_id_func			= sc1145_probe_id,	//use IIC bus
	.sensor_get_resolution_func		= sc1145_get_resolution,
	.sensor_get_mclk_func			= sc1145_get_mclk,
	.sensor_get_fps_func			= sc1145_get_fps,
	.sensor_get_valid_coordinate_func	= sc1145_get_valid_coordinate,
	.sensor_get_bus_type_func		= sc1145_get_bus_type,

	.sensor_set_power_on_func		= sc1145_set_power_on,
	.sensor_set_power_off_func		= sc1145_set_power_off,
	.sensor_set_fps_func			= sc1145_set_fps,
	.sensor_set_standby_in_func		= sc1145_set_standby_in,
	.sensor_set_standby_out_func	= sc1145_set_standby_out
};

AK_SENSOR_MODULE(sc1145_callback, sc1145)

///////////////////////////////////////////////////////////////////////////////////////////
static int a_gain_2x_now=1;
static int a_gain_4x_now=1;
static int a_gain_2x_prev=1;
static int a_gain_4x_prev=1;
//static int a_gain_2x_4x_locked=0;
static unsigned int _cmos_gains_convert(unsigned int a_gain, unsigned int d_gain,
		unsigned int *a_gain_out ,unsigned int *d_gain_out)
{
	//ģ??????????ת???????????㷨??????????ת????sensor????????ʽ
	unsigned int ag_interget_table[61] = {0,  1, 2, 2, 4, 4, 4, 4, 8, 8, 8, 8,
        8, 8, 8, 8,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,32,32,32,32,
        32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32,32};
    unsigned int ag_integer, ag_fraction, tmp;
	unsigned int ag_return;
 	tmp =  a_gain>>4;
	if(tmp == 0)
	{
		tmp = 1;
		ak_sensor_print("Div ERROR 000!\n");
	}
	//ak_sensor_print(" a_gain = %d ",a_gain);
    tmp =  OV_CLIP3(0, 60, tmp);
    ag_integer =  ag_interget_table[tmp];
    ag_fraction = (a_gain*17)/(16*ag_integer)- 17;
    ag_fraction =  OV_CLIP3(0, 15, ag_fraction);
    //ak_sensor_print(" ag_integer = %d ",ag_integer);
	//ak_sensor_print(" ag_fraction = %d\n", ag_fraction);
    
    switch (ag_integer)
    {
        case 1 :
            //ak_sensor_print("case 1\n");
        	//sc1145_sensor_write_register(0x3610,0x03);
        	//sc1145_sensor_write_register(0x3605,0x0c);
            ag_integer = 0x00;
			a_gain_2x_now = 1;
			a_gain_4x_now = 1;
            break;
        case 2 :
        	//ak_sensor_print("case 2\n");
        	//sc1145_sensor_write_register(0x3610,0x03);
        	//sc1145_sensor_write_register(0x3605,0x0c);
            ag_integer = 0x10;
			a_gain_2x_now = 1;
			a_gain_4x_now = 1;
            break;
        case 4 :
        	//ak_sensor_print("case 4\n");
        	//sc1145_sensor_write_register(0x3610,0x03);
        	//sc1145_sensor_write_register(0x3605,0x0c);
            ag_integer = 0x30;
			a_gain_2x_now = 1;
			a_gain_4x_now = 1;
            break;
        case 8 :
        	//ak_sensor_print("case 8\n");
        	//sc1145_sensor_write_register(0x3610,0x03);
        	//sc1145_sensor_write_register(0x3605,0x0c);
            ag_integer = 0x70;
			a_gain_2x_now = 1;
			a_gain_4x_now = 1;
            break;
        case 16 :
        	//ak_sensor_print("case 16\n");
        	//sc1145_sensor_write_register(0x3610,0x1b);
        	//sc1145_sensor_write_register(0x3605,0x0c);
            ag_integer = 0x70;
			a_gain_2x_now = 2;
			a_gain_4x_now = 1;
            break;
        case 32 :
        	//ak_sensor_print("case 32\n");
       	 	//sc1145_sensor_write_register(0x3610,0x1b);
       	 	//sc1145_sensor_write_register(0x3605,0x3c);
            ag_integer = 0x70;
			a_gain_2x_now = 2;
			a_gain_4x_now = 2;
            break;
        default:
        	//ak_sensor_print("case default\n");
        	//sc1145_sensor_write_register(0x3610,0x03);
        	//sc1145_sensor_write_register(0x3605,0x0c);
            ag_integer = 0x00;
            break;
    }
    //ak_sensor_print("ag_integer=%x ag_fraction=%x a_gain_2x_now=%x, a_gain_4x_now=%x\n",
	//	  ag_integer,ag_fraction,a_gain_2x_now,a_gain_4x_now);
    ag_return = ag_integer|ag_fraction;
	*a_gain_out = ag_return ; 
 
	return ag_return;
}

static int sc1145_cmos_updata_d_gain(const unsigned int d_gain)
{
	//?????????????Ļص?????
	return 0;
}

/*
static int sc1145_updata_2x_4x_a_gain(void)
{
     //a_gain_2x_now = 2;
	// a_gain_4x_now = 2;
	int all_2x_4x_gain=0;
	    a_gain_2x_4x_locked=0;
	 if((a_gain_2x_prev==1)&&(a_gain_2x_now==2))
	 {
		a_gain_2x_4x_locked =AK_TRUE;
		all_2x_4x_gain = a_gain_2x_prev*a_gain_4x_now;
	 }		
	 else if((a_gain_4x_prev==1)&&(a_gain_4x_now==2))
	 {
		a_gain_2x_4x_locked = AK_TRUE;
		all_2x_4x_gain = a_gain_2x_now*a_gain_4x_prev;
	 }
	 else if((a_gain_2x_prev==2)&&(a_gain_2x_now==1))
	 {
		a_gain_2x_4x_locked = AK_TRUE;
		all_2x_4x_gain = a_gain_2x_prev*a_gain_4x_now;
	 }
	 else if((a_gain_4x_prev==2)&&(a_gain_4x_now==1))
	 {
		a_gain_2x_4x_locked = AK_TRUE;
		all_2x_4x_gain = a_gain_2x_now*a_gain_4x_prev;
	 }
	 
	   if(a_gain_2x_4x_locked!=AK_TRUE)
	   {
	     all_2x_4x_gain=a_gain_2x_now*a_gain_4x_now;
	     switch(all_2x_4x_gain)
	     {
		 	case 1:
			sc1145_sensor_write_register(0x3610,0x03);
        	sc1145_sensor_write_register(0x3605,0x0c);
		
			  break;
			case 2:
			sc1145_sensor_write_register(0x3610,0x1b);
        	sc1145_sensor_write_register(0x3605,0x0c);
			
			  break;
			case 4:
			sc1145_sensor_write_register(0x3610,0x1b);
       	 	sc1145_sensor_write_register(0x3605,0x3c);
			
			   break;
			default:
			   break;
	     }
	   
	   }
	   else   //дȥֵһεֵ;
	   {
	     //all_2x_4x_gain=a_gain_2x_now*a_gain_4x_now;
	     switch(all_2x_4x_gain)
	     {
		 	case 1:
			sc1145_sensor_write_register(0x3610,0x03);
        	sc1145_sensor_write_register(0x3605,0x0c);
			
			  break;
			case 2:
			sc1145_sensor_write_register(0x3610,0x1b);
        	sc1145_sensor_write_register(0x3605,0x0c);
		
			  break;
			case 4:
			sc1145_sensor_write_register(0x3610,0x1b);
       	 	sc1145_sensor_write_register(0x3605,0x3c);
			
			   break;
			default:
			   break;
	     }
	      
	   }

	 a_gain_2x_prev = a_gain_2x_now;
	 a_gain_4x_prev = a_gain_4x_now;
	 return 0;	 
}*/
static int sc1145_updata_2x_4x_a_gain(void)
{
    int   all_2x_4x_gain;
	   {
	     all_2x_4x_gain=a_gain_2x_now*a_gain_4x_now;
	     switch(all_2x_4x_gain)
	     {
		 	case 1:
			sc1145_sensor_write_register(0x3610,0x03);
        	sc1145_sensor_write_register(0x3605,0x0c);
		
			  break;
			case 2:
			sc1145_sensor_write_register(0x3610,0x1b);
        	sc1145_sensor_write_register(0x3605,0x0c);
			
			  break;
			case 4:
			sc1145_sensor_write_register(0x3610,0x1b);
       	 	sc1145_sensor_write_register(0x3605,0x3c);
			
			   break;
			default:
			   break;
	     }
	   
	   }
	   
	 a_gain_2x_prev = a_gain_2x_now;
	 a_gain_4x_prev = a_gain_4x_now;
	 return 0;	 
}
//for SC1145
#define SENSOR_BLC_TOP_VALUE (0x1c)
#define SENSOR_BLC_BOT_VALUE (0x19)
#define SENSOR_AGAIN_ADAPT_STEP (1)
#define SENSOR_MAX_AGAIN (0x7f)
static const unsigned short sensor_gain_map[48] = {
	0x10, 0x11, 0x12, 0x13, 0x14,0x15, 0x16, 0x17, 0x18, 0x19,
	0x1a, 0x1b, 0x1c,0x1d, 0x1e, 0x1f, 0x30, 0x31,
	0x32, 0x33, 0x34, 0x35,0x36, 0x37, 0x38, 0x39,
	0x3a, 0x3b, 0x3c, 0x3d,0x3e, 0x3f, 0x70, 0x71,
	0x72, 0x73, 0x74, 0x75,0x76, 0x77, 0x78, 0x79,
	0x7a, 0x7b, 0x7c, 0x7d,0x7e, 0x7f
};

static unsigned short gainmap_gain2index(unsigned short Gain)
{
	unsigned short i = 0;
	for(i = 0; i < sizeof(sensor_gain_map)/sizeof(sensor_gain_map[0]); i++){
		if(sensor_gain_map[i] == Gain){
			break;
		}
	}
	return i;
}

//return limited Again
//param:Again that ISP calculated
unsigned int Again_limit(unsigned int Again)
{
	unsigned int ret_Again;
	unsigned short BLC_top = SENSOR_BLC_TOP_VALUE, BLC_bot = SENSOR_BLC_BOT_VALUE, BLC_reg/*, Again_step = SENSOR_AGAIN_ADAPT_STEP*/, gain_index;
	static unsigned short MaxAgain = SENSOR_MAX_AGAIN;

	//I2C read reg of sensor
	BLC_reg = sc1145_sensor_read_register(0x3911);

	gain_index = gainmap_gain2index(MaxAgain);

	if(BLC_reg > BLC_top){//>0x58
		if(gain_index>0){
			//limit max Again by step
			gain_index -= SENSOR_AGAIN_ADAPT_STEP;
		}
		MaxAgain = sensor_gain_map[gain_index];
	}else if(BLC_reg < BLC_bot){//<0x45
		//release Again limit by step
		gain_index += SENSOR_AGAIN_ADAPT_STEP;
		if(gain_index > sizeof(sensor_gain_map)/sizeof(sensor_gain_map[0])-1){
			gain_index = sizeof(sensor_gain_map)/sizeof(sensor_gain_map[0]) - 1;
		}
		MaxAgain = sensor_gain_map[gain_index];
	}else{//0x45 < BLC_reg < 0x58
		//do nothing
	}
	ret_Again = Again > MaxAgain ? MaxAgain : Again;
	//ak_sensor_print("limit gain:ret_Again=%d;Again=%d;MaxAgain=%d\n", ret_Again, Again, MaxAgain);
	return ret_Again;
}

#define SKIP_FRAME_NUM	2
static int sc1145_cmos_updata_a_gain(const unsigned int a_gain)
{
	//??????ģ???????Ļص?????
	unsigned short ag_value;
	unsigned int tmp_a_gain;
	unsigned int tmp_d_gain;
	unsigned int tmp_a_gain_out;
	unsigned int tmp_d_gain_out;
	int ret = 0;
	
	g_a_gain  = a_gain;

	tmp_a_gain = a_gain>>4;
	tmp_d_gain = 0;

	//ag_value = cmos_gains_update(isp,isp->aec_param.current_a_gain);
	_cmos_gains_convert(tmp_a_gain, tmp_d_gain,  &tmp_a_gain_out ,&tmp_d_gain_out);
	
	if(a_gain_2x_prev!=a_gain_2x_now || a_gain_4x_now!=a_gain_4x_prev)
		ret = SKIP_FRAME_NUM;
	ag_value = tmp_a_gain_out;
	sc1145_updata_2x_4x_a_gain();
	sc1145_sensor_write_register(0x3e09, Again_limit(tmp_a_gain_out));
	//ak_sensor_print(" 0x3e09 = %d\n", sc1145_sensor_read_register(0x3e09));
	//ak_sensor_print(" 0x3610 = %d\n", sc1145_sensor_read_register(0x3610));
	//ak_sensor_print(" 0x3605 = %d\n", sc1145_sensor_read_register(0x3605));
	//ak_sensor_print(" tmp_a_gain = %d ",tmp_a_gain);
	//ak_sensor_print(" tmp_a_gain_out = %d\n", tmp_a_gain_out);
	//updata_auto_exposure_num++;
	//isp->aec_param.current_exposure_time = tmp_exposure_time;
	return ret;
}


static int sc1145_cmos_updata_exp_time(unsigned int exp_time)
{
    //?????ع?ʱ???Ļص?????
    unsigned char exposure_time_msb;
	unsigned char exposure_time_lsb;
	unsigned char tem;
 	//ak_sensor_print("exp_time=%d ",exp_time);
    exposure_time_msb =(exp_time>>4)&0xff;
    exposure_time_lsb =((exp_time)&0xf)<<4;
    tem=sc1145_sensor_read_register(0x3e02)&0x0f;
    exposure_time_lsb =exposure_time_lsb|tem;
	//ak_sensor_print("msb = %d  ",exposure_time_msb);
	//ak_sensor_print("lsb = %d\n",exposure_time_lsb);
	sc1145_sensor_write_register(0x3e01,exposure_time_msb);
	sc1145_sensor_write_register(0x3e02,exposure_time_lsb);
    return 0;
}

static int sc1145_cmos_update_a_gain_timer(void)
{
	#define UPDATE_A_GAIN_TIMER_PERIOD	(5)	//s
	static struct timeval stv = {.tv_sec = 0,};
	struct timeval tv;

	do_gettimeofday(&tv);

	if (tv.tv_sec >= stv.tv_sec + UPDATE_A_GAIN_TIMER_PERIOD) {
		stv.tv_sec = tv.tv_sec;
		sc1145_cmos_updata_a_gain(g_a_gain);
	}

	return 0;
}
