 /**
 *  @file      /driver/mtd/devices/ak_spi_nand.c
 *  @brief     SPI Flash driver for Anyka AK39e platform.
 *  Copyright C 2016 Anyka CO.,LTD
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *  @author    luoyongchuang 
 *  @date      2016-05-17
 */


#include <linux/init.h>
#include <linux/module.h>
#include <linux/device.h>
#include <linux/interrupt.h>
#include <linux/mutex.h>
#include <linux/math64.h>
#include <linux/sched.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/partitions.h>

#include <linux/spi/spi.h>
#include <linux/spi/flash.h>
#include <linux/delay.h>
#include <mach-anyka/anyka_types.h>
#include <linux/mm.h>

//#define SPINAND_DEBUG  
      
#ifdef SPINAND_DEBUG
#define PDEBUG(fmt, args...) printk( KERN_ALERT fmt,## args)
#define DEBUG(n, args...) printk(KERN_ALERT args)
#else
#define PDEBUG(fmt, args...)
#define DEBUG(n, args...) 
#endif

#define SPI_FLASH_READ		1
#define SPI_FLASH_WRITE		2

//#define CONFIG_SPINAND_USE_FAST_READ 1

/*mtd layer allocate memory use for 'vmalloc' interface, need to convert.*/
#define SPINAND_USE_MTD_BLOCK_LAYER  

#define OPCODE_RESET            0xff   
#define OPCODE_WREN             0x06    /* Write Enable */ 
#define OPCODE_WRDI             0x04    /* Write Disable */ 

#define OPCODE_RDSR1            0x0f    /* Read Status Register1 */
#define OPCODE_RDSR2            0x35    /* Read Status Register2 */ 
#define OPCODE_RDSR3            0x15    /* Read Status Register3 */ 
#define OPCODE_WRSR1          	0x1f    /* Write Status Register */ 
#define OPCODE_WRSR2          	0x31    /* Write Status2 Register*/ 
#define OPCODE_WRSR3          	0x11    /* Write Status3 Register*/ 

#define OPCODE_READ_TO_CACHE  	0x13
#define OPCODE_NORM_READ        0x03    /* Read Data Bytes */ 
#define OPCODE_FAST_READ      	0x0b    /* Read Data Bytes at Higher Speed */ 
#define OPCODE_FAST_D_READ     	0x3b    /* Read Data Bytes at Dual output */ 
#define OPCODE_FAST_Q_READ     	0x6b    /* Read Data Bytes at Quad output */ 
#define OPCODE_FAST_D_IO        0xbb    /* Read Data Bytes at Dual i/o */ 
#define OPCODE_FAST_Q_IO        0xeb    /* Read Data Bytes at Quad i/o */ 

#define OPCODE_PP               0x02    /* Page Program */
#define OPCODE_PP_DUAL          0x12		/* Dual Page Program*/
#define OPCODE_PP_QUAD          0x32		/* Quad Page Program*/
#define OPCODE_2IO_PP           0x18		/* 2I/O Page Program (tmp)*/
#define OPCODE_4IO_PP           0x38		/* 4I/O Page Program*/
#define OPCODE_PP_EXEC          0x10

#define OPCODE_BE_4K            0x20    /* Sector (4K) Erase */ 
#define OPCODE_BE_32K           0x52    /* Block (32K) Erase */
#define OPCODE_BE_64K          	0xd8    /* Block (64K) Erase */ 
#define	OPCODE_SE               0xd8		/* Sector erase (usually 64KiB) */
#define OPCODE_ERASE_BLOCK      0xd8    /* block Erase */ 
#define	OPCODE_RDID             0x9f		/* Read JEDEC ID */
#define OPCODE_DP               0xb9    /* Deep Power-down */ 
#define OPCODE_RES              0xab    /* Release from DP, and Read Signature */ 

#define SPI_STATUS_REG1	1
#define SPI_STATUS_REG2	2

/* Define max times to check status register before we give up. */
#define	MAX_READY_WAIT_JIFFIES	(40 * HZ)	/* 40s max chip erase */

#define	CMD_SIZE		(1)
#define ADDR_SIZE		(2)
#define CMD_ADDR_SIZE	(CMD_SIZE + ADDR_SIZE)
#define MAX_DUMMY_SIZE	(4)

#define MTD_PART_NAME_LEN (4)

#ifdef CONFIG_SPINAND_USE_FAST_READ
#define OPCODE_READ 	OPCODE_FAST_READ
#define FAST_READ_DUMMY_BYTE 1
#else
#define OPCODE_READ 	OPCODE_NORM_READ
#define FAST_READ_DUMMY_BYTE 0
#endif

#define SPINAND_OOB_LEN(info) 	\
	((info.oob_up_skiplen + info.oob_seglen + info.oob_down_skiplen)*info.oob_seg_perpage)

#define ALIGN_DOWN(a, b)  (((a) / (b)) * (b))

static char *spinand_bad_block_table = NULL;

/*
 * SPI device driver setup and teardown
 */
struct flash_info {
	char		*name;

	/* JEDEC id zero means "no ID" (most older chips); otherwise it has
	 * a high byte of zero plus three data bytes: the manufacturer id,
	 * then a two byte device id.
	 */
	u32			jedec_id;
	u16			ext_id;

	u32 		page_size;
	u32 		page_per_block;

	/* The size listed here is what works with OPCODE_SE, which isn't
	 * necessarily called a "sector" by the vendor.
	 */
	unsigned	block_size;
	u16			n_blocks;

	/*|--------------------64bytes------------------------------|*/
	/*|---12---|-4--|---12----|-4--|---12---|-4--|---12----|-4--|*/
	/*|-seglen-|skip|-segllen-|skip|-seglen-|skip|-segllen-|skip|*/
	u32 		oob_size;        
	u16 		oob_up_skiplen;
	u16 		oob_seglen;
	u16 		oob_down_skiplen;	
	u16 		oob_seg_perpage;
	u16			oob_vail_data_offset;

	/**
	 *  chip character bits:
	 *  bit 0: under_protect flag, the serial flash under protection or not when power on
	 *  bit 1: fast read flag, the serial flash support fast read or not(command 0Bh)
	 *  bit 2: AAI flag, the serial flash support auto address increment word programming
	 *  bit 3: support dual write or no
	 *  bit 4: support dual read or no
	 *  bit 5: support quad write or no
	 *  bit 6: support quad read or no
	 *  bit 7: the second status command (35h) flag,if use 4-wire(quad) mode,the bit must be is enable
	 */
	u16			flags;
#define	SFLAG_UNDER_PROTECT			(1<<0)
#define SFLAG_FAST_READ           	(1<<1)
#define SFLAG_AAAI                	(1<<2)
#define SFLAG_COM_STATUS2         	(1<<3)

#define SFLAG_DUAL_IO_READ         	(1<<4)
#define SFLAG_DUAL_READ           	(1<<5)
#define SFLAG_QUAD_IO_READ         	(1<<6)
#define SFLAG_QUAD_READ           	(1<<7)

#define SFLAG_DUAL_IO_WRITE        	(1<<8)
#define SFLAG_DUAL_WRITE          	(1<<9)
#define SFLAG_QUAD_IO_WRITE        	(1<<10)
#define SFLAG_QUAD_WRITE          	(1<<11)

#define SFLAG_SECT_4K       		(1<<12)
};

/**
  *because of some spi nand is difference of status register difinition.
  *this structure use mapping the status reg function and corresponding.
*/
struct flash_status_reg
{
	u32		jedec_id;	
	u16		ext_id;
	unsigned b_wip:4;		/*write in progress*/
	unsigned b_wel:4;		/*wrute ebabke latch*/
	unsigned b_bp0:4;		/*block protected 0*/
	unsigned b_bp1:4;		/*block protected 1*/
	unsigned b_bp2:4;		/*block protected 2*/
	unsigned b_bp3:4;		/*block protected 3*/
	unsigned b_bp4:4;		/*block protected 4*/
	unsigned b_srp0:4;		/*status register protect 0*/
	
	unsigned b_srp1:4;		/*status register protect 1*/
	unsigned b_qe:4;		/*quad enable*/
	unsigned b_lb:4;		/*write protect control and status to the security reg.*/
/*
	unsigned b_reserved0:4;
	unsigned b_reserved1:4;
	unsigned b_reserved2:4;
*/
	unsigned b_cmp:4;		/*conjunction bp0-bp4 bit*/
	unsigned b_sus:4;		/*exec an erase/program suspend cmd_buf*/
	unsigned b_efail:4;		/**/
	unsigned b_pfail:4;		/**/
};

struct ak_spinand {
	struct spi_device	*spi;
	struct mutex		lock;
	struct flash_info	info;
	struct mtd_info		mtd;
	unsigned			partitioned:1;
	
	u8		bus_width;
	unsigned char 		*buf;
	u8		cmd_buf[CMD_ADDR_SIZE + MAX_DUMMY_SIZE];
	u8		dummy_len;

	u8		erase_opcode;
	u8		tx_opcode;
	u8		rx_opcode;
	u8		txd_bus_width;
	u8		rxd_bus_width;
	
	u8		txa_bus_width;
	u8		rxa_bus_width;	
	u32 	page_shift;
	struct flash_status_reg stat_reg;
};

/*
 * feature cmd list ,reference by spec.
 * */
static int feature_cmd[3] = {0xC0, 0xB0, 0xA0};

static inline struct ak_spinand *mtd_to_spiflash(struct mtd_info *mtd)
{
	return container_of(mtd, struct ak_spinand, mtd);
}

static 	struct flash_status_reg __devinitdata  status_reg_list[] = {
		/*normal status reg define*/
		{
			.jedec_id = 0xc8f4c8f4,	.ext_id = 0,
			.b_wip = 0,	.b_wel = 1,	.b_efail= 2,	.b_pfail= 3,
			.b_bp2 = 4,	.b_bp3 = 5,	.b_bp4 = 6,	.b_srp0 = 7,

			.b_qe = 8,	.b_srp1 = 9,.b_lb = 10,	.b_cmp = 14,
			.b_sus = 15,
		},
		/*normal status reg define*/
		{
			.jedec_id = 0xc8f2c8f2,	.ext_id = 0,
			.b_wip = 0,	.b_wel = 1,	.b_efail= 2,	.b_pfail= 3,
			.b_bp2 = 4,	.b_bp3 = 5,	.b_bp4 = 6,	.b_srp0 = 7,

			.b_qe = 8,	.b_srp1 = 9,.b_lb = 10,	.b_cmp = 14,
			.b_sus = 15,
		},
		/*normal status reg define*/
		{
			.jedec_id = 0xc8f1c8f1,	.ext_id = 0,
			.b_wip = 0,	.b_wel = 1,	.b_efail= 2,	.b_pfail= 3,
			.b_bp2 = 4,	.b_bp3 = 5,	.b_bp4 = 6,	.b_srp0 = 7,

			.b_qe = 8,	.b_srp1 = 9,.b_lb = 10,	.b_cmp = 14,
			.b_sus = 15,
		},
		/*normal status reg define*/
		{
			.jedec_id = 0xc8d1c8d1,	.ext_id = 0,
			.b_wip = 0,	.b_wel = 1,	.b_efail= 2,	.b_pfail= 3,
			.b_bp2 = 4,	.b_bp3 = 5,	.b_bp4 = 6,	.b_srp0 = 7,

			.b_qe = 8,	.b_srp1 = 9,.b_lb = 10,	.b_cmp = 14,
			.b_sus = 15,
		},	
		/*normal status reg define*/
		{
			.jedec_id = 0xc8d2c8d2,	.ext_id = 0,
			.b_wip = 0,	.b_wel = 1,	.b_efail= 2,	.b_pfail= 3,
			.b_bp2 = 4,	.b_bp3 = 5,	.b_bp4 = 6,	.b_srp0 = 7,

			.b_qe = 8,	.b_srp1 = 9,.b_lb = 10,	.b_cmp = 14,
			.b_sus = 15,
		},				
		/*normal status reg define*/
		{
			.jedec_id = 0xb148c8b1,	.ext_id = 0,
			.b_wip = 0,	.b_wel = 1,	.b_efail= 2,	.b_pfail= 3,
			.b_bp2 = 4,	.b_bp3 = 5,	.b_bp4 = 6,	.b_srp0 = 7,

			.b_qe = 8,	.b_srp1 = 9,.b_lb = 10,	.b_cmp = 14,
			.b_sus = 15,
		},
		/*normal status reg define*/
		{
			.jedec_id = 0xc8217f7f,	.ext_id = 0,
			.b_wip = 0,	.b_wel = 1,	.b_efail= 2,	.b_pfail= 3,
			.b_bp2 = 4,	.b_bp3 = 5,	.b_bp4 = 6,	.b_srp0 = 7,

			.b_qe = 8,	.b_srp1 = 9,.b_lb = 10,	.b_cmp = 14,
			.b_sus = 15,
		},
		/*normal status reg define*/
		{
			.jedec_id = 0xc831C831,	.ext_id = 0,
			.b_wip = 0,	.b_wel = 1,	.b_efail= 2,	.b_pfail= 3,
			.b_bp2 = 4,	.b_bp3 = 5,	.b_bp4 = 6,	.b_srp0 = 7,

			.b_qe = 8,	.b_srp1 = 9,.b_lb = 10,	.b_cmp = 14,
			.b_sus = 15,
		},	
		/*normal status reg define*/
		{
			.jedec_id = 0xa1e1a1e1,	.ext_id = 0,
			.b_wip = 0,	.b_wel = 1,	.b_efail= 2,	.b_pfail= 3,
			.b_bp2 = 4,	.b_bp3 = 5,	.b_bp4 = 6,	.b_srp0 = 7,

			.b_qe = 8,	.b_srp1 = 9,.b_lb = 10,	.b_cmp = 14,
			.b_sus = 15,
		},
		{
			.jedec_id = 0xc212c212,	.ext_id = 0,
			.b_wip = 0,	.b_wel = 1,	.b_efail= 2,	.b_pfail= 3,
			.b_bp2 = 4,	.b_bp3 = 5,	.b_bp4 = 6,	.b_srp0 = 7,

			.b_qe = 8,	.b_srp1 = 9,.b_lb = 10,	.b_cmp = 14,
			.b_sus = 15,
		},		
		{
			.jedec_id = 0xefaa2100,	.ext_id = 0,
			.b_wip = 0,	.b_wel = 1,	.b_efail= 2,	.b_pfail= 3,
			.b_bp2 = 4,	.b_bp3 = 5,	.b_bp4 = 6,	.b_srp0 = 7,

			.b_qe = 8,	.b_srp1 = 9,.b_lb = 10,	.b_cmp = 14,
			.b_sus = 15,
		},	
		/*normal status reg define*/
		{
			.jedec_id = 0,	.ext_id = 0,
			.b_wip = 0,	.b_wel = 1,	.b_bp0 = 2,	.b_bp1 = 3,
			.b_bp2 = 4,	.b_bp3 = 5,	.b_bp4 = 6,	.b_srp0 = 7,
			
			.b_srp1 = 8,.b_qe = 9,	.b_lb = 10,	.b_cmp = 14,
			.b_sus = 15,
		},
};

/* This is fake ecc layout to keep avoid oob used without ecc protection
   for spi nand flash GD5F1GQ4U
 */
static struct nand_ecclayout nand_oob_ecclayout = {
	.eccbytes = 16,
	.eccpos = { 0, 1, 2, 3, 
	            16, 17, 18, 19,
	            32, 33, 34, 35,
	            48, 49, 50, 51 },
	.oobfree = {
        {.offset = 4,
         .length = 12},
        {.offset = 20,
         .length = 12},
        {.offset = 36,
         .length = 12},
        {.offset = 52,
         .length = 12} }
};

/* NOTE: double check command sets and memory organization when you add
 * more flash chips.  This current list focusses on newer chips, which
 * have been converging on command sets which including JEDEC ID.
 */
static struct flash_info __devinitdata ak_spinand_supportlist [] = {
    { "GD5G1GQ4U",      0xc8f1c8f1, 0, 2048, 64, 128*1024, 1024, 64, 0, 12, 4, 4, 4, SFLAG_DUAL_READ|SFLAG_QUAD_READ|SFLAG_QUAD_WRITE,},
    { "GD5F1GQ4UB",     0xc8d1c8d1, 0, 2048, 64, 128*1024, 1024, 64, 0, 12, 4, 4, 4, SFLAG_DUAL_READ|SFLAG_QUAD_READ|SFLAG_QUAD_WRITE,},  
    { "GD5G2GQ4U",      0xc8f2c8f2, 0, 2048, 64, 128*1024, 2048, 64, 0, 12, 4, 4, 4, SFLAG_DUAL_READ|SFLAG_QUAD_READ|SFLAG_QUAD_WRITE,},
    { "GD5F2GQ4UB",     0xc8d2c8d2, 0, 2048, 64, 128*1024, 2048, 64, 0, 16, 0, 4, 4, SFLAG_DUAL_READ|SFLAG_QUAD_READ|SFLAG_QUAD_WRITE,},
    { "GD5G4GQ4U",      0xc8f4c8f4, 0, 2048, 64, 128*1024, 4096, 64, 0, 12, 4, 4, 4, SFLAG_DUAL_READ|SFLAG_QUAD_READ|SFLAG_QUAD_WRITE,},	
    { "PSU1GQ4U",       0xc8217f7f, 0, 2048, 64, 128*1024, 1024, 64, 8, 8,  0, 4, 8, SFLAG_DUAL_READ|SFLAG_QUAD_READ|SFLAG_QUAD_WRITE,},
    { "AFS1GQ4UAAWC2",  0xc831C831, 0, 2048, 128,256*1024, 512,  64, 0, 6,  10,4, 4, SFLAG_DUAL_READ|SFLAG_QUAD_READ|SFLAG_QUAD_WRITE,},	
    { "PN26G01AWSIUG",  0xa1e1a1e1, 0, 2048, 64, 128*1024, 1024, 64, 0, 8,  0, 4, 6, SFLAG_DUAL_READ|SFLAG_QUAD_READ|SFLAG_QUAD_WRITE,},
    { "MX35LF1GE4AB",   0xc212c212, 0, 2048, 64, 128*1024, 1024, 64, 0, 16, 0, 4, 4, SFLAG_DUAL_READ|SFLAG_QUAD_READ|SFLAG_QUAD_WRITE,},
    { "W25N01GV",       0xefaa2100, 0, 2048, 64, 128*1024, 1024, 64, 0, 4,  12,4, 4, SFLAG_DUAL_READ|SFLAG_QUAD_READ|SFLAG_QUAD_WRITE,},
};

#ifdef SPINAND_USE_MTD_BLOCK_LAYER
/**
* @brief: because of the _read() function call by mtd block layer, the buffer be
* allocate by vmalloc() in mtd layer, spi driver layer may use this buffer that 
* intents of use for DMA transfer, so, add this function to transition buffer.
* call this function at before real read/write data.
* 
* @author lixinhai
* @date 2013-04-10
* @param[in] flash  spiflash handle.
* @param[in] buffer.
* @param[in] buffer len
* @param[in] read/write
* @retval return the transition buffer
*/
static void *flash_buf_bounce_pre(struct ak_spinand *flash,
				void *buf, u32 len, int dir)
{
	if(!is_vmalloc_addr(buf)) {
		return buf;
	}

	if(dir == SPI_FLASH_WRITE) {
		memcpy(flash->buf, buf, len);
	}
	return flash->buf;
}

/**
* @brief: because of the _read() function call by mtd block layer, the buffer be
* allocate by vmalloc() in mtd layer, spi driver layer may use this buffer that 
* intents of use for DMA transfer, so, add this function to transition buffer.
* call this function at after real read/write data
* 
* @author lixinhai
* @date 2013-04-10
* @param[in] flash  spiflash handle.
* @param[in] buffer.
* @param[in] buffer len
* @param[in] read/write
* @retval return the transition buffer
*/
static void flash_buf_bounce_post(struct ak_spinand *flash,
				void *buf, u32 len, int dir)
{
	if(!is_vmalloc_addr(buf)) {
		return;
	}

	if(dir == SPI_FLASH_READ) {
		memcpy(buf, flash->buf, len);
	}
}
#else
static inline void *flash_buf_bounce_pre(struct ak_spinand *flash,
				void *buf, u32 len, int dir)
{
	return buf;
}

static inline void flash_buf_bounce_post(struct ak_spinand *flash,
				void *buf, u32 len, int dir)
{
}
#endif

/*
 * Internal helper functions
 */

/**
* @brief Read the status register.
* 
*  returning its value in the location
* @author lixinhai
* @date 2014-03-20
* @param[in] spiflash handle.
* @return int Return the status register value.
*/
static int read_sr(struct ak_spinand *flash, u32 addr)
{
	u8 			st_tmp= 0;
	int			status;
	struct spi_message	message;
	struct spi_transfer	x[2];

	spi_message_init(&message);
	memset(x, 0, sizeof x);
	
	flash->cmd_buf[0]= OPCODE_RDSR1;
	flash->cmd_buf[1]= addr;	
	
	x[0].tx_buf = flash->cmd_buf;
	x[0].len = 2;
	spi_message_add_tail(&x[0], &message);


	x[1].rx_buf = flash->cmd_buf + 2;
	x[1].len = 1;
	spi_message_add_tail(&x[1], &message);

	/* do the i/o */
	status = spi_sync(flash->spi, &message);
	if (status == 0)
		memcpy(&st_tmp, x[1].rx_buf, 1);

	return st_tmp;	
}

/**
* @brief Write status register
* 
*  Write status register 1 byte.
* @author lixinhai
* @date 2014-03-20
* @param[in] flash  spiflash handle.
* @param[in] val  register value to be write.
* @return int return write success or failed
* @retval returns zero on success
* @retval return a negative error code if failed
*/
static int write_sr(struct ak_spinand *flash, u32 addr, u16 val)
{
	struct spi_transfer t[1];
	struct spi_message m;
	spi_message_init(&m);
	memset(t, 0, (sizeof t));
		
	flash->cmd_buf[0] = OPCODE_WRSR1;
	flash->cmd_buf[1] = addr;
	flash->cmd_buf[2] = val;
	
	t[0].tx_buf = flash->cmd_buf;
	t[0].len = 3;	
	spi_message_add_tail(&t[0], &m);
	
	return spi_sync(flash->spi, &m);
}

static inline int sflash_reset(struct ak_spinand *flash)
{
	u8	code = OPCODE_RESET;
	int ret;

	ret = spi_write_then_read(flash->spi, &code, 1, NULL, 0);
	ret |= write_sr(flash, 0xa0, 0x0);
	return ret;
}

/**
* @brief Set write enable latch.
* 
*  Set write enable latch with Write Enable command.
* @author lixinhai
* @date 2014-03-20
* @param[in] flash  spiflash handle.
* @return int return write success or failed
* @retval returns zero on success
* @retval return a negative error code if failed
*/
static inline int write_enable(struct ak_spinand *flash)
{
	int			status;
	struct spi_message	message;
	struct spi_transfer	x[2];

	spi_message_init(&message);
	memset(x, 0, sizeof x);
	
	flash->cmd_buf[0]= OPCODE_WREN;

	x[0].tx_buf = flash->cmd_buf;
	x[0].len = 1;
	spi_message_add_tail(&x[0], &message);

	/* do the i/o */
	status = spi_sync(flash->spi, &message);

	return status;		
}

/**
* @brief Set write disble
* 
*  Set write disble instruction to the chip.
* @author lixinhai
* @date 2014-03-20
* @param[in] flash	spiflash handle.
* @return int return write success or failed
* @retval returns zero on success
* @retval return a negative error code if failed
*/
static inline int write_disable(struct ak_spinand *flash)
{
	u8	code = OPCODE_WRDI;

	return spi_write_then_read(flash->spi, &code, 1, NULL, 0);
}

/**
* @brief  Wait for SPI flash ready.
* 
*  Service routine to read status register until ready, or timeout occurs.
* @author lixinhai
* @date 2014-03-20
* @param[in] flash	spiflash handle.
* @return int return write success or failed
* @retval returns zero on success
* @retval return a non-zero error code if failed
*/
static int wait_till_ready(struct ak_spinand *flash)
{
	unsigned long deadline;
	int idx, shift;
	u32 sr;
	u8 addr;
	struct flash_status_reg *fsr = &flash->stat_reg;

	deadline = jiffies + MAX_READY_WAIT_JIFFIES;

	shift = fsr->b_wip / 8;
	idx = fsr->b_wip % 8;
	addr = feature_cmd[shift];
	do {
		if ((sr = read_sr(flash, addr)) < 0)
			break;
		else if(((sr >> 4) & 0x3) == 2)
		{
			printk(KERN_ERR "ecc error sr:%d\n", sr );
			return 1;
		}
		else if (!(sr & (1<<(fsr->b_wip%8))))
			return 0;

		cond_resched();

	} while (!time_after_eq(jiffies, deadline));

	return 1;
}


/**
* @brief: enable 4 wire transfer mode.
* 
* @author lixinhai
* @date 2014-04-10
* @param[in] flash  spiflash handle.
*/
static int quad_mode_enable(struct ak_spinand *flash)
{
	int ret, idx, shift;
	u32 regval;
	u8 addr;
	struct flash_status_reg *fsr = &flash->stat_reg;

	shift = fsr->b_qe / 8;
	idx = fsr->b_qe % 8;
	
	addr = feature_cmd[shift];
	ret = wait_till_ready(flash);
	if (ret)
		return -EBUSY;

	write_enable(flash);
	
	regval = read_sr(flash, addr);
	regval |= 1<<(fsr->b_qe % 8);
	write_sr(flash, addr, regval);


	regval = read_sr(flash, addr);

	write_disable(flash);
	return 0;
}

/**
* @brief: disable 4 wire transfer mode.
* 
* @author lixinhai
* @date 2014-04-10
* @param[in] flash  spiflash handle.
*/
static int quad_mode_disable(struct ak_spinand *flash)
{
	int ret, idx, shift;
	u32 regval;
	u8 addr;
	struct flash_status_reg *fsr = &flash->stat_reg;

	shift = fsr->b_qe / 8;
	idx = fsr->b_qe % 8; 
	addr = feature_cmd[shift];
	ret = wait_till_ready(flash);
	if (ret)
		return -EBUSY;
	
	write_enable(flash);
	
	regval = read_sr(flash, addr);
	regval &= ~(1<<(fsr->b_qe%8));
	write_sr(flash, addr, regval);

	regval = read_sr(flash, addr);
	write_disable(flash);
	return 0;
}

/**
* @brief  Erase sector
* 
*  Erase a sector specialed by user.
* @author lixinhai
* @date 2014-03-20
* @param[in] flash	    spiflash handle.
* @param[in] offset    which is any address within the sector which should be erased.
* @return int return write success or failed
* @retval returns zero on success
* @retval return a non-zero error code if failed
*/
static int erase_block(struct ak_spinand *flash, u32 offset)
{
	u32 row;	
	struct spi_transfer t[1];
	struct spi_message m;
	
	spi_message_init(&m);
	memset(t, 0, (sizeof t));
		
	DEBUG(MTD_DEBUG_LEVEL3, "%s: %s %dKiB at 0x%08x\n",
			dev_name(&flash->spi->dev), __func__,
			flash->mtd.erasesize / 1024, offset);

	row = ((offset>>flash->page_shift) & 0xffffff);

	/* Wait until finished previous write command. */
	if (wait_till_ready(flash))
	{
		printk(KERN_ERR "kernel: error erase_block previous write command fail\n" );
		return -EBUSY;
	}

	/* Send write enable, then erase commands. */
	write_enable(flash);

	/* Set up command buffer. */
	flash->cmd_buf[0] = flash->erase_opcode;
	flash->cmd_buf[1] = row >> 16;
	flash->cmd_buf[2] = row >> 8;
	flash->cmd_buf[3] = row;

	t[0].tx_buf = flash->cmd_buf;
	t[0].len = 4;	
	spi_message_add_tail(&t[0], &m);
	
    spi_sync(flash->spi, &m);

    //spi_write(flash->spi, flash->cmd_buf, 4);

	if (wait_till_ready(flash)) {
		printk(KERN_ERR "kernel: error erase_block write command fail\n" );
		/* REVISIT status return?? */		
		return -EBUSY;
	}

	return 0;
}


/**
* @brief  MTD Erase
* 
* Erase an address range on the flash chip.
* @author luoyongchuang
* @date 2015-05-17
* @param[in] mtd    mtd info handle.
* @param[in] instr   erase info.
* @return int return write success or failed
* @retval returns zero on success
* @retval return a non-zero error code if failed
*/
static int ak_spinand_erase(struct mtd_info *mtd, struct erase_info *instr)
{
	struct ak_spinand *flash = mtd_to_spiflash(mtd);
	u32 addr,len;
	uint32_t rem;

	DEBUG(MTD_DEBUG_LEVEL2, "%s: %s %s 0x%llx, len %lld\n",
	      dev_name(&flash->spi->dev), __func__, "at",
	      (long long)instr->addr, (long long)instr->len);

	/* sanity checks */
	if (instr->addr + instr->len > mtd->size)
	{
		printk(KERN_ERR "ak_spinand_erase:instr->addr[0x%llx] + instr->len[%lld] > mtd->size[%lld]\n",
			instr->addr, instr->len, mtd->size );
		return -EINVAL;
	}
	div_u64_rem(instr->len, mtd->erasesize, &rem);
	if (rem != 0)
	{
		printk(KERN_ERR "ak_spinand_erase:rem!=0 [%u]\n", rem );
		return -EINVAL;
	}

	addr = instr->addr;
	len = instr->len;

	mutex_lock(&flash->lock);

	//printk("ak_spinand_erase:%d:%d\n", addr, len);
	while (len) {
		if (erase_block(flash, addr)) {
			instr->state = MTD_ERASE_FAILED;
			mutex_unlock(&flash->lock);
			return -EIO;
		}

		addr += mtd->erasesize;
		len -= mtd->erasesize;
	}

	mutex_unlock(&flash->lock);

	instr->state = MTD_ERASE_DONE;
	mtd_erase_callback(instr);

	return 0;
}

/**
* @brief initilize spi nand flash read/write param. 
* 
* @author lixinhai
* @date 2014-04-20
* @param[in] spiflash info handle.
* @return int return config success or failed
* @retval returns zero on success
* @retval return a non-zero error code if failed
*/
static int init_spiflash_rw_info(struct ak_spinand *flash)
{
	/**default param.*/
	flash->rx_opcode = OPCODE_READ;
	flash->rxd_bus_width = XFER_1DATAWIRE;
	flash->rxa_bus_width = XFER_1DATAWIRE;
	flash->tx_opcode = OPCODE_PP;
	flash->txd_bus_width = XFER_1DATAWIRE;
	flash->txa_bus_width = XFER_1DATAWIRE;
	flash->dummy_len = 1;

	if(flash->bus_width & FLASH_BUS_WIDTH_2WIRE){
		if(flash->info.flags & SFLAG_DUAL_READ) {
			flash->rx_opcode = OPCODE_FAST_D_READ;
			flash->rxd_bus_width = XFER_2DATAWIRE;
			flash->rxa_bus_width = XFER_1DATAWIRE;
			flash->dummy_len = 1;
		} else if (flash->info.flags & SFLAG_DUAL_IO_READ) {
			flash->rx_opcode = OPCODE_FAST_D_IO;
			flash->rxd_bus_width = XFER_2DATAWIRE;
			flash->rxa_bus_width = XFER_2DATAWIRE;
			flash->dummy_len = 1;
		}

		if(flash->info.flags & SFLAG_DUAL_WRITE) {
			flash->tx_opcode = OPCODE_PP_DUAL;
			flash->txd_bus_width = XFER_2DATAWIRE;
			flash->txa_bus_width = XFER_1DATAWIRE;
		} else if(flash->info.flags & SFLAG_DUAL_IO_WRITE) {
			flash->tx_opcode = OPCODE_2IO_PP;
			flash->txd_bus_width = XFER_2DATAWIRE;
			flash->txa_bus_width = XFER_2DATAWIRE;
		}	
	}

	if(flash->bus_width & FLASH_BUS_WIDTH_4WIRE){
		if(flash->info.flags & SFLAG_QUAD_READ) {
			flash->rx_opcode = OPCODE_FAST_Q_READ;
			flash->rxd_bus_width = XFER_4DATAWIRE;
			flash->rxa_bus_width = XFER_1DATAWIRE;
			flash->dummy_len = 1;
		}else if(flash->info.flags & SFLAG_QUAD_IO_READ){
			flash->rx_opcode = OPCODE_FAST_Q_IO;
			flash->rxd_bus_width = XFER_4DATAWIRE;
			flash->rxa_bus_width = XFER_4DATAWIRE;
			flash->dummy_len = 3;
		}

		if(flash->info.flags & SFLAG_QUAD_WRITE) {
			flash->tx_opcode = OPCODE_PP_QUAD;
			flash->txd_bus_width = XFER_4DATAWIRE;			
			flash->txa_bus_width = XFER_1DATAWIRE;
		}else if(flash->info.flags & SFLAG_QUAD_IO_WRITE) {
			flash->tx_opcode = OPCODE_4IO_PP;
			flash->txd_bus_width = XFER_4DATAWIRE;
			flash->txa_bus_width = XFER_4DATAWIRE;
		}
	}
    
	return 0;
}

/**
* @brief configure spi nandflash transfer mode according to flags. 
* 
* @author lixinhai
* @date 2014-04-20
* @param[in] spiflash info handle.
* @return int return config success or failed
* @retval returns zero on success
* @retval return a non-zero error code if failed
*/
static int ak_spinand_cfg_quad_mode(struct ak_spinand *flash)
{
	int ret = 0;
	
	if((flash->bus_width & FLASH_BUS_WIDTH_4WIRE) && 
		(flash->info.flags & (SFLAG_QUAD_WRITE|SFLAG_QUAD_IO_WRITE|
			SFLAG_DUAL_READ|SFLAG_DUAL_IO_READ))) {		
		ret = quad_mode_enable(flash);
		if(ret) {
			flash->bus_width &= ~FLASH_BUS_WIDTH_4WIRE;
			printk("config the spiflash quad enable fail. transfer use 1 wire.\n");
		}
	}
	else {
		quad_mode_disable(flash);
    }

	return ret;
}

#define FILL_CMD(c, val) do{c[0] = (val);}while(0)
#define FILL_ADDR(c, val) do{	\
		c[CMD_SIZE] = ((val) >> 8) & 0xf;	\
		c[CMD_SIZE+1] = ((val) & 0xff);	\
		}while(0)
		
#define FILL_DUMMY_DATA(c, val) do{	\
			c[CMD_ADDR_SIZE] = val >> 16;	\
			c[CMD_ADDR_SIZE+1] = 0;	\
			c[CMD_ADDR_SIZE+2] = 0;	\
			c[CMD_ADDR_SIZE+3] = 0;	\
			}while(0)

/**
* @brief configure spi nandflash transfer mode according to flags. 
* 
* @author lixinhai
* @date 2014-04-20
* @param[in] mtd info handle.
* @param[in] row address.
* @param[in] column address.
* @param[in] transfer len.
* @param[out] transfer result len.
* @param[out] result buffer.
* @return int return config success or failed
* @retval returns zero on success
* @retval return a non-zero error code if failed
*/
static int spinand_read(struct mtd_info *mtd, int row, int col, size_t len,
		size_t *retlen, u_char *buf)
{
	struct ak_spinand *flash = mtd_to_spiflash(mtd);
	struct spi_transfer t[4];
	struct spi_message m;
	struct spi_message m1;	
	void *bounce_buf;
  
	spi_message_init(&m);
	spi_message_init(&m1);	
	memset(t, 0, (sizeof t));
	
	mutex_lock(&flash->lock);
	bounce_buf = flash_buf_bounce_pre(flash, buf, len, SPI_FLASH_READ);

	/*fill command*/
	flash->cmd_buf[0] = OPCODE_READ_TO_CACHE;
	flash->cmd_buf[1] = (row >> 16) & 0xff;
	flash->cmd_buf[2] = (row >> 8) & 0xff;
	flash->cmd_buf[3] = row & 0xff;
	t[3].tx_buf = flash->cmd_buf;
	t[3].len = 4;	
	spi_message_add_tail(&t[3], &m1);
	spi_sync(flash->spi, &m1);
	
//	spi_write(flash->spi, flash->cmd_buf, 4);

	t[0].tx_buf = flash->cmd_buf;
	t[0].len = CMD_SIZE;
	spi_message_add_tail(&t[0], &m);

	t[1].tx_buf = &flash->cmd_buf[CMD_SIZE];
	t[1].len = ADDR_SIZE + flash->dummy_len;
	t[1].xfer_mode = flash->rxa_bus_width;
	spi_message_add_tail(&t[1], &m);

	t[2].rx_buf = bounce_buf;
	t[2].len = len;	
	t[2].cs_change = 1;	
	t[2].xfer_mode = flash->rxd_bus_width;

	spi_message_add_tail(&t[2], &m);

	/* Byte count starts at zero. */
	if (retlen)
		*retlen = 0;

	/* Wait till previous write/erase is done. */
	if (wait_till_ready(flash)) {
		/* REVISIT status return?? */		
		mutex_unlock(&flash->lock);
		printk(KERN_ERR"kernel: error spinand_read fail, row:%d\n", row);
		return -EBUSY;
	}

	/* Set up the write data buffer. */
	FILL_CMD(flash->cmd_buf, flash->rx_opcode);
	FILL_ADDR(flash->cmd_buf, col);
	FILL_DUMMY_DATA(flash->cmd_buf, 0x00);

	spi_sync(flash->spi, &m);

	*retlen = m.actual_length - CMD_ADDR_SIZE - flash->dummy_len;
	
	flash_buf_bounce_post(flash, buf, len, SPI_FLASH_READ);
	
	mutex_unlock(&flash->lock);

	return 0;
}

/**
* @brief configure spi nandflash transfer mode according to flags. 
* 
* @author lixinhai
* @date 2014-04-20
* @param[in] mtd info handle.
* @param[in] from: address.
* @param[in] transfer len.
* @param[out] transfer result len.
* @param[out] result buffer.
* @return int return config success or failed
* @retval returns zero on success
* @retval return a non-zero error code if failed
*/
static int ak_spinand_read(struct mtd_info *mtd, loff_t from, size_t len,
		size_t *retlen, u_char *buf)
{
	int ret = 0;
	size_t rlen = 0;
	u32 xfer_len;
	u32 offset = 0;
	u32 count = len;
	int row, column;
	uint8_t buftmp[2048];
	struct ak_spinand *flash = mtd_to_spiflash(mtd);

	*retlen = 0;

	/*decode row/column in address param*/
	row = ((from>>flash->page_shift) & 0xffffff);
	column = (from & 0x7ff);

    if ( (column & (flash->info.page_size - 1)) && count > 0 )
    {
		xfer_len = flash->info.page_size;

		ret = spinand_read(mtd, row, 0, xfer_len, &rlen, buftmp);
		if(unlikely(ret)) {
			ret = -EBUSY;
			goto out;
		}

		if ( ( count + column ) >= flash->info.page_size )
			rlen = flash->info.page_size - column;
        else
            rlen = count;
        
		memcpy(buf, buftmp + column, rlen);
        
		row++;
		column = 0;
		*retlen += rlen;
		count -= rlen;		
		offset += rlen;
    }
  
	while(count >= flash->info.page_size) {
		xfer_len = flash->info.page_size;
    
		ret = spinand_read(mtd, row, column, xfer_len, &rlen, buf + offset);
		if(unlikely(ret)) {
			ret = -EBUSY;
			goto out;
		}
		row++;
		column = 0;
		*retlen += rlen;
		count -= rlen;		
		offset += rlen;
	}	

    if ( count > 0 )
    {
		xfer_len = flash->info.page_size;

		ret = spinand_read(mtd, row, 0, xfer_len, &rlen, buftmp);
		if(unlikely(ret)) {
			ret = -EBUSY;
			goto out;
		}

		memcpy(buf + offset, buftmp, count);
        
		*retlen += count;
    }
out:
	return ret;
}

static int spinand_do_read_page(struct mtd_info *mtd, loff_t from, struct mtd_oob_ops *ops)
{
	int ret = 0;
	size_t rlen = 0;
	u32 xfer_len,read_len,read_len_oob,oob_size,oob_xfer_len;
	u32 offset = 0;
	u32 oob_offset = 0;	
	u32 count = ops->len;
	int row, column;
	uint8_t buftmp[2112] = {0};
	uint8_t *buf = NULL;	
	uint8_t *oobbuf = ops->oobbuf;
	uint8_t *datbuf = ops->datbuf;
	struct ak_spinand *flash = mtd_to_spiflash(mtd);
	
	/*decode row/column in address param*/
	row = ((from>>flash->page_shift) & 0xffffff);
	column = (from & 0x7ff);
  
	read_len_oob = ops->ooblen;
	oob_size  = flash->info.oob_size;

	while(count > 0) {
		xfer_len = (count > flash->info.page_size) ? flash->info.page_size : count;

		/*transfer len not greater than page size*/
		if(xfer_len > flash->info.page_size)
			xfer_len = ALIGN_DOWN(xfer_len, flash->info.page_size);
		if(xfer_len+column >= flash->info.page_size)
			xfer_len = flash->info.page_size - column;
    
    	read_len = xfer_len + oob_size;
    
		ret = spinand_read(mtd, row, column, read_len, &rlen, buftmp);
		if(unlikely(ret)) {
			ret = -EBUSY;
			goto out;
		}
		memcpy(datbuf + offset, buftmp, xfer_len);
		
	  	buf	 = buftmp + rlen - oob_size;

        if(read_len_oob > 0)
        {
            oob_xfer_len = (read_len_oob > oob_size) ? oob_size : read_len_oob;

            memcpy(oobbuf + oob_offset, buf, oob_xfer_len);

            read_len_oob -= oob_xfer_len;		
            oob_offset += oob_xfer_len;
            ops->oobretlen += oob_xfer_len;			 
    	}
        
		row++;
		column = 0;
		ops->retlen += (rlen - oob_size);
		count -= (rlen - oob_size);		
		offset += (rlen - oob_size);	
   }
    
out:
	return ret;
}

/**
* @brief   MTD write
* 
* Write an address range to the flash chip.
* @author lixinhai
* @date 2014-03-20
* @param[in] mtd	mtd info handle.
* @param[in] to 	write start address.
* @param[in] len	write length.
* @param[out] retlen  write length at actually.
* @param[out] buf	   the pointer to write data.
* @return int return write success or failed
* @retval returns zero on success
* @retval return a non-zero error code if failed
*/
static int spinand_write(struct mtd_info *mtd, int row, int col, size_t len,
	size_t *retlen, const u_char *buf)
{
	struct ak_spinand *flash = mtd_to_spiflash(mtd);
	struct spi_transfer t[4];
	struct spi_message m;
	struct spi_message m1;	
	void *bounce_buf;

	DEBUG(MTD_DEBUG_LEVEL2, "%s: %s %s 0x%08x, len %zd\n",
			dev_name(&flash->spi->dev), __func__, "to(page_shift)",
			(u32)row, len);

	if (retlen)
		*retlen = 0;

	/* sanity checks */
	if (!len)
		return(0);

	spi_message_init(&m);
	spi_message_init(&m1);	
	memset(t, 0, (sizeof t));
	
	mutex_lock(&flash->lock);
	bounce_buf = flash_buf_bounce_pre(flash, (void*)buf, len, SPI_FLASH_WRITE);

	t[0].tx_buf = flash->cmd_buf;
	t[0].len = CMD_SIZE;
	spi_message_add_tail(&t[0], &m);

	t[1].tx_buf = &flash->cmd_buf[CMD_SIZE];
	t[1].len = ADDR_SIZE;
	t[1].xfer_mode = flash->txa_bus_width;
	spi_message_add_tail(&t[1], &m);

	t[2].tx_buf = bounce_buf;
	t[2].cs_change = 1;
	t[2].xfer_mode = flash->txd_bus_width;

	spi_message_add_tail(&t[2], &m);

	/* Wait until finished previous write command. */
	if (wait_till_ready(flash)) {
		mutex_unlock(&flash->lock);
		printk(KERN_ERR"kernel: error spinand_write write cmd fail\n");
		return -EBUSY;
	}

	write_enable(flash);
	/* Set up the opcode in the write buffer. */
	FILL_CMD(flash->cmd_buf, flash->tx_opcode);
	FILL_ADDR(flash->cmd_buf, col);

	t[2].len = len;

	spi_sync(flash->spi, &m);
	*retlen = m.actual_length - CMD_ADDR_SIZE;

	flash->cmd_buf[0] = OPCODE_PP_EXEC;
	flash->cmd_buf[1] = (row >> 16) & 0xff;
	flash->cmd_buf[2] = (row >> 8) & 0xff;
	flash->cmd_buf[3] = row & 0xff;
	
	t[3].tx_buf = flash->cmd_buf;
	t[3].len = 4;
	spi_message_add_tail(&t[3], &m1);
	spi_sync(flash->spi, &m1);		
	
    //spi_write(flash->spi, flash->cmd_buf, 4);

	/* Wait until finished previous write command. */
	if (wait_till_ready(flash))
	{
		printk(KERN_ERR"kernel: error spinand_write write data fail\n");
		return -EBUSY;
	}

	PDEBUG("ak_spinand_write: retlen=%ld\n", *retlen);
	flash_buf_bounce_post(flash, (void*)buf, len, SPI_FLASH_WRITE);

	mutex_unlock(&flash->lock);

	return 0;
}

#define FLASH_OOB_SIZE  8 
static int ak_spinand_write(struct mtd_info *mtd, loff_t to, size_t len,
		size_t *retlen, const u_char *buf)
{
	int ret = 0;
	size_t rlen = 0;
	u32 xfer_len;
	u32 offset = 0;
	u32 count = len;
	int row, column;
	struct ak_spinand *flash = mtd_to_spiflash(mtd);

	*retlen = 0;

	/*decode row/column in address param*/
	row = ((to>>flash->page_shift) & 0xffffff);
	column = (to & 0x7ff);

	while(count > 0) {
		xfer_len = (count > flash->info.page_size) ? flash->info.page_size : count;

		/*transfer len not greater than page size*/
		if(xfer_len > flash->info.page_size)
			xfer_len = ALIGN_DOWN(xfer_len, flash->info.page_size);
		if(xfer_len+column >= flash->info.page_size)
			xfer_len = flash->info.page_size - column;

		ret = spinand_write(mtd, row, column, xfer_len, &rlen, buf + offset);
		if(unlikely(ret)) {
			ret = -EBUSY;
			goto out;
		}
		
		row++;
		column = 0;
		*retlen += rlen;
		count -= rlen;		
		offset += rlen;
    }	
out:
	return ret;
}

#if 0
/**
* @brief adjust transfer len according to readlen and column. 
* 
* @author lixinhai
* @date 2014-04-20
* @param[in] spiflash info handle.
* @param[in] column pos.
* @param[in] need read length.
* @retval return transfer len 
*/
static int adjust_xfer_len(struct ak_spinand *flash, int column, int readlen)
{
	int seg_oob;
	int xfer_len;
	int ofs = flash->mtd.writesize;
	int start = column - ofs;
	int end;

	/*|--------------------64bytes------------------------------|*/
	/*|---12---|-4--|---12----|-4--|---12---|-4--|---12----|-4--|*/
	/*|-seglen-|skip|-segllen-|skip|-seglen-|skip|-segllen-|skip|*/

	xfer_len = (readlen > flash->info.oob_seglen) ? flash->info.oob_seglen : readlen;
	end = start + xfer_len;
	seg_oob = flash->info.oob_up_skiplen + flash->info.oob_seglen + flash->info.oob_down_skiplen;
	if(start/seg_oob != end/seg_oob)
		end = (start/seg_oob + 1)*seg_oob;

	xfer_len = end - start;

	return xfer_len;	
}
#endif

/**
* @brief convert oob offset and addr pos to row/column coord. 
* 
* @author lixinhai
* @date 2014-04-20
* @param[in] spiflash info handle.
* @param[in] read from addr
* @param[in] offset to read from
* @return int return write success or failed
* @retval returns zero on success
* @retval return a non-zero error code if failed
*/
static int spinand_oobpos_to_coord(struct ak_spinand *flash,
	   	loff_t addr, uint32_t ooboffs, int *row, int *column)
{
	*row = ((addr >> flash->page_shift) & 0xffffff);
	*column = (addr & (flash->info.page_size - 1));

	*row += ooboffs / flash->mtd.oobsize;
	*column += ooboffs % flash->mtd.oobsize;

	*column += flash->info.page_size;

	return 0;
}

static int spinand_do_read_oob(struct mtd_info *mtd, loff_t from,
			    struct mtd_oob_ops *ops)
{

	struct ak_spinand *flash = mtd_to_spiflash(mtd);
	int ret = 0;
	size_t rlen = 0;
	u32 xfer_len;
	u32 offset = 0;
	int row, column;
	uint8_t oobrw_buf[64];	
	int readlen ;			
	uint8_t *buf = NULL ;
	ops->oobretlen = 0;	
	readlen = ops->ooblen;
	buf = ops->oobbuf;
	
	DEBUG(MTD_DEBUG_LEVEL3, "%s: from = 0x%08Lx, len = %i\n",
			__func__, (unsigned long long)from, readlen);

	if (unlikely(ops->ooboffs >= mtd->oobsize)) {
		DEBUG(MTD_DEBUG_LEVEL0, "%s: Attempt to start read "
					"outside oob\n", __func__);
		return -EINVAL;
	}

	/* Do not allow reads past end of device */
	if (unlikely(from >= mtd->size ||
		     ops->ooboffs + readlen > ((mtd->size >> flash->page_shift) -
					(from >> flash->page_shift)) * mtd->oobsize)) {
		DEBUG(MTD_DEBUG_LEVEL0, "%s: Attempt read beyond end "
					"of device\n", __func__);
		return -EINVAL;
	}
                    
    spinand_oobpos_to_coord(flash, from, ops->ooboffs, &row, &column);
    memset(oobrw_buf,0, sizeof(oobrw_buf));

    xfer_len = sizeof(oobrw_buf) - (column - flash->info.page_size);
    
    ret = spinand_read(mtd, row, column, xfer_len, &rlen, oobrw_buf);
    if(unlikely(ret)) {
        ret = -EBUSY;
        goto out;
    }

    xfer_len = (readlen > xfer_len) ? xfer_len : readlen;
    memcpy(buf + offset, oobrw_buf, xfer_len);
    ops->oobretlen = xfer_len;

out:
	return ret;
}

static int ak_spinand_read_oob(struct mtd_info *mtd, loff_t from,
			 struct mtd_oob_ops *ops)
{
	int ret = -ENOTSUPP;

	ops->retlen = 0;

	/* Do not allow reads past end of device */
	if (ops->datbuf && (from + ops->len) > mtd->size) {
		DEBUG(MTD_DEBUG_LEVEL0, "%s: Attempt read "
				"beyond end of device\n", __func__);
		return -EINVAL;
	}

	switch(ops->mode) {
	case MTD_OOB_PLACE:
	case MTD_OOB_AUTO:
	case MTD_OOB_RAW:
		break;

	default:
		goto out;
	}

	if (!ops->datbuf){
		ret = spinand_do_read_oob(mtd, from, ops);
	} else {
		ret = spinand_do_read_page(mtd, from, ops);
	}

 out:
	return ret;
}

static int spinand_do_write_oob(struct mtd_info *mtd, loff_t to,
			    struct mtd_oob_ops *ops)
{
	int ret = 0;
	size_t rlen = 0;
	u32 xfer_len;
	int row, column;
	int writelen = ops->ooblen;
	uint8_t *buf = ops->oobbuf;	
	struct ak_spinand *flash = mtd_to_spiflash(mtd);

	DEBUG(MTD_DEBUG_LEVEL3, "%s: from = 0x%08Lx, len = %i\n",
			__func__, (unsigned long long)to, writelen);

	if (unlikely(ops->ooboffs + ops->ooblen >= mtd->oobsize)) {
		DEBUG(MTD_DEBUG_LEVEL0, "%s: Attempt to write "
					"outside oob\n", __func__);
		return -EINVAL;
	}

	/* Do not allow reads past end of device */
	if (unlikely(to >= mtd->size ||
		     ops->ooboffs + writelen > ((mtd->size >> flash->page_shift) -
					(to >> flash->page_shift)) * mtd->oobsize)) {
		DEBUG(MTD_DEBUG_LEVEL0, "%s: Attempt write beyond end "
					"of device\n", __func__);
		return -EINVAL;
	}
                    
	spinand_oobpos_to_coord(flash, to, ops->ooboffs, &row, &column);
	xfer_len = ops->ooblen;

	ret = spinand_write(mtd, row, column, xfer_len, &rlen, buf);
	if(unlikely(ret)) {
		ret = -EBUSY;
		goto out;
	}
    ops->oobretlen = rlen;
    
out:	
	return 0;		
}

static int spinand_do_write_page(struct mtd_info *mtd, loff_t to, struct mtd_oob_ops *ops)
{
	int ret = 0;
	size_t rlen = 0;
	u32 xfer_len,read_len_oob,oob_size,oob_xfer_len;
	u32 offset = 0;
	u32 oob_offset = 0;	
	u32 count = ops->len;
	int row, column;
	uint8_t buftmp_oob[64]={0};		
	uint8_t buftmp[2112]={0};
	uint8_t *oobbuf = ops->oobbuf;
	uint8_t *datbuf = ops->datbuf;
	struct ak_spinand *flash = mtd_to_spiflash(mtd);

    ops->oobretlen = 0;
    ops->retlen = 0;

	/*decode row/column in address param*/
	row = ((to>>flash->page_shift) & 0xffffff);
	column = (to & 0x7ff);
  
	read_len_oob = ops->ooblen;
	oob_size  = flash->info.oob_size;

	while(count > 0) {
		xfer_len = (count > flash->info.page_size) ? flash->info.page_size : count;

		/*transfer len not greater than page size*/
		if(xfer_len > flash->info.page_size)
			xfer_len = ALIGN_DOWN(xfer_len, flash->info.page_size);
		if(xfer_len+column >= flash->info.page_size)
			xfer_len = flash->info.page_size - column;

		memcpy(buftmp, datbuf+offset, xfer_len);
        
        oob_xfer_len = (read_len_oob > oob_size) ? oob_size : read_len_oob;       
        memset(buftmp_oob, 0xff, sizeof(buftmp_oob));
        memcpy(buftmp_oob, oobbuf + oob_offset, oob_xfer_len);
		memcpy(buftmp+xfer_len, buftmp_oob, oob_size);

		ret = spinand_write(mtd, row, column, xfer_len+oob_size, &rlen, buftmp);
		if(unlikely(ret)) {
			ret = -EBUSY;
			goto out;
		}
		
		row++;
		column = 0;
		count -= xfer_len;		
		offset += xfer_len;
        oob_offset += oob_xfer_len;
        read_len_oob -= oob_xfer_len;
        
        ops->retlen += xfer_len;
		ops->oobretlen += oob_xfer_len;
	}
    
out:
	return 0;//ret;	
}

static int ak_spinand_write_oob(struct mtd_info *mtd, loff_t to,
			 struct mtd_oob_ops *ops)
{
	int ret = -ENOTSUPP;

	ops->retlen = 0;

	/* Do not allow writes past end of device */
	if (ops->datbuf && (to + ops->len) > mtd->size) {
		DEBUG(MTD_DEBUG_LEVEL0, "%s: Attempt write beyond "
				"end of device\n", __func__);
		return -EINVAL;
	}

	switch(ops->mode) {
	case MTD_OOB_PLACE:
	case MTD_OOB_AUTO:
	case MTD_OOB_RAW:
		break;

	default:
		goto out;
	}

	if (!ops->datbuf)
		ret = spinand_do_write_oob(mtd, to, ops);
	else
		ret = spinand_do_write_page(mtd, to, ops);

 out:
	return 0;//ret;
}

static int _ak_spinand_read(struct mtd_info *mtd, loff_t from, size_t len,
			size_t *retlen, u_char *buf)
{
	ak_spinand_read(mtd,from,len,retlen,buf);
	return 0;
}

static int _ak_spinand_write(struct mtd_info *mtd, loff_t to, size_t len,
		size_t *retlen, const u_char *buf)
{
	ak_spinand_write(mtd,to,len,retlen,buf);
	return 0;			
}

static int spinand_set_badblock(struct mtd_info *mtd, int block)
{
    unsigned long byte_loc, byte_offset;
    unsigned char buf[2];
    int retlen;
	struct ak_spinand *flash = mtd_to_spiflash(mtd);

    if ( spinand_bad_block_table == NULL )
    {
        printk(KERN_ERR"Bad block table is not available\n");
        return -EINVAL;
    }

    if(NULL != spinand_bad_block_table)
    {
        byte_loc = block / 8;
        byte_offset = 7 - block % 8;
        spinand_bad_block_table[byte_loc] |= 1 << byte_offset;
    }

    buf[0] = 0x00;
    if( spinand_write(mtd, block*flash->info.page_per_block, flash->info.page_size, 1, &retlen, (const void *)buf) == -1 )
    {
        printk(KERN_ERR"spinand_set_badblock: set bad bclok failed，block：%d\n", block);
        return -EINVAL;
    }
    
    return 0;
}

static int  spinand_is_badblock(struct mtd_info *mtd, unsigned long block)
{
    unsigned long byte_loc, byte_offset;

    if ( spinand_bad_block_table == NULL )
    {
        printk(KERN_ERR"Bad block table is not available\n");
        return -EINVAL;
    }

    byte_loc = block / 8;
    byte_offset = 7 - block % 8;

    if ( spinand_bad_block_table[byte_loc] & (1 << byte_offset) )
    {
        return 1;
    }

    return 0;
}

static int check_block(struct mtd_info *mtd, unsigned long block)
{
    unsigned char data[64];
    int retlen;
	struct ak_spinand *flash = mtd_to_spiflash(mtd);

    memset(data, 0, sizeof(data));
    
    spinand_read(mtd, block * flash->info.page_per_block, flash->info.page_size, 64, &retlen, data);
    if(data[0] != 0xFF)
    {
        return 1;
    }
    
    return 0;
}

static int spinand_scan_all_block(struct mtd_info *mtd)
{
    unsigned int block;
    unsigned int byte_loc;
    unsigned int byte_offset;
    unsigned int i;
	struct ak_spinand *flash = mtd_to_spiflash(mtd);
    unsigned long block_num_total= flash->info.n_blocks;

	spinand_bad_block_table = kzalloc(block_num_total/8+1, GFP_KERNEL);
    if( spinand_bad_block_table == NULL )
		return -ENOMEM;
    
    for(i = 0; i < block_num_total; i++)
    {
        block = i % block_num_total;

        byte_loc = i/8;
        byte_offset = 7 - i%8;     
        if( check_block(mtd, block) )
        {
            printk("Bad block: 0x%08x\n", i * flash->info.block_size);
            spinand_bad_block_table[byte_loc] |= 1 << byte_offset;
        }
    }

    return 0;
}

/**
 * nand_block_bad - [anyka] Read bad block marker in anyka bbt
 * @mtd:	MTD device structure
 * @ofs:	offset from device start
 *
 * Check, if the block is bad.
 */
int ak_spinand_block_isbad(struct mtd_info *mtd, loff_t offs)
{
	u32 blk_no = 0;

	if (offs > mtd->size)
	{
	    printk("spinand %s %d offs=0x%llx erase:%08x Error!!!\n!!!\n", __func__, __LINE__, offs, mtd->erasesize);
		return -EINVAL;
	}

	blk_no = div_u64(offs, mtd->erasesize);
	return spinand_is_badblock(mtd, blk_no);
}

/**
 * ak_nand_block_markbad - [anyka] mark a block bad in anyka bbt
 * @mtd:	MTD device structure
 * @ofs:	offset from device start
 *
*/
int ak_spinand_block_markbad(struct mtd_info *mtd, loff_t offs)
{
	u32 blk_no = 0;

	if (offs > mtd->size)
	{
	    printk("spinand %s %d offs=0x%llx erase:%08x Error!!!\n!!!\n", __func__, __LINE__, offs, mtd->erasesize);
		return -EINVAL;
	}

	blk_no = div_u64(offs, mtd->erasesize);
	return spinand_set_badblock(mtd, blk_no);
}

/**
* @brief	MTD get device ID
* 
* get the device ID of  the spi nand chip.
* @author lixinhai
* @date 2014-03-20
* @param[in] mtd	 mtd info handle.
* @return int return device ID of  the spi nand chip.
*/
static int ak_spinand_get_devid(struct mtd_info *mtd)
{
	struct ak_spinand *flash = mtd_to_spiflash(mtd);
	int			ret;
	u8			id[5];
	u32			jedec;

	/* Wait until finished previous write command. */
	if (wait_till_ready(flash))
		return -EBUSY;

	flash->cmd_buf[0]  = OPCODE_RDID;
	flash->cmd_buf[1]  = 0x0;
	/* JEDEC also defines an optional "extended device information"
	 * string for after vendor-specific data, after the three bytes
	 * we use here.  Supporting some chips might require using it.
	 */
	ret = spi_write_then_read(flash->spi, flash->cmd_buf, 2, id, 4);
	if (ret < 0) {
		DEBUG(MTD_DEBUG_LEVEL0, "%s: error %d ak_spinand_get_devid\n",
			dev_name(&flash->spi->dev), ret);
		return AK_FALSE;
	}

	jedec = id[0] | (id[1]<<8) | (id[2]<<16) | (id[3]<<24) ;
	printk("spi nand ID: 0x%08x\n", jedec);

	return jedec;
}

/**
* @brief	 get_flash_info
* 
* Read the device ID and identify that it was supported or not.
* @author 	luoyongchuang
* @date 		2016-03-17
* @param[in] mtd	  spi device handle.
* @return int return the device info.
*/

static struct flash_info *__devinit get_flash_info(struct spi_device *spi)
{
	int			tmp;
	u8			cmd[2];
	u8			id[5];
	u32			jedec;
	u16                     ext_jedec = 0;
	struct flash_info	*info;
	/* JEDEC also defines an optional "extended device information"
	 * string for after vendor-specific data, after the three bytes
	 * we use here.  Supporting some chips might require using it.
	 */
	cmd[0] = OPCODE_RDID; 
	cmd[1] = 0x0; 
	tmp = spi_write_then_read(spi, cmd, 2, id, 4);
	if (tmp < 0) {
		DEBUG(MTD_DEBUG_LEVEL0, "%s: error %d reading JEDEC ID\n",
			dev_name(&spi->dev), tmp);
		return NULL;
	}
	jedec = id[0];
	jedec = jedec << 8;
	jedec |= id[1];
	jedec = jedec << 8;
	jedec |= id[2];
	jedec = jedec << 8;
	jedec |= id[3];

	printk("akspi nand ID: 0x%08x\n", jedec);

	//ext_jedec = id[3] << 8 | id[4];
	for (tmp = 0, info = ak_spinand_supportlist;
			tmp < ARRAY_SIZE(ak_spinand_supportlist);
			tmp++, info++) {
		if (info->jedec_id == jedec) {
			if (info->ext_id != 0 && info->ext_id != ext_jedec)
				continue;
			return info;
		}
	}
	dev_err(&spi->dev, "get_flash_info() unrecognized flash id %06x\n", jedec);
	return NULL;
}

static int ak_spinand_init_stat_reg(struct ak_spinand *flash)
{
	int i;
	struct flash_status_reg *sr;
	struct flash_info *info = &flash->info;

	for(i=0, sr=status_reg_list; i<ARRAY_SIZE(status_reg_list); i++, sr++) {
		if (sr->jedec_id == info->jedec_id) {
			if (info->ext_id != 0 && info->ext_id != sr->ext_id)
				continue;
			flash->stat_reg = *sr;
			return 0;
		}
	}
	
	flash->stat_reg = status_reg_list[i-1];
	return 0;
}

/**
* @brief	 spi nand probe
* 
* Initial the spi nand device driver to kernel.
* @author luoyongchuang
* @date 2016-03-17
* @param[in] mtd	  spi device handle.
* @return int return write success or failed
* @retval returns zero on success
* @retval return a non-zero error code if failed
*/
static int __devinit ak_spinand_probe(struct spi_device *spi)
{
	unsigned										i;
	unsigned										ret = 0;
	struct flash_info						*info;
	struct ak_spinand						*flash;	
	struct flash_platform_data	*data;	
	
	printk("ak spinandflash probe enter.\n");
	
	/* Platform data helps sort out which chip type we have, as
	 * well as how this board partitions it.  If we don't have
	 * a chip ID, try the JEDEC id commands; they'll work for most
	 * newer chips, even if we don't recognize the particular chip.
	 */
	 
	data = spi->dev.platform_data;
	info = get_flash_info(spi);
	if (!info)
		return -ENODEV;

	flash = kzalloc(sizeof *flash, GFP_KERNEL);
	if (!flash)
		return -ENOMEM;

    memset(flash, 0, sizeof *flash);
  
	flash->spi = spi;
	flash->info = *info;
	mutex_init(&flash->lock);
	dev_set_drvdata(&spi->dev, flash);
	
	sflash_reset(flash);

	if (data && data->name)
		flash->mtd.name = data->name;
	else
		flash->mtd.name = dev_name(&spi->dev);

	flash->mtd.type = MTD_NANDFLASH;
	flash->mtd.writesize = flash->info.page_size;
    flash->mtd.writebufsize = flash->info.page_size;
	flash->mtd.flags = MTD_WRITEABLE;
	flash->mtd.size = info->block_size * info->n_blocks;
	flash->mtd._erase = ak_spinand_erase;
	flash->mtd._write = _ak_spinand_write;
	flash->mtd._read = _ak_spinand_read;
	flash->mtd._read_oob = ak_spinand_read_oob;
	flash->mtd._write_oob = ak_spinand_write_oob;
	flash->mtd._block_isbad = ak_spinand_block_isbad;
	flash->mtd._block_markbad = ak_spinand_block_markbad;
	flash->mtd.get_device_id = ak_spinand_get_devid;
	printk("%s, info->block_size = %d, info->n_blocks = %d\n", info->name, info->block_size, info->n_blocks);

	flash->erase_opcode = OPCODE_ERASE_BLOCK;
	flash->mtd.erasesize = info->block_size;
	flash->mtd.oobsize = info->oob_size;
	flash->mtd.oobavail = info->oob_seglen * info->oob_seg_perpage;
    // We should populated ecclayout from flash info
    flash->mtd.ecclayout = &nand_oob_ecclayout;
	flash->mtd.ecc_strength = 8;

	flash->bus_width = data->bus_width;
	flash->page_shift = ffs(flash->mtd.writesize)-1;
	
#ifdef SPINAND_USE_MTD_BLOCK_LAYER
	/*pre-allocation buffer use for spi nand data transfer.*/
	flash->buf = kzalloc(flash->info.page_size, GFP_KERNEL);
	if (!flash->buf) {
		printk("Allocate buf for spi page failed\n");
		kfree(flash);
		return -ENOMEM;
	}
#endif

	ak_spinand_init_stat_reg(flash);
	ak_spinand_cfg_quad_mode(flash);
	init_spiflash_rw_info(flash);

	flash->mtd.dev.parent = &spi->dev;

    spinand_scan_all_block(&flash->mtd);
    
	dev_info(&spi->dev, "%s (%lld Kbytes)\n", info->name,
			(long long)flash->mtd.size >> 10);

	DEBUG(MTD_DEBUG_LEVEL0,
		"mtd .name = %s, .size = 0x%llx (%lldMiB) "
			".erasesize = 0x%.8x (%uKiB) .numeraseregions = %d\n",
		flash->mtd.name,
		(long long)flash->mtd.size, (long long)(flash->mtd.size >> 20),
		flash->mtd.erasesize, flash->mtd.erasesize / 1024,
		flash->mtd.numeraseregions);

	if (flash->mtd.numeraseregions)
		for (i = 0; i < flash->mtd.numeraseregions; i++)
			DEBUG(MTD_DEBUG_LEVEL0,
				"mtd.eraseregions[%d] = { .offset = 0x%llx, "
				".erasesize = 0x%.8x (%uKiB), "
				".numblocks = %d }\n",
				i, (long long)flash->mtd.eraseregions[i].offset,
				flash->mtd.eraseregions[i].erasesize,
				flash->mtd.eraseregions[i].erasesize / 1024,
				flash->mtd.eraseregions[i].numblocks);

	/* partitions should match sector boundaries; and it may be good to
	 * use readonly partitions for writeprotected sectors (BP2..BP0).
	 */
	ret = mtd_device_parse_register(&flash->mtd, NULL, NULL, NULL, 0);
	if (ret) {
		printk("Add root MTD device failed\n");
		kfree(flash->buf);
		kfree(flash);
		return -EINVAL;
	}

    printk("Init AK SPI Flash finish.\n"); 
	return 0;
}

/**
* @brief	  spi nand remove
* 
* Remove the spi nand device driver from kernel.
* @author lixinhai
* @date 2014-03-20
* @param[in] mtd	   spi device handle.
* @return int return write success or failed
* @retval returns zero on success
* @retval return a non-zero error code if failed
*/
static int __devexit ak_spinand_remove(struct spi_device *spi)
{
	struct ak_spinand	*flash = dev_get_drvdata(&spi->dev);
	int		status;
	
	status = mtd_device_unregister(&flash->mtd);
/*	
	if (mtd_has_partitions() && flash->partitioned)
		status = del_mtd_partitions(&flash->mtd);

	status = del_mtd_device(&flash->mtd);
*/
	if (status == 0) {
		kfree(flash->buf);
		kfree(flash);
	}
	return 0;
}

static struct spi_driver ak_spinand_driver = {
	.driver = {
		.name	= "ak-spiflash",
		.bus	= &spi_bus_type,
		.owner	= THIS_MODULE,
	},
	.probe	= ak_spinand_probe,
	.remove	= __devexit_p(ak_spinand_remove),

	/* REVISIT: many of these chips have deep power-down modes, which
	 * should clearly be entered on suspend() to minimize power use.
	 * And also when they're otherwise idle...
	 */
};

/**
* @brief spi nand device init
* 
*  Moudle initial.
* @author luoyongchuang
* @date 2016-05-17
* @return int return write success or failed
* @retval returns zero on success
* @retval return a non-zero error code if failed
*/
static int __init ak_spinand_init(void)
{
	printk("init Anyka SPI Nand Flash driver\n");
	return spi_register_driver(&ak_spinand_driver);
}


/**
* @brief spi nand device exit
* 
*  Moudle exit.
* @author luoyongchuang
* @date 2016-05-17
* @return None
*/
static void __exit ak_spinand_exit(void)
{
	printk("exit Anyka SPI Nand Flash driver\n");	
	spi_unregister_driver(&ak_spinand_driver);
}


module_init(ak_spinand_init);
module_exit(ak_spinand_exit);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Luoyongchuang");
MODULE_DESCRIPTION("MTD SPI Nand flash driver for Anyka chips");

