/* $Id: mdns-scan.c 56 2004-12-21 17:04:19Z lennart $ */

/***
  This file is part of mdns-scan.
 
  mdns-scan is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
 
  mdns-scan is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
  General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with mdns-scan; if not, write to the Free Software Foundation,
  Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
***/

#include <stdlib.h>
#include <unistd.h>
#include <assert.h>
#include <stdio.h>
#include <string.h>
#include "syslog.h"

#include "dns.h"
#include "query.h"
#include "util.h"
#include "nvram.h"

#define SERVICE_RR  "_services._dns-sd._udp.local"
#define SENTRY_RR	"_sentry._tcp.local"
#define AIR_PLAY	"_airplay._tcp.local"
#define AFPOVERTCP 	"_afpovertcp._tcp.local"

#define HASH_TABLE_SIZE 1023

struct hash_entry {
    unsigned hash;
    char *name;
    struct hash_entry *next;
};

typedef struct recv_msg_s{
	int			number; 	/*  */
	char		name[256];
}recv_msg_t;

typedef struct recv_msg_list_s{
	recv_msg_t	m[5];
}recv_msg_list_t;

struct hash_entry *hash_table[HASH_TABLE_SIZE];


int get_http_port(void);
int get_rtsp_port(void);


static unsigned hash_func(const char *c) {
    unsigned hash = 0;
    
    for (; *c; c++)
        hash = 31 * hash + *c;

    return hash;
}

static void handle_service(const char *service) {
    struct hash_entry *e;
    unsigned h, i;

    i = (h = hash_func(service)) % HASH_TABLE_SIZE;

    if (hash_table[i]) {
        for (e = hash_table[i]; e; e = e->next)
            if (e->hash == h)
                if (!strcmp(e->name, service))
                    return;
    }

    e = malloc(sizeof(struct hash_entry));
    assert(e);
    e->hash = h;
    e->name = strdup(service);
    assert(e->name);
    e->next = hash_table[i];
    hash_table[i] = e;

	//printf("find %s \n", service);
    fprintf(stderr, "+ %s\n", service);
}

static void free_hash_table(void) {
    unsigned i;
    struct hash_entry **e;
    
    for (i = 0, e = hash_table; i < HASH_TABLE_SIZE; i++, e++) {
        while (*e) {
            struct hash_entry *n = (*e)->next;

            free((*e)->name);
            free(*e);
            *e = n;
        }
    }
}

static int send_query(int fd, const char*name, uint16_t type) {
    int ret = -1;
    struct dns_packet *p = NULL;

    assert(fd >= 0);

    if (!(p = dns_packet_new())) {
        fprintf(stderr, "Failed to allocate DNS packet.\n");
        goto finish;
    }

    dns_packet_set_field(p, DNS_FIELD_FLAGS, DNS_FLAGS(0, 0, 0, 0, 0, 0, 0, 0, 0, 0));

    if (!dns_packet_append_name(p, name)) {
        fprintf(stderr, "Bad host name\n");
        goto finish;
    }
    
    dns_packet_append_uint16(p, type);
    dns_packet_append_uint16(p, DNS_CLASS_IN);
    dns_packet_set_field(p, DNS_FIELD_QDCOUNT, 1);
    
    if (mdns_send_dns_packet(fd, p) < 0)
        goto finish;
    
    ret = 0;
    
finish:
    if (p)
        dns_packet_free(p);

    return ret;
}

static int handle_packet(int fd, struct dns_packet *p) {
    for (;;) {
        char pname[256];
        uint16_t type, class;
        uint32_t rr_ttl;
        uint16_t rdlength;
        
        if (dns_packet_consume_name(p, pname, sizeof(pname)) < 0 ||
            dns_packet_consume_uint16(p, &type) < 0 ||
            dns_packet_consume_uint16(p, &class) < 0 ||
            dns_packet_consume_uint32(p, &rr_ttl) < 0 ||
            dns_packet_consume_uint16(p, &rdlength) < 0) {
            break;
        }
        
        /* Remove mDNS cache flush bit */
        class &= ~0x8000;
        
        if (type == DNS_TYPE_PTR &&
            class == DNS_CLASS_IN &&
            strcmp(pname, SERVICE_RR) == 0) {
            char service[256];

            if (dns_packet_consume_name(p, service, sizeof(service)) < 0)
                break;

            //if (send_query(fd, service, DNS_TYPE_PTR) < 0)
            //    return -1;

        } else if (type == DNS_TYPE_PTR &&
            class == DNS_CLASS_IN &&
            (strstr(pname, "._tcp.") || strstr(pname, "._udp."))) {
            char service[256];
            
            if (dns_packet_consume_name(p, service, sizeof(service)) < 0)
                break;
			printf("handle_service ... [%s] \n", service);
            handle_service(service);
        } else
            if (dns_packet_consume_seek(p, rdlength) < 0)
                break;
    }

    return 0;
}

static char rotdash(void) {
    static const char *rd = "/-\\|\\-";
    static const char*c = NULL;

    if (!c || !*c)
        c = rd;
    
    return *(c++);
}

#define MDNS_LOCAL_VIDEO	"video._tcp.local"
#define MDNS_LOCAL_HTTP		"http._tcp.local"
#define MDNS_LOCAL_RTSP		"rtsp._tcp.local"

#if 0
#define GS_IPC_gsvideo		"_ipc-video._tcp.local"
#define GS_IPC_httptcp		"_ipc-http._tcp.local"
#define GS_IPC_rtsptcp		"_ipc-rtsp._tcp.local"
#else
#define GS_IPC_gsvideo		local_video
#define GS_IPC_httptcp		local_http
#define GS_IPC_rtsptcp		local_rtsp

#endif
#define GS_IPC_NAME			"Grandstream Door System"

#if 0
#define GS_IPC_RTSP		"GS GXV3672 - 000b82d8a3b1._gs-rtsp._tcp.local"
#define GS_IPC_VIDEO	"GS GXV3672 - 000b82d8a3b1._gs-video._tcp.local"
#define GS_IPC_HTTP		"GS GXV3672 - 000b82d8a3b1._gs-http._tcp.local"
#else
#define GS_IPC_RTSP		gsipc_rtsp
#define GS_IPC_VIDEO	gsipc_video
#define GS_IPC_HTTP		gsipc_http

#endif
//#define GS_IPC_MAC		"GS-000b82d8a3b1.local"
#define GS_IPC_MAC 			gsipc_mac_str
#define GS_IPC_N_MAC	"GS GXV3672 - 000b82d8a3b1"
//#define GS_MAC_ADDR		"macaddress=000b82d8a3b1"
#define GS_MAC_ADDR			gsipc_macaddr
#define GS_MEDIA_STR	"path=/"

unsigned char gsipc_mac_str[128];
unsigned char gsipc_rtsp[128];
unsigned char gsipc_video[128];
unsigned char gsipc_http[128];
unsigned char gsipc_macaddr[64];

unsigned char gsipc_rtsp_h264[128];

unsigned char vendor[32] = {0};
unsigned char local_video[128];
unsigned char local_http[128];
unsigned char local_rtsp[128];




typedef struct OemInfo_s{
	unsigned int 	oem_id;
	char			*name;
}OemInfo_t;

enum{
	enOemId_Gs = 0,		//* 
	enOemId_Neutral,	//* ͽ
	enOemId_Safer,		//* 
	enOemId_CNHE,		//*
	enOemId_NATURE,
	enOemId_YSJD,
	enOemId_Telecom,
	enOemId_HUACAM,
	enOemId_Legrand,
};
/*
00 00# -GS?
01 00# -NEUTRAL?
02 00# - SAFER?
03 00# - CNHE?
04 00# - NATURE?
05 00# - YSJD?
06 00# - TELECOM?
07 00# - HUACAM?
08 00# - LEGRAND
 */
 
OemInfo_t gst_oem_info[] = {
	{0x00, 	"Grandstream"},
	{0x01,	"Neutral"},
	{0x02,  "Safer"},		//* 
	{0x03,  "CNHE"},
	{0x04,  "Nature"},
	{0x05,  "IES"},			//* »
	{0x06,  "FJ-Telecom"},	//* 
	{0x07,  "HuaCam"},
	{0x08, 	"Legrand"},
};

int get_cfg_oem_name(char *name, unsigned int oem_id)
{
	int max = sizeof(gst_oem_info)/sizeof(OemInfo_t);

	if( oem_id > max || oem_id < 0 ){
		DEBUG(LOG_ERR,  "oem id error \n");
		return -1; 
	}
	strcpy(name, gst_oem_info[oem_id].name);
	return 0; 
}

#include <netinet/in.h>

void dns_get_domain_name(struct dns_packet *p, char *n, unsigned short *t)
{
	int i;
	unsigned char ch, pos; 
	char *pn = NULL; 
	char *dst = NULL; 
	char *pstr = NULL; 
	unsigned short type = 0; 

	
	if( p == NULL || n == NULL ){
		printf("input params error \n");
		return -1; 
	}

	pn = &p->data[12];

	pstr = pn; 
	dst = n; 
	while(1){
		ch=pstr[0];
		//printf("ch = %d \n", ch);
		if( ch == 0 || ch >= p->size){
			//printf("break \n");
			break; 
		}
		memcpy(dst, pstr+1, ch);
		if( pstr[ch+1] != 0 ){
			dst[ch]  = '.';
		}
		dst  += ch+1;
		pstr += ch+1; 
		
	}
	memcpy(&type, pstr+1, sizeof(type));

	*t = ntohs(type);
	//printf("type = %x \n", *t);
	#if 0
	printf("\n==================start=====================\n");
	printf("pn = [%s] , len = %d\n", pn, p->size);
	for(i=0; i<p->size; i++){
		printf("%x ", pn[i]);
		if( i+1 %16 == 0 ){
			printf("\n");
		}
	}
	printf("\n===================end====================\n");
	#endif 
}


void dns_check_answer_domain(
	struct dns_packet *p, 
	char *n, 
	unsigned short qcnt,
	unsigned short answer_cnt)
{
	int i;
	unsigned char ch, pos; 
	char *pn = NULL; 
	char *dst = NULL; 
	char *pstr = NULL; 
	unsigned short type = 0; 
	unsigned short quarycnt = qcnt;
	unsigned short ans_cnt = 0 ; 

	
	if( p == NULL || n == NULL ){
		printf("input params error \n");
		return -1; 
	}

	pn = &p->data[12];

	pstr = pn; 
	dst = n; 
	while(1){
		if( quarycnt >  0 ){
		
			ch=pstr[0];
			if( ch == 0 || ch >= p->size){
				break; 
			}
			//memcpy(dst, pstr+1, ch);
			//if( pstr[ch+1] != 0 ){
			//	dst[ch]  = '.';
			//}
			//dst  += ch+1;
			pstr += ch+1; 
			pstr += 4; 
			quarycnt --;
		}

		if( ans_cnt != answer_cnt ){
			ch=pstr[0];
			if( ch == 0 || ch >= p->size){
				break; 
			}
			memcpy(dst, pstr+1, ch);
			if( pstr[ch+1] != 0 ){
				dst[ch]  = '.';
			}
			dst  += ch+1;
			pstr += ch+1; 
			memcpy(&type, pstr+1, sizeof(type));
			if( type == DNS_TYPE_PTR ){
				
			}
			pstr += 4; 
			quarycnt ++;
		}
	}
	

}

int dns_answer(
		int fd, 
		char *req, 
		char *re_name, 
		int type
)	
{
    int ret = -1;
    struct dns_packet *p = NULL;

    assert(fd >= 0);

    if (!(p = dns_packet_new())) {
        fprintf(stderr, "Failed to allocate DNS packet.\n");
        goto finish;
    }

    dns_packet_set_field(p, DNS_FIELD_FLAGS, DNS_FLAGS(1, 0, 1, 0, 0, 0, 0, 0, 0, 0));
	dns_packet_set_field(p, DNS_FIELD_QDCOUNT, 0);
	dns_packet_set_field(p, DNS_FIELD_ANCOUNT, 1);
	
    if (!dns_packet_append_name(p, req)) {
        fprintf(stderr, "Bad host name\n");
        goto finish;
    }

	/* http */
    dns_packet_append_uint16(p, type);
    dns_packet_append_uint16(p, DNS_CLASS_IN);
	dns_packet_append_uint32(p, 120);	//* 2 min
	dns_packet_append_uint16(p, strlen(re_name)+2);
    
	if (!dns_packet_append_name(p, re_name)) {
        fprintf(stderr, "Bad host name\n");
        goto finish;
    }

	/* video */
    dns_packet_append_uint16(p, type);
    dns_packet_append_uint16(p, DNS_CLASS_IN);
	dns_packet_append_uint32(p, 120);	//* 2 min
	dns_packet_append_uint16(p, strlen(GS_IPC_gsvideo)+2);
    
	if (!dns_packet_append_name(p, GS_IPC_gsvideo)) {
        fprintf(stderr, "Bad host name\n");
        goto finish;
    }

	/* rtsp */
    dns_packet_append_uint16(p, type);
    dns_packet_append_uint16(p, DNS_CLASS_IN);
	dns_packet_append_uint32(p, 120);	//* 2 min
	dns_packet_append_uint16(p, strlen(GS_IPC_rtsptcp)+2);
    
	if (!dns_packet_append_name(p, GS_IPC_rtsptcp)) {
        fprintf(stderr, "Bad host name\n");
        goto finish;
    }
	
    if (mdns_send_dns_packet(fd, p) < 0)
        goto finish;
    
    ret = 0;
    
finish:
    if (p)
        dns_packet_free(p);

    return ret;
}

int dns_answer_n(
		int fd, 
		char *req, 
		char **re_name, 
		int		name_cnt,
		int type
)	
{
    int ret = -1;
	int n; 
    struct dns_packet *p = NULL;

    assert(fd >= 0);

    if (!(p = dns_packet_new())) {
        fprintf(stderr, "Failed to allocate DNS packet.\n");
        goto finish;
    }

    dns_packet_set_field(p, DNS_FIELD_FLAGS, DNS_FLAGS(1, 0, 1, 0, 0, 0, 0, 0, 0, 0));
	dns_packet_set_field(p, DNS_FIELD_QDCOUNT, 0);
	dns_packet_set_field(p, DNS_FIELD_ANCOUNT, 3);
	
    

	for(n=0; n<name_cnt; n++){
		if( n == 0 ){
			if (!dns_packet_append_name(p, req)) {
		        fprintf(stderr, "Bad host name\n");
		        goto finish;
		    }
		}else{
			dns_packet_append_uint16(p, 0xc00c);
		}
		printf("rename[%d] = [%s] \n", n, re_name[n]);
	    dns_packet_append_uint16(p, type);
	    dns_packet_append_uint16(p, DNS_CLASS_IN);
		dns_packet_append_uint32(p, 120);	//* 2 min
		dns_packet_append_uint16(p, strlen(re_name[n])+2);

		if (!dns_packet_append_name(p, re_name[n])) {
	        fprintf(stderr, "Bad host name\n");
	        goto finish;
	    }
	}
    if (mdns_send_dns_packet(fd, p) < 0)
        goto finish;
    
    ret = 0;
    
finish:
    if (p)
        dns_packet_free(p);

    return ret;
}

int dns_ipstr2int(
	unsigned int *ip_addr,
	char *ip
)
{
	int ret; 
	unsigned int  ipint = 0;
	unsigned int ipaddr[4];

	if( ip ){
		if( strstr(ip, ".")){
			ret = sscanf(ip, "%d.%d.%d.%d", &ipaddr[0], &ipaddr[1],&ipaddr[2],&ipaddr[3]);
			if( ret != 4 ){
				printf("sscanf %s error, ret = %d \n", ip, ret);
				perror("sscanf ");
				return -1; 
			}
		}
	}
	ipint = (((ipaddr[0]&0xff) << 24) | ((ipaddr[1]&0xff) << 16) | 
			((ipaddr[2]&0xff) << 8 )| (ipaddr[3]&0xff));
	*ip_addr = ipint;
	return 0; 
}


int dns_answer1(
		int fd, 
		char *req, 
		char *re_name, 
		char *ip,
		int type
)	
{
    int ret = -1;
	unsigned int ipaddr;
	char strip[32] ;
    struct dns_packet *p = NULL;

	unsigned short rtspptr, httpptr, videoptr, targetptr;
	
    assert(fd >= 0);
	

	memset(strip, 0, sizeof(strip));
	strcpy(strip, ip);

    if (!(p = dns_packet_new())) {
        fprintf(stderr, "Failed to allocate DNS packet.\n");
        goto finish;
    }

    dns_packet_set_field(p, DNS_FIELD_FLAGS, DNS_FLAGS(1, 0, 1, 0, 0, 0, 0, 0, 0, 0));
	dns_packet_set_field(p, DNS_FIELD_QDCOUNT, 0);
	dns_packet_set_field(p, DNS_FIELD_ANCOUNT, 3);	//* 3 ANSWER
	dns_packet_set_field(p, DNS_FIELD_ARCOUNT, 7);	//* 7 additions

	/* ==== answer ==== */

	/* http */
    if (!dns_packet_append_name(p, GS_IPC_httptcp)) {
        fprintf(stderr, "Bad host name\n");
        goto finish;
    }
    
    dns_packet_append_uint16(p, DNS_TYPE_PTR);
    dns_packet_append_uint16(p, DNS_CLASS_IN);
	dns_packet_append_uint32(p, 120);
	dns_packet_append_uint16(p, strlen(GS_IPC_HTTP)+2);
    
	if (!dns_packet_append_name(p, GS_IPC_HTTP)) {
        fprintf(stderr, "Bad host name\n");
        goto finish;
    }
	httpptr = p->size & 0xffff;

	/* GS VIDEO */
    if (!dns_packet_append_name(p, GS_IPC_gsvideo)) {
        fprintf(stderr, "Bad host name\n");
        goto finish;
    }
    
    dns_packet_append_uint16(p, DNS_TYPE_PTR);
    dns_packet_append_uint16(p, DNS_CLASS_IN);
	dns_packet_append_uint32(p, 120);
	dns_packet_append_uint16(p, strlen(GS_IPC_VIDEO)+2);
    
	if (!dns_packet_append_name(p, GS_IPC_VIDEO)) {
        fprintf(stderr, "Bad host name\n");
        goto finish;
    }
	videoptr = p->size & 0xffff;
	
	/* RTSP */
    if (!dns_packet_append_name(p, GS_IPC_rtsptcp)) {
        fprintf(stderr, "Bad host name\n");
        goto finish;
    }
    
    dns_packet_append_uint16(p, DNS_TYPE_PTR);
    dns_packet_append_uint16(p, DNS_CLASS_IN);
	dns_packet_append_uint32(p, 120);
	dns_packet_append_uint16(p, strlen(GS_IPC_RTSP)+2);
    
	if (!dns_packet_append_name(p, GS_IPC_RTSP)) {
        fprintf(stderr, "Bad host name\n");
        goto finish;
    }

	rtspptr = p->size & 0xffff;
	/* ============= additional record =========== */
	
	#if 1
	/* mac */
	if (!dns_packet_append_name(p, GS_IPC_MAC)) {
        fprintf(stderr, "Bad host name\n");
        goto finish;
    }
	
	targetptr = ((p->size & 0xffff) - (strlen(GS_IPC_MAC)+2) );
	printf("targetptr = %x , %x\n", targetptr, p->size);
	dns_packet_append_uint16(p, DNS_TYPE_A);
	dns_packet_append_uint16(p, DNS_CLASS_IN | 0x8000);
	dns_packet_append_uint32(p, 120);
	dns_packet_append_uint16(p, 4);

	dns_ipstr2int(&ipaddr, strip);
	dns_packet_append_uint32(p, ipaddr);


	/* HTTP SRV */
	httpptr |= 0xc000;
	//printf("%x \n",httpptr);
	dns_packet_append_uint16(p, httpptr);
	#if 0
	if (!dns_packet_append_name(p, GS_IPC_N_MAC)) {
        fprintf(stderr, "Bad host name\n");
        goto finish;
    }
	#endif 
 	dns_packet_append_uint16(p, DNS_TYPE_SRV);
	dns_packet_append_uint16(p, DNS_CLASS_IN | 0x8000);
	dns_packet_append_uint32(p, 120);	//* time
	dns_packet_append_uint16(p, 8);	//* LEN
	dns_packet_append_uint16(p, 0);	//* priority
	dns_packet_append_uint16(p, 0);	//* weight
	dns_packet_append_uint16(p, 80);	//* port
	targetptr |= 0xc000;
	//printf("%x \n",targetptr);
	dns_packet_append_uint16(p, targetptr);


	/* HTTP TXT */
	httpptr |= 0xc000;
	//printf("%x \n",httpptr);
	dns_packet_append_uint16(p, httpptr);
 	dns_packet_append_uint16(p, DNS_TYPE_TXT);
	dns_packet_append_uint16(p, DNS_CLASS_IN | 0x8000);
	dns_packet_append_uint32(p, 4500);
	dns_packet_append_uint16(p, 1);	//* len 
	dns_packet_append_uint8(p, 0);	//* txtlen

	

	/* video SRV */
	videoptr |= 0xc000;
	//printf("%x \n",videoptr);
	dns_packet_append_uint16(p, videoptr);

 	dns_packet_append_uint16(p, DNS_TYPE_SRV);
	dns_packet_append_uint16(p, DNS_CLASS_IN | 0x8000);
	dns_packet_append_uint32(p, 120);	//* time
	dns_packet_append_uint16(p, 8);	//* LEN
	dns_packet_append_uint16(p, 0);	//* priority
	dns_packet_append_uint16(p, 0);	//* weight
	dns_packet_append_uint16(p, 80);	//* port
	targetptr |= 0xc000;
	//printf("%x \n",targetptr);
	dns_packet_append_uint16(p, targetptr);


	/* video TXT */
	videoptr |= 0xc000;
	//printf("%x \n",videoptr);
	dns_packet_append_uint16(p, videoptr);
 	dns_packet_append_uint16(p, DNS_TYPE_TXT);
	dns_packet_append_uint16(p, DNS_CLASS_IN | 0x8000);
	dns_packet_append_uint32(p, 4500);
	dns_packet_append_uint16(p, strlen(GS_MAC_ADDR)+1);	//* len 
	dns_packet_append_uint8(p, strlen(GS_MAC_ADDR));	//* txtlen
	dns_packet_append_str(p, GS_MAC_ADDR,strlen(GS_MAC_ADDR) );
	
	/* rtsp SRV */
	rtspptr |= 0xc000;
	//printf("%x \n",rtspptr);
	dns_packet_append_uint16(p, rtspptr);

 	dns_packet_append_uint16(p, DNS_TYPE_SRV);
	dns_packet_append_uint16(p, DNS_CLASS_IN | 0x8000);
	dns_packet_append_uint32(p, 120);	//* time
	dns_packet_append_uint16(p, 8);	//* LEN
	dns_packet_append_uint16(p, 0);	//* priority
	dns_packet_append_uint16(p, 0);	//* weight
	dns_packet_append_uint16(p, 554);	//* port
	targetptr |= 0xc000;
	//printf("%x \n",targetptr);
	dns_packet_append_uint16(p, targetptr);


	/* rtsp TXT */
	rtspptr |= 0xc000;
	//printf("%x \n",rtspptr);
	dns_packet_append_uint16(p, rtspptr);
 	dns_packet_append_uint16(p, DNS_TYPE_TXT);
	dns_packet_append_uint16(p, DNS_CLASS_IN | 0x8000);
	dns_packet_append_uint32(p, 4500);
	dns_packet_append_uint16(p, strlen(GS_MEDIA_STR)+1);	//* len 
	dns_packet_append_uint8(p, strlen(GS_MEDIA_STR));	//* txtlen
	dns_packet_append_str(p, GS_MEDIA_STR, strlen(GS_MEDIA_STR));
	
	#endif 
	
    if (mdns_send_dns_packet(fd, p) < 0)
        goto finish;
    
    ret = 0;
    
finish:
    if (p)
        dns_packet_free(p);

    return ret;
}


int dns_answer1_BAK(
		int fd, 
		char *req, 
		char *re_name, 
		char *ip,
		int type
)	
{
    int ret = -1;
	unsigned int ipaddr;
	char strip[32] ;
    struct dns_packet *p = NULL;

    assert(fd >= 0);

	memset(strip, 0, sizeof(strip));
	strcpy(strip, ip);

    if (!(p = dns_packet_new())) {
        fprintf(stderr, "Failed to allocate DNS packet.\n");
        goto finish;
    }

    dns_packet_set_field(p, DNS_FIELD_FLAGS, DNS_FLAGS(1, 0, 1, 0, 0, 0, 0, 0, 0, 0));
	dns_packet_set_field(p, DNS_FIELD_QDCOUNT, 0);
	dns_packet_set_field(p, DNS_FIELD_ANCOUNT, 1);
	dns_packet_set_field(p, DNS_FIELD_ARCOUNT, 2);
	
    if (!dns_packet_append_name(p, req)) {
        fprintf(stderr, "Bad host name\n");
        goto finish;
    }
    
    dns_packet_append_uint16(p, type);
    dns_packet_append_uint16(p, DNS_CLASS_IN);
	dns_packet_append_uint32(p, 120);
	dns_packet_append_uint16(p, strlen(re_name)+2);
    
	if (!dns_packet_append_name(p, re_name)) {
        fprintf(stderr, "Bad host name\n");
        goto finish;
    }

	#if 1

	if (!dns_packet_append_name(p, GS_IPC_MAC)) {
        fprintf(stderr, "Bad host name\n");
        goto finish;
    }
	dns_packet_append_uint16(p, DNS_TYPE_A);
	dns_packet_append_uint16(p, DNS_CLASS_IN | 0x8000);
	dns_packet_append_uint32(p, 120);
	dns_packet_append_uint16(p, 4);

	dns_ipstr2int(&ipaddr, strip);
	dns_packet_append_uint32(p, ipaddr);
	
	if (!dns_packet_append_name(p, re_name)) {
        fprintf(stderr, "Bad host name\n");
        goto finish;
    }
 	dns_packet_append_uint16(p, DNS_TYPE_A);
	dns_packet_append_uint16(p, DNS_CLASS_IN | 0x8000);
	dns_packet_append_uint32(p, 120);
	dns_packet_append_uint16(p, 4);

	dns_ipstr2int(&ipaddr, strip);
	dns_packet_append_uint32(p, ipaddr);
	#endif 
	
    if (mdns_send_dns_packet(fd, p) < 0)
        goto finish;
    
    ret = 0;
    
finish:
    if (p)
        dns_packet_free(p);

    return ret;
}

int dns_answer1_n(
		int 		fd, 
		char 		*req, 
		char 		**doname, 
		char 		*ip,
		int 		doname_cnt,
		int 		type
)	
{
    int ret = -1, n;
	unsigned int ipaddr;
	char strip[32] ;
    struct dns_packet *p = NULL;

    assert(fd >= 0);

	memset(strip, 0, sizeof(strip));
	strcpy(strip, ip);

    if (!(p = dns_packet_new())) {
        fprintf(stderr, "Failed to allocate DNS packet.\n");
        goto finish;
    }

    dns_packet_set_field(p, DNS_FIELD_FLAGS, DNS_FLAGS(1, 0, 1, 0, 0, 0, 0, 0, 0, 0));
	dns_packet_set_field(p, DNS_FIELD_QDCOUNT, 0);
	dns_packet_set_field(p, DNS_FIELD_ANCOUNT, 1);
	dns_packet_set_field(p, DNS_FIELD_ARCOUNT, 2);
	
    if (!dns_packet_append_name(p, req)) {
        fprintf(stderr, "Bad host name\n");
        goto finish;
    }

#if 0
	
    dns_packet_append_uint16(p, type);
    dns_packet_append_uint16(p, DNS_CLASS_IN);
	dns_packet_append_uint32(p, 120);
	dns_packet_append_uint16(p, strlen(re_name)+2);
    
	if (!dns_packet_append_name(p, re_name)) {
        fprintf(stderr, "Bad host name\n");
        goto finish;
    }
#endif 
	for(n=0; n<doname_cnt; n++){
		printf("rename[%d] = [%s] \n", n, doname[n]);
		if (!dns_packet_append_name(p, doname[n])) {
	        fprintf(stderr, "Bad host name\n");
	        goto finish;
	    }
		dns_packet_append_uint16(p, DNS_TYPE_A);
		dns_packet_append_uint16(p, DNS_CLASS_IN | 0x8000);
		dns_packet_append_uint32(p, 120);
		dns_packet_append_uint16(p, 4);

		dns_ipstr2int(&ipaddr, strip);
		dns_packet_append_uint32(p, ipaddr);
	}

#if 0

	if (!dns_packet_append_name(p, GS_IPC_MAC)) {
        fprintf(stderr, "Bad host name\n");
        goto finish;
    }
	dns_packet_append_uint16(p, DNS_TYPE_A);
	dns_packet_append_uint16(p, DNS_CLASS_IN | 0x8000);
	dns_packet_append_uint32(p, 120);
	dns_packet_append_uint16(p, 4);

	dns_ipstr2int(&ipaddr, strip);
	dns_packet_append_uint32(p, ipaddr);
	
	if (!dns_packet_append_name(p, re_name)) {
        fprintf(stderr, "Bad host name\n");
        goto finish;
    }
 	dns_packet_append_uint16(p, DNS_TYPE_A);
	dns_packet_append_uint16(p, DNS_CLASS_IN | 0x8000);
	dns_packet_append_uint32(p, 120);
	dns_packet_append_uint16(p, 4);

	dns_ipstr2int(&ipaddr, strip);
	dns_packet_append_uint32(p, ipaddr);
	
#endif 
	
    if (mdns_send_dns_packet(fd, p) < 0)
        goto finish;
    
    ret = 0;
    
finish:
    if (p)
        dns_packet_free(p);

    return ret;
}


int dns_answer_A(
		int fd, 
		char *req, 
		char *ip, 
		int type
)	
{
    int ret = -1;
    struct dns_packet *p = NULL;
	char ipstr[32];
	unsigned int ipaddr;

    assert(fd >= 0);

	memset(ipstr, 0, sizeof(ipstr));
	strcpy(ipstr, ip);

    if (!(p = dns_packet_new())) {
        fprintf(stderr, "Failed to allocate DNS packet.\n");
        goto finish;
    }

    dns_packet_set_field(p, DNS_FIELD_FLAGS, DNS_FLAGS(1, 0, 1, 0, 0, 0, 0, 0, 0, 0));
	dns_packet_set_field(p, DNS_FIELD_QDCOUNT, 0);
	dns_packet_set_field(p, DNS_FIELD_ANCOUNT, 1);
	
    if (!dns_packet_append_name(p, req)) {
        fprintf(stderr, "Bad host name\n");
        goto finish;
    }
 	dns_packet_append_uint16(p, DNS_TYPE_A);
	dns_packet_append_uint16(p, DNS_CLASS_IN | 0x8000);
	dns_packet_append_uint32(p, 120);
	dns_packet_append_uint16(p, 4);

	dns_ipstr2int(&ipaddr, ipstr);
	dns_packet_append_uint32(p, ipaddr);

    if (mdns_send_dns_packet(fd, p) < 0)
        goto finish;
    
    ret = 0;
    
finish:
    if (p)
        dns_packet_free(p);

    return ret;
}


char *doname[5] = {0};

#if 0
static int event_loop(int fd) {
    struct timeval tv;
	char	name[256];
	int done = 1; 
	int rcnt = 0;
	unsigned short type;

	fprintf(stderr, "Browsing ... %c           \r", rotdash());
        //* change by lwx, send mdns type any packet 
        //if (send_query(fd, GS_IPC_MDNS, DNS_TYPE_ANY) < 0)
        //    return -1;
		
    while(done) {
 
        gettimeofday(&tv, NULL);
        timeval_add(&tv, 1000000);
        
      //  for (;;) {
            struct dns_packet *p = NULL;
            uint8_t ttl;
            int r;

            if ((r = mdns_recv_dns_packet(fd, &p, &ttl, &tv)) < 0)
                return -1;

            if (!p) /* Timeout */
                break;

            assert(p);
			printf("ttl = %x, size = %d , rindex = %d \n", ttl, p->size, p->rindex);
			memset(name, 0, sizeof(name));
			dns_get_domain_name(p, name, &type);
			
			//printf("name = [%s] \n",name );
			if( strcmp(name, SERVICE_RR) == 0 ){
				//printf("RECV [%s] \n", name);
				//dns_answer(fd, SERVICE_RR,GS_IPC_httptcp, DNS_TYPE_PTR);
				dns_answer_n(fd, SERVICE_RR, doname, 3, DNS_TYPE_PTR);
				//printf("answer ptr ");
			}
			#if 1
			if( strcmp(name,GS_IPC_gsvideo) == 0 ||
				strcmp(name,GS_IPC_rtsptcp) == 0 ||
				strcmp(name,GS_IPC_httptcp) == 0){
				rcnt ++;
			} 
			if( rcnt == 3 ){
				//printf("RECV [%s] \n", name);
				dns_answer1(fd, GS_IPC_httptcp,GS_IPC_HTTP, "192.168.86.29", DNS_TYPE_PTR);
				dns_answer_A(fd, GS_IPC_MAC, "192.168.86.29", DNS_TYPE_A);
				done = 0;
				rcnt = 0;
			}
			#endif /* test */
			#if 0
			//printf("1111 \n");
			/* for client, just find MDNS DEVICE */
            if (ttl == 0xFF && dns_packet_check_valid_response(p) >= 0) {
                /* valid packet */
				printf("handle_packet \n");
                handle_packet(fd, p);
            }

			/* for server, answer mdns request  */
			
			#endif 
            dns_packet_free(p);
			usleep(100000);
			//usleep(1000000);
			//sleep(1);
        //}
    }
    
}
    

int main(int argc, char*argv[]) {
    int fd = -1;
    int ret = 1;

while(1){
    memset(&hash_table, 0, sizeof(hash_table));

	printf("mdns_open_socket .. \n");
    if ((fd = mdns_open_socket()) < 0)
        goto finish;

	printf("event_loop... \n");
    ret = event_loop(fd) < 0 ? 1 : 0;
    
finish:

    if (fd >= 0)
        close(fd);

    free_hash_table();
}
	
    return ret;
}

#else

#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <netdb.h>
#include <errno.h>

#define BUFLEN 9000
#define MDNS_IPADDR		"224.0.0.251"
#define MDNS_PORT		5353
#define GS_ID			0

int dns_answer_packet(
		struct dns_packet *p,
		char *req, 
		char *re_name, 
		int type
)	
{

	dns_packet_set_field(p, DNS_FIELD_ID, GS_ID);
    dns_packet_set_field(p, DNS_FIELD_FLAGS, DNS_FLAGS(1, 0, 1, 0, 0, 0, 0, 0, 0, 0));
	dns_packet_set_field(p, DNS_FIELD_QDCOUNT, 0);
	dns_packet_set_field(p, DNS_FIELD_ANCOUNT, 1);
	//dns_packet_set_field(p, DNS_FIELD_ARCOUNT, 3);	//* 7 additions
	

	if (!dns_packet_append_name(p, req)) {
        fprintf(stderr, "Bad host name\n");
		printf("dns_packet_append_name [%s] error\n",req);
        return -1;
    }

    dns_packet_append_uint16(p, type);
    dns_packet_append_uint16(p, DNS_CLASS_IN);
	dns_packet_append_uint32(p, 120);	//* 2 min
	dns_packet_append_uint16(p, strlen(re_name)+2);

	if (!dns_packet_append_name(p, re_name)) {
        fprintf(stderr, "Bad host name\n");
		printf("dns_packet_append_name [%s] error \n", re_name);
        return -1;
    }

    return 0;
}

int dns_answer_n_packet(
		struct dns_packet *p,
		char *req, 
		char **re_name, 
		unsigned short		name_cnt,
		int type
)	
{
	int n; 
	
	dns_packet_set_field(p, DNS_FIELD_ID, GS_ID);
    dns_packet_set_field(p, DNS_FIELD_FLAGS, DNS_FLAGS(1, 0, 1, 0, 0, 0, 0, 0, 0, 0));
	dns_packet_set_field(p, DNS_FIELD_QDCOUNT, 0);
	dns_packet_set_field(p, DNS_FIELD_ANCOUNT, name_cnt);
	
	for(n=0; n<name_cnt; n++){
		if( n == 0 ){
			if (!dns_packet_append_name(p, req)) {
		        DEBUG(LOG_ERR,  "dns_packet_append_name :Bad host name\n");
				DEBUG(LOG_ERR,  "dns_packet_append_name [%s] error\n",req);
		        return -1;
		    }
		}else{
			dns_packet_append_uint16(p, 0xc00c);
		}

	    dns_packet_append_uint16(p, type);
	    dns_packet_append_uint16(p, DNS_CLASS_IN);
		dns_packet_append_uint32(p, 4500);	//* 2 min
		dns_packet_append_uint16(p, strlen(re_name[n])+2);

		if (!dns_packet_append_name(p, re_name[n])) {
	        DEBUG(LOG_ERR,  "Bad host name\n");
			DEBUG(LOG_ERR,  "dns_packet_append_name [%s] error \n", 
				re_name[n]);
	        return -1;
	    }
	}

    
    return 0;
}


int dns_answer_txt_packet(
		struct dns_packet *p, 
		char *req, 
		char *name,
		char *doname, 
		char *ip,
		char	*txtstring,
		unsigned short srv_port,
		int answer_type,
		int type
)	
{
    int ret = -1;
	unsigned int ipaddr;
	char strip[32] ;

	unsigned short donameptr, targetptr;
	
	

	memset(strip, 0, sizeof(strip));
	strcpy(strip, ip);

	dns_packet_set_field(p, DNS_FIELD_ID, GS_ID);
    dns_packet_set_field(p, DNS_FIELD_FLAGS, DNS_FLAGS(1, 0, 1, 0, 0, 0, 0, 0, 0, 0));
	dns_packet_set_field(p, DNS_FIELD_QDCOUNT, 0);
	dns_packet_set_field(p, DNS_FIELD_ANCOUNT, 1);	//* 3 ANSWER
	dns_packet_set_field(p, DNS_FIELD_ARCOUNT, 7);	//* 7 additions

	/* ==== answer ==== */
	//* request name
    if (!dns_packet_append_name(p, req)) {
        DEBUG(LOG_ERR, "Bad host name [%s]\n", req);
        return -1;
    }

	//* answer type
    dns_packet_append_uint16(p, answer_type);
    dns_packet_append_uint16(p, DNS_CLASS_IN);	//* class type
	dns_packet_append_uint32(p, 4500);			//* time 
	dns_packet_append_uint16(p, strlen(doname)+2);	//* name len 

	//* answer name 
	if (!dns_packet_append_name(p, doname)) {
        DEBUG(LOG_ERR,  "Bad host name [%s]\n", doname);
        return -1;
    }
	//* answer name ptr 
	donameptr = p->size & 0xffff;


	/* ============= additional record =========== */
	
	//* additional record name 
	if (!dns_packet_append_name(p, name)) {
        DEBUG(LOG_ERR,  "Bad host name [%s]\n", name);
        return -1;
    }
	//* additional record name ptr 
	targetptr = ((p->size & 0xffff) - (strlen(GS_IPC_MAC)+2) );

	//* responses type, A Record
	dns_packet_append_uint16(p, DNS_TYPE_A);
	dns_packet_append_uint16(p, DNS_CLASS_IN | 0x8000);
	dns_packet_append_uint32(p, 4500);
	dns_packet_append_uint16(p, 4);	//* A record len, IPV4

	dns_ipstr2int(&ipaddr, strip);
	dns_packet_append_uint32(p, ipaddr); //* IP


	/* SRV */
	donameptr |= 0xc000;
	//* use ptr , point to answer name  
	dns_packet_append_uint16(p, donameptr);

 	dns_packet_append_uint16(p, DNS_TYPE_SRV);
	dns_packet_append_uint16(p, DNS_CLASS_IN | 0x8000);
	dns_packet_append_uint32(p, 4500);	//* time
	dns_packet_append_uint16(p, 8);	//* LEN
	dns_packet_append_uint16(p, 0);	//* priority
	dns_packet_append_uint16(p, 0);	//* weight
	dns_packet_append_uint16(p, get_http_port());	//* port
	targetptr |= 0xc000;	//* use ptr, point to record name 
	dns_packet_append_uint16(p, targetptr);

	/* TXT */
	donameptr |= 0xc000;
	dns_packet_append_uint16(p, donameptr);
 	dns_packet_append_uint16(p, DNS_TYPE_TXT);
	dns_packet_append_uint16(p, DNS_CLASS_IN | 0x8000);
	dns_packet_append_uint32(p, 4500);		//* time 
	dns_packet_append_uint16(p, strlen(txtstring)+1);	//* len 
	dns_packet_append_uint8(p, strlen(txtstring));	//* txtlen
	dns_packet_append_str(p, txtstring, strlen(txtstring));	
    
    return 0; 
}


int dns_answer1_packet(
		struct dns_packet *p, 
		char *req, 
		char *re_name, 
		char *ip,
		int type
)	
{
    int ret = -1;
	unsigned int ipaddr;
	char strip[32] ;

	unsigned short rtspptr, httpptr, videoptr, targetptr;
	
	

	memset(strip, 0, sizeof(strip));
	strcpy(strip, ip);

	dns_packet_set_field(p, DNS_FIELD_ID, GS_ID);
    dns_packet_set_field(p, DNS_FIELD_FLAGS, DNS_FLAGS(1, 0, 1, 0, 0, 0, 0, 0, 0, 0));
	dns_packet_set_field(p, DNS_FIELD_QDCOUNT, 0);
	dns_packet_set_field(p, DNS_FIELD_ANCOUNT, 3);	//* 3 ANSWER
	dns_packet_set_field(p, DNS_FIELD_ARCOUNT, 7);	//* 7 additions

	/* ==== answer ==== */

	/* http */
    if (!dns_packet_append_name(p, GS_IPC_httptcp)) {
        DEBUG(LOG_ERR,  "Bad host name [%s]\n", GS_IPC_httptcp);
        return -1;
    }
    
    dns_packet_append_uint16(p, DNS_TYPE_PTR);
    dns_packet_append_uint16(p, DNS_CLASS_IN);
	dns_packet_append_uint32(p, 4500);
	dns_packet_append_uint16(p, strlen(GS_IPC_HTTP)+2);
    
	if (!dns_packet_append_name(p, GS_IPC_HTTP)) {
        DEBUG(LOG_ERR,  "Bad host name [%s]\n", GS_IPC_HTTP);
        return -1;
    }
	httpptr = p->size & 0xffff;

	/* GS VIDEO */
    if (!dns_packet_append_name(p, GS_IPC_gsvideo)) {
        DEBUG(LOG_ERR,  "Bad host name [%s]\n", GS_IPC_gsvideo);
        return -1;
    }
    
    dns_packet_append_uint16(p, DNS_TYPE_PTR);
    dns_packet_append_uint16(p, DNS_CLASS_IN);
	dns_packet_append_uint32(p, 4500);
	dns_packet_append_uint16(p, strlen(GS_IPC_VIDEO)+2);
    
	if (!dns_packet_append_name(p, GS_IPC_VIDEO)) {
        DEBUG(LOG_ERR,  "Bad host name [%s]\n", GS_IPC_VIDEO);
        return -1;
    }
	videoptr = p->size & 0xffff;
	
	/* RTSP */
    if (!dns_packet_append_name(p, GS_IPC_rtsptcp)) {
        DEBUG(LOG_ERR,   "Bad host name [%s]\n", GS_IPC_rtsptcp);
        return -1;
    }
    
    dns_packet_append_uint16(p, DNS_TYPE_PTR);
    dns_packet_append_uint16(p, DNS_CLASS_IN);
	dns_packet_append_uint32(p, 120);
	dns_packet_append_uint16(p, strlen(GS_IPC_RTSP)+2);
    
	if (!dns_packet_append_name(p, GS_IPC_RTSP)) {
        DEBUG(LOG_ERR,  "Bad host name [%s]\n", GS_IPC_RTSP);
        return -1;
    }

	rtspptr = p->size & 0xffff;
	/* ============= additional record =========== */
	
	#if 1
	/* mac */
	if (!dns_packet_append_name(p, GS_IPC_MAC)) {
        DEBUG(LOG_ERR,  "Bad host name[%s]\n", GS_IPC_MAC);
        return -1;
    }
	
	targetptr = ((p->size & 0xffff) - (strlen(GS_IPC_MAC)+2) );
	//printf("targetptr = %x , %x\n", targetptr, p->size);
	dns_packet_append_uint16(p, DNS_TYPE_A);
	dns_packet_append_uint16(p, DNS_CLASS_IN | 0x8000);
	dns_packet_append_uint32(p, 4500);
	dns_packet_append_uint16(p, 4);

	dns_ipstr2int(&ipaddr, strip);
	dns_packet_append_uint32(p, ipaddr);

#if 0
	/* test, add 169.254.200.100 */
	targetptr |= 0xc000;
	//printf("%x \n",httpptr);
	dns_packet_append_uint16(p, targetptr);
	dns_packet_append_uint16(p, DNS_TYPE_A);
	dns_packet_append_uint16(p, DNS_CLASS_IN | 0x8000);
	dns_packet_append_uint32(p, 120);
	dns_packet_append_uint16(p, 4);
	memset(strip, 0, sizeof(strip));
	strcpy(strip, "169.254.200.100");
	dns_ipstr2int(&ipaddr, strip);
	dns_packet_append_uint32(p, ipaddr);
#endif 

	/* HTTP SRV */
	httpptr |= 0xc000;
	//printf("%x \n",httpptr);
	dns_packet_append_uint16(p, httpptr);
	#if 0
	if (!dns_packet_append_name(p, GS_IPC_N_MAC)) {
        fprintf(stderr, "Bad host name\n");
        goto finish;
    }
	#endif 
 	dns_packet_append_uint16(p, DNS_TYPE_SRV);
	dns_packet_append_uint16(p, DNS_CLASS_IN | 0x8000);
	dns_packet_append_uint32(p, 4500);	//* time
	dns_packet_append_uint16(p, 8);	//* LEN
	dns_packet_append_uint16(p, 0);	//* priority
	dns_packet_append_uint16(p, 0);	//* weight
	dns_packet_append_uint16(p, get_http_port());	//* port
	targetptr |= 0xc000;
	//printf("%x \n",targetptr);
	dns_packet_append_uint16(p, targetptr);


	/* HTTP TXT */
	httpptr |= 0xc000;
	//printf("%x \n",httpptr);
	dns_packet_append_uint16(p, httpptr);
 	dns_packet_append_uint16(p, DNS_TYPE_TXT);
	dns_packet_append_uint16(p, DNS_CLASS_IN | 0x8000);
	dns_packet_append_uint32(p, 4500);
	dns_packet_append_uint16(p, 1);	//* len 
	dns_packet_append_uint8(p, 0);	//* txtlen

	

	/* video SRV */
	videoptr |= 0xc000;
	//printf("%x \n",videoptr);
	dns_packet_append_uint16(p, videoptr);

 	dns_packet_append_uint16(p, DNS_TYPE_SRV);
	dns_packet_append_uint16(p, DNS_CLASS_IN | 0x8000);
	dns_packet_append_uint32(p, 4500);	//* time
	dns_packet_append_uint16(p, 8);	//* LEN
	dns_packet_append_uint16(p, 0);	//* priority
	dns_packet_append_uint16(p, 0);	//* weight
	dns_packet_append_uint16(p, get_http_port());	//* port
	targetptr |= 0xc000;
	//printf("%x \n",targetptr);
	dns_packet_append_uint16(p, targetptr);


	/* video TXT */
	videoptr |= 0xc000;
	//printf("%x \n",videoptr);
	dns_packet_append_uint16(p, videoptr);
 	dns_packet_append_uint16(p, DNS_TYPE_TXT);
	dns_packet_append_uint16(p, DNS_CLASS_IN | 0x8000);
	dns_packet_append_uint32(p, 4500);
	dns_packet_append_uint16(p, strlen(GS_MAC_ADDR)+1);	//* len 
	dns_packet_append_uint8(p, strlen(GS_MAC_ADDR));	//* txtlen
	dns_packet_append_str(p, GS_MAC_ADDR,strlen(GS_MAC_ADDR) );
	
	/* rtsp SRV */
	rtspptr |= 0xc000;
	//printf("%x \n",rtspptr);
	dns_packet_append_uint16(p, rtspptr);

 	dns_packet_append_uint16(p, DNS_TYPE_SRV);
	dns_packet_append_uint16(p, DNS_CLASS_IN | 0x8000);
	dns_packet_append_uint32(p, 4500);	//* time
	dns_packet_append_uint16(p, 8);	//* LEN
	dns_packet_append_uint16(p, 0);	//* priority
	dns_packet_append_uint16(p, 0);	//* weight
	dns_packet_append_uint16(p, get_rtsp_port());	//* port
	targetptr |= 0xc000;
	//printf("%x \n",targetptr);
	dns_packet_append_uint16(p, targetptr);


	/* rtsp TXT */
	rtspptr |= 0xc000;
	//printf("%x \n",rtspptr);
	dns_packet_append_uint16(p, rtspptr);
 	dns_packet_append_uint16(p, DNS_TYPE_TXT);
	dns_packet_append_uint16(p, DNS_CLASS_IN | 0x8000);
	dns_packet_append_uint32(p, 4500);
	dns_packet_append_uint16(p, strlen(GS_MEDIA_STR)+1);	//* len 
	dns_packet_append_uint8(p, strlen(GS_MEDIA_STR));	//* txtlen
	dns_packet_append_str(p, GS_MEDIA_STR, strlen(GS_MEDIA_STR));
	
	#endif 
	
    
    return 0; 
}



int dns_answer_VIDEO_packet_change(
		struct dns_packet *p, 
		char *req, 
		char *re_name, 
		char *ip,
		int type
)	
{
    int ret = -1;
	unsigned int ipaddr;
	char strip[32] ;

	unsigned short rtspptr, httpptr, videoptr, targetptr;
	
	

	memset(strip, 0, sizeof(strip));
	strcpy(strip, ip);

	dns_packet_set_field(p, DNS_FIELD_ID, GS_ID);
    dns_packet_set_field(p, DNS_FIELD_FLAGS, DNS_FLAGS(1, 0, 1, 0, 0, 0, 0, 0, 0, 0));
	dns_packet_set_field(p, DNS_FIELD_QDCOUNT, 0);
	dns_packet_set_field(p, DNS_FIELD_ANCOUNT, 1);	//* 3 ANSWER
	dns_packet_set_field(p, DNS_FIELD_ARCOUNT, 3);	//* 7 additions

	/* ==== answer ==== */

	/* GS VIDEO */
    if (!dns_packet_append_name(p, GS_IPC_gsvideo)) {
        DEBUG(LOG_ERR,   "dns_packet_append_name: Bad host name [%s]\n",
					GS_IPC_gsvideo);
        return -1;
    }
    dns_packet_append_uint16(p, DNS_TYPE_PTR);
    dns_packet_append_uint16(p, DNS_CLASS_IN);
	dns_packet_append_uint32(p, 4500);
	dns_packet_append_uint16(p, strlen(GS_IPC_VIDEO)+2);
    videoptr = p->size & 0xffff; 
	if (!dns_packet_append_name(p, GS_IPC_VIDEO)) {
        DEBUG(LOG_ERR,   "dns_packet_append_name: Bad host name [%s]\n",
					GS_IPC_VIDEO);
        return -1;
    }
	
	/* ============= additional record =========== */
	
	#if 1

	/* video SRV */
	videoptr |= 0xc000;
	//printf("%x \n",videoptr);
	dns_packet_append_uint16(p, videoptr);

 	dns_packet_append_uint16(p, DNS_TYPE_SRV);
	dns_packet_append_uint16(p, DNS_CLASS_IN | 0x8000);
	dns_packet_append_uint32(p, 4500);	//* time
	dns_packet_append_uint16(p, strlen(GS_IPC_MAC)+8);	//* LEN
	dns_packet_append_uint16(p, 0);	//* priority
	dns_packet_append_uint16(p, 0);	//* weight
	dns_packet_append_uint16(p, get_http_port());	//* port
	targetptr = (p->size & 0xffff );
	if (!dns_packet_append_name(p, GS_IPC_MAC)) {
        DEBUG(LOG_ERR,   "dns_packet_append_name: Bad host name [%s]\n",
					GS_IPC_MAC);
        return -1;
    }


	/* video TXT */
	videoptr |= 0xc000;
	dns_packet_append_uint16(p, videoptr);
 	dns_packet_append_uint16(p, DNS_TYPE_TXT);
	dns_packet_append_uint16(p, DNS_CLASS_IN | 0x8000);
	dns_packet_append_uint32(p, 4500);
	dns_packet_append_uint16(p, strlen(GS_MAC_ADDR)+1);	//* len 
	dns_packet_append_uint8(p, strlen(GS_MAC_ADDR));	//* txtlen
	dns_packet_append_str(p, GS_MAC_ADDR,strlen(GS_MAC_ADDR) );


	targetptr |= 0xc000;
	dns_packet_append_uint16(p, targetptr);
	dns_packet_append_uint16(p, DNS_TYPE_A);
	dns_packet_append_uint16(p, DNS_CLASS_IN | 0x8000);
	dns_packet_append_uint32(p, 4500);
	dns_packet_append_uint16(p, 4);

	dns_ipstr2int(&ipaddr, strip);
	dns_packet_append_uint32(p, ipaddr);

	
	#endif 
	
    
    return 0; 
}


int dns_answer_VIDEO_packet(
		struct dns_packet *p, 
		char *req, 
		char *re_name, 
		char *ip,
		int type
)	
{
    int ret = -1;
	unsigned int ipaddr;
	char strip[32] ;

	unsigned short rtspptr, httpptr, videoptr, targetptr;
	
	

	memset(strip, 0, sizeof(strip));
	strcpy(strip, ip);

	dns_packet_set_field(p, DNS_FIELD_ID, GS_ID);
    dns_packet_set_field(p, DNS_FIELD_FLAGS, DNS_FLAGS(1, 0, 1, 0, 0, 0, 0, 0, 0, 0));
	dns_packet_set_field(p, DNS_FIELD_QDCOUNT, 0);
	dns_packet_set_field(p, DNS_FIELD_ANCOUNT, 1);	//* 3 ANSWER
	dns_packet_set_field(p, DNS_FIELD_ARCOUNT, 3);	//* 7 additions

	/* ==== answer ==== */

	/* GS VIDEO */
    if (!dns_packet_append_name(p, GS_IPC_gsvideo)) {
        DEBUG(LOG_ERR,   "dns_packet_append_name: Bad host name [%s]\n",
					GS_IPC_gsvideo);
        return -1;
    }
    //videoptr = p->size & 0xffff;
    dns_packet_append_uint16(p, DNS_TYPE_PTR);
    dns_packet_append_uint16(p, DNS_CLASS_IN);
	dns_packet_append_uint32(p, 4500);
	dns_packet_append_uint16(p, strlen(GS_IPC_VIDEO)+2);
    videoptr = p->size & 0xffff; 
	if (!dns_packet_append_name(p, GS_IPC_VIDEO)) {
        DEBUG(LOG_ERR,   "dns_packet_append_name: Bad host name [%s]\n",
					GS_IPC_VIDEO);
        return -1;
    }
	//videoptr = p->size & 0xffff; //org
	
	/* ============= additional record =========== */
	
	#if 1
	/* mac */
	if (!dns_packet_append_name(p, GS_IPC_MAC)) {
        DEBUG(LOG_ERR,   "dns_packet_append_name: Bad host name [%s]\n",
					GS_IPC_MAC);
        return -1;
    }
	
	targetptr = ((p->size & 0xffff) - (strlen(GS_IPC_MAC)+2) );
	dns_packet_append_uint16(p, DNS_TYPE_A);
	dns_packet_append_uint16(p, DNS_CLASS_IN | 0x8000);
	dns_packet_append_uint32(p, 4500);
	dns_packet_append_uint16(p, 4);

	dns_ipstr2int(&ipaddr, strip);
	dns_packet_append_uint32(p, ipaddr);

	/* video SRV */
	videoptr |= 0xc000;
	//printf("%x \n",videoptr);
	dns_packet_append_uint16(p, videoptr);

 	dns_packet_append_uint16(p, DNS_TYPE_SRV);
	dns_packet_append_uint16(p, DNS_CLASS_IN | 0x8000);
	dns_packet_append_uint32(p, 4500);	//* time
	dns_packet_append_uint16(p, 8);	//* LEN
	dns_packet_append_uint16(p, 0);	//* priority
	dns_packet_append_uint16(p, 0);	//* weight
	dns_packet_append_uint16(p, get_http_port());	//* port
	targetptr |= 0xc000;
	dns_packet_append_uint16(p, targetptr);


	/* video TXT */
	videoptr |= 0xc000;
	dns_packet_append_uint16(p, videoptr);
 	dns_packet_append_uint16(p, DNS_TYPE_TXT);
	dns_packet_append_uint16(p, DNS_CLASS_IN | 0x8000);
	dns_packet_append_uint32(p, 4500);
	dns_packet_append_uint16(p, strlen(GS_MAC_ADDR)+1);	//* len 
	dns_packet_append_uint8(p, strlen(GS_MAC_ADDR));	//* txtlen
	dns_packet_append_str(p, GS_MAC_ADDR,strlen(GS_MAC_ADDR) );
	
	#endif 
	
    
    return 0; 
}



int dns_answer_VIDEO_srv_packet(
		struct dns_packet *p, 
		char *req, 
		char *re_name, 
		char *ip,
		int type
)	
{
    int ret = -1;
	unsigned int ipaddr;
	char strip[32] ;

	unsigned short rtspptr, httpptr, videoptr, targetptr;
	
	

	memset(strip, 0, sizeof(strip));
	strcpy(strip, ip);

	dns_packet_set_field(p, DNS_FIELD_ID, GS_ID);
    dns_packet_set_field(p, DNS_FIELD_FLAGS, DNS_FLAGS(1, 0, 1, 0, 0, 0, 0, 0, 0, 0));
	dns_packet_set_field(p, DNS_FIELD_QDCOUNT, 0);
	dns_packet_set_field(p, DNS_FIELD_ANCOUNT, 1);	//* 3 ANSWER
	dns_packet_set_field(p, DNS_FIELD_ARCOUNT, 1);	//* 7 additions

	/* ==== answer ==== */

	/* http */
    if (!dns_packet_append_name(p, GS_IPC_VIDEO)) {
        DEBUG(LOG_ERR,   "dns_packet_append_name: Bad host name [%s]\n",
					GS_IPC_VIDEO);
        return -1;
    }
    
	dns_packet_append_uint16(p, DNS_TYPE_SRV);
	dns_packet_append_uint16(p, DNS_CLASS_IN | 0x8000);
	dns_packet_append_uint32(p, 4500);	//* time
	dns_packet_append_uint16(p, 8+strlen(GS_IPC_MAC));	//* LEN
	dns_packet_append_uint16(p, 0);	//* priority
	dns_packet_append_uint16(p, 0);	//* weight
	dns_packet_append_uint16(p, get_http_port());	//* port

	if (!dns_packet_append_name(p, GS_IPC_MAC)) {
        DEBUG(LOG_ERR,   "dns_packet_append_name: Bad host name [%s]\n",
					GS_IPC_MAC);
        return -1;
    }
	httpptr = p->size & 0xffff;


	/* ============= additional record =========== */
	
	#if 1
	/* mac */
	if (!dns_packet_append_name(p, GS_IPC_MAC)) {
        DEBUG(LOG_ERR,   "dns_packet_append_name: Bad host name [%s]\n",
					GS_IPC_MAC);
        return -1;
    }
	
	targetptr = ((p->size & 0xffff) - (strlen(GS_IPC_MAC)+2) );
	//printf("targetptr = %x , %x\n", targetptr, p->size);
	dns_packet_append_uint16(p, DNS_TYPE_A);
	dns_packet_append_uint16(p, DNS_CLASS_IN | 0x8000);
	dns_packet_append_uint32(p, 4500);
	dns_packet_append_uint16(p, 4);

	dns_ipstr2int(&ipaddr, strip);
	dns_packet_append_uint32(p, ipaddr);

#if 0
	/* test, add 169.254.200.100 */
	targetptr |= 0xc000;
	//printf("%x \n",httpptr);
	dns_packet_append_uint16(p, targetptr);
	dns_packet_append_uint16(p, DNS_TYPE_A);
	dns_packet_append_uint16(p, DNS_CLASS_IN | 0x8000);
	dns_packet_append_uint32(p, 120);
	dns_packet_append_uint16(p, 4);
	memset(strip, 0, sizeof(strip));
	strcpy(strip, "169.254.200.100");
	dns_ipstr2int(&ipaddr, strip);
	dns_packet_append_uint32(p, ipaddr);
#endif 

	#endif 
	
    
    return 0; 
}


int dns_answer_VIDEO_txt_packet(
		struct dns_packet *p, 
		char *req, 
		char *re_name, 
		char *ip,
		int type
)	
{
    int ret = -1;
	unsigned int ipaddr;
	char strip[32] ;

	unsigned short rtspptr, httpptr, videoptr, targetptr;
	
	

	memset(strip, 0, sizeof(strip));
	strcpy(strip, ip);

	dns_packet_set_field(p, DNS_FIELD_ID, GS_ID);
    dns_packet_set_field(p, DNS_FIELD_FLAGS, DNS_FLAGS(1, 0, 1, 0, 0, 0, 0, 0, 0, 0));
	dns_packet_set_field(p, DNS_FIELD_QDCOUNT, 0);
	dns_packet_set_field(p, DNS_FIELD_ANCOUNT, 1);	//* 3 ANSWER
	dns_packet_set_field(p, DNS_FIELD_ARCOUNT, 1);	//* 7 additions

	/* ==== answer ==== */

	/* http */
    if (!dns_packet_append_name(p, GS_IPC_VIDEO)) {
        DEBUG(LOG_ERR,   "dns_packet_append_name: Bad host name [%s]\n",
					GS_IPC_VIDEO);
        return -1;
    }
    
 	dns_packet_append_uint16(p, DNS_TYPE_TXT);
	dns_packet_append_uint16(p, DNS_CLASS_IN | 0x8000);
	dns_packet_append_uint32(p, 4500);
	dns_packet_append_uint16(p, strlen(GS_MAC_ADDR)+1);	//* len 
	dns_packet_append_uint8(p, strlen(GS_MAC_ADDR));	//* txtlen
	dns_packet_append_str(p, GS_MAC_ADDR,strlen(GS_MAC_ADDR) );
	httpptr = p->size & 0xffff;


	/* ============= additional record =========== */
	
	#if 1
	/* mac */
	if (!dns_packet_append_name(p, GS_IPC_MAC)) {
        DEBUG(LOG_ERR,   "dns_packet_append_name: Bad host name [%s]\n",
					GS_IPC_MAC);
        return -1;
    }
	
	targetptr = ((p->size & 0xffff) - (strlen(GS_IPC_MAC)+2) );
	//printf("targetptr = %x , %x\n", targetptr, p->size);
	dns_packet_append_uint16(p, DNS_TYPE_A);
	dns_packet_append_uint16(p, DNS_CLASS_IN | 0x8000);
	dns_packet_append_uint32(p, 4500);
	dns_packet_append_uint16(p, 4);

	dns_ipstr2int(&ipaddr, strip);
	dns_packet_append_uint32(p, ipaddr);

#if 0
	/* test, add 169.254.200.100 */
	targetptr |= 0xc000;
	//printf("%x \n",httpptr);
	dns_packet_append_uint16(p, targetptr);
	dns_packet_append_uint16(p, DNS_TYPE_A);
	dns_packet_append_uint16(p, DNS_CLASS_IN | 0x8000);
	dns_packet_append_uint32(p, 120);
	dns_packet_append_uint16(p, 4);
	memset(strip, 0, sizeof(strip));
	strcpy(strip, "169.254.200.100");
	dns_ipstr2int(&ipaddr, strip);
	dns_packet_append_uint32(p, ipaddr);
#endif 

	#endif 
	
    
    return 0; 
}


int dns_answer_RTSP_packet(
		struct dns_packet *p, 
		char *req, 
		char *re_name, 
		char *ip,
		int type
)	
{
    int ret = -1;
	unsigned int ipaddr;
	char strip[32] ;

	unsigned short rtspptr, httpptr, videoptr, targetptr;
	
	

	memset(strip, 0, sizeof(strip));
	strcpy(strip, ip);

	dns_packet_set_field(p, DNS_FIELD_ID, GS_ID);
    dns_packet_set_field(p, DNS_FIELD_FLAGS, DNS_FLAGS(1, 0, 1, 0, 0, 0, 0, 0, 0, 0));
	dns_packet_set_field(p, DNS_FIELD_QDCOUNT, 0);
	dns_packet_set_field(p, DNS_FIELD_ANCOUNT, 1);	//* 3 ANSWER
	dns_packet_set_field(p, DNS_FIELD_ARCOUNT, 3);	//* 7 additions

	/* ==== answer ==== */


	
	/* RTSP */
    if (!dns_packet_append_name(p, GS_IPC_rtsptcp)) {
        DEBUG(LOG_ERR,   "dns_packet_append_name: Bad host name [%s]\n",
					GS_IPC_rtsptcp);
        return -1;
    }
    
    dns_packet_append_uint16(p, DNS_TYPE_PTR);
    dns_packet_append_uint16(p, DNS_CLASS_IN);
	dns_packet_append_uint32(p, 4500);
	dns_packet_append_uint16(p, strlen(GS_IPC_RTSP)+2);
    rtspptr = p->size & 0xffff;
	if (!dns_packet_append_name(p, GS_IPC_RTSP)) {
        DEBUG(LOG_ERR,   "dns_packet_append_name: Bad host name [%s]\n",
					GS_IPC_RTSP);
        return -1;
    }

	//rtspptr = p->size & 0xffff;
	/* ============= additional record =========== */
	
	#if 1
	/* mac */
	if (!dns_packet_append_name(p, GS_IPC_MAC)) {
        DEBUG(LOG_ERR,   "dns_packet_append_name: Bad host name [%s]\n",
					GS_IPC_MAC);
        return -1;
    }
	
	targetptr = ((p->size & 0xffff) - (strlen(GS_IPC_MAC)+2) );
	//printf("targetptr = %x , %x\n", targetptr, p->size);
	dns_packet_append_uint16(p, DNS_TYPE_A);
	dns_packet_append_uint16(p, DNS_CLASS_IN | 0x8000);
	dns_packet_append_uint32(p, 4500);
	dns_packet_append_uint16(p, 4);

	dns_ipstr2int(&ipaddr, strip);
	dns_packet_append_uint32(p, ipaddr);

#if 0
	/* test, add 169.254.200.100 */
	targetptr |= 0xc000;
	//printf("%x \n",httpptr);
	dns_packet_append_uint16(p, targetptr);
	dns_packet_append_uint16(p, DNS_TYPE_A);
	dns_packet_append_uint16(p, DNS_CLASS_IN | 0x8000);
	dns_packet_append_uint32(p, 120);
	dns_packet_append_uint16(p, 4);
	memset(strip, 0, sizeof(strip));
	strcpy(strip, "169.254.200.100");
	dns_ipstr2int(&ipaddr, strip);
	dns_packet_append_uint32(p, ipaddr);
#endif 
	
	/* rtsp SRV */
	rtspptr |= 0xc000;
	//printf("%x \n",rtspptr);
	dns_packet_append_uint16(p, rtspptr);

 	dns_packet_append_uint16(p, DNS_TYPE_SRV);
	dns_packet_append_uint16(p, DNS_CLASS_IN | 0x8000);
	dns_packet_append_uint32(p, 4500);	//* time
	dns_packet_append_uint16(p, 8);	//* LEN
	dns_packet_append_uint16(p, 0);	//* priority
	dns_packet_append_uint16(p, 0);	//* weight
	dns_packet_append_uint16(p, get_rtsp_port());	//* port
	targetptr |= 0xc000;
	//printf("%x \n",targetptr);
	dns_packet_append_uint16(p, targetptr);


	/* rtsp TXT */
	rtspptr |= 0xc000;
	//printf("%x \n",rtspptr);
	dns_packet_append_uint16(p, rtspptr);
 	dns_packet_append_uint16(p, DNS_TYPE_TXT);
	dns_packet_append_uint16(p, DNS_CLASS_IN | 0x8000);
	dns_packet_append_uint32(p, 4500);
	dns_packet_append_uint16(p, strlen(GS_MEDIA_STR)+1);	//* len 
	dns_packet_append_uint8(p, strlen(GS_MEDIA_STR));	//* txtlen
	dns_packet_append_str(p, GS_MEDIA_STR, strlen(GS_MEDIA_STR));
	
	#endif 
	
    
    return 0; 
}




int dns_answer_RTSP_srv_packet(
		struct dns_packet *p, 
		char *req, 
		char *re_name, 
		char *ip,
		int type
)	
{
    int ret = -1;
	unsigned int ipaddr;
	char strip[32] ;

	unsigned short rtspptr, httpptr, videoptr, targetptr;
	
	memset(strip, 0, sizeof(strip));
	strcpy(strip, ip);

	dns_packet_set_field(p, DNS_FIELD_ID, GS_ID);
    dns_packet_set_field(p, DNS_FIELD_FLAGS, DNS_FLAGS(1, 0, 1, 0, 0, 0, 0, 0, 0, 0));
	dns_packet_set_field(p, DNS_FIELD_QDCOUNT, 0);
	dns_packet_set_field(p, DNS_FIELD_ANCOUNT, 1);	//* 3 ANSWER
	dns_packet_set_field(p, DNS_FIELD_ARCOUNT, 1);	//* 7 additions

	/* ==== answer ==== */

	/* http */
    if (!dns_packet_append_name(p, GS_IPC_RTSP)) {
        DEBUG(LOG_ERR,   "dns_packet_append_name: Bad host name [%s]\n",
					GS_IPC_RTSP);
        return -1;
    }
    
	dns_packet_append_uint16(p, DNS_TYPE_SRV);
	dns_packet_append_uint16(p, DNS_CLASS_IN | 0x8000);
	dns_packet_append_uint32(p, 4500);	//* time
	dns_packet_append_uint16(p, 8+strlen(GS_IPC_MAC));	//* LEN
	dns_packet_append_uint16(p, 0);	//* priority
	dns_packet_append_uint16(p, 0);	//* weight
	dns_packet_append_uint16(p, get_rtsp_port());	//* port

	if (!dns_packet_append_name(p, GS_IPC_MAC)) {
        DEBUG(LOG_ERR,   "dns_packet_append_name: Bad host name [%s]\n",
					GS_IPC_MAC);
        return -1;
    }
	httpptr = p->size & 0xffff;


	/* ============= additional record =========== */
	
	#if 1
	/* mac */
	if (!dns_packet_append_name(p, GS_IPC_MAC)) {
        DEBUG(LOG_ERR,   "dns_packet_append_name: Bad host name [%s]\n",
					GS_IPC_MAC);
        return -1;
    }
	
	targetptr = ((p->size & 0xffff) - (strlen(GS_IPC_MAC)+2) );
	//printf("targetptr = %x , %x\n", targetptr, p->size);
	dns_packet_append_uint16(p, DNS_TYPE_A);
	dns_packet_append_uint16(p, DNS_CLASS_IN | 0x8000);
	dns_packet_append_uint32(p, 4500);
	dns_packet_append_uint16(p, 4);

	dns_ipstr2int(&ipaddr, strip);
	dns_packet_append_uint32(p, ipaddr);

#if 0
	/* test, add 169.254.200.100 */
	targetptr |= 0xc000;
	//printf("%x \n",httpptr);
	dns_packet_append_uint16(p, targetptr);
	dns_packet_append_uint16(p, DNS_TYPE_A);
	dns_packet_append_uint16(p, DNS_CLASS_IN | 0x8000);
	dns_packet_append_uint32(p, 120);
	dns_packet_append_uint16(p, 4);
	memset(strip, 0, sizeof(strip));
	strcpy(strip, "169.254.200.100");
	dns_ipstr2int(&ipaddr, strip);
	dns_packet_append_uint32(p, ipaddr);
#endif 

	#endif 
	
    
    return 0; 
}



int dns_answer_RTSP_txt_packet(
		struct dns_packet *p, 
		char *req, 
		char *re_name, 
		char *ip,
		int type
)	
{
    int ret = -1;
	unsigned int ipaddr;
	char strip[32] ;

	unsigned short rtspptr, httpptr, videoptr, targetptr;
	
	

	memset(strip, 0, sizeof(strip));
	strcpy(strip, ip);

	dns_packet_set_field(p, DNS_FIELD_ID, GS_ID);
    dns_packet_set_field(p, DNS_FIELD_FLAGS, DNS_FLAGS(1, 0, 1, 0, 0, 0, 0, 0, 0, 0));
	dns_packet_set_field(p, DNS_FIELD_QDCOUNT, 0);
	dns_packet_set_field(p, DNS_FIELD_ANCOUNT, 1);	//* 3 ANSWER
	dns_packet_set_field(p, DNS_FIELD_ARCOUNT, 1);	//* 7 additions

	/* ==== answer ==== */

	/* http */
    if (!dns_packet_append_name(p, GS_IPC_RTSP)) {
        DEBUG(LOG_ERR,   "dns_packet_append_name: Bad host name [%s]\n",
					GS_IPC_RTSP);
        return -1;
    }
    
 	dns_packet_append_uint16(p, DNS_TYPE_TXT);
	dns_packet_append_uint16(p, DNS_CLASS_IN | 0x8000);
	dns_packet_append_uint32(p, 4500);
	dns_packet_append_uint16(p, strlen(GS_MEDIA_STR)+1);	//* len 
	dns_packet_append_uint8(p, strlen(GS_MEDIA_STR));	//* txtlen
	dns_packet_append_str(p, GS_MEDIA_STR, strlen(GS_MEDIA_STR));
	httpptr = p->size & 0xffff;


	/* ============= additional record =========== */
	
	#if 1
	/* mac */
	if (!dns_packet_append_name(p, GS_IPC_MAC)) {
        DEBUG(LOG_ERR,   "dns_packet_append_name: Bad host name [%s]\n",
					GS_IPC_MAC);
        return -1;
    }
	
	targetptr = ((p->size & 0xffff) - (strlen(GS_IPC_MAC)+2) );
	//printf("targetptr = %x , %x\n", targetptr, p->size);
	dns_packet_append_uint16(p, DNS_TYPE_A);
	dns_packet_append_uint16(p, DNS_CLASS_IN | 0x8000);
	dns_packet_append_uint32(p, 4500);
	dns_packet_append_uint16(p, 4);

	dns_ipstr2int(&ipaddr, strip);
	dns_packet_append_uint32(p, ipaddr);

#if 0
	/* test, add 169.254.200.100 */
	targetptr |= 0xc000;
	//printf("%x \n",httpptr);
	dns_packet_append_uint16(p, targetptr);
	dns_packet_append_uint16(p, DNS_TYPE_A);
	dns_packet_append_uint16(p, DNS_CLASS_IN | 0x8000);
	dns_packet_append_uint32(p, 120);
	dns_packet_append_uint16(p, 4);
	memset(strip, 0, sizeof(strip));
	strcpy(strip, "169.254.200.100");
	dns_ipstr2int(&ipaddr, strip);
	dns_packet_append_uint32(p, ipaddr);
#endif 

	#endif 
	
    
    return 0; 
}


int dns_answer_http_packet(
		struct dns_packet *p, 
		char *req, 
		char *re_name, 
		char *ip,
		int type
)	
{
    int ret = -1;
	unsigned int ipaddr;
	char strip[32] ;

	unsigned short rtspptr, httpptr, videoptr, targetptr;
	
	

	memset(strip, 0, sizeof(strip));
	strcpy(strip, ip);

	dns_packet_set_field(p, DNS_FIELD_ID, GS_ID);
    dns_packet_set_field(p, DNS_FIELD_FLAGS, DNS_FLAGS(1, 0, 1, 0, 0, 0, 0, 0, 0, 0));
	dns_packet_set_field(p, DNS_FIELD_QDCOUNT, 0);
	dns_packet_set_field(p, DNS_FIELD_ANCOUNT, 1);	//* 3 ANSWER
	dns_packet_set_field(p, DNS_FIELD_ARCOUNT, 3);	//* 7 additions

	/* ==== answer ==== */

	/* http */
    if (!dns_packet_append_name(p, GS_IPC_httptcp)) {
        DEBUG(LOG_ERR,   "dns_packet_append_name: Bad host name [%s]\n",
					GS_IPC_httptcp);
        return -1;
    }
    
    dns_packet_append_uint16(p, DNS_TYPE_PTR);
    dns_packet_append_uint16(p, DNS_CLASS_IN);
	dns_packet_append_uint32(p, 4500);
	dns_packet_append_uint16(p, strlen(GS_IPC_HTTP)+2);
    httpptr = p->size & 0xffff;
	if (!dns_packet_append_name(p, GS_IPC_HTTP)) {
        DEBUG(LOG_ERR,   "dns_packet_append_name: Bad host name [%s]\n",
					GS_IPC_HTTP);
        return -1;
    }
	//httpptr = p->size & 0xffff;


	/* ============= additional record =========== */
	
	#if 1
	/* mac */
	if (!dns_packet_append_name(p, GS_IPC_MAC)) {
        DEBUG(LOG_ERR,   "dns_packet_append_name: Bad host name [%s]\n",
					GS_IPC_MAC);
        return -1;
    }
	
	targetptr = ((p->size & 0xffff) - (strlen(GS_IPC_MAC)+2) );
	//printf("targetptr = %x , %x, offset = %d \n", targetptr, p->size, 
	//	strlen(GS_IPC_MAC)+2);
	dns_packet_append_uint16(p, DNS_TYPE_A);
	dns_packet_append_uint16(p, DNS_CLASS_IN | 0x8000);
	dns_packet_append_uint32(p, 4500);
	dns_packet_append_uint16(p, 4);

	dns_ipstr2int(&ipaddr, strip);
	dns_packet_append_uint32(p, ipaddr);

#if 0
	/* test, add 169.254.200.100 */
	targetptr |= 0xc000;
	//printf("%x \n",httpptr);
	dns_packet_append_uint16(p, targetptr);
	dns_packet_append_uint16(p, DNS_TYPE_A);
	dns_packet_append_uint16(p, DNS_CLASS_IN | 0x8000);
	dns_packet_append_uint32(p, 120);
	dns_packet_append_uint16(p, 4);
	memset(strip, 0, sizeof(strip));
	strcpy(strip, "169.254.200.100");
	dns_ipstr2int(&ipaddr, strip);
	dns_packet_append_uint32(p, ipaddr);
#endif 

	/* HTTP SRV */
	httpptr |= 0xc000;
	//printf("%x \n",httpptr);
	dns_packet_append_uint16(p, httpptr);
	#if 0
	if (!dns_packet_append_name(p, GS_IPC_N_MAC)) {
        fprintf(stderr, "Bad host name\n");
        goto finish;
    }
	#endif 
 	dns_packet_append_uint16(p, DNS_TYPE_SRV);
	dns_packet_append_uint16(p, DNS_CLASS_IN | 0x8000);
	dns_packet_append_uint32(p, 4500);	//* time
	dns_packet_append_uint16(p, 8);	//* LEN
	dns_packet_append_uint16(p, 0);	//* priority
	dns_packet_append_uint16(p, 0);	//* weight
	dns_packet_append_uint16(p, get_http_port());	//* port
	targetptr |= 0xc000;
	//printf("%x \n",targetptr);
	dns_packet_append_uint16(p, targetptr);


	/* HTTP TXT */
	httpptr |= 0xc000;
	//printf("%x \n",httpptr);
	dns_packet_append_uint16(p, httpptr);
 	dns_packet_append_uint16(p, DNS_TYPE_TXT);
	dns_packet_append_uint16(p, DNS_CLASS_IN | 0x8000);
	dns_packet_append_uint32(p, 4500);
	dns_packet_append_uint16(p, 1);	//* len 
	dns_packet_append_uint8(p, 0);	//* txtlen

	
	#endif 
	
    
    return 0; 
}


int dns_answer_http_srv_packet(
		struct dns_packet *p, 
		char *req, 
		char *re_name, 
		char *ip,
		int type
)	
{
    int ret = -1;
	unsigned int ipaddr;
	char strip[32] ;

	unsigned short rtspptr, httpptr, videoptr, targetptr;
	
	

	memset(strip, 0, sizeof(strip));
	strcpy(strip, ip);

	dns_packet_set_field(p, DNS_FIELD_ID, GS_ID);
    dns_packet_set_field(p, DNS_FIELD_FLAGS, DNS_FLAGS(1, 0, 1, 0, 0, 0, 0, 0, 0, 0));
	dns_packet_set_field(p, DNS_FIELD_QDCOUNT, 0);
	dns_packet_set_field(p, DNS_FIELD_ANCOUNT, 1);	//* 3 ANSWER
	dns_packet_set_field(p, DNS_FIELD_ARCOUNT, 1);	//* 7 additions

	/* ==== answer ==== */

	/* http */
    if (!dns_packet_append_name(p, GS_IPC_HTTP)) {
        DEBUG(LOG_ERR,   "dns_packet_append_name: Bad host name [%s]\n",
					GS_IPC_HTTP);
        return -1;
    }
    
	dns_packet_append_uint16(p, DNS_TYPE_SRV);
	dns_packet_append_uint16(p, DNS_CLASS_IN | 0x8000);
	dns_packet_append_uint32(p, 4500);	//* time
	dns_packet_append_uint16(p, 8+strlen(GS_IPC_MAC));	//* LEN
	dns_packet_append_uint16(p, 0);	//* priority
	dns_packet_append_uint16(p, 0);	//* weight
	dns_packet_append_uint16(p, get_http_port());	//* port
	printf("strlen(GS_IPC_MAC) = %d , [%s]\n", strlen(GS_IPC_MAC), GS_IPC_MAC);
	if (!dns_packet_append_name(p, GS_IPC_MAC)) {
        DEBUG(LOG_ERR,   "dns_packet_append_name: Bad host name [%s]\n",
					GS_IPC_MAC);
        return -1;
    }
	httpptr = p->size & 0xffff;


	/* ============= additional record =========== */
	
	#if 1
	/* mac */
	if (!dns_packet_append_name(p, GS_IPC_MAC)) {
        DEBUG(LOG_ERR,   "dns_packet_append_name: Bad host name [%s]\n",
					GS_IPC_MAC);
        return -1;
    }
	
	targetptr = ((p->size & 0xffff) - (strlen(GS_IPC_MAC)+2) );
	//printf("targetptr = %x , %x\n", targetptr, p->size);
	dns_packet_append_uint16(p, DNS_TYPE_A);
	dns_packet_append_uint16(p, DNS_CLASS_IN | 0x8000);
	dns_packet_append_uint32(p, 4500);
	dns_packet_append_uint16(p, 4);

	dns_ipstr2int(&ipaddr, strip);
	dns_packet_append_uint32(p, ipaddr);


	#endif 
	
    
    return 0; 
}

int dns_answer_http_txt_packet(
		struct dns_packet *p, 
		char *req, 
		char *re_name, 
		char *ip,
		int type
)	
{
    int ret = -1;
	unsigned int ipaddr;
	char strip[32] ;

	unsigned short rtspptr, httpptr, videoptr, targetptr;
	
	

	memset(strip, 0, sizeof(strip));
	strcpy(strip, ip);

	dns_packet_set_field(p, DNS_FIELD_ID, GS_ID);
    dns_packet_set_field(p, DNS_FIELD_FLAGS, DNS_FLAGS(1, 0, 1, 0, 0, 0, 0, 0, 0, 0));
	dns_packet_set_field(p, DNS_FIELD_QDCOUNT, 0);
	dns_packet_set_field(p, DNS_FIELD_ANCOUNT, 1);	//* 3 ANSWER
	dns_packet_set_field(p, DNS_FIELD_ARCOUNT, 1);	//* 7 additions

	/* ==== answer ==== */

	/* http */
    if (!dns_packet_append_name(p, GS_IPC_HTTP)) {
        DEBUG(LOG_ERR,   "dns_packet_append_name: Bad host name [%s]\n",
					GS_IPC_HTTP);
        return -1;
    }
    
 	dns_packet_append_uint16(p, DNS_TYPE_TXT);
	dns_packet_append_uint16(p, DNS_CLASS_IN | 0x8000);
	dns_packet_append_uint32(p, 4500);
	dns_packet_append_uint16(p, 1);	//* len 
	dns_packet_append_uint8(p, 0);	//* txtlen
	httpptr = p->size & 0xffff;


	/* ============= additional record =========== */
	
	#if 1
	/* mac */
	if (!dns_packet_append_name(p, GS_IPC_MAC)) {
        DEBUG(LOG_ERR,   "dns_packet_append_name: Bad host name [%s]\n",
					GS_IPC_MAC);
        return -1;
    }
	
	targetptr = ((p->size & 0xffff) - (strlen(GS_IPC_MAC)+2) );
	//printf("targetptr = %x , %x\n", targetptr, p->size);
	dns_packet_append_uint16(p, DNS_TYPE_A);
	dns_packet_append_uint16(p, DNS_CLASS_IN | 0x8000);
	dns_packet_append_uint32(p, 4500);
	dns_packet_append_uint16(p, 4);

	dns_ipstr2int(&ipaddr, strip);
	dns_packet_append_uint32(p, ipaddr);

	#endif 
	
    
    return 0; 
}


int dns_answer_A_packet(
		struct dns_packet *p, 
		char *req, 
		char *ip, 
		int type
)	
{
	char ipstr[32];
	unsigned int ipaddr;
	unsigned short targetptr = 0; 

	memset(ipstr, 0, sizeof(ipstr));
	strcpy(ipstr, ip);

	dns_packet_set_field(p, DNS_FIELD_ID, GS_ID);
    dns_packet_set_field(p, DNS_FIELD_FLAGS, DNS_FLAGS(1, 0, 1, 0, 0, 0, 0, 0, 0, 0));
	dns_packet_set_field(p, DNS_FIELD_QDCOUNT, 0);
	dns_packet_set_field(p, DNS_FIELD_ANCOUNT, 1);
	
    if (!dns_packet_append_name(p, req)) {
        DEBUG(LOG_ERR,   "dns_packet_append_name: Bad host name [%s]\n",
					req);
        return -1;
    }
	targetptr = ((p->size & 0xffff) - (strlen(GS_IPC_MAC)+2) );
	
 	dns_packet_append_uint16(p, DNS_TYPE_A);
	dns_packet_append_uint16(p, DNS_CLASS_IN | 0x8000);
	dns_packet_append_uint32(p, 4500);
	dns_packet_append_uint16(p, 4);

	dns_ipstr2int(&ipaddr, ipstr);
	dns_packet_append_uint32(p, ipaddr);


    return 0;
}

#define QU 	(0X8000)


int mdns_broadcast_QU_QM(struct dns_packet *p, char *ipstr, int q )
{
	unsigned short ptr_mac = 0;
	unsigned short ptr_http = 0; 
	unsigned short ptr_rtsp = 0; 
	unsigned short ptr_video = 0;
	unsigned int	ipaddr = 0; 
	unsigned int	flag_qu = 0;

	if( p == NULL ){
		DEBUG(LOG_ERR,  "input params error \n");
		return -1; 
	}

	if( q == 1 ){
		flag_qu = QU; 
	}
	//* HEAD 
	dns_packet_set_field(p, DNS_FIELD_ID, GS_ID);
    dns_packet_set_field(p, DNS_FIELD_FLAGS, DNS_FLAGS(0, 0, 0, 0, 0, 0, 0, 0, 0, 0));
	dns_packet_set_field(p, DNS_FIELD_QDCOUNT, 4);
	dns_packet_set_field(p, DNS_FIELD_NSCOUNT, 4);


	// QUERIES =========================================
	//* mac
	if (!dns_packet_append_name(p, GS_IPC_MAC)) {
        DEBUG(LOG_ERR, "Bad host name [%s]\n", GS_IPC_MAC);
        return -1;
    }
	ptr_mac = (((p->size & 0xffff) - (strlen(GS_IPC_MAC)+2) ) | 0xc000);
    dns_packet_append_uint16(p, DNS_TYPE_ANY);
    dns_packet_append_uint16(p, DNS_CLASS_IN | flag_qu);	//* class type

	//* http
	ptr_http = ((p->size & 0xffff)|0xc000);
	if (!dns_packet_append_name(p, GS_IPC_HTTP)) {
        DEBUG(LOG_ERR, "Bad host name [%s]\n", GS_IPC_HTTP);
        return -1;
    }
	//ptr_http = p->size & 0xffff;
    dns_packet_append_uint16(p, DNS_TYPE_ANY);
    dns_packet_append_uint16(p, DNS_CLASS_IN | flag_qu);	//* class type

	//* RTSP
	ptr_rtsp = ((p->size & 0xffff)|0xc000);
	if (!dns_packet_append_name(p, GS_IPC_RTSP)) {
        DEBUG(LOG_ERR, "Bad host name [%s]\n", GS_IPC_RTSP);
        return -1;
    }
	//ptr_rtsp = p->size & 0xffff;
    dns_packet_append_uint16(p, DNS_TYPE_ANY);
    dns_packet_append_uint16(p, DNS_CLASS_IN | flag_qu);	//* class type

	//* VIDEO
	ptr_video = ((p->size & 0xffff)|0xc000);
	if (!dns_packet_append_name(p, GS_IPC_VIDEO)) {
        DEBUG(LOG_ERR, "Bad host name [%s]\n", GS_IPC_VIDEO);
        return -1;
    }
	//ptr_video = p->size & 0xffff;
    dns_packet_append_uint16(p, DNS_TYPE_ANY);
    dns_packet_append_uint16(p, DNS_CLASS_IN | flag_qu);	//* class type


	// AUTHORITATIVE NAMESERVER   =========================
	//* A record
	ptr_mac |= 0xc000;
	dns_packet_append_uint16(p, ptr_mac);
 	dns_packet_append_uint16(p, DNS_TYPE_A);
	dns_packet_append_uint16(p, DNS_CLASS_IN );
	dns_packet_append_uint32(p, 120);
	dns_packet_append_uint16(p, 4);
	dns_ipstr2int(&ipaddr, ipstr);
	dns_packet_append_uint32(p, ipaddr);


	//* http srv 
	dns_packet_append_uint16(p, ptr_http);
 	dns_packet_append_uint16(p, DNS_TYPE_SRV);
	dns_packet_append_uint16(p, DNS_CLASS_IN );
	dns_packet_append_uint32(p, 120);	//* time
	dns_packet_append_uint16(p, 8);	//* LEN
	dns_packet_append_uint16(p, 0);	//* priority
	dns_packet_append_uint16(p, 0);	//* weight
	dns_packet_append_uint16(p, get_http_port());	//* port
	dns_packet_append_uint16(p, ptr_mac);

	//* rtsp srv 
	dns_packet_append_uint16(p, ptr_rtsp);
 	dns_packet_append_uint16(p, DNS_TYPE_SRV);
	dns_packet_append_uint16(p, DNS_CLASS_IN );
	dns_packet_append_uint32(p, 120);	//* time
	dns_packet_append_uint16(p, 8);	//* LEN
	dns_packet_append_uint16(p, 0);	//* priority
	dns_packet_append_uint16(p, 0);	//* weight
	dns_packet_append_uint16(p, get_rtsp_port());	//* port
	ptr_mac |= 0xc000;
	dns_packet_append_uint16(p, ptr_mac);

	//* video srv 
	dns_packet_append_uint16(p, ptr_video );
 	dns_packet_append_uint16(p, DNS_TYPE_SRV);
	dns_packet_append_uint16(p, DNS_CLASS_IN );
	dns_packet_append_uint32(p, 120);	//* time
	dns_packet_append_uint16(p, 8);	//* LEN
	dns_packet_append_uint16(p, 0);	//* priority
	dns_packet_append_uint16(p, 0);	//* weight
	dns_packet_append_uint16(p, get_http_port());	//* port
	dns_packet_append_uint16(p, ptr_mac);

	return 0; 
}

int mdns_get_ipstr(char *str, unsigned int ip)
{
	if( str == NULL || ip == 0 ){
		DEBUG(LOG_ERR,  "input params error \n");
		return -1;
	}

	sprintf(str, "%d.%d.%d.%d.in-addr.arap", ip&0xff, ((ip>>8)&0xff),
		((ip>>16)&0xff),((ip>>24)&0xff));

	return 0; 
}

int mdns_broadcast_A(struct dns_packet *p, char *ipstr )
{
	unsigned short ptr_mac = 0;
	unsigned short ptr_http = 0; 
	unsigned short ptr_rtsp = 0; 
	unsigned short ptr_video = 0; 
	unsigned short ptr_http_tcp = 0; 
	unsigned short ptr_rtsp_tcp = 0; 
	unsigned short ptr_video_tcp = 0; 
	unsigned short	ptr_rr = 0; 
	unsigned int   ipaddr = 0 ;
	char			ip_in_addr[64] = {0};

	if( p == NULL || ipstr == NULL ){
		DEBUG(LOG_ERR,  "input params error \n");
		return -1; 
	}

	//* head
	dns_packet_set_field(p, DNS_FIELD_ID, GS_ID);
    dns_packet_set_field(p, DNS_FIELD_FLAGS, DNS_FLAGS(1, 0, 1, 0, 0, 0, 0, 0, 0, 0));
	dns_packet_set_field(p, DNS_FIELD_QDCOUNT, 0);
	dns_packet_set_field(p, DNS_FIELD_ANCOUNT, 14);


	//* answer 
	//* mac 
    if (!dns_packet_append_name(p, GS_IPC_MAC)) {
        DEBUG(LOG_ERR,   "dns_packet_append_name: Bad host name [%s]\n",
					GS_IPC_MAC);
        return -1;
    }
	ptr_mac = ((p->size & 0xffff) - (strlen(GS_IPC_MAC)+2) );
	ptr_mac |= 0xc000;
	
 	dns_packet_append_uint16(p, DNS_TYPE_A);
	dns_packet_append_uint16(p, DNS_CLASS_IN | QU);
	dns_packet_append_uint32(p, 120);
	dns_packet_append_uint16(p, 4);
	dns_ipstr2int(&ipaddr, ipstr);
	dns_packet_append_uint32(p, ipaddr);

	//* ip-in_addr
	memset(ip_in_addr, 0, sizeof(ip_in_addr));
	mdns_get_ipstr(ip_in_addr, ipaddr);
	printf("ip-inaddr = [%s] \n", ip_in_addr);
	 if (!dns_packet_append_name(p, ip_in_addr)) {
        DEBUG(LOG_ERR,   "dns_packet_append_name: Bad host name [%s]\n",
					GS_IPC_MAC);
        return -1;
    }
	
 	dns_packet_append_uint16(p, DNS_TYPE_PTR);
	dns_packet_append_uint16(p, DNS_CLASS_IN | QU);
	dns_packet_append_uint32(p, 120);
	dns_packet_append_uint16(p, 2);	//* len 
	dns_packet_append_uint16(p, ptr_mac);

//* ======================= http ===============
	//* http srv 
	ptr_http = (p->size & 0xffff) | 0xc000;
	ptr_http_tcp = ptr_http + (strlen(GS_IPC_HTTP)-strlen(GS_IPC_httptcp));
	
	if (!dns_packet_append_name(p, GS_IPC_HTTP)) {
        DEBUG(LOG_ERR,   "dns_packet_append_name: Bad host name [%s]\n",
					GS_IPC_HTTP);
        return -1;
    }
	
 	dns_packet_append_uint16(p, DNS_TYPE_SRV);
	dns_packet_append_uint16(p, DNS_CLASS_IN | QU );
	dns_packet_append_uint32(p, 120);	//* time
	dns_packet_append_uint16(p, 8);	//* LEN
	dns_packet_append_uint16(p, 0);	//* priority
	dns_packet_append_uint16(p, 0);	//* weight
	dns_packet_append_uint16(p, get_http_port());	//* port
	dns_packet_append_uint16(p, ptr_mac);

	/* http txt */

	dns_packet_append_uint16(p, ptr_http);
 	dns_packet_append_uint16(p, DNS_TYPE_TXT);
	dns_packet_append_uint16(p, DNS_CLASS_IN | QU);
	dns_packet_append_uint32(p, 4500);
	dns_packet_append_uint16(p, 1);	//* len 
	dns_packet_append_uint8(p, 0);	//* txtlen

	//* answer key [_services._dns-sd._udp.local] 
	ptr_rr = (p->size & 0xffff) | 0xc000;
	if (!dns_packet_append_name(p, SERVICE_RR)) {
        DEBUG(LOG_ERR,   "dns_packet_append_name: Bad host name [%s]\n",
					SERVICE_RR);
        return -1;
    }
	
	
	dns_packet_append_uint16(p, DNS_TYPE_PTR);
    dns_packet_append_uint16(p, DNS_CLASS_IN);
	dns_packet_append_uint32(p, 4500);	//* 2 min

#if 1
	dns_packet_append_uint16(p, 2); //len 2
	dns_packet_append_uint16(p, ptr_http_tcp);
	
#else
	dns_packet_append_uint16(p, strlen(GS_IPC_httptcp)+2);
	if (!dns_packet_append_name(p, GS_IPC_httptcp)) {
		DEBUG(LOG_ERR,  "dns_packet_append_name [%s] error \n", 
			GS_IPC_httptcp);
        return -1;
    }
#endif 

	//* answer key [_http._tcp.local] 
	dns_packet_append_uint16(p, ptr_http_tcp);
	dns_packet_append_uint16(p, DNS_TYPE_PTR);
    dns_packet_append_uint16(p, DNS_CLASS_IN);
	dns_packet_append_uint32(p, 4500);	//* 2 min
	dns_packet_append_uint16(p, 2); //len 2
	dns_packet_append_uint16(p, ptr_http);
	
//* ==========================RTSP=====================================

	//* rtsp srv 
	ptr_rtsp = (p->size & 0xffff) | 0xc000;
	ptr_rtsp_tcp = ptr_rtsp + (strlen(GS_IPC_RTSP)-strlen(GS_IPC_RTSP));
	if (!dns_packet_append_name(p, GS_IPC_RTSP)) {
        DEBUG(LOG_ERR,   "dns_packet_append_name: Bad host name [%s]\n",
					GS_IPC_RTSP);
        return -1;
    }
	
 	dns_packet_append_uint16(p, DNS_TYPE_SRV);
	dns_packet_append_uint16(p, DNS_CLASS_IN | QU );
	dns_packet_append_uint32(p, 120);	//* time
	dns_packet_append_uint16(p, 8);	//* LEN
	dns_packet_append_uint16(p, 0);	//* priority
	dns_packet_append_uint16(p, 0);	//* weight
	dns_packet_append_uint16(p, get_rtsp_port());	//* port
	dns_packet_append_uint16(p, ptr_mac);

	/* rtsp txt */
	dns_packet_append_uint16(p, ptr_rtsp);
 	dns_packet_append_uint16(p, DNS_TYPE_TXT);
	dns_packet_append_uint16(p, DNS_CLASS_IN | QU);
	dns_packet_append_uint32(p, 4500);
	dns_packet_append_uint16(p, 1);	//* len 
	dns_packet_append_uint8(p, 0);	//* txtlen

	//* answer key [_services._dns-sd._udp.local] 
	dns_packet_append_uint16(p, ptr_rr);
	dns_packet_append_uint16(p, DNS_TYPE_PTR);
    dns_packet_append_uint16(p, DNS_CLASS_IN);
	dns_packet_append_uint32(p, 4500);	//* 2 min

#if 1
	dns_packet_append_uint16(p, 2); //len 2
	dns_packet_append_uint16(p, ptr_rtsp_tcp);
	
#else
	dns_packet_append_uint16(p, strlen(GS_IPC_httptcp)+2);
	if (!dns_packet_append_name(p, GS_IPC_httptcp)) {
		DEBUG(LOG_ERR,  "dns_packet_append_name [%s] error \n", 
			GS_IPC_httptcp);
        return -1;
    }
#endif 

	//* answer key [_rtsp._tcp.local] 
	dns_packet_append_uint16(p, ptr_rtsp_tcp);
	dns_packet_append_uint16(p, DNS_TYPE_PTR);
    dns_packet_append_uint16(p, DNS_CLASS_IN);
	dns_packet_append_uint32(p, 4500);	//* 2 min
	dns_packet_append_uint16(p, 2); //len 2
	dns_packet_append_uint16(p, ptr_rtsp);

//* ========================VIDEO=======================================

	//* video srv 
	ptr_video = (p->size & 0xffff) | 0xc000;
	ptr_video_tcp = ptr_rtsp + (strlen(GS_IPC_VIDEO)-strlen(GS_IPC_VIDEO));
	
	if (!dns_packet_append_name(p, GS_IPC_VIDEO)) {
        DEBUG(LOG_ERR,   "dns_packet_append_name: Bad host name [%s]\n",
					GS_IPC_VIDEO);
        return -1;
    }
	
 	dns_packet_append_uint16(p, DNS_TYPE_SRV);
	dns_packet_append_uint16(p, DNS_CLASS_IN | QU );
	dns_packet_append_uint32(p, 120);	//* time
	dns_packet_append_uint16(p, 8);	//* LEN
	dns_packet_append_uint16(p, 0);	//* priority
	dns_packet_append_uint16(p, 0);	//* weight
	dns_packet_append_uint16(p, get_http_port());	//* port
	dns_packet_append_uint16(p, ptr_mac);

	/* video txt */
	dns_packet_append_uint16(p, ptr_video);
 	dns_packet_append_uint16(p, DNS_TYPE_TXT);
	dns_packet_append_uint16(p, DNS_CLASS_IN | QU);
	dns_packet_append_uint32(p, 4500);
	dns_packet_append_uint16(p, 1);	//* len 
	dns_packet_append_uint8(p, 0);	//* txtlen

	//* answer key [_services._dns-sd._udp.local] 
	dns_packet_append_uint16(p, ptr_rr);
	dns_packet_append_uint16(p, DNS_TYPE_PTR);
    dns_packet_append_uint16(p, DNS_CLASS_IN);
	dns_packet_append_uint32(p, 4500);	//* 2 min

#if 1
	dns_packet_append_uint16(p, 2); //len 2
	dns_packet_append_uint16(p, ptr_video_tcp);
	
#else
	dns_packet_append_uint16(p, strlen(GS_IPC_httptcp)+2);
	if (!dns_packet_append_name(p, GS_IPC_httptcp)) {
		DEBUG(LOG_ERR,  "dns_packet_append_name [%s] error \n", 
			GS_IPC_httptcp);
        return -1;
    }
#endif 

	//* answer key [_Grandstream-video._tcp.local] 
	dns_packet_append_uint16(p, ptr_video_tcp);
	dns_packet_append_uint16(p, DNS_TYPE_PTR);
    dns_packet_append_uint16(p, DNS_CLASS_IN);
	dns_packet_append_uint32(p, 4500);	//* 2 min
	dns_packet_append_uint16(p, 2); //len 2
	dns_packet_append_uint16(p, ptr_video);

	return 0; 

}




struct dns_packet * dns_packet_init(struct dns_packet *p)
{
	memset(p, 0, sizeof(struct dns_packet));
	p->size = p->rindex = 2*6;
	return p;
}

int dns_send(int fd, struct sockaddr_in *dst, struct dns_packet *p)
{
	struct sockaddr_in	*sdst = NULL; 
	struct sockaddr_in mdns_addr;

	if( !dst ){
		memset(&mdns_addr, 0, sizeof(mdns_addr));
		mdns_addr.sin_family = AF_INET;
		mdns_addr.sin_port = htons(MDNS_PORT);
		if (inet_pton(AF_INET, MDNS_IPADDR, &mdns_addr.sin_addr) <= 0) {
			DEBUG(LOG_ERR,  "Wrong dest IP address [%s]!\n", MDNS_IPADDR);
			return -1;
		}
		sdst = &mdns_addr;
	}else{
		sdst = dst;
	}
	if (sendto(fd, p->data, p->size, 0,(struct sockaddr *) sdst, 
		sizeof(struct sockaddr_in)) < 0) {
		DEBUG(LOG_ERR,  "sendto error! \n");
		return -1;
	}
	return 0; 
}

int check_peer(struct sockaddr_in *p, char *ip)
{
	unsigned char ipstr[32];
	
	memset(ipstr, 0, sizeof(ipstr));
	inet_ntop(AF_INET, &p->sin_addr, ipstr, 16);

	if( 0 == strcmp(ipstr, ip) ){
		return 1;
	}
	return 0; 
}

int mdns_answer_QM_packet(
	struct dns_packet *p,
		char *req, 
		char *re_name, 
		int type
)
{
	if( p == NULL || req == NULL || re_name == NULL ){
		printf("input params error \n");
		return -1; 
	}

	dns_packet_set_field(p, DNS_FIELD_ID, GS_ID);
    dns_packet_set_field(p, DNS_FIELD_FLAGS, DNS_FLAGS(1, 0, 1, 0, 0, 0, 0, 0, 0, 0));
	dns_packet_set_field(p, DNS_FIELD_QDCOUNT, 0);
	dns_packet_set_field(p, DNS_FIELD_ANCOUNT, 1);
	

	if (!dns_packet_append_name(p, req)) {
        DEBUG(LOG_ERR,   "dns_packet_append_name: Bad host name [%s]\n",
					req);
        return -1;
    }

    dns_packet_append_uint16(p, type);
    dns_packet_append_uint16(p, DNS_CLASS_IN);
	dns_packet_append_uint32(p, 4500);	//* 2 min
	dns_packet_append_uint16(p, strlen(re_name)+2);

	if (!dns_packet_append_name(p, re_name)) {
        DEBUG(LOG_ERR,   "dns_packet_append_name: Bad host name [%s]\n",
					re_name);
        return -1;
    }


	return 0;
}

#include <stdio.h>      
#include <sys/types.h>
#include <ifaddrs.h>
#include <netinet/in.h> 
#include <string.h> 
#include <arpa/inet.h>

int get_local_ip (void) {
    struct ifaddrs * ifAddrStruct=NULL, *h = NULL;
    void * tmpAddrPtr=NULL;

    getifaddrs(&ifAddrStruct);
	//DEBUG(LOG_DEBUG, "######>>>> getifaddrs \n");
	h = ifAddrStruct;
    while (ifAddrStruct!=NULL) {
        if ((NULL != ifAddrStruct->ifa_addr)&&
			(ifAddrStruct->ifa_addr->sa_family==AF_INET)) { // check it is IP4
            // is a valid IP4 Address
            tmpAddrPtr=&((struct sockaddr_in *)ifAddrStruct->ifa_addr)->sin_addr;
            char addressBuffer[INET_ADDRSTRLEN];
            inet_ntop(AF_INET, tmpAddrPtr, addressBuffer, INET_ADDRSTRLEN);
			if( strcmp(ifAddrStruct->ifa_name, "lo") != 0 ){
            	//printf("%s IP Address %s \n", ifAddrStruct->ifa_name, addressBuffer); 
			}
        } 
#if 0	// IPV6
		else if (ifAddrStruct->ifa_addr->sa_family==AF_INET6) { // check it is IP6
            // is a valid IP6 Address
            tmpAddrPtr=&((struct sockaddr_in *)ifAddrStruct->ifa_addr)->sin_addr;
            char addressBuffer[INET6_ADDRSTRLEN];
            inet_ntop(AF_INET6, tmpAddrPtr, addressBuffer, INET6_ADDRSTRLEN);
            printf("%s IP Address %s/n", ifAddrStruct->ifa_name, addressBuffer); 
        } 
#endif 
        ifAddrStruct=ifAddrStruct->ifa_next;
    }
	if( h != NULL ){
		freeifaddrs(h);
		ifAddrStruct = NULL; 
		h = NULL;
		///DEBUG(LOG_DEBUG, "$$$$$$$$>>>> freeifaddrs \n");

	}
    return 0;
}

#define MAC_PATH		"/proc/gxp/dev_info/dev_mac"
#define NAME_PATH		"/proc/gxp/dev_info/dev_id"
#define OEMID_PATH		"/proc/gs/cfg/oem_id"

int xhtoi(char *str,unsigned int *l_value,unsigned int *h_value) 
{ 
 	int i = 0,num = 0; 
   	unsigned int low_value = 0;
   	unsigned int up_value = 0;
	#if 0
	if (s[0] == '0' && (s[1]=='x' || s[1]=='X')) //жǷǰ0x0X
   	{ 
		 i = 2; 
	} 
	else
	{ 
 		i = 0; 
	} 
	#endif
	if((NULL == str) || ((num = strlen(str))<1) || (NULL == l_value))
	{
		DEBUG(LOG_ERR,   "str is NULL \n");
		return (-1);
	}
	
	if(num > 16)
	{
		DEBUG(LOG_ERR,   "str is to length \n");
		return (-1);
	}
	if(num > 8)
	{
		if(NULL == h_value)
		{
			DEBUG(LOG_ERR,   "hight_value is NULL \n");
			return (-1);
		}
		for (; (str[i] >= '0' && str[i] <= '9')|| (str[i] >= 'a' && str[i] <= 'z') || (str[i] >='A' && str[i] <= 'Z');++i)
		{  
			if(i >= (num-8))
				break;
			if (tolower(str[i]) > '9') 
			{
				up_value = 16 * up_value + (10 + tolower(str[i]) - 'a');
			}
			else 
			{ 
				up_value = 16 * up_value + (tolower(str[i]) - '0'); 
			}
		}
		*h_value = up_value;
		i = num-8;
		for (; (str[i] >= '0' && str[i] <= '9')|| (str[i] >= 'a' && str[i] <= 'z') || (str[i] >='A' && str[i] <= 'Z');++i)
		{   
			if(i >= 16)
				break;
			if (tolower(str[i]) > '9') 
			{
				low_value = 16 * low_value + (10 + tolower(str[i]) - 'a');
			}
			else 
			{ 
				low_value = 16 * low_value + (tolower(str[i]) - '0'); 
			}
		}
	}
	else
	{
		for (; (str[i] >= '0' && str[i] <= '9')|| (str[i] >= 'a' && str[i] <= 'z') || (str[i] >='A' && str[i] <= 'Z');++i)
		{   
			if(i >= 8)
				break;
			if (tolower(str[i]) > '9') 
			{
				low_value = 16 * low_value + (10 + tolower(str[i]) - 'a');
			}
			else 
			{ 
				low_value = 16 * low_value + (tolower(str[i]) - '0'); 
			}
		}
	}
	*l_value = low_value;
	return 0;
} 


/*ȥո*/
char *alltrim(char *str) 
{ 
	char *p=str;
	char *q = NULL;
	while(*p!=0){
		if(*p==':' || *p == ' '){ 
			q = p;
			while(*q!=0)
			{ 
				*q=*(q+1); 
				q++; 
			}  
		} 
		else{ 
	 		p++;
		}
	}
	return str;
} 



/*ȥոַ*/
int get_cfg_params(char *cfg,char *buf)
{
	char buft[128] = {0};	 
	FILE *pPn = NULL;
	if(NULL == cfg || NULL == buf){
		DEBUG(LOG_ERR,   "cfg or buf is NULL \n");
		return -1; 
	}
	
	pPn = fopen(cfg, "r");
	if(pPn == NULL){
		DEBUG(LOG_ERR,   "fopen cfg:%s error \n",cfg);
		return -1;
	}

	memset(buft, 0, 128);
	if(NULL == fgets(buft, 128, pPn)){
		fclose(pPn);
		pPn = NULL; 
		DEBUG(LOG_ERR,   "fgets cfg:%s error \n",cfg);
		return -1; 
	}
	alltrim(buft);
	strcat(buf,buft);
	fclose(pPn);
	return 0;
}

int get_cfg_2bytes(char *filename,int *first,int *last)
{
	int ret = 0;
	unsigned int value = 0;

	char buf[128]={0};
	if((NULL == filename)||(NULL == first)||(NULL == last))
	{
		DEBUG(LOG_ERR,   "Input value is NULL\n");
		return -1; 
	}
	/*80 03 05 00 03 C0#*/
	ret = get_cfg_params(filename, buf);
	if(0 != ret)
	{
		DEBUG(LOG_ERR,   "Proc_Get_CFG error \n");
		return -1; 
	}
	ret = xhtoi(buf,&value,NULL);
	if(0 != ret)
	{
		DEBUG(LOG_ERR,   "xhtoi error \n");
		return -1; 
	}
	/*01 00#: 01=>first 00=>last*/
	*last = (value&0x0ff);
	*first = ((value&0xff00)>>8);
	return 0;
}

#define LIMIT_OEMID		0x08


int get_cfg_oemid(unsigned int *oemId,unsigned int *cfg_loc)
{
	int ret = 0;
	int before = 0,last = 0;
	*oemId = 0;
	return 0;
	ret = get_cfg_2bytes(OEMID_PATH, &before, &last);
	if(ret != 0){
		DEBUG(LOG_ERR,   "Get2ByteValue_ByCfg error \n");
		return (-1);
	}
	if(NULL != oemId)
	{
		*oemId = before;
		//DEBUG(LOG_INFO,  "Get OemId is 0x%x\n", before);
	}
	if(NULL != cfg_loc)
		*cfg_loc = last;
	
	return 0;
}


int get_dev_mac(unsigned char *vendor)
{
	char buft[128] = {0};	 
	FILE *pPn = NULL;

	unsigned int oemid;
	
	get_cfg_oemid(&oemid, NULL);

	pPn = fopen(MAC_PATH, "r");
	if(pPn == NULL){
		DEBUG(LOG_ERR,   "fopen cfg:%s error \n", MAC_PATH);
		return -1;
	}

	memset(gsipc_mac_str, 0, sizeof(gsipc_mac_str));
	memset(buft, 0, 128);
	//if( oemid == LIMIT_OEMID ){
	if(0){
		get_cfg_params(NAME_PATH, gsipc_mac_str);
	}else{
		strcpy(gsipc_mac_str, vendor);
	}
	strcat(gsipc_mac_str, "-");
	if(NULL == fgets(buft, 128, pPn)){
		fclose(pPn);
		pPn = NULL; 
		DEBUG(LOG_ERR,   "fgets cfg:%s error \n",MAC_PATH);
		return -1; 
	}
	alltrim(buft);
	strcat(gsipc_mac_str,buft);
	strcat(gsipc_mac_str, ".local");
	fclose(pPn);
	
	return 0; 
}

void get_macaddr(void)
{
	memset(gsipc_macaddr, 0, sizeof(gsipc_macaddr));
	strcpy(gsipc_macaddr, "macaddress=");
	get_cfg_params(MAC_PATH, gsipc_macaddr);
}

void get_params_rtsph264(unsigned char *p, char *vendor, const char *end)
{
	unsigned int oemid;
	get_cfg_oemid(&oemid, NULL);
//	if( oemid != LIMIT_OEMID ){
	//if( strstr(end, "video") ){
	if( 1){
		strcpy(p, vendor);
		strcat(p, " ");
	}
	
	get_cfg_params(NAME_PATH, p);
	strcat(p, " - ");
	get_cfg_params(MAC_PATH, p);

	strcat(p, " H.264.");
	strcat(p, end);
	
}

void get_params(unsigned char *p, char *vendor, const char *end)
{
	unsigned int oemid;
	get_cfg_oemid(&oemid, NULL);
//	if( oemid != LIMIT_OEMID ){
	//if( strstr(end, "video") ){
	if( 1){
		strcpy(p, vendor);
		strcat(p, " ");
	}
	
	get_cfg_params(NAME_PATH, p);
	strcat(p, " - ");
	get_cfg_params(MAC_PATH, p);

	strcat(p, ".");

	strcat(p, end);
	
}



//"_ipc-video._tcp.local"
void get_local(
	unsigned char *p, 
	unsigned char *vendor, 
	const char *end
)
{
	unsigned int oemid = 0;
	get_cfg_oemid(&oemid, NULL);
	strcpy(p, "_");
	//if( oemid != LIMIT_OEMID ){
	if(strstr(end, "video")){
		strcat(p, vendor);
		strcat(p, "-");
	}
	strcat(p, end);
	
}

typedef struct ServerConf_s{
	int enable;
	char name[16];
	char *buf;
}ServerConf_t;

typedef struct WebCfg_s{
	int webPort;	/* web˿ */
	int useSSL;		/* SSLܿ */
	int	rtsp_port; 	/* RTSP PORT */
#ifdef __ENABLE_USER_DEFINE_RTSP_PORT__
	int rtspPortSelfDefineEnable;	/* ԶRTSP ˿ñ־ */
	int	rtspPort;	/* RTSP˿ */
#endif 
}WebCfg_t;

WebCfg_t 	gweb_config;
#define 	DEFAULT_RTSP_PORT			554
#define 	RTSP_PORT_OFFSET			2000

#define 	P_WEB_PORT					"P901"
#define 	P_RTSP_PORT					"P903"

int get_web_config(void)
{
	char *p;
	p = nvram_get(P_WEB_PORT);
	if(p == NULL){
		gweb_config.webPort = 80;
	}else{
		gweb_config.webPort = atoi(p);
		if(gweb_config.webPort <= 80 || gweb_config.webPort >= 65535){
			gweb_config.webPort = 80;
		}
	}

	p = nvram_get(P_RTSP_PORT);
	if(p == NULL){
		gweb_config.rtsp_port= 554;
	}else{
		gweb_config.rtsp_port = atoi(p);
		if(gweb_config.rtsp_port <= 0 || gweb_config.webPort >= 65535){
			gweb_config.rtsp_port = 554;
		}
	}

	

	//DEBUG(LOG_DEBUG,  "Http.port = %d, Rtsp.port = %d \n", 
	//	gweb_config.webPort, gweb_config.rtsp_port);

	//printf("Http.port = %d, Rtsp.port = %d \n", 
	//	gweb_config.webPort, gweb_config.rtsp_port);

//	DB_Close();
	return 0; 
}

int get_http_port(void)
{
	return gweb_config.webPort;
}

int get_rtsp_port(void)
{
	return gweb_config.rtsp_port;
}

void dev_info_init(void)
{
	unsigned int oemid;

	memset(vendor, 0, sizeof(vendor));
	memset(gsipc_rtsp, 0, sizeof(gsipc_rtsp));
	memset(gsipc_video, 0, sizeof(gsipc_video));
	memset(gsipc_http, 0, sizeof(gsipc_http));
	memset(local_http, 0, sizeof(local_http));
	memset(local_video, 0, sizeof(local_video));
	memset(local_rtsp, 0, sizeof(local_rtsp));
	memset(gsipc_rtsp_h264, 0, sizeof(gsipc_rtsp_h264));

	get_web_config();
	get_cfg_oemid(&oemid, NULL);
	get_cfg_oem_name(vendor, oemid);
	get_macaddr();
	get_dev_mac(vendor);

	get_local(local_http, vendor, MDNS_LOCAL_HTTP);
	get_local(local_rtsp, vendor, MDNS_LOCAL_RTSP);
	get_local(local_video, vendor, MDNS_LOCAL_VIDEO);

	get_params(gsipc_rtsp, vendor, GS_IPC_rtsptcp);
	get_params(gsipc_video, vendor, GS_IPC_gsvideo);
	get_params(gsipc_http, vendor, GS_IPC_httptcp);
	get_params_rtsph264(gsipc_rtsp_h264, vendor, GS_IPC_rtsptcp);

#if 0
	if( oemid == 0x08 ){
		//* legrand oem id 
		//memset(gsipc_http, 0, sizeof(gsipc_http));
		memset(local_http, 0, sizeof(local_http));
		strcpy(local_http, "_http._tcp.local");
	}
#endif 

#if 1
	doname[0] = GS_IPC_httptcp;
	doname[1] = GS_IPC_rtsptcp;
	doname[2] = GS_IPC_gsvideo;
	//doname[2] = GS_IPC_gsvideo;
	//doname[1] = GS_IPC_gsvideo;
	//doname[2] = GS_IPC_rtsptcp;
#else
	doname[0] = GS_IPC_gsvideo;
#endif 

	DEBUG(LOG_DEBUG,  "DEVINFO: \n"
		"MAC: [%s] \n"
		"MACADDR: [%s] \n"
		"GS_IPC_RTSP: [%s] \n"
		"GS_IPC_VIDEO: [%s] \n"
		"GS_IPC_HTTP: [%s] \n"
		"LOCAL_VIDEO: [%s] \n"
		"LOCAL_RTSP: [%s] \n"
		"LOCAL_HTTP: [%s] \n"
		"gsipc_rtsp_h264: [%s]\n"
		, GS_IPC_MAC, 
		GS_MAC_ADDR,
		GS_IPC_RTSP, 
		GS_IPC_VIDEO, 
		GS_IPC_HTTP,
		GS_IPC_gsvideo,
		GS_IPC_rtsptcp, 
		GS_IPC_httptcp, 
		gsipc_rtsp_h264);
#if 0
	printf( "DEVINFO: \n"
		"MAC: [%s] \n"
		"MACADDR: [%s] \n"
		"GS_IPC_RTSP: [%s] \n"
		"GS_IPC_VIDEO: [%s] \n"
		"GS_IPC_HTTP: [%s] \n"
		"LOCAL_VIDEO: [%s] \n"
		"LOCAL_RTSP: [%s] \n"
		"LOCAL_HTTP: [%s] \n"
		"gsipc_rtsp_h264: [%s] \n"
		, GS_IPC_MAC, 
		GS_MAC_ADDR,
		GS_IPC_RTSP, 
		GS_IPC_VIDEO, 
		GS_IPC_HTTP,
		GS_IPC_gsvideo,
		GS_IPC_rtsptcp, 
		GS_IPC_httptcp, 
		gsipc_rtsp_h264);
#endif 
}

#define 	MASK_HTTP		(1<<0)
#define 	MASK_RTSP		(1<<1)
#define 	MASK_VIDEO		(1<<2)

char * get_ptr_str(
		char **dest, 
		unsigned char offset, 
		struct dns_packet *p
)
{
	int done = 1;
	char ch;
	char *str = NULL, off_ptr = 0; 
	char *dst = *dest;
	
	
	if( dst == NULL || p == NULL ){
		printf("Input params error \n");
		DEBUG(LOG_ERR,  "input params error \n");
		return dst; 
	}

	
	str = p->data + offset; 
	
	while(done == 1 ){
		ch=str[0];
		if (ch == 0xc0){
			off_ptr = str[1];
			printf("getptr: if ptr too, rerun get_ptr_str %d\n", off_ptr);
			return get_ptr_str(dst, off_ptr, p);
		}
		if( ch == 0 ){
			printf("ch is 0 \n");
			return dst; 
		}
		//dst[0] = '.';
		printf("getptr: copy %d [%s] \n", ch, str);
		memcpy(dst, str+1, ch);
		str += ch+1; 
		dst += ch;
		printf("dest: [%s] \n", dest);
		if( strstr(dest, "local")){
			printf("get local \n");
			return dst; 
		}
		dst[0] = '.';
		dst += 1; 
	}
	
	return dst; 
}

char * mdns_get_name_type(
	char *buf, 
	char *name, 
	unsigned short *type,
	struct dns_packet *p)
{
	unsigned char ch;
	unsigned short t;
	char 		*pstr = buf;
	char		*dst = name;
	char		*org = name;
	char 		*ptr_dst = NULL; 

#if 0
	printf("000Start: \n"
		"0x%02x 0x%02x 0x%02x %02x \n"
		"0x%02x 0x%02x 0x%02x %02x \n"
		"0x%02x 0x%02x 0x%02x %02x \n\n", 
		pstr[0], pstr[1], pstr[2], pstr[3],
		pstr[4], pstr[5], pstr[6], pstr[7],
		pstr[8], pstr[9], pstr[10], pstr[11]);
#endif 

	while(1){
		ch=pstr[0];
		if( ch == 0 || strstr(org, "local")){
			//printf("org = [%s] \n", org);
			break; 
		}
		/* ָ */
		if(pstr[0] != 0xc0 ) {
			memcpy(dst, pstr+1, ch);
			if( pstr[ch+1] != 0 ){
				dst[ch]  = '.';
			}
			dst  += ch+1;
			pstr += ch+1; 
		}else{
		#if 0
			if( pstr[0] == 0xc0 && pstr[1] == 0x23 ){
				memcpy(dst, "local", sizeof("local"));
				pstr += 1; 
				break; 
			}
			if( pstr[1] == 0x43 || pstr[1] == 0x40){
				memcpy(dst, "_tcp.local", sizeof("_tcp.local"));
				pstr += 1;
				break; 
			}
		#endif 
		#if 0
		dst = get_ptr_str(&dst, pstr[1],p);
		#else
			ptr_dst = p->data + pstr[1];
			ch = ptr_dst[0];
			//printf("00ptrdst_ch = %d, pstr[1] = %d, dst=[%s] \n",
			//	ch, pstr[1], ptr_dst+1);
			memcpy(dst, ptr_dst+1, ch);
			dst += ch;
			
			ptr_dst = ptr_dst +1+ch;
			if( ptr_dst[0] ==  0xc0 ){
				ptr_dst = p->data + ptr_dst[1];
				ch = ptr_dst[0];
				dst[0] = '.';
				//printf("00.11ptrdst_ch = %d, pstr[1] = %d, dst=[%s] \n",
				//ch, pstr[1], ptr_dst+1);
				memcpy(dst+1, ptr_dst+1, ch);
			}
		#endif 
			pstr += 2; 
		}
		
	}
	memcpy(&t, pstr+1, sizeof(t));
	*type = ntohs(t);
	pstr += 2; //* type
	pstr += 2; //* class
	ch=pstr[0];

#if 0
	printf("000End: \n"
		"0x%02x 0x%02x 0x%02x %02x \n"
		"0x%02x 0x%02x 0x%02x %02x \n"
		"0x%02x 0x%02x 0x%02x %02x \n\n", 
		pstr[0], pstr[1], pstr[2], pstr[3],
		pstr[4], pstr[5], pstr[6], pstr[7],
		pstr[8], pstr[9], pstr[10], pstr[11]);
#endif 
	return pstr;
}

char * mdns_get_doman_type(
		char *buf, 
		char *name, 
		unsigned short *type,
		struct dns_packet *p
)
{
	int ret; 
	unsigned char ch;
	unsigned short t, data_len = 0, l;
	char 		*pstr = buf;
	char		*dst = name;
	char		*org = name; 
	char		*ptr_dst = NULL; 

#if 0
	printf("Start: \n"
		"0x%02x 0x%02x 0x%02x %02x \n"
		"0x%02x 0x%02x 0x%02x %02x \n"
		"0x%02x 0x%02x 0x%02x %02x \n\n", 
		pstr[0], pstr[1], pstr[2], pstr[3],
		pstr[4], pstr[5], pstr[6], pstr[7],
		pstr[8], pstr[9], pstr[10], pstr[11]);
#endif 
	pstr += 2; //* jump ptr
	//printf("JPTR: %x%x %x%x \n", pstr[0], pstr[1], pstr[2], pstr[3]);
	memcpy(&t,pstr, sizeof(t));
	pstr += 2; //* jump type
	//printf("JTYPE: %x%x %x%x \n", pstr[0], pstr[1], pstr[2], pstr[3]);
	pstr += 2; //* jump class type
	//printf("JCLASSTYPE: %x%x %x%x \n", pstr[0], pstr[1], pstr[2], pstr[3]);
	pstr += 4; //* jump time 
	//printf("JTIME: %x%x %x%x \n", pstr[0], pstr[1], pstr[2], pstr[3]);
	memcpy(&l, pstr, sizeof(l));
	data_len = ntohs(l);
	//printf("DATALEN = %d \n", data_len);
	pstr += 2; //* jump datalen
#if 0
	printf("StartWhile:\n"
		"0x%02x 0x%02x 0x%02x %02x \n"
		"0x%02x 0x%02x 0x%02x %02x \n"
		"0x%02x 0x%02x 0x%02x %02x \n\n",
		pstr[0], pstr[1], pstr[2], pstr[3],
		pstr[4], pstr[5], pstr[6], pstr[7],
		pstr[8], pstr[9], pstr[10], pstr[11]);
#endif 

	while(1){
		//printf("WW: %x%x %x%x \n", pstr[0], pstr[1], pstr[2], pstr[3]);
		ch = pstr[0];
		if( ch == 0xc0 ){ 
			#if 0
			printf("chisC0:\n"
		"0x%02x 0x%02x 0x%02x %02x \n"
		"0x%02x 0x%02x 0x%02x %02x \n"
		"0x%02x 0x%02x 0x%02x %02x \n\n",  
		pstr[0], pstr[1], pstr[2], pstr[3],
		pstr[4], pstr[5], pstr[6], pstr[7],
		pstr[8], pstr[9], pstr[10], pstr[11]);
			#endif 
			#if 0
			if(pstr[1] == 0x23 ){
				memcpy(dst, "local", sizeof("local"));
				
			}
			if( pstr[1] == 0x43 || pstr[1] == 0x40){
				memcpy(dst, "_tcp.local", sizeof("_tcp.local"));
			}
			#endif 

		#if 0
		printf("Before: get_ptr_str ch = %x , [%s]\n", ch,org);
		dst = get_ptr_str(&dst, pstr[1],p);
		printf("After: get_ptr_str ch = %x , [%s]\n", ch,org);
		#else
			ptr_dst = p->data + pstr[1] ;
			ch = ptr_dst[0];
			//printf("11ptrdst_ch = %d, pstr[1] = %d , ptrdst = [%s]\n",
			//	ch, pstr[1], ptr_dst+1);
			memcpy(dst, ptr_dst+1, ch);
			
			dst += ch;
			ptr_dst = ptr_dst +1+ch;
			if( ptr_dst[0] ==  0xc0 ){
				ptr_dst = p->data + ptr_dst[1];
				ch = ptr_dst[0];
				dst[0] = '.';
				//printf("11.11ptrdst_ch = %d, pstr[1] = %d , ptrdst = [%s]\n",
				//ch, pstr[1], ptr_dst+1);
				memcpy(dst+1, ptr_dst+1, ch);
			}

		#endif 	
			pstr += 2; 
			//pstr += 2;
			
			break; 
			
		}
			//printf("CH = %x , [%s]\n", ch,name);
			memcpy(dst, pstr+1, ch);
			if( pstr[ch+1] != 0 ){
				dst[ch]  = '.';
			}
			
			dst  += ch+1;
			pstr += ch+1; 
		
	}

	#if 0
	printf("End: \n"
		"0x%02x 0x%02x 0x%02x %02x \n"
		"0x%02x 0x%02x 0x%02x %02x \n"
		"0x%02x 0x%02x 0x%02x %02x \n\n",  
		pstr[0], pstr[1], pstr[2], pstr[3],
		pstr[4], pstr[5], pstr[6], pstr[7],
		pstr[8], pstr[9], pstr[10], pstr[11]);
	#endif 
	return pstr;
}

unsigned short str_find(char *sig, int siglen, char *str)
{
	int n=0;
	unsigned short offset = 0; 
	char *orgstr =str;
	char *pfind = NULL; 

	while(orgstr != NULL ){
		pfind = strstr(orgstr, "sig");
		if( pfind != NULL ){
			printf("find sig[%s], orgstr[%s] \n", sig, orgstr);
			break; 
		}
		orgstr ++;
		n ++;
		if( n> 8000){
			return 0; 
		}
	}
	offset = orgstr - str;
	printf("offset= %d \n", offset);
	return offset;
}

int find_local_tcplocal_ptr( 
	unsigned short *ptr_local, 
	unsigned short *ptr_tcp_loca,
	struct dns_packet *p
)
{
	char *pstr = NULL; 
	char *data = p->data+12; 
	unsigned short acnt, qcnt;

	/* answer count */
	acnt = dns_packet_get_field(p, DNS_FIELD_ANCOUNT);
	
	/* quary count */
	qcnt = dns_packet_get_field(p, DNS_FIELD_QDCOUNT);
	
	printf("acnt = %d, qcnt = %d \n", acnt, qcnt);

	
}

int mdns_find_answer(struct dns_packet *p, char *name)
{
	unsigned short n=0;
	unsigned short acnt = 0, qcnt = 0, cnt = 0 , type;
	char	r_name[128];
	char *data = NULL, *ptr = NULL;
	unsigned short ptr_local=0, ptr_tcp_local=0;
	

	if( p == NULL || name == NULL ){
		DEBUG(LOG_ERR,  "input params error \n");
		return -1; 
	}

#if 0 
	if ( strstr(name, "_http._tcp.local")){
		return 0; 
	}
#endif 

	data = p->data;

	//find_local_tcplocal_ptr(&ptr_local, &ptr_tcp_local, p);
	
	/* answer count */
	acnt = dns_packet_get_field(p, DNS_FIELD_ANCOUNT);
	
	/* quary count */
	qcnt = dns_packet_get_field(p, DNS_FIELD_QDCOUNT);

	//DEBUG(LOG_DEBUG,  "QCNT = %d, ACNT = %d \n", qcnt, acnt);
	//printf("QCNT = %d, ACNT = %d \n", qcnt, acnt);
	cnt = acnt + qcnt; 
	n = 0;
	if( qcnt > 0 ){
		ptr = (char *)(p->data+12); //* ȥͷ
		//DEBUG(LOG_DEBUG,  "PraseQ: \n");
		//printf("PraseQ: %d \n", qcnt);
		while(n<qcnt){
			memset(r_name, 0, sizeof(r_name));
			//printf("PraseQ: mdns_get_name_type [%s] \n", ptr );
			ptr = mdns_get_name_type(ptr, r_name, &type, p);
			//printf("QNEED[%s], RN[%s] \n", name, r_name);
			//DEBUG(LOG_DEBUG,  "QNEED[%s], RN[%s] \n", name, r_name);
			if( strcmp(r_name, name) == 0 ){
				//printf("QFind it %s \n", r_name);
				return 1;
			}
			n++;
			//printf("n=%d, qcnt=%d \n", n, qcnt);
			if(  n<qcnt ){
				ptr ++;
			}
			
			
			
			#if 0
			printf("Q_N = %d \n", n);
			printf("After_mdns_get_name_type: \n"
			"0x%02x 0x%02x 0x%02x %02x \n"
			"0x%02x 0x%02x 0x%02x %02x \n"
			"0x%02x 0x%02x 0x%02x %02x \n\n", 
			ptr[0], ptr[1], ptr[2], ptr[3],
			ptr[4], ptr[5], ptr[6], ptr[7],
			ptr[8], ptr[9], ptr[10], ptr[11]);
			#endif 
			
		}
		if( qcnt<2){
			ptr++;
		}
		n = 0; 
		//DEBUG(LOG_DEBUG,  "PraseA: \n");
		#if 0
		printf("PraseA: %d \n", acnt);
		printf("PraseAStart: \n"
		"0x%02x 0x%02x 0x%02x %02x \n"
		"0x%02x 0x%02x 0x%02x %02x \n"
		"0x%02x 0x%02x 0x%02x %02x \n\n", 
		ptr[0], ptr[1], ptr[2], ptr[3],
		ptr[4], ptr[5], ptr[6], ptr[7],
		ptr[8], ptr[9], ptr[10], ptr[11]);
		#endif 
		while(n < acnt ) {
			memset(r_name, 0, sizeof(r_name));
			//printf("PraseA: mdns_get_doman_type [%s] \n", ptr );
			ptr = mdns_get_doman_type(ptr, r_name, &type, p);
			//DEBUG(LOG_DEBUG,  "ANeed[%s], RN[%s] \n", name, r_name);
			//printf("ANeed[%s], RN[%s] \n", name, r_name);
			if( strcmp(r_name, name) == 0 ){
				//DEBUG(LOG_DEBUG,  "AFind it %s \n", r_name);
				//printf("AFind it %s \n", r_name);
				return 1;
			}
			n++;
			//ptr ++;
			//printf("A_N = %d \n", n);
		}
	}

	return 0; 
}

int check_mdns_enable(void)
{
	char *p;
	int enable = 0; 
	#define P_MDNS_ENABLE	"mdns_enable"

	p = nvram_get(P_MDNS_ENABLE);
	if(p==NULL){
		return 0; 
	}
	enable = atoi(p);
	if(enable < 0 || enable > 1 ){
		return 0; 
	}
	return enable; 
}

int main(int argc, char **argv)
{
	struct sockaddr_in peeraddr;
	struct in_addr ia;
	int sockfd;
	char recmsg[BUFLEN + 1];
	char	ipstr[32];
	char	rr[64];
	unsigned int socklen;
	int n;
	struct hostent *group;
	struct ip_mreq mreq;
	struct dns_packet *p = NULL;
	unsigned short id = 0, type, answer_cnt,quary_cnt, qdcount = 0;
	char name[256];
	int rcnt = 0; 
	int domain_n = 0;

	struct ifaddrs * ifAddrStruct=NULL, *h=NULL;
    void * tmpAddrPtr=NULL;
	char addressBuffer[INET_ADDRSTRLEN];
	syslog_init("mdns_discover");
	
	dev_info_init();
	/*  socket UDPͨѶ */
	sockfd = socket(AF_INET, SOCK_DGRAM, 0);
	if (sockfd < 0) {
		DEBUG(LOG_ERR,  "socket creating err in udptalk \n");
		exit(1);
	}

	/* Ҫ鲥ĵַ */
	bzero(&mreq, sizeof(struct ip_mreq));

	if ((group = gethostbyname(MDNS_IPADDR)) == (struct hostent *) 0) {
		DEBUG(LOG_ERR,  "gethostbyname error \n");
		close(sockfd);
		exit(errno);
	}
	
	bcopy((void *) group->h_addr, (void *) &ia, group->h_length);
	/* ַ */
	bcopy(&ia, &mreq.imr_multiaddr.s_addr, sizeof(struct in_addr));

	/* ÷鲥ϢԴĵַϢ */
	mreq.imr_interface.s_addr = htonl(INADDR_ANY);

	/* ѱ鲥ַΪ鲥Աֻмյ鲥Ϣ */
	if (setsockopt(sockfd, IPPROTO_IP, IP_ADD_MEMBERSHIP, &mreq,
			sizeof(struct ip_mreq)) == -1) {
		DEBUG(LOG_ERR,  "setsockopt IP_ADD_MEMBERSHIP error \n");
		close(sockfd);
		exit(-1);
	}

	socklen = sizeof(struct sockaddr_in);
	memset(&peeraddr, 0, socklen);
	peeraddr.sin_family = AF_INET;
	peeraddr.sin_port = htons(MDNS_PORT);
	if (inet_pton(AF_INET, MDNS_IPADDR, &peeraddr.sin_addr) <= 0) {
		DEBUG(LOG_ERR,  "Wrong dest IP address! \n");
		close(sockfd);
		exit(0);
	}


	/* ԼĶ˿ںIPϢsocket */
	if (bind(sockfd, (struct sockaddr *) &peeraddr,
		sizeof(struct sockaddr_in)) == -1) {
		DEBUG(LOG_ERR,  "Bind error \n");
		close(sockfd);
		exit(0);
	}

	p = (struct dns_packet *)malloc(sizeof(struct dns_packet));
	if( p == NULL ){
		DEBUG(LOG_ERR,  "malloc error \n");
		close(sockfd);
		exit(0);
	}
	
	
	getifaddrs(&ifAddrStruct);
	//DEBUG(LOG_DEBUG, "######>>>> getifaddrs \n");
	h = ifAddrStruct;

	while (ifAddrStruct!=NULL) {
		if ((NULL != ifAddrStruct->ifa_addr)&&
			(ifAddrStruct->ifa_addr->sa_family==AF_INET)) {
			tmpAddrPtr=&((struct sockaddr_in *)ifAddrStruct->ifa_addr)->sin_addr;
			if( strcmp(ifAddrStruct->ifa_name, "lo") == 0 ){
            	//printf("%s IP Address %s, is lo \n", ifAddrStruct->ifa_name, addressBuffer); 
				ifAddrStruct=ifAddrStruct->ifa_next;
				//DEBUG(LOG_DEBUG, "lo addr,continue \n");
            	continue;
			}
			//* ȡIPaddressBuffer
			inet_ntop(AF_INET, tmpAddrPtr, addressBuffer, INET_ADDRSTRLEN);
			//DEBUG(LOG_DEBUG,  "DEVIP: %s IP Address %s  \n", 
			//		ifAddrStruct->ifa_name, addressBuffer); 
			ifAddrStruct=ifAddrStruct->ifa_next;
			
		}else{
			//* IPV6, PASSD
			ifAddrStruct=ifAddrStruct->ifa_next;
			//DEBUG(LOG_DEBUG,  "ipv6 address, drop it ... \n");
			//DEBUG(LOG_DEBUG, "IPV6 ADDRESS , DROP IT , continue \n");
			continue; 
		}
		p = dns_packet_init(p);
		mdns_broadcast_QU_QM(p, addressBuffer, 1);
		dns_send(sockfd, NULL, p);
		p = dns_packet_init(p);
		mdns_broadcast_QU_QM(p, addressBuffer, 0);
		dns_send(sockfd, NULL, p);
		n=5;
		while(n>1){
			p = dns_packet_init(p);
			mdns_broadcast_A(p, addressBuffer);
			dns_send(sockfd, NULL, p);
			usleep(300000);
			n --;
		}
	}
	if( h != NULL ){
		freeifaddrs(h);
		//DEBUG(LOG_DEBUG, "$$$$$$$$>>>> freeifaddrs \n");
		h = NULL; 
		ifAddrStruct = NULL; 
	}

	//DEBUG(LOG_DEBUG,  "run loop, start recv mdns packet  ... \n");
	/* ѭ鲥Ϣ */
	for (;;) {
		memset(p, 0, sizeof(struct dns_packet));
		n = recvfrom(sockfd, p->data, BUFLEN, 0,
				(struct sockaddr *) &peeraddr, &socklen);
		if (n < 0) {
			DEBUG(LOG_ERR,  "recvfrom err in udptalk! \n");
			//exit(4);
		} else {
			memset(ipstr, 0, sizeof(ipstr));
			inet_ntop(AF_INET, &peeraddr.sin_addr, ipstr, 16);
			getifaddrs(&ifAddrStruct);
			//DEBUG(LOG_DEBUG, "######>>>> getifaddrs \n");
			h = ifAddrStruct;
			while (ifAddrStruct!=NULL) {
				if ((NULL != ifAddrStruct->ifa_addr)&&
					(ifAddrStruct->ifa_addr->sa_family==AF_INET)) {
					tmpAddrPtr=&((struct sockaddr_in *)ifAddrStruct->ifa_addr)->sin_addr;
					if( strcmp(ifAddrStruct->ifa_name, "lo") == 0 ){
		            	//printf("%s IP Address %s, is lo \n", ifAddrStruct->ifa_name, addressBuffer); 
						ifAddrStruct=ifAddrStruct->ifa_next;
						//DEBUG(LOG_DEBUG, "lo addr, continue \n");
		            	continue;
					}
					//* ȡIPaddressBuffer
					inet_ntop(AF_INET, tmpAddrPtr, addressBuffer, INET_ADDRSTRLEN);
					//DEBUG(LOG_DEBUG,  "DEVIP: %s IP Address %s  \n", 
					//		ifAddrStruct->ifa_name, addressBuffer); 
					ifAddrStruct=ifAddrStruct->ifa_next;
					
				}else{
					//* IPV6, PASSD
					ifAddrStruct=ifAddrStruct->ifa_next;
					//DEBUG(LOG_DEBUG,  "ipv6 address, drop it ... \n");
					//DEBUG(LOG_DEBUG, "ipv6 addr, drop it,continue \n");
					continue; 
				}

				if( check_peer(&peeraddr, addressBuffer)  ){
					//printf("drop ... \n");
					//DEBUG(LOG_DEBUG, "self addr, drop, continue \n");
					continue; 
				}
				/* ɹյݱ */
				p->size = n; 
				memset(name, 0, sizeof(name));
				dns_get_domain_name(p, name, &type);
				answer_cnt = dns_packet_get_field(p, DNS_FIELD_ANCOUNT);
				quary_cnt = dns_packet_get_field(p, DNS_FIELD_QDCOUNT);
				qdcount = dns_packet_get_field(p, DNS_FIELD_QDCOUNT);
				//DEBUG(LOG_DEBUG,  "recv data form ip[%s], type = %x,"
				//	" name[%s],nscount = %x, quary_cnt = %x, qdcnt = %d \n",
				//	ipstr, 	type, name, answer_cnt,quary_cnt, qdcount);

				#if 0
				printf("recv data form ip[%s], type = %x,"
					" name[%s],nscount = %x, quary_cnt = %x, qdcnt = %d \n",
					ipstr, 	type, name, answer_cnt,quary_cnt, qdcount);
				#endif 
				/* request _services._dns-sd._udp.local QM */
				if( qdcount > 0 ){
					if( strcmp(name, SERVICE_RR) == 0 ||
						strcmp(name, SENTRY_RR) == 0) {
						memset(rr, 0, sizeof(rr));
						if(strcmp(name, SENTRY_RR) == 0){
							strcpy(rr, SENTRY_RR);
						}else{
							strcpy(rr, SERVICE_RR);
						}
						#if 0
						if(  mdns_find_answer(p, doname[0]) != 1 &&
							mdns_find_answer(p, doname[1]) != 1 &&
							mdns_find_answer(p, doname[2]) != 1){
							printf("Have no this product , send all ... \n");
							p = dns_packet_init(p);
							dns_answer_n_packet(p, SERVICE_RR, doname, 3, DNS_TYPE_PTR);
							dns_send(sockfd, NULL, p);
							
						}else{
							printf("one of this have register, just find unregister item... \n");
							for(domain_n=0; domain_n<3; domain_n++){
								if(  mdns_find_answer(p, doname[domain_n]) != 1 ){
									printf("send pkt %s \n", doname[domain_n]);
									p = dns_packet_init(p);
									dns_answer_n_packet(p, SERVICE_RR, &doname[domain_n], 1, DNS_TYPE_PTR);
									dns_send(sockfd, NULL, p);
								}
								//continue; 
							}
						}
						#else
						if(  mdns_find_answer(p, doname[2]) != 1 ){
							//printf("find answer %s \n", doname[0]);
							//continue;
						
							p = dns_packet_init(p);
							dns_answer_n_packet(p, rr, doname, 3, DNS_TYPE_PTR);
							dns_send(sockfd, NULL, p);
						}
						continue; 

						#endif 
					}else if (strcmp(name, AIR_PLAY) == 0 ) {
						if(  mdns_find_answer(p, doname[0]) == 1 ){
							continue;
						}
						p = dns_packet_init(p);
						dns_answer_n_packet(p, AIR_PLAY, doname, 1, DNS_TYPE_PTR);
						dns_send(sockfd, NULL, p);
						continue; 
					}else if (strcmp(name, AFPOVERTCP) == 0 ){
						if(  mdns_find_answer(p, doname[0]) == 1 ){
							continue;
						}
						p = dns_packet_init(p);
						dns_answer_n_packet(p, AFPOVERTCP, doname, 1, DNS_TYPE_PTR);
						dns_send(sockfd, NULL, p);
						continue; 
					}
					if( strcmp(name, GS_IPC_httptcp) == 0 ){
						//printf("HTTP[%s]\n", name);
						if( mdns_find_answer(p, GS_IPC_HTTP) == 1 ){
							continue; 
						}
						p = dns_packet_init(p);
						dns_answer_http_packet(p, GS_IPC_httptcp, GS_IPC_HTTP, addressBuffer, DNS_TYPE_PTR);
						dns_send(sockfd,NULL, p);
						
						p = dns_packet_init(p);
						dns_answer_A_packet(p, GS_IPC_MAC, addressBuffer, DNS_TYPE_A);
						dns_send(sockfd, NULL, p);
						
					}else if ( strcmp(name, GS_IPC_rtsptcp) == 0 ){
						//printf("RTSP[%s]\n", name);
						if( mdns_find_answer(p, GS_IPC_RTSP) == 1 ){
							continue; 
						}
						//printf("111 RTSP \n");
						p = dns_packet_init(p);
						dns_answer_RTSP_packet(p, GS_IPC_rtsptcp, GS_IPC_RTSP, addressBuffer, DNS_TYPE_PTR);
						dns_send(sockfd,NULL, p);

						p = dns_packet_init(p);
						dns_answer_A_packet(p, GS_IPC_MAC, addressBuffer, DNS_TYPE_A);
						dns_send(sockfd, NULL, p);
					
					}else if (  strcmp(name, GS_IPC_gsvideo) == 0){
						//printf("VIDEO[%s]\n", name);
						if( mdns_find_answer(p, GS_IPC_VIDEO) == 1 ){
							continue; 
						}
						p = dns_packet_init(p);
						dns_answer_VIDEO_packet(p, GS_IPC_gsvideo, GS_IPC_VIDEO, addressBuffer, DNS_TYPE_PTR);
						//dns_answer_VIDEO_packet_change(p, GS_IPC_gsvideo, GS_IPC_VIDEO, addressBuffer, DNS_TYPE_PTR);
						dns_send(sockfd,NULL, p);

						p = dns_packet_init(p);
						dns_answer_A_packet(p, GS_IPC_MAC, addressBuffer, DNS_TYPE_A);
						dns_send(sockfd, NULL, p);
					}else if (strcmp(name, GS_IPC_HTTP) == 0){
						if ( type == DNS_TYPE_SRV ){
							p = dns_packet_init(p);
							dns_answer_http_srv_packet(p, GS_IPC_HTTP, GS_IPC_MAC, addressBuffer, DNS_TYPE_PTR);
							dns_send(sockfd,NULL, p);
							
							p = dns_packet_init(p);
							dns_answer_A_packet(p, GS_IPC_MAC, addressBuffer, DNS_TYPE_A);
							dns_send(sockfd, NULL, p);
						}else if ( type == DNS_TYPE_TXT ){
							p = dns_packet_init(p);
							dns_answer_http_txt_packet(p, GS_IPC_HTTP, GS_IPC_MAC, addressBuffer, DNS_TYPE_PTR);
							dns_send(sockfd,NULL, p);

							p = dns_packet_init(p);
							dns_answer_A_packet(p, GS_IPC_MAC, addressBuffer, DNS_TYPE_A);
							dns_send(sockfd, NULL, p);
						}
					}else if (strcmp(name, GS_IPC_RTSP) == 0){
						//printf("222 RTSP \n");
						if ( type == DNS_TYPE_SRV ){
							
							//printf("222 RTSP SRV \n");
							p = dns_packet_init(p);
							dns_answer_RTSP_srv_packet(p, GS_IPC_HTTP, GS_IPC_MAC, addressBuffer, DNS_TYPE_PTR);
							dns_send(sockfd,NULL, p);
							
							p = dns_packet_init(p);
							dns_answer_A_packet(p, GS_IPC_MAC, addressBuffer, DNS_TYPE_A);
							dns_send(sockfd, NULL, p);
						}else if ( type == DNS_TYPE_TXT ){
							//printf("222 RTSP TXT \n");
							p = dns_packet_init(p);
							dns_answer_RTSP_txt_packet(p, GS_IPC_HTTP, GS_IPC_MAC, addressBuffer, DNS_TYPE_PTR);
							dns_send(sockfd,NULL, p);

							p = dns_packet_init(p);
							dns_answer_A_packet(p, GS_IPC_MAC, addressBuffer, DNS_TYPE_A);
							dns_send(sockfd, NULL, p);
						}
					}else if (strcmp(name, GS_IPC_VIDEO) == 0){
						if ( type == DNS_TYPE_SRV ){
							p = dns_packet_init(p);
							dns_answer_VIDEO_srv_packet(p, GS_IPC_HTTP, GS_IPC_MAC, addressBuffer, DNS_TYPE_PTR);
							dns_send(sockfd,NULL, p);
							
							p = dns_packet_init(p);
							dns_answer_A_packet(p, GS_IPC_MAC, addressBuffer, DNS_TYPE_A);
							dns_send(sockfd, NULL, p);
						}else if ( type == DNS_TYPE_TXT ){
							p = dns_packet_init(p);
							dns_answer_VIDEO_txt_packet(p, GS_IPC_HTTP, GS_IPC_MAC, addressBuffer, DNS_TYPE_PTR);
							dns_send(sockfd,NULL, p);

							p = dns_packet_init(p);
							dns_answer_A_packet(p, GS_IPC_MAC, addressBuffer, DNS_TYPE_A);
							dns_send(sockfd, NULL, p);
						}
					}
				}//*if( qdcount > 0 )
			}//* while (ifAddrStruct!=NULL)
			//DEBUG(LOG_DEBUG, "end of while(ifAddrStruct !=NULL ) \n");
			if( h != NULL ){
				freeifaddrs(h);
				//DEBUG(LOG_DEBUG, "$$$$$$$$>>>> freeifaddrs \n");
				h = NULL; 
				ifAddrStruct = NULL; 
			}
		}
	}
	if( p != NULL ){
		free(p);
		p = NULL; 
	}
	if(sockfd > 0){
		close(sockfd);
	}
	syslog_exit();
}

#endif 


