/*
 * (C) Copyright 2013 Grandstream Networks, Inc.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <common.h>
#include <command.h>
#include <nand.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/partitions.h>
#include <ubi_uboot.h>
#include <jffs2/load_kernel.h>
#define CONFIG_MMU 
#include <asm/arch/factory.h>
#include <jffs2/jffs2.h>
#include <asm/io.h>
#include <gs_rsa/gs_rsa.h>

static int read_image_data(char *part_name, int offset, int len, char *buf)
{
    int ret = 0;
    struct spi_flash *spi;
    struct mtd_device *mtd_dev;
    struct part_info *part;
    u8 pnum;
    
    mtdparts_init();
    
    if (find_dev_and_part(part_name, &mtd_dev, &pnum, &part)) {
        printf("Partition %s not found!\n", part_name);
        return -1;
    }

    
    spi = spi_flash_probe(CONFIG_ENV_SPI_BUS, CONFIG_ENV_SPI_CS,
            CONFIG_ENV_SPI_MAX_HZ, CONFIG_ENV_SPI_MODE);
    if (!spi) {
        printf("%s: spi_flash probe error.\n", __func__);
        return -1;
    }

    ret = spi_flash_read(spi, (int)part->offset + offset, len, (u_char *)buf);
    if (ret) {
        printf("%s: spi_flash read error.\n", __func__);
        spi_flash_free(spi);
        return 1;
    }
    
    spi_flash_free(spi);
    return 0;
}

static int calculate_checksum( unsigned char *buf, unsigned int size, unsigned short header_checksum )
{
    int i;
    unsigned short check_sum = 0;
    unsigned short *ptr = ( unsigned short *)buf;

    for (i = 0; i < size/sizeof( unsigned short); i++)
    {
        check_sum += gs_endian_conv16( ptr[i] );
    }

    /* verify the checksum */
    if (((header_checksum + check_sum) & 0xffff) != 0)
        return 1;
    return 0;    
}

//check partition, UBI corrupt bit, signature, get compatibility versions
int check_part(char* mtd_partname, 
               unsigned int *version, 
               int *provision_counter, 
               unsigned int *compatibility,
               unsigned int *fw_mask)
{
    char *buf = NULL;
    size_t size = 0;
    ImageHeader *imgHeader = NULL;
    unsigned int signature;
	uint8_t sha256_digest[32];
    GS_FACTORYPARA* fact_para;
    
    printf("Checking %s partition...\n", mtd_partname);

    size = sizeof(ImageHeader);
    buf = (char *)0x82208000;   // loadaddr

    read_image_data( mtd_partname, 0, size, buf);
    imgHeader = ( ImageHeader *)buf;

    signature = gs_endian_conv32( imgHeader->Signature );
    if ( signature != 0x23C97AF9 ) //check Grandstream Signature
    {
        printf( "%s not found grandstream image signature\n", mtd_partname );
        return -1;
    }

    read_image_data( mtd_partname, 0, gs_endian_conv32(imgHeader->SizeUsed) + HEADER_BLOCK_SIZE, buf);

    /* if rsa signature verified successful, 
       checksum check is not necessary */
    if ( calculate_checksum( (unsigned char *)buf + HEADER_BLOCK_SIZE, 
                             gs_endian_conv32(imgHeader->SizeUsed), 
                             gs_endian_conv16(imgHeader->Checksum) ) != 0 )
    {
        printf( "%s checksum error\n", mtd_partname );
        return -1;
    }

    /*gs_sha256(sha256_digest,(unsigned char *)buf + HEADER_BLOCK_SIZE, gs_endian_conv32(imgHeader->SizeUsed));
    if ( gs_rsa_verify( sha256_digest, imgHeader->RsaSignature, sizeof(imgHeader->RsaSignature)) != 0 )
    {
        printf( "%s signature check failed\n", mtd_partname );
        fact_para = fact_para_get(0);
        if ( fact_para->hw_production == 0 ) 
        {
            printf("     RSA-signature check bypassed\n");
        }
        else 
        {
            return -1;
        }
    }*/
    
    *version = gs_endian_conv32(imgHeader->Version);
    if ( provision_counter )
    {
        // use maxium image header ProvisionCounter as its ProvisionCounter
        int temp_counter = gs_endian_conv32( imgHeader->ProvisionCounter );
        if ( *provision_counter == -1 )
            *provision_counter = temp_counter;
        else if ( *provision_counter !=  temp_counter )
        {
            printf( "%s provision counter is not same\n", mtd_partname );
            return -1;
        }

    }

    *fw_mask |= gs_endian_conv32( imgHeader->FwVerMask );
    return 0;
}

int compare_versions( unsigned int Version1, unsigned int Version2 )
{
    unsigned short v[2] ;

    v[0] = ( Version1 >> 24) & 0xff;
    v[1] = ( Version2 >> 24) & 0xff;
    if ( v[0] > v[1] )
        return -1;
    else if ( v[1] > v[0] )
        return 1;

    v[0] = ( Version1 >> 16) & 0xff;
    v[1] = ( Version2 >> 16) & 0xff;
    if ( v[0] > v[1] )
        return -1;
    else if ( v[1] > v[0] )
        return 1;

    v[0] = ( Version1 >> 8) & 0xff;
    v[1] = ( Version2 >> 8) & 0xff;
    if ( v[0] > v[1] )
        return -1;
    else if ( v[1] > v[0] )
        return 1;

    v[0] = Version1 & 0xff;
    v[1] = Version2 & 0xff;
    if ( v[0] > v[1] )
        return -1;
    else if ( v[1] > v[0] )
        return 1;

    return 0;
}

#define CORE              0
#define BASE              1

#define MAIN_PART         0
#define BACKUP_PART       1

//check all partitions
int check_parts(void)
{
    unsigned int version[2][3] = { { 0, 0, 0 }, { 0, 0, 0 } };
    char partition_ok[2][3] = { { 0, 0, 0 }, { 0, 0, 0 } };
    char compatible[2] = { 0 };
    int retval = 0;
    int provision_counter[2] = { -1, -1 };
    unsigned int fw_mask[2] = { 0, 0 };

    partition_ok[MAIN_PART][CORE] = check_part("core", &(version[MAIN_PART][CORE]), &(provision_counter[MAIN_PART]), NULL, &fw_mask[MAIN_PART] );  
    partition_ok[MAIN_PART][BASE] = check_part("base", &(version[MAIN_PART][BASE]), &(provision_counter[MAIN_PART]), NULL, &fw_mask[MAIN_PART] ); 
    partition_ok[BACKUP_PART][CORE] = check_part("core2",&(version[BACKUP_PART][CORE]), &(provision_counter[BACKUP_PART]), NULL, &fw_mask[BACKUP_PART] );  
    partition_ok[BACKUP_PART][BASE] = check_part("base2",&(version[BACKUP_PART][BASE]), &(provision_counter[BACKUP_PART]), NULL, &fw_mask[BACKUP_PART] ); 
    
    //if all partitions are OK
    if( !partition_ok[MAIN_PART][CORE] && !partition_ok[MAIN_PART][BASE] &&  
        !partition_ok[BACKUP_PART][CORE] && !partition_ok[BACKUP_PART][BASE]  
      ) 
    {
        printf( "All partitions look ok\n" );
        
        //check provision counter
        if ( provision_counter[MAIN_PART] > provision_counter[BACKUP_PART] )
        {
            return 0;
        }
        else if ( provision_counter[MAIN_PART] < provision_counter[BACKUP_PART] )    
        {
            return 1;
        }
       
        //compare versions
        //base first
        retval = compare_versions( version[MAIN_PART][BASE], version[BACKUP_PART][BASE] ) ;
        printf( "Compare base, %s\n", ( retval == 0 ) ? "both are equal" : ( ( retval == 1 ) ? "base2 is selected" : "base is selected" ) );
        if( retval == 1 )
        {
            return 1;
        }
        else if ( retval == -1 )
        {
            return 0;
        }
        //core second    
        retval = compare_versions( version[MAIN_PART][CORE], version[BACKUP_PART][CORE] ) ;
        printf( "Compare cores, %s\n", ( retval == 0 ) ? "both are equal" : ( ( retval == 1 ) ? "core2 is selected" : "core is selected" ) );
        if( retval == 1 )
        {
            return 1;
        }
        else if ( retval == -1 )
        {
            return 0;
        }

        return 0;
    }
    //if core, base and prog are ok
    if( !partition_ok[MAIN_PART][CORE] && !partition_ok[MAIN_PART][BASE] ) 
    {
        printf ( "Only core, base are OK.\n" );
        return 0;
    }
    //if core2, base2 and prog2 are ok
    if( !partition_ok[BACKUP_PART][CORE] && !partition_ok[BACKUP_PART][BASE] ) 
    {
        printf ( "Only core2, base2 are OK.\n" );
        return 1;
    }

    // will not start if no valid image set
    printf ( "No valid image found.\n" );

    return -1;
}

int do_gsboot(cmd_tbl_t *cmdtp, int flag, int argc, char * const argv[])
{
    if( check_parts() )
    {
        printf( "Selecting core2, base2.\n" );
        setenv("boot_partition", "1");
        setenv("mtd_root", "/dev/mtdblock8");
        setenv("kernel_addr", "0xfb0200");
        setenv("mtdparts", "mtdparts=spi0.0:" \
                            "256k(boot)," \
                            "28k(factory)," \
                            "4K(env)," \
                            "96k(nvram)," \
                            "2048K(core2)," \
                            "13504k(base2)," \
                            "128k(oem2)," \
                            "2048K(core)," \
                            "13504k(base)," \
                            "128k(oem)," \
                            "-(data)\0");
    }
    else
    {
        printf( "Selecting core, root.\n" );
        setenv("boot_partition", "0");
        setenv("mtd_root", "/dev/mtdblock5");
        setenv("kernel_addr", "0x60200");
        setenv("mtdparts", "mtdparts=spi0.0:" \
                            "256k(boot)," \
                            "28k(factory)," \
                            "4K(env)," \
                            "96k(nvram)," \
                            "2048K(core)," \
                            "13504k(base)," \
                            "128k(oem)," \
                            "2048K(core2)," \
                            "13504k(base2)," \
                            "128k(oem2)," \
                            "-(data)\0");
    }

    run_command("setenv bootargs console=${console} ${gxp_console} bootpartition=${bootpartition} root=${mtd_root} rootfstype=${rootfstype} init=${init} mem=${memsize} ${mtdparts}; run read_kernel; bootm ${loadaddr}", 0);

    return 0;
}

/**************************************************/
U_BOOT_CMD(
    gsboot, 4, 1, do_gsboot, "gs boot\n",
		"    --boot from flash\n"
);



