/*
 * SPI flash probing
 *
 * Copyright (C) 2008 Atmel Corporation
 * Copyright (C) 2010 Reinhard Meyer, EMK Elektronik
 * Copyright (C) 2013 Jagannadha Sutradharudu Teki, Xilinx Inc.
 *
 * SPDX-License-Identifier:	GPL-2.0+
 */

#include <common.h>
#include <fdtdec.h>
#include <malloc.h>
#include <spi.h>
#include <spi_flash.h>
#include "sf_internal.h"

//#include <common.h>
#include <command.h>
#include <s_record.h>
#include <net.h>
#include <exports.h>
#include <xyzModem.h>
#ifdef CONFIG_PARTITION_TABLE_BY_CUSTOMER_ENV
#include <linux/mtd/mtd.h>
struct mtd_info spinand_mtd;
#endif

#define FEATURE_ECC_EN        (1<<4)

unsigned char m_pBuf_BadBlk[2048/8] = {0};

#define SPI_FLASH_COM_RDID        0x9f
#define SPI_FLASH_COM_READ        0x03
#define SPI_FLASH_COM_PROGRAM     0x02
#define SPI_FLASH_COM_WR_EN       0x06

#define SPI_NAND_READ_ID            0x9F

#define SPI_NAND_WRITE_ENABLE       0x06
#define SPI_NAND_WRITE_DISABLE      0x04

#define SPI_NAND_SET_FEATURE        0x1F
#define SPI_NAND_GET_FEATURE        0x0F

#define SPI_NAND_PAGE_READ          0x13
#define SPI_NAND_READ_CACHE         0x03
#define SPI_NAND_READ_CACHE_X2      0x3B
#define SPI_NAND_READ_CACHE_X4      0x6B

#define SPI_NAND_PROGRAM_LOAD       0x02
#define SPI_NAND_PROGRAM_LOAD_X4    0x32
#define SPI_NAND_PROGRAM_EXECUTE    0x10

#define SPI_NAND_BLOCK_ERASE        0xD8

#define ADDR_PROTECTION 0xA0
#define ADDR_FEATURE    0xB0
#define ADDR_STATUS     0xC0

#define STATUS_ECCS1    (1<<5)
#define STATUS_ECCS0    (1<<4)
#define STATUS_P_FAIL   (1<<3)
#define STATUS_E_FAIL   (1<<2)
#define STATUS_OIP      (1<<0)

static int feature_cmd[3] = {0xC0, 0xB0, 0xA0};

int  spinand_isbadblock(struct spinand_flash *flash, unsigned long block);
int  spinand_setbadblock(struct spinand_flash *flash, unsigned long offset);

DECLARE_GLOBAL_DATA_PTR;

struct flash_info {
	char		*name;

	/* JEDEC id zero means "no ID" (most older chips); otherwise it has
	 * a high byte of zero plus three data bytes: the manufacturer id,
	 * then a two byte device id.
	 */
	u32			jedec_id;
	u16			ext_id;

	u32 		page_size;
	u32 		page_per_block;

	/* The size listed here is what works with OPCODE_SE, which isn't
	 * necessarily called a "sector" by the vendor.
	 */
	unsigned	block_size;
	u16			n_blocks;

	/*|--------------------64bytes------------------------------|*/
	/*|---12---|-4--|---12----|-4--|---12---|-4--|---12----|-4--|*/
	/*|-seglen-|skip|-segllen-|skip|-seglen-|skip|-segllen-|skip|*/
	u32 		oob_size;        
	u16 		oob_up_skiplen;
	u16 		oob_seglen;
	u16 		oob_down_skiplen;	
	u16 		oob_seg_perpage;
	u16			oob_vail_data_offset;

	/**
	 *  chip character bits:
	 *  bit 0: under_protect flag, the serial flash under protection or not when power on
	 *  bit 1: fast read flag, the serial flash support fast read or not(command 0Bh)
	 *  bit 2: AAI flag, the serial flash support auto address increment word programming
	 *  bit 3: support dual write or no
	 *  bit 4: support dual read or no
	 *  bit 5: support quad write or no
	 *  bit 6: support quad read or no
	 *  bit 7: the second status command (35h) flag,if use 4-wire(quad) mode,the bit must be is enable
	 */
	u16			flags;
#define	SFLAG_UNDER_PROTECT			(1<<0)
#define SFLAG_FAST_READ           	(1<<1)
#define SFLAG_AAAI                	(1<<2)
#define SFLAG_COM_STATUS2         	(1<<3)

#define SFLAG_DUAL_IO_READ         	(1<<4)
#define SFLAG_DUAL_READ           	(1<<5)
#define SFLAG_QUAD_IO_READ         	(1<<6)
#define SFLAG_QUAD_READ           	(1<<7)

#define SFLAG_DUAL_IO_WRITE        	(1<<8)
#define SFLAG_DUAL_WRITE          	(1<<9)
#define SFLAG_QUAD_IO_WRITE        	(1<<10)
#define SFLAG_QUAD_WRITE          	(1<<11)

#define SFLAG_SECT_4K       		(1<<12)
};

static 	struct spinand_flash_status_reg status_reg_list[] = {
		{
			.jedec_id = 0xc8f4c8f4,	.ext_id = 0,
			.b_wip = 0,	.b_wel = 1,	.b_efail= 2,	.b_pfail= 3,
			.b_bp2 = 4,	.b_bp3 = 5,	.b_bp4 = 6,	.b_srp0 = 7,

			.b_qe = 8,	.b_srp1 = 9,.b_lb = 10,	.b_cmp = 14,
			.b_sus = 15,
		},
		/*normal status reg define*/
		{
			.jedec_id = 0xc8f2c8f2,	.ext_id = 0,
			.b_wip = 0,	.b_wel = 1,	.b_efail= 2,	.b_pfail= 3,
			.b_bp2 = 4,	.b_bp3 = 5,	.b_bp4 = 6,	.b_srp0 = 7,

			.b_qe = 8,	.b_srp1 = 9,.b_lb = 10,	.b_cmp = 14,
			.b_sus = 15,
		},
		/*normal status reg define*/
		{
			.jedec_id = 0xc8f1c8f1,	.ext_id = 0,
			.b_wip = 0,	.b_wel = 1,	.b_efail= 2,	.b_pfail= 3,
			.b_bp2 = 4,	.b_bp3 = 5,	.b_bp4 = 6,	.b_srp0 = 7,

			.b_qe = 8,	.b_srp1 = 9,.b_lb = 10,	.b_cmp = 14,
			.b_sus = 15,
		},
		/*normal status reg define*/
		{
			.jedec_id = 0xc8d1c8d1,	.ext_id = 0,
			.b_wip = 0,	.b_wel = 1,	.b_efail= 2,	.b_pfail= 3,
			.b_bp2 = 4,	.b_bp3 = 5,	.b_bp4 = 6,	.b_srp0 = 7,

			.b_qe = 8,	.b_srp1 = 9,.b_lb = 10,	.b_cmp = 14,
			.b_sus = 15,
		},	
		/*normal status reg define*/
		{
			.jedec_id = 0xc8d2c8d2,	.ext_id = 0,
			.b_wip = 0,	.b_wel = 1,	.b_efail= 2,	.b_pfail= 3,
			.b_bp2 = 4,	.b_bp3 = 5,	.b_bp4 = 6,	.b_srp0 = 7,

			.b_qe = 8,	.b_srp1 = 9,.b_lb = 10,	.b_cmp = 14,
			.b_sus = 15,
		},				
		/*normal status reg define*/
		{
			.jedec_id = 0xb148c8b1,	.ext_id = 0,
			.b_wip = 0,	.b_wel = 1,	.b_efail= 2,	.b_pfail= 3,
			.b_bp2 = 4,	.b_bp3 = 5,	.b_bp4 = 6,	.b_srp0 = 7,

			.b_qe = 8,	.b_srp1 = 9,.b_lb = 10,	.b_cmp = 14,
			.b_sus = 15,
		},
		/*normal status reg define*/
		{
			.jedec_id = 0xc8217f7f,	.ext_id = 0,
			.b_wip = 0,	.b_wel = 1,	.b_efail= 2,	.b_pfail= 3,
			.b_bp2 = 4,	.b_bp3 = 5,	.b_bp4 = 6,	.b_srp0 = 7,

			.b_qe = 8,	.b_srp1 = 9,.b_lb = 10,	.b_cmp = 14,
			.b_sus = 15,
		},
		/*normal status reg define*/
		{
			.jedec_id = 0xc831C831,	.ext_id = 0,
			.b_wip = 0,	.b_wel = 1,	.b_efail= 2,	.b_pfail= 3,
			.b_bp2 = 4,	.b_bp3 = 5,	.b_bp4 = 6,	.b_srp0 = 7,

			.b_qe = 8,	.b_srp1 = 9,.b_lb = 10,	.b_cmp = 14,
			.b_sus = 15,
		},	
		/*normal status reg define*/
		{
			.jedec_id = 0xa1e1a1e1,	.ext_id = 0,
			.b_wip = 0,	.b_wel = 1,	.b_efail= 2,	.b_pfail= 3,
			.b_bp2 = 4,	.b_bp3 = 5,	.b_bp4 = 6,	.b_srp0 = 7,

			.b_qe = 8,	.b_srp1 = 9,.b_lb = 10,	.b_cmp = 14,
			.b_sus = 15,
		},
		{
			.jedec_id = 0xc212c212,	.ext_id = 0,
			.b_wip = 0,	.b_wel = 1,	.b_efail= 2,	.b_pfail= 3,
			.b_bp2 = 4,	.b_bp3 = 5,	.b_bp4 = 6,	.b_srp0 = 7,

			.b_qe = 8,	.b_srp1 = 9,.b_lb = 10,	.b_cmp = 14,
			.b_sus = 15,
		},		
		/*normal status reg define*/
		{
			.jedec_id = 0,	.ext_id = 0,
			.b_wip = 0,	.b_wel = 1,	.b_bp0 = 2,	.b_bp1 = 3,
			.b_bp2 = 4,	.b_bp3 = 5,	.b_bp4 = 6,	.b_srp0 = 7,
			
			.b_srp1 = 8,.b_qe = 9,	.b_lb = 10,	.b_cmp = 14,
			.b_sus = 15,
		},
};

/* NOTE: double check command sets and memory organization when you add
 * more flash chips.  This current list focusses on newer chips, which
 * have been converging on command sets which including JEDEC ID.
 */
static struct flash_info ak_spinand_supportlist [] = {
    { "GD5G1GQ4U",      0xc8f1c8f1, 0, 2048, 64, 128*1024, 1024, 64, 0, 12, 4, 4, 4, SFLAG_DUAL_READ|SFLAG_QUAD_READ|SFLAG_QUAD_WRITE,},
    { "GD5F1GQ4UB",     0xc8d1c8d1, 0, 2048, 64, 128*1024, 1024, 64, 0, 12, 4, 4, 4, SFLAG_DUAL_READ|SFLAG_QUAD_READ|SFLAG_QUAD_WRITE,},  
    { "GD5G2GQ4U",      0xc8f2c8f2, 0, 2048, 64, 128*1024, 2048, 64, 0, 12, 4, 4, 4, SFLAG_DUAL_READ|SFLAG_QUAD_READ|SFLAG_QUAD_WRITE,},
    { "GD5F2GQ4UB",     0xc8d2c8d2, 0, 2048, 64, 128*1024, 2048, 64, 0, 16, 0, 4, 4, SFLAG_DUAL_READ|SFLAG_QUAD_READ|SFLAG_QUAD_WRITE,},
    { "GD5G4GQ4U",      0xc8f4c8f4, 0, 2048, 64, 128*1024, 4096, 64, 0, 12, 4, 4, 4, SFLAG_DUAL_READ|SFLAG_QUAD_READ|SFLAG_QUAD_WRITE,},	
    { "PSU1GQ4U",       0xc8217f7f, 0, 2048, 64, 128*1024, 1024, 64, 8, 8,  0, 4, 8, SFLAG_DUAL_READ|SFLAG_QUAD_READ|SFLAG_QUAD_WRITE,},
    { "AFS1GQ4UAAWC2",  0xc831C831, 0, 2048, 128,256*1024, 512,  64, 0, 6,  10,4, 4, SFLAG_DUAL_READ|SFLAG_QUAD_READ|SFLAG_QUAD_WRITE,},	
    { "PN26G01AWSIUG",  0xa1e1a1e1, 0, 2048, 64, 128*1024, 1024, 64, 0, 8,  0, 4, 6, SFLAG_DUAL_READ|SFLAG_QUAD_READ|SFLAG_QUAD_WRITE,},
    { "MX35LF1GE4AB",   0xc212c212, 0, 2048, 64, 128*1024, 1024, 64, 0, 16, 0, 4, 4, SFLAG_DUAL_READ|SFLAG_QUAD_READ|SFLAG_QUAD_WRITE,},
    { "W25N01GV",       0xefaa2100, 0, 2048, 64, 128*1024, 1024, 64, 0, 4,  12,4, 4, SFLAG_DUAL_READ|SFLAG_QUAD_READ|SFLAG_QUAD_WRITE,},
};

/* cdh:check ok 2016.10.17 */
static int init_spinand_flash_rw_info(struct spinand_flash *flash)
{
	struct spi_xfer_ctl *xfer_ctl = &flash->xfer_ctl;

	/**default param.*/
	xfer_ctl->rx_opcode = OPCODE_READ;
	xfer_ctl->rxd_bus_width = SPI_XFER_1DATAWIRE;
	xfer_ctl->rxa_bus_width = SPI_XFER_1DATAWIRE;
	xfer_ctl->tx_opcode = CMD_PAGE_PROGRAM;
	xfer_ctl->txd_bus_width = SPI_XFER_1DATAWIRE;
	xfer_ctl->txa_bus_width = SPI_XFER_1DATAWIRE;

	/** 
	* normal slow read cmd=0x03, dummy_len=0
	* normal fast read cmd= 0x0B, dummy_len=1
	*/
	xfer_ctl->dummy_len = FAST_READ_DUMMY_BYTE;
	
	if(flash->bus_width & FLASH_BUS_WIDTH_2WIRE){
		if(flash->flags & SFLAG_DUAL_READ) {
			xfer_ctl->rx_opcode = CMD_READ_ARRAY_DUAL;
			xfer_ctl->rxd_bus_width = SPI_XFER_2DATAWIRE;
			xfer_ctl->rxa_bus_width = SPI_XFER_1DATAWIRE;
			xfer_ctl->dummy_len = 1; // cdh:check ok
		} else if (flash->flags & SFLAG_DUAL_IO_READ) {
			xfer_ctl->rx_opcode = CMD_READ_ARRAY_DUAL_IO;
			xfer_ctl->rxd_bus_width = SPI_XFER_2DATAWIRE;
			xfer_ctl->rxa_bus_width = SPI_XFER_2DATAWIRE;
			xfer_ctl->dummy_len = 1; // cdh:check ok
		}

		if(flash->flags & SFLAG_DUAL_WRITE) {
			xfer_ctl->tx_opcode = CMD_PAGE_PROGRAM_DUAL; // cdh:not support dual write
			xfer_ctl->txd_bus_width = SPI_XFER_2DATAWIRE;
			xfer_ctl->txa_bus_width = SPI_XFER_1DATAWIRE;
		} else if(flash->flags & SFLAG_DUAL_IO_WRITE) {
			xfer_ctl->tx_opcode = CMD_PAGE_PROGRAM_2IO;
			xfer_ctl->txd_bus_width = SPI_XFER_2DATAWIRE;
			xfer_ctl->txa_bus_width = SPI_XFER_2DATAWIRE;
		}	
	}

	if(flash->bus_width & FLASH_BUS_WIDTH_4WIRE){
		if(flash->flags & SFLAG_QUAD_READ) {
			xfer_ctl->rx_opcode = CMD_READ_ARRAY_QUAD;
			xfer_ctl->rxd_bus_width = SPI_XFER_4DATAWIRE;
			xfer_ctl->rxa_bus_width = SPI_XFER_1DATAWIRE;
			xfer_ctl->dummy_len = 1; // cdh:check ok
		}else if(flash->flags & SFLAG_QUAD_IO_READ){
			xfer_ctl->rx_opcode = CMD_READ_ARRAY_QUAD_IO;
			xfer_ctl->rxd_bus_width = SPI_XFER_4DATAWIRE;
			xfer_ctl->rxa_bus_width = SPI_XFER_4DATAWIRE;
			xfer_ctl->dummy_len = 3; // cdh:modify from winbond 32MB spi norflash
		}

		if(flash->flags & SFLAG_QUAD_WRITE) {
			xfer_ctl->tx_opcode = CMD_PAGE_PROGRAM_QUAD;  // cdh:check ok
			xfer_ctl->txd_bus_width = SPI_XFER_4DATAWIRE;			
			xfer_ctl->txa_bus_width = SPI_XFER_1DATAWIRE;
		}else if(flash->flags & SFLAG_QUAD_IO_WRITE) {  
			xfer_ctl->tx_opcode = CMD_PAGE_PROGRAM_4IO;  // cdh:not support dual write
			xfer_ctl->txd_bus_width = SPI_XFER_4DATAWIRE;
			xfer_ctl->txa_bus_width = SPI_XFER_4DATAWIRE;
		}
	}
	return 0;
}

int write_enable_spinand(struct spi_slave *spi)
{
    u8 buf1[1];
    
    //write enable
    buf1[0] = SPI_NAND_WRITE_ENABLE;
	return spi_flash_cmd_write(spi, buf1,1, NULL, 0);
}

int write_disable_spinand(struct spi_slave *spi)
{
    u8 buf1[1];
    
    //write enable
    buf1[0] = SPI_NAND_WRITE_DISABLE;
	return spi_flash_cmd_write(spi, buf1,1, NULL, 0);
}

static int get_feature(struct spi_slave *spi, u8 addr, u8 *val)
{
	u8 buf[2];

	buf[0] = SPI_NAND_GET_FEATURE;
	buf[1] = addr;

	if (spi_flash_cmd_read(spi, buf,2, val, 1) == -1)
    {
        printf("erase: spi_flash_cmd_write fail\n");
        return -1;
    }
	
	return 0;
}

static int set_feature(struct spi_slave *spi, unsigned char addr, unsigned char val)
{
    unsigned char buf[3] = {0};
	unsigned char data[3] = {0};

    buf[0] = SPI_NAND_SET_FEATURE;
    buf[1] = addr;

	data[0] = val;

    if (spi_flash_cmd_write(spi, buf,2, data, 1) == -1)
    {
        printf("erase: spi_flash_cmd_write fail\n");
        return -1;
    }
       
    return 0;
}

static int wait_cmd_finish(struct spi_slave *spi, u8 *status)
{
	u32 timeout = 0;

	while(timeout++ < 1000000)
	{
		if(get_feature(spi, ADDR_STATUS, status) == 0)
		{
			//printf("get_feature status:%x, STATUS_OIP:%x\n", *status, STATUS_OIP);
			if(!(*status&STATUS_OIP))
				return 0;
		}
	}

	printf("get_feature timeout status:%x\n", *status);
	return -1;
}

static int check_ECC_status(struct spi_slave *spi)
{
	u8 ecc = 0;
	u8 status = 0;

	if(get_feature(spi, ADDR_STATUS, &status) == -1)
	{
		printf("ECC get_feature fail \n");
        return -1;
	}
	
	ecc = (status >> 4) & 0x3;
    if(2 == ecc)    //bit error and not corrected
    {
    	printf("ECC error sr:%x\n", status);
        return -1;
    }

	return 0;
}

int spi_nand_erase(struct spinand_flash *flash, u32 page)
{
    unsigned char cmdbuf[4];
    unsigned char status;

    //write enable
    write_enable_spinand(flash->spi);

    //erase
    cmdbuf[0] = SPI_NAND_BLOCK_ERASE;
    cmdbuf[1] = (page >> 16) & 0xFF;
    cmdbuf[2] = (page >> 8) & 0xFF;
    cmdbuf[3] = (page >> 0) & 0xFF;

    //erase
    if (spi_flash_cmd_write(flash->spi, cmdbuf,4, NULL, 0) == -1)
    {
        printf("erase: spi_flash_cmd_write fail\n");
        goto ERASE_EXIT;
    }
    //us_delay(10);
    //check status
   if (wait_cmd_finish(flash->spi, &status) == 0)
    {
    	//printf("status:%x\n", status);
        if(!(status & STATUS_E_FAIL))
        {
             return 0;
        }
        else
        {
            printf("wait_cmd_finish fail status == STATUS_E_FAIL\n");
        }
    }

ERASE_EXIT:    
    
    return -1;
}

bool ecc_ctl(struct spinand_flash *flash, bool bEnable)
{
    unsigned char feature;

    if(!get_feature(flash, ADDR_FEATURE, &feature))
    {
        return false;
    }

    if(bEnable)
        feature |= FEATURE_ECC_EN;
    else
        feature &= ~FEATURE_ECC_EN;

    if(!set_feature(flash, ADDR_FEATURE, feature))
    {
        return false;
    }

    return true;
}

int spi_nand_read_raw(struct spinand_flash *flash, u32 page, u32 column, u8 *buf, u32 len)
{
	u8 cmdbuf[4];
	u8 status;

	//set to no ecc
    /*if (ecc_ctl(flash->spi, false) == -1)
    {
        printf("ecc_ctl fail\n");  
        return -1;
    }*/
    
	cmdbuf[0] = SPI_NAND_PAGE_READ;
    cmdbuf[1] = (page >> 16) & 0xFF;
    cmdbuf[2] = (page >> 8) & 0xFF;
    cmdbuf[3] = (page >> 0) & 0xFF;
	
	if (spi_flash_cmd_write(flash->spi, cmdbuf, 4, NULL, 0) == -1)
    {
        printf("spi_nand_read_raw: spi_flash_cmd_write fail\n");
        return -1;
    }

	//printf("read page:%d.\n", page);
	if(wait_cmd_finish(flash->spi, &status) == -1)
	{
		printf("spi_nand_read_raw: wait_cmd_finish error, page: %d\n", page);
		return -1;
	}

	cmdbuf[0] = SPI_NAND_READ_CACHE;   
	cmdbuf[1] = (column >> 8) & 0x0F;
	cmdbuf[2] = column & 0xFF;
	cmdbuf[3] = 0x0;

	if (spi_flash_cmd_read(flash->spi, cmdbuf, 4, buf, len) == -1)
    {
        printf("spi_nand_read_raw: : spi_flash_cmd_write fail\n");
        return -1;
    }

	/*if (ecc_ctl(flash->spi, true) == -1)
    {
        printf("spi_nand_read_raw ecc_ctl fail\n");
        return -1;
    }*/

	return 0;
}

int spi_nand_read(struct spinand_flash *flash, u32 page, u32 column, u8 *buf, u32 len)
{
	u8 cmdbuf[4];
	u8 status;

	cmdbuf[0] = SPI_NAND_PAGE_READ;
    cmdbuf[1] = (page >> 16) & 0xFF;
    cmdbuf[2] = (page >> 8) & 0xFF;
    cmdbuf[3] = (page >> 0) & 0xFF;
	
	if (spi_flash_cmd_write(flash->spi, cmdbuf, 4, NULL, 0) == -1)
    {
        printf("erase: read spi_flash_cmd_write fail\n");
        return -1;
    }

	//printf("read page:%d.\n", page);
	if(wait_cmd_finish(flash->spi, &status) == -1)
	{
		printf("wait_cmd_finish error, page:%d\n", page);
		return -1;
	}

	cmdbuf[0] = SPI_NAND_READ_CACHE;   
	cmdbuf[1] = (column >> 8) & 0x0F;
	cmdbuf[2] = column & 0xFF;
	cmdbuf[3] = 0x0;

	if (spi_flash_cmd_read(flash->spi, cmdbuf, 4, buf, len) == -1)
    {
        printf("erase: spi_flash_cmd_write fail\n");
        return -1;
    }

	if (check_ECC_status(flash->spi) == -1)
	{
		printf("check_ECC_status error, page:%x\n", page);
		return -1;
	}

	return 0;
}

int spi_nand_write(struct spinand_flash *flash, u32 page, u32 column, const void *buf, u32 len)
{
	unsigned char cmdbuf[4];
    unsigned char status;

    //write enable
    write_enable_spinand(flash->spi);

    //program load
    cmdbuf[0] = SPI_NAND_PROGRAM_LOAD;    
    cmdbuf[1] = (column >> 8) & 0x0F;
    cmdbuf[2] = column & 0xFF;

    if (spi_flash_cmd_write(flash->spi, cmdbuf, 3, buf, len) == -1)
    {
        printf("write: write spi_flash_cmd_write fail\n");
        return -1;
    }

    //program execute
    cmdbuf[0] = SPI_NAND_PROGRAM_EXECUTE;
    cmdbuf[1] = (page >> 16) & 0xFF;
    cmdbuf[2] = (page >> 8) & 0xFF;
    cmdbuf[3] = (page >> 0) & 0xFF;

	if (spi_flash_cmd_write(flash->spi, cmdbuf, 4, NULL, 0) == -1)
    {
        printf("write: write spi_flash_cmd_write fail\n");
        return -1;
    }

    //check status
    if (wait_cmd_finish(flash->spi, &status) == 0)
    {
        if(!(status & STATUS_P_FAIL))
        {
        	//printf("status:%x\n", status);
            return 0;
        }
        else
        {
        	printf("fail status == STATUS_P_FAIL\n");
        }
    }

    return -1;
}

int spinand_flash_erase_inter(struct spinand_flash *flash, u32 offset, size_t len)
{
	u32 addr = offset;
	u32 page = addr/flash->page_size;

	if(flash == NULL)
	{
		printf("error:flash == NULL\n");
		return -1;
	}

	if(offset % flash->erase_size != 0)
	{
		printf("error: erase offset % erase_size != 0\n");
		return -1;
	}

	if(len % flash->erase_size != 0)
	{
		printf("error: erase len % erase_size != 0\n");
		return -1;
	}

	while (len) {
		if (spi_nand_erase(flash, page) == -1) {
			printf("error: spi_nand_erase fail, addr:%d\n", addr);
			return -1;
		}

		len -= flash->erase_size;
	}

	return 0;
}

int spinand_flash_write_inter(struct spinand_flash *flash, u32 offset, size_t len, const void *data,  void *oob, size_t oob_len)
{
	u32 page = 0;
	u8  oob_buf[64] = {0};
	u8  buf[2112] = {0};
	u32 write_len = 0;
	int ret = -1;

	if(flash == NULL)
	{
		printf("error:flash == NULL\n");
		return -1;
	}

	page = offset/flash->page_size;

	if(len > flash->page_size )
	{
		printf("error:spinand_flash_write_inter more than page size\n");
		return -1;
	}

	//fill buf
	memcpy(buf, data, len);
	write_len = flash->page_size;
	if(oob != NULL && oob_len != 0)
	{
		//read page and oob
		if(spi_nand_read(flash, page, flash->page_size, oob_buf, flash->oob_size) == -1)
		{
			printf("error:spi_nand_read fail, page:%d\n", page);
			return -1;
		}
        memcpy(oob_buf, oob, oob_len);

		memcpy(&buf[flash->page_size], oob_buf, flash->oob_size);
		write_len = flash->page_size + flash->oob_size;
	}

	//write
	ret = spi_nand_write(flash, page, 0,  buf, write_len);
	if(ret == -1)
	{
		printf("error:spi_nand_write fail, page:%d\n", page);
	}
	
    return ret; 
}

int spinand_flash_read_inter(struct spinand_flash *flash, u32 offset, size_t len, void *data,  void *oob, size_t oob_len)
{
	u32 page = 0, i = 0;
	u32 column = 0;
	u32 readlen = oob_len;
	u32 xfer_len = 0;
	u32 idex = flash->badblock_offset;
	u32 oob_idex = 0;
	u8  buf[2112] = {0};
	u32 will_read_len = 0;

	if(flash == NULL)
	{
		printf("error:flash == NULL\n");
		return -1;
	}

	page = offset/flash->page_size;
	will_read_len = flash->page_size + flash->oob_size;	

	if( len > flash->page_size )
	{
		printf("error:spinand_flash_read_inter more than page size, %x, len: %d\n", offset, len);
		return -1;
	}
	
	if(spi_nand_read(flash, page, column, buf, will_read_len) == -1)
	{
		printf("error:spi_nand_read fail, page:%d\n", page);
		return -1;
	}

	memcpy(data, buf, len);
	if(oob != NULL && oob_len != 0)
	{
		memcpy(oob, &buf[flash->page_size], oob_len);
	}

	return 0;
}

/**
* @brief: enable 4 wire transfer mode.
* 
* @author lixinhai
* @date 2014-04-10
* @param[in] flash  spiflash handle.
*/
static int quad_mode_enable(struct spinand_flash *flash)
{
	int shift;
	//int idx;
	u32 regval;
	u8 addr;
	u8 status = 0;
	struct flash_status_reg *fsr = &flash->spinand_stat_reg;

	shift = fsr->b_qe / 8;
	//idx = fsr->b_qe % 8;
	
	addr = feature_cmd[shift];
	if(!wait_cmd_finish(flash->spi, &status))
	{
		return -1;
	}

	write_enable_spinand(flash->spi);

	if(get_feature(flash->spi, addr, &status) == -1)
	{
		write_disable_spinand(flash->spi);
		printf("quad get_feature fail \n");
        return -1;
	}
	
	regval = (u32)status;
	regval |= 1<<(fsr->b_qe % 8);
	
	if(set_feature(flash->spi, addr, (u8)regval) == -1)
	{
		write_disable_spinand(flash->spi);
		printf("quad set_feature fail \n");
        return -1;
	}

	write_disable_spinand(flash->spi);
	
	return 0;
}

/**
* @brief: disable 4 wire transfer mode.
* 
* @author lixinhai
* @date 2014-04-10
* @param[in] flash  spiflash handle.
*/
static int quad_mode_disable(struct spinand_flash *flash)
{
	//int idx; 
	int	shift;
	u32 regval;
	u8 addr;
	u8 status = 0;
	struct flash_status_reg *fsr = &flash->spinand_stat_reg;

	shift = fsr->b_qe / 8;
	//idx = fsr->b_qe % 8; 
	addr = feature_cmd[shift];

	if(!wait_cmd_finish(flash->spi, &status))
	{
		return -1;
	}
	
	write_enable_spinand(flash->spi);
	
	if(get_feature(flash->spi, addr, &status) == -1)
	{
		write_disable_spinand(flash->spi);
		printf("quad get_feature fail \n");
        return -1;
	}
	
	regval = (u32)status;
	regval &= ~(1<<(fsr->b_qe%8));
	if(set_feature(flash->spi, addr, (u8)regval) == -1)
	{
		write_disable_spinand(flash->spi);
		printf("quad set_feature fail \n");
        return -1;
	}
	
	write_disable_spinand(flash->spi);
	return 0;
}

static int spinand_cfg_quad_mode(struct spinand_flash *flash)
{
	int ret = 0;
	
	if((flash->bus_width & FLASH_BUS_WIDTH_4WIRE) && 
		(flash->flags & (SFLAG_QUAD_WRITE|SFLAG_QUAD_IO_WRITE|
			SFLAG_DUAL_READ|SFLAG_DUAL_IO_READ))) {	
		ret = quad_mode_enable(flash);
		if(ret) {
			flash->bus_width &= ~FLASH_BUS_WIDTH_4WIRE;
			printf("config the spiflash quad enable fail. transfer use 1 wire.\n");
		}
	}
	else
		quad_mode_disable(flash);

	return ret;
}

static int spinand_mtd_erase(struct mtd_info *mtd, struct erase_info *instr)
{
    struct spinand_flash *spi = mtd->priv;

	return spinand_erase_flash(spi, instr->addr, instr->len);
}

static int spinand_mtd_read(struct mtd_info *mtd, loff_t from, size_t len,
	size_t *retlen, u_char *buf)
{
	struct spinand_flash *spi = mtd->priv;

    spinand_read_flash(spi, (int)from, (u_char *)buf, len);
	*retlen = len;

	return 0;
}

static int spinand_mtd_write(struct mtd_info *mtd, loff_t to, size_t len,
	size_t *retlen, const u_char *buf)
{
    struct spinand_flash *spi = mtd->priv;
    int error;
    
    error = spinand_write_flash(spi, (int)to, (u_char *)buf, len);
    *retlen = len;
    
	if (!error) {
		*retlen = len;
		return 0;
	}

	return -EIO;
}

static int spinand_mtd_isbad(struct mtd_info *mtd, loff_t ofs)
{
    struct spinand_flash *spi = mtd->priv;

    return spinand_isbadblock(spi, (unsigned long)ofs);
}

static int spinand_mtd_markbad(struct mtd_info *mtd, loff_t ofs)
{
    struct spinand_flash *spi = mtd->priv;

    return spinand_setbadblock(spi, (unsigned long)ofs);
}    
    
static struct flash_info *get_flash_info(u8 *id)
{
    int         i;
    u32         jedec;
    u16         ext_jedec = 0;
    struct flash_info   *info;

    jedec = id[0];
    jedec = jedec << 8;
    jedec |= id[1];
    jedec = jedec << 8;
    jedec |= id[2];
    jedec = jedec << 8;
    jedec |= id[3];

    //ext_jedec = id[3] << 8 | id[4];
    for (i = 0, info = ak_spinand_supportlist;
            i < ARRAY_SIZE(ak_spinand_supportlist);
            i ++, info ++) 
    {
        if (info->jedec_id == jedec) {
            if (info->ext_id != 0 && info->ext_id != ext_jedec)
                continue;
            return info;
        }
    }

    return NULL;
}

static struct spinand_flash *spinand_flash_validate_params(struct spi_slave *spi, u8 *idcode)
{
	struct spinand_flash_status_reg *sr;
	struct spinand_flash *flash;
	int i;
    struct flash_info   *info;

    info = get_flash_info(idcode);
    if ( info )
    {
        printf("Found nand flash: %s\n", info->name);
    }
    else
    {
        return NULL;
    }
     
	flash = malloc(sizeof(*flash));
	if (!flash) {
		debug("SF: Failed to allocate spi_flash\n");
		return NULL;
	}
	memset(flash, '\0', sizeof(*flash));

	/* Assign spi data */
	flash->spi = spi;
	flash->name = "SPINAND";
	flash->memory_map = spi->memory_map;
	flash->bus_width = spi->bus_width;
	flash->flags = info->flags;
	flash->flashid = info->jedec_id;
	flash->oob_size = info->oob_size;
	
	for (i=0; i<ARRAY_SIZE(status_reg_list); i++) {
		sr = &status_reg_list[i];
		if (sr->jedec_id == flash->flashid) {
			//printf("SF: status_reg_list supported flash IDs:0x%x\n", params->jedec);
			break;
		}
	}

	flash->spinand_stat_reg = *sr;

	flash->bus_width = FLASH_BUS_WIDTH_1WIRE;
	debug("anyka spi flash. bus_width:1\n");

	/* Assign spi_flash ops */
	flash->spinand_write = spinand_flash_write_inter;
	flash->spinand_erase = spinand_flash_erase_inter;
	flash->spinand_read = spinand_flash_read_inter;
	flash->spinand_isbadblock = spinand_isbadblock;
	flash->spinand_setbadblock = spinand_setbadblock;

	/* Compute the flash size */
	flash->page_size = info->page_size;
	flash->sector_size = info->block_size;
	flash->erase_size = info->block_size;
	flash->size = info->block_size * info->n_blocks;

	/* Poll cmd seclection */
	//flash->poll_cmd = CMD_READ_STATUS;

	// cdh:intial spi operation data wire and cmd
	init_spinand_flash_rw_info(flash);
    spinand_cfg_quad_mode(flash);
	
#if defined(CONFIG_PARTITION_TABLE_BY_CUSTOMER_ENV)
    spinand_mtd.name = "nand0";
    spinand_mtd.type = MTD_DATAFLASH;
    spinand_mtd.writesize = flash->page_size;
    spinand_mtd.size = flash->size;
    spinand_mtd.erasesize = flash->erase_size;
    spinand_mtd.numeraseregions = 0;

    spinand_mtd._erase   = spinand_mtd_erase;
    spinand_mtd._read    = spinand_mtd_read;
    spinand_mtd._write   = spinand_mtd_write;
    spinand_mtd._block_isbad   = spinand_mtd_isbad;
    spinand_mtd._block_markbad = spinand_mtd_markbad;
    spinand_mtd.priv     = flash;
    spinand_mtd.flags   |= MTD_WRITEABLE;
    // move add_mtd_device to board_init_r because of spl not supporting mtd  
    //add_mtd_device(&spinand_mtd);
#endif

	debug("spi_flash_validate_params finish.\n");
	return flash;
}

struct spinand_flash *spinand_flash_probe(unsigned int bus, unsigned int cs,
		unsigned int max_hz, unsigned int spi_mode)
{
	struct spi_slave *spi;
	struct spinand_flash *flash = NULL;
	u8 idcode[4];
    u8 cmd[2];
	int ret;
	
	/* Setup spi_slave */
	spi = spi_setup_slave(bus, cs, max_hz, spi_mode);
	if (!spi) {
		printf("SF: Failed to set up slave\n");
		return NULL;
	}

	/* Claim spi bus */
	ret = spi_claim_bus(spi);
	if (ret) {
		printf("SF: Failed to claim SPI bus: %d\n", ret);
		goto err_claim_bus;
	}

	/* Read the ID codes */
    memset(idcode, 0, sizeof(idcode));
    cmd[0] = CMD_READ_ID;
    cmd[1] = 0;
	ret = spi_flash_cmd_read(spi, cmd, 2, idcode, 4);
	if (ret) {
		printf("SF: Failed to get idcodes\n");
		goto err_read_id;
	}
	//printf("idcode:%02x, %02x, %02x, %02x, %02x \n", idcode[0], idcode[1], idcode[2], idcode[3], idcode[4]);

	/* Validate params from spi_flash_params table */
	flash = spinand_flash_validate_params(spi, idcode);
	if (!flash)
		goto err_read_id;

	/* Release spi bus */
	spi_release_bus(spi);

	return flash;

err_read_id:
	spi_release_bus(spi);
err_claim_bus:
	spi_free_slave(spi);
	return NULL;
}

void spinand_set_no_protection(struct spinand_flash *flash)
{
	if (-1 == set_feature(flash->spi, ADDR_PROTECTION, 0))
    {
        printf("set_feature fail, and no write \n");
    }
}

void spinand_flash_free(struct spinand_flash *flash)
{
	//spi_free_slave(flash->spi);
	//free(flash);
}

int spinand_isbadblock(struct spinand_flash *flash, unsigned long offset)
{
	unsigned long block;
	unsigned long byte_loc, byte_offset;

	block = offset/flash->erase_size;

    byte_loc = block / 8;
    byte_offset = 7 - block % 8;

    if(m_pBuf_BadBlk[byte_loc] & (1 << byte_offset))
    {
        return 1;
    }

    return 0;
}

int  spinand_setbadblock(struct spinand_flash *flash, unsigned long offset)
{
	unsigned long block;
    unsigned long byte_loc, byte_offset;
    unsigned char buf[1] = {0};

	if(offset % flash->erase_size != 0)
	{
		printf("error: offset % flash->erase_size != 0\n");
		return -1;
	}

	block = offset/flash->erase_size;

    if(NULL != m_pBuf_BadBlk)
    {
        byte_loc = block / 8;
        byte_offset = 7 - block % 8;
        m_pBuf_BadBlk[byte_loc] |= 1 << byte_offset;
    }

    buf[0] = 0x00;
    
    if(spi_nand_write(flash, offset, flash->page_size,(const void *)buf, 1) == -1)
    {
        printf("error: set bad bclok in oob fail， block：%d\n", block);
        return -1;
    
    }
    
    return 0;
}

int  spinand_erase_flash(struct spinand_flash *flash,  u32 start_pos, u32 partition_size)
{	
	u32 blockcnt = 0;
	u32 b_idex = 0, good_block = 0;
	int ret = 0;
	
	blockcnt = (partition_size + flash->erase_size - 1)/flash->erase_size;

	while(1)
	{
		if(b_idex == blockcnt)
		{
			break;
		}

		if(spinand_flash_isbadblock(flash, start_pos + b_idex*flash->erase_size))
		{
			printf("bad block at offset 0x%08x\n", (start_pos + b_idex*flash->erase_size));
			b_idex++;
			continue;
		}
	
		ret = spinand_flash_erase(flash, start_pos + b_idex*flash->erase_size, flash->erase_size);
		if (ret == -1)
		{
			b_idex++;
			spinand_flash_setbadblock(flash, start_pos + b_idex*flash->erase_size);
			printf("erase fail, set bad blcok at offset 0x%08x\n", start_pos + b_idex*flash->erase_size);
			continue;
		}

		b_idex++;
		good_block++;
	}

	if(good_block == 0)
	{
		printf("error: all block is bad, pls chk\n");
		return -1;
	}
	
	return 0;
}

int  spinand_read_flash(struct spinand_flash *flash, u32 start_pos, u8 *buf, u32 buf_len)
{
	u32 read_len = buf_len;
	u32 write_idex = 0;
	u32 page_idex = 0;
	u8 *tmp_buf = NULL;
	u32 will_read_len = 0;
	u32 buf_idex = 0, b_idex = 0, idex = 0;
	int ret = -1;

	//printf("start_pos:%d, partition_size:%d, buf_len:%d\n", start_pos, partition_size, buf_len);
	while(1)
	{
		if(read_len == 0)
		{
			break;
		}
        
		if(spinand_flash_isbadblock(flash, start_pos + b_idex*flash->erase_size))
		{
			printf("bad block at offset 0x%08x\n", (start_pos + b_idex*flash->erase_size));
			b_idex++;
			page_idex += (flash->erase_size/flash->page_size);
			continue;
		}

		tmp_buf = buf + write_idex*flash->page_size;
		buf_idex = 0;
		while(1)
		{
			
			if(read_len > flash->page_size)
			{
				will_read_len = flash->page_size;
			}
			else
			{
				will_read_len = read_len;
			}

			//printf("will_write_len:%d, page_pos:%d\n", will_read_len, start_pos+page_idex*flash->page_size);
			ret = spinand_flash_read(flash, start_pos+page_idex*flash->page_size, will_read_len, &tmp_buf[buf_idex], NULL, 0);
			if (ret == -1)
			{
				printf("error: spinand_flash_read at offset 0x%08x\n", start_pos+page_idex*flash->page_size);
				return -1;
			}
			page_idex++;
			write_idex++;
			read_len -= will_read_len;
			buf_idex += will_read_len;

			//printf("will_write_len:%d, read_len:%d, write_idex:%d, page_idex:%d, %d, %d\n", buf_idex, read_len, write_idex, page_idex, flash->erase_size, flash->page_size);
			
			if(read_len == 0)
			{
				break;
			}

			if((start_pos+page_idex*flash->page_size)%flash->erase_size == 0)
			{
				//printf("write next block\n");
				break;
			}
		
}
		b_idex++;
	}
	return 0;	
}

int  spinand_write_flash(struct spinand_flash *flash, u32 start_pos, u8 *buf, u32 buf_len)
{
	u32 write_len = buf_len;
	u32 write_idex = 0;
	u32 page_idex = 0;
	u8 *tmp_buf = NULL;
	u32 will_write_len = 0;
	u32 buf_idex = 0, b_idex = 0, idex = 0;
	int ret = -1;

	while(1)
	{
		if(write_len == 0)
		{
			break;
		}
		
		if(spinand_flash_isbadblock(flash, start_pos + b_idex*flash->erase_size))
		{
			printf("bad block at offset 0x%08x\n", (start_pos + b_idex*flash->erase_size));
			b_idex++;
			page_idex += (flash->erase_size/flash->page_size);
			continue;
		}

		idex = 0;
		tmp_buf = buf + write_idex*flash->page_size;
		buf_idex = 0;
		while(1)
		{
			
			if(write_len > flash->page_size)
			{
				will_write_len = flash->page_size;
			}
			else
			{
				will_write_len = write_len;
			}

			//printf("will_write_len:%d, page_pos:%d\n", will_write_len, start_pos+page_idex*flash->page_size);
			ret = spinand_flash_write(flash, start_pos+page_idex*flash->page_size, will_write_len, &tmp_buf[buf_idex], NULL, 0);
			if (ret == -1)
			{
				printf("spinand write fail, set badblock at offset 0x%08x\n", start_pos+page_idex*flash->page_size);
				spinand_flash_setbadblock(flash, start_pos + b_idex*flash->erase_size);
				write_idex = write_idex - idex;
				page_idex += (flash->erase_size/flash->page_size);
				break;
			}
			idex++;
			page_idex++;
			write_idex++;
			write_len-= will_write_len;
			buf_idex += will_write_len;
			
			if(write_len == 0)
			{
				break;
			}

			if((start_pos+page_idex*flash->page_size)%flash->erase_size == 0)
			{
				//printf("write next block\n");
				break;
			}
		
}
		b_idex++;
	}

	return 0;
}

static int check_block(struct spinand_flash *flash, unsigned long block)
{
    unsigned char data[64];

    memset(data, 0, sizeof(data));    
    spi_nand_read_raw(flash, block * (flash->erase_size/flash->page_size), flash->page_size,  data, 64);
    if(data[0] != 0xFF)
    {
        return -1;
    }
    
    return 0;
}

static int spinand_scan_all_block(struct spinand_flash *flash, unsigned char pBuf[], unsigned long len)
{
    unsigned long block;
    unsigned long byte_loc;
    unsigned long byte_offset;
    unsigned long i;
	unsigned long block_num_total= flash->size/flash->erase_size;

    if(NULL == pBuf)
    {
    	printf("error : NULL == pBuf\n");
        return -1;
    }
    
    if(block_num_total > len*8)
    {
    	printf("error: len is too less\n");
        return -1;
    }

    memset(pBuf, 0, len);
	
  	for(i = 0; i < block_num_total; i++)
    {
        block = i % block_num_total;

        byte_loc = i/8;
        byte_offset = 7 - i%8;     
        if(-1 == check_block(flash, block))
        {
            printf("Bad block: 0x%08x\n", i * flash->erase_size);
            pBuf[byte_loc] |= 1 << byte_offset;
        }
    }

    return 0;
}

int scan_all_block_spinand(struct spinand_flash *flash)
{
	if(spinand_scan_all_block(flash, m_pBuf_BadBlk, sizeof(m_pBuf_BadBlk)) == -1)
	{
		printf("spinand_scan_all_block fail\n");
		return -1;
	}

	return 0;
}

// called from command "spi nand xxx"
int do_spi_nand(struct spinand_flash *flash, int argc, char * const argv[])
{
	unsigned int offset;
	char buf[2112];
    int i;
    u32 page;

	if (argc < 2)
		return -1;
		
	if ( strncmp(argv[1], "dump", 4) == 0 ) {
		offset = simple_strtoul(argv[2], NULL, 16);
        page = offset / flash->page_size;

        memset(buf, 0, sizeof(buf));

         if ( spi_nand_read_raw(flash, page, 0, buf, flash->page_size + flash->oob_size) ) {
            printf("spi nand read offset(0x%x)error\n", offset);
            return -1;
        }

        printf("Page 0x%08x dump:\n", offset);
        for ( i = 0; i < flash->page_size; i ++  ) {
            if ( (i % 16) == 0 )
                printf("\t");
            else if ( (i % 16) == 8 )
                printf(" ");
            printf("%02x ", buf[i]);
            if ( (i % 16) == 15 ) 
                printf("\n");
        }
        
        printf("OOB:\n");
        for ( i = 0; i < flash->oob_size; i ++  ) {
            if ( (i % 8) == 0 )
                printf("\t");
            printf("%02x ", buf[i + flash->page_size]);
            if ( (i % 8) == 7 )
                printf("\n");
        }
	}
    else if ( strncmp(argv[1], "bad", 3) == 0 ) {
        printf("Bad blocks:\n");

        for ( i = 0; i < flash->size; i += flash->erase_size ) {
            if ( spinand_isbadblock(flash, i) )
                printf("\t%08x\n", i);
        }
        printf("\n");
    }
    else if ( strncmp(argv[1], "setbad", 6) == 0 ) {
		offset = simple_strtoul(argv[2], NULL, 16);

        if ( spinand_setbadblock(flash, offset) == 0 )
            printf("Set bad block %x successfull\n", offset);
        else
            printf("Set bad block %x failed\n", offset);
    }
    else {
        printf("spi nand %s not supported\n", argv[1]);
    }

	return 0;
}


