/*
 * Copyright (C) 2011 OMICRON electronics GmbH
 *
 * based on drivers/mtd/nand/nand_spl_load.c
 *
 * Copyright (C) 2011
 * Heiko Schocher, DENX Software Engineering, hs@denx.de.
 *
 * SPDX-License-Identifier:	GPL-2.0+
 */

#include <common.h>
#include <spi_flash.h>
#include <spl.h>
#include <asm/arch/factory.h>
#include <gs_rsa/gs_rsa.h>

#ifdef CONFIG_SPINAND_FLASH
//=======================================================================
static int 
spinand_spl_load_image(
    uint32_t offs, 
    unsigned int size, 
    void *vdst
)
//=======================================================================
{
	struct spinand_flash *flash;

	flash = (struct spinand_flash *)spi_flash_probe(CONFIG_SPL_SPI_BUS, CONFIG_SPL_SPI_CS,
				CONFIG_SF_DEFAULT_SPEED, SPI_MODE_3);
	if (!flash) {
		puts("spinand probe failed.\n");
		hang();
	}
    
	if( spinand_read_flash(flash, offs, vdst, size) == -1)
	{
		puts("spinand_read_flash failed.\n");
		hang();
	}

    return 0;
}

//=======================================================================
static unsigned char 
compare_uboot_versions( 
    unsigned int Signature1, 
    unsigned int Version1, 
    unsigned int Signature2, 
    unsigned int Version2 
)
//=======================================================================
{
    unsigned short v[2] ;

    if ( Signature1 != GS_PROD_SIGNATURE && Signature2 != GS_PROD_SIGNATURE )
        return 0;
    if ( Signature1 != GS_PROD_SIGNATURE )
        return 1;
    if ( Signature2 != GS_PROD_SIGNATURE )
        return 0;

    v[0] = ( Version1 >> 24) & 0xff;
    v[1] = ( Version2 >> 24) & 0xff;
    if ( v[0] > v[1] )
        return 0;
    else if ( v[1] > v[0] )
        return 1;

    v[0] = ( Version1 >> 16) & 0xff;
    v[1] = ( Version2 >> 16) & 0xff;
    if ( v[0] > v[1] )
        return 0;
    else if ( v[1] > v[0] )
        return 1;

    v[0] = ( Version1 >> 8) & 0xff;
    v[1] = ( Version2 >> 8) & 0xff;
    if ( v[0] > v[1] )
        return 0;
    else if ( v[1] > v[0] )
        return 1;

    v[0] = Version1 & 0xff;
    v[1] = Version2 & 0xff;
    if ( v[0] > v[1] )
        return 0;
    else if ( v[1] > v[0] )
        return 1;

    return 0;
}

//=======================================================================
static int
calculate_image_checksum(
    unsigned int offset, 
    unsigned int size, 
    unsigned short header_checksum
)
//=======================================================================
{
    int i;
    unsigned short check_sum = 0;
    void * addr = (void *)CONFIG_SYS_TEXT_BASE;
    unsigned short *ptr = ( unsigned short *)addr;
    int ret; 

    // needs page-aligned
    ret = spinand_spl_load_image(offset, size + HEADER_BLOCK_SIZE, (void*)((char *)addr - HEADER_BLOCK_SIZE));
    if ( ret != 0 ) 
    {
        printf("failed to load image from offset: 0x%x\n", offset);
        return -1;
    }
    
    for (i = 0; i < size/sizeof( unsigned short); i++)
    {
        check_sum += gs_endian_conv16( ptr[i] );
    }

    /* verify the checksum */
    if (((header_checksum + check_sum) & 0xffff) != 0)
        return 1;
    return 0;    
}

//=======================================================================
static int
image_signature_verification(
    char *name,
    unsigned int offset, 
    unsigned int size, 
    unsigned char *sig, 
    unsigned int siglen
)
//=======================================================================
{
	uint8_t sha256_digest[32];
    int ret; 
    void * addr = (void *)CONFIG_SYS_TEXT_BASE;
    unsigned char *ptr = (unsigned char *)addr + siglen;
    ImageHeader *pHeader = (ImageHeader *)addr;
    GS_FACTORYPARA* fact_para;

    // needs page-aligned
    ret = spinand_spl_load_image(offset, size + HEADER_BLOCK_SIZE, addr);
    if ( ret != 0 ) 
    {
        printf("failed to load image from offset: 0x%x\n", offset);
        return -1;
    }

    // Change ProvisionCounter to 0 for signature verification
    pHeader->ProvisionCounter = 0;

    size = size + HEADER_BLOCK_SIZE - siglen;
    gs_sha256( sha256_digest, ptr, size);
    
    ret = gs_rsa_verify( sha256_digest, sig, siglen );
    printf("%s RSA-signature check %s at offset 0x%08x\n", name, ret ? "failed" : "passed", offset);

    if ( ret )
    {
        fact_para = fact_para_get(0);
        if ( fact_para && fact_para->hw_production == 0 ) 
        {
            printf("     RSA-signature check bypassed\n");
            
            ret = calculate_image_checksum( offset, 
                                            gs_endian_conv32(pHeader->SizeUsed), 
                                            gs_endian_conv16(pHeader->Checksum) );
        }
    }

    return ret;
}

//=======================================================================
int
gs_uboot_check(
    ImageHeader *image
)
//=======================================================================
{
#define UBOOTA_OFFSET 0x400000
#define UBOOTB_OFFSET 0x500000

    ImageHeader header[2];
    uint32_t offset[2] = { UBOOTA_OFFSET, UBOOTB_OFFSET }; 
    unsigned char uboot_index = 0;
    int found = 0;

    memset( &header[0], 0, sizeof(header) );

    // read headers
    spinand_spl_load_image(offset[0], sizeof ( ImageHeader ), (void*)&header[0]);
    spinand_spl_load_image(offset[1], sizeof ( ImageHeader ), (void*)&header[1]);
    
    if ( gs_endian_conv32(header[0].Signature) == GS_PROD_SIGNATURE && 
         gs_endian_conv32(header[1].Signature) == GS_PROD_SIGNATURE )
    {
        //compare provision counters
        if ( gs_endian_conv32(header[0].ProvisionCounter) == gs_endian_conv32(header[1].ProvisionCounter) )
        {
            //compare versions
            uboot_index = compare_uboot_versions( gs_endian_conv32(header[0].Signature), header[0].Version, gs_endian_conv32(header[1].Signature), header[1].Version );
        }
        else
            uboot_index = ( gs_endian_conv32(header[0].ProvisionCounter) < gs_endian_conv32(header[1].ProvisionCounter) )?1:0;

    }
    else if ( gs_endian_conv32(header[1].Signature) == GS_PROD_SIGNATURE )
    {
        uboot_index = 1;
    }

    //calculate checksum
    if ( gs_endian_conv32(header[uboot_index].Signature) == GS_PROD_SIGNATURE &&
         ( image_signature_verification( "UBOOT", 
                                         offset[uboot_index], 
                                         gs_endian_conv32(header[uboot_index].SizeUsed), 
                                         header[uboot_index].RsaSignature, 
                                         sizeof( header[uboot_index].RsaSignature ) ) == 0 ) ) 
    {
        found = 1;
    }
    else
    {
       uboot_index = ( uboot_index == 0 ) ? 1 : 0;

        //calculate checksum
        if( gs_endian_conv32(header[uboot_index].Signature) == GS_PROD_SIGNATURE &&
            ( image_signature_verification( "UBOOT", 
                                            offset[uboot_index], 
                                            gs_endian_conv32(header[uboot_index].SizeUsed), 
                                            header[uboot_index].RsaSignature, 
                                            sizeof( header[uboot_index].RsaSignature ) ) == 0 ) ) 
        {
            found = 1;
        }
    }

    if ( found )
    {
        printf(" Found a valid GS uboot image\n");
    }
    else
    {
        printf(" ERROR!!! No GS uboot image found\n");
        while( 1 );
    }
    
    memcpy( image, &header[uboot_index], sizeof( ImageHeader ) );

    return ( uboot_index == 0 ) ? UBOOTA_OFFSET : UBOOTB_OFFSET;
}


/*
 * The main entry for SPI booting. It's necessary that SDRAM is already
 * configured and available since this code loads the main U-Boot image
 * from SPI into SDRAM and starts it from there.
 */
void spl_spinand_load_image(void)
{
	struct spinand_flash *flash;
    ImageHeader image_header;
	u32 uboot_load_addr;

	/*
	 * Load U-Boot image from SPI flash into RAM
	 */

	flash = (struct spinand_flash *)spi_flash_probe(CONFIG_SPL_SPI_BUS, CONFIG_SPL_SPI_CS,
				CONFIG_SF_DEFAULT_SPEED, SPI_MODE_3);
	if (!flash) {
		puts("spinand probe failed.\n");
		hang();
	}

	if( scan_all_block_spinand(flash) == -1 )
	{
		puts("scan_all_block_spinand failed.\n");
		hang();
	}

    uboot_load_addr = gs_uboot_check(&image_header);

    spl_image.size = image_header.SizeUsed;
    spl_image.entry_point = CONFIG_SYS_TEXT_BASE;
    spl_image.load_addr = CONFIG_SYS_TEXT_BASE;
    spl_image.os = IH_OS_U_BOOT;
    spl_image.name = "U-Boot";

	if ( spinand_spl_load_image( uboot_load_addr, spl_image.size + HEADER_BLOCK_SIZE, (void *)(spl_image.load_addr - HEADER_BLOCK_SIZE) ) == -1)
	{
		puts("spinand_read_flash failed.\n");
		hang();
	}
}

#endif
